(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Import CoreDec.
Require Import SignCorrectness.

Set Implicit Arguments.
Unset Strict Implicit.

  Section ICaseDecOk.

    Context {sg : sigma}.
    Context {sgn_ok : sign_ok}.


    Transparent sort.

    Lemma elim_sort_sat :
      forall sc si s1 s2,
        le_sort s1 s2 -> elim_sort sc si s2 -> elim_sort sc si s1.
    Proof.
      intros.
      apply elim_v6_sat with (2 := H0).
      apply inv_sort with ( ); simple apply H.
    Qed.

    Lemma elim_sort_dec : forall sc si se, decide (elim_sort sc si se).
    Proof elim_v6_dec.

    Lemma elim_sort_typed :
      forall sc si se, elim_sort sc si se -> typed_sort se.
    Proof v6_elim_typed.


    Lemma every_sort_typed:
      forall s, typed_sort s.
    Proof.
      destruct s.
      unfold typed_sort.
      exists (Stype O); apply ax_prop.
      exists (Stype (S n)); apply ax_type; auto.
    Qed.

    Lemma so_dec : forall s, { s = so }+{ ~ s = so }.
    Proof.
      unfold so.
      destruct s as [[|]|].
      right; discriminate.
      left; auto.
      right; discriminate.
    Qed.
    Opaque sort.

    (* Typing case is decidable *)

    Lemma coerce_to_pred_dec :
      forall e t is par,
        wf_type e t ->
        ⊦
          ((e; x_ : (i_arg is)[0:=par]); x_ : Ind{i_name is // ^(1)par,,(#0,,[])}) ->
          ppal_dec (fun se => e ⊦ t <= (case_predicate_form is se)[0:=par]) le_sort.
    Proof.
      intros.
      generalize (inv_wf _ _ H0); intro.
      elim red_to_prd_dec with e t; trivial.
      2: right; red in |- *; intros.
      intros ((x, (a, b)), redt).
      generalize (subj_red_wft H redt); intro wf_pr1.
      elim subtype_dec with e ((i_arg is)[0:=par]) a; intros; try solve [ eapply ok_gs_inj; eassumption ].
      2: right; red in |- *; intros.
      elim red_to_prd_dec with (e; x_ : (i_arg is)[0:=par]) b.
      2: right; red in |- *; intros.
      intros ((x', (a', b')), redb).
      elim
        subtype_dec
        with (e; x_ : (i_arg is)[0:=par]) Ind{i_name is // ^(1)par,,(#0,,[])} a';
          intros; try solve [ eapply ok_gs_inj; eassumption ].
      2: right; red in |- *; intros.
      elim
        red_to_sort_dec
        with
          ((e; x_ : (i_arg is)[0:=par]); x_
            : Ind{i_name is // ^(1)par,,(#0,,[])})
          b'.
      2: right; red in |- *; intros.
      intros (se, redb').
      left.
      split with se.
      split; intros.
      change_notation; apply cci_trans with (Forall x : a, b); auto.
      apply cci_trans with (Forall x_ : a, b).
      apply eq_term_cci_sub.
      apply eq_term_step.
      do 2 right; left; constructor.

      simpl in |- *.
      apply cci_sub_prd; trivial.
      apply cci_trans with (Forall x' : a', b'); auto.
      apply cci_trans with (Forall x_ : a', b'); auto.
      apply eq_term_cci_sub.
      apply eq_term_step.
      do 2 right; left; constructor.

      generalize (inv_prod_r (cci_trans (eq_term_sym_cci_sub (redn_conv redt)) H2));
        intro.
      generalize (inv_prod_r (cci_trans (eq_term_sym_cci_sub (redn_conv redb)) H3));
        intro.
      generalize (cci_trans (eq_term_sym_cci_sub (redn_conv redb')) H4); intro.
      intro; change_notation; apply sub_sort_env_indep with (1 := H5).

      apply (b0 x0).
      apply inv_prod_r with x' a'.
      apply cci_trans with b; auto.
      apply inv_prod_r with x a.
      apply cci_trans with t; auto.

      inversion_clear wf_pr1.
      Inversion_typ H2.
      cut ((e; x_ : (i_arg is)[0:=par]) ⊦ b : %%s2); intros.
      generalize (subject_reduction redb H7); intro wfprd2.
      Inversion_typ wfprd2.
      left with s4.
      apply subtype_env with (1 := H9); trivial.
      unfold add_var in |- *; constructor; constructor; simpl in |- *; trivial.

      apply subtype_env with (1 := H4); trivial.
      unfold add_var in |- *; constructor; constructor; simpl in |- *; trivial.

      apply b0.
      apply inv_prod_l with x' x_ b' (%%x0).
      apply cci_trans with b; auto.
      apply inv_prod_r with x a.
      apply cci_trans with t; auto.

      apply ok_normalize.
      apply var_wft; trivial.

      apply ok_normalize.
      inversion_clear wf_pr1.
      Inversion_typ H2.
      cut ((e; x_ : (i_arg is)[0:=par]) ⊦ b : %%s2); intros.
      generalize (subject_reduction redb H7); intro wfprd2.
      Inversion_typ wfprd2.
      left with s0; trivial.

      apply subtype_env with (1 := H4); trivial.
      unfold add_var in |- *; constructor; constructor; simpl in |- *; trivial.

      apply (b0 x_ Ind{i_name is // ^(1)par,,(#0,,[])} %%x0).
      apply inv_prod_r with x a.
      apply cci_trans with t; auto.

      inversion_clear wf_pr1.
      Inversion_typ H2.
      left with s2.
      apply subtype_env with (1 := H4); trivial.
      unfold add_var in |- *; constructor; constructor; simpl in |- *; trivial.

      apply b0.
      simpl in *.
      eapply inv_prod_l; eapply cci_trans with t.
      eauto; fail.
      change_notation; eassumption; fail.

      apply ok_normalize.
      apply var_wft; trivial.

      apply ok_normalize.
      inversion_clear wf_pr1.
      Inversion_typ H2.
      left with s1; trivial.

      simpl in H2.
      elim b with (1 := H2).
    Qed.

    Lemma new_elim_sort_dec : forall is se (pats:patterns) (b:bool),
      { if b then elim_sort (i_ckind is) (i_kind is) se
        else List.length pats = 0 \/ List.length pats = 1 /\ i_ckind is = so
      } + { ~ if b then elim_sort (i_ckind is) (i_kind is) se
        else List.length pats = 0 \/ List.length pats = 1 /\ i_ckind is = so }.
    Proof.
      intros.
      destruct b.
      destruct (elim_sort_dec (i_ckind is) (i_kind is) se); intuition; fail.
      destruct pats.
      intuition; fail.
      destruct pats; simpl.
      destruct (so_dec (i_ckind is)); intuition; fail.
      intuition; fail.
    Qed.


    Lemma sign_dec_case :
      forall pats e m t b,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat (MutCase b pats) e m t u)
          (cci_subtype e).
    Proof.
      intros.
      elim is_pair_dec with m;
        [ intros ((c, pbr), eqp) | right; red in |- *; intros ].
      rewrite eqp.
      rewrite eqp in H.
      clear eqp.
      elim is_pair_dec with pbr;
        [ intros ((p, br), eqp) | right; red in |- *; intros ].
      rewrite eqp.
      rewrite eqp in H.
      clear eqp pbr m.
      elim red_to_sum_dec with e t;
        [ intros ((tyi, typbr), redt) | right; red in |- *; intros | idtac ].
      elim red_to_sum_dec with e typbr;
        [ intros ((typr, tybr), redtypbr) | right; red in |- *; intros | idtac ].
      cut (red redn_term e t (tyi ** (typr ** tybr)));
        [ intros | apply red_trans with (1 := redt); auto ].
      clear redtypbr redt typbr.
      elim red_to_ind_dec with e tyi;
        [ intros ((ind, (par, (arg, (lm, l)))), redtyi)
          | right; red in |- *; intros
          | idtac ].
      elim search_global with ind;
        [ intros ([co| inds is| is cs], sg_gs);
          [ right; red in |- *; intros | idtac | right; red in |- *; intros ]
          | right; red in |- *; intros ].
      InvSignSat H1 ipattern:ty.
      elim H2; intros.
      generalize H1.
      generalize (inv_sum_l (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H6));
        intro.
      elim (inv_ind_i (cci_trans (eq_term_sym_cci_sub (redn_conv redtyi)) H8)).
      intro.
      generalize (ok_gs_inj sg_gs H9).
      discriminate.
      2: InvSignSat H1 ipattern:ty.
      2: case H2; intros.
      2: generalize H1.
      2: generalize (inv_sum_l (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H6));
        intro.
      2: elim (inv_ind_i (cci_trans (eq_term_sym_cci_sub (redn_conv redtyi)) H8)).
      2: intro.
      2: generalize (ok_gs_inj sg_gs H9).
      2: discriminate.
      cut (⊦ (e; x_ : Ind{ind; lm // par,,(arg,,l)})).
      intro wf_ind'.
      elim
        coerce_to_pred_dec
        with (t := typr) (2 := wf_case_predicate1 sg_gs (var_wft _ _ wf_ind'));
          [ intros (se, (le_pr, lst_pr))
            | RightStr err_bad_predicate; red in |- *; intros
            | idtac ].
      elim (new_elim_sort_dec is se pats b); intros;
        [ intros | right; red in |- *; intros ].
      elim full_patt_dec with is pats;
        [ intros | RightStr err_not_exhaustive; red in |- *; intros ].
      elim (mark_branch_types inds (is:=is) (e; x_ : i_param is) p l pats);
        [ intros | right; red in |- *; intros | idtac ].
      inversion_clear a1.
      cut
        (⊦
          (e; x_
            : Ind{ind; lm // par,,(arg,,l)} **
            ((case_predicate_form is se)[0:=par] ** x[0:=par]))).
      intro wf_triple.
      elim subtype_dec with e tybr (x[0:=par]); intros.
      left.
      split with ((p @ arg) @ c).
      split; intros.
      split
        with
          (Ind{ind; lm // par,,(arg,,l)} **
            ((case_predicate_form is se)[0:=par] ** x[0:=par])).
      simpl in |- *.
      split.
      change_notation; apply cci_trans with (tyi ** (typr ** tybr)); auto.

      trivial.

      split.
      simpl in |- *; red in |- *.
      rewrite (glob_ind_eq_name sg_gs).
      apply Mem_case with inds is.
      rewrite <- (glob_ind_eq_name sg_gs); trivial.

      elim (pp_ok H1); intros.
      rewrite H2.
      change
        (type_case inds is pats p c b
          (Ind{i_name is; lm // par,,(arg,,l)} **
            ((case_predicate_form is se)[0:=par] ** (mkvect x0)[0:=par]))
          ((p @ arg) @ c)) in |- *.
      constructor; trivial.

      apply
        ty_case_ok
        with
          inds
          is
          pats
          p
          c
          br b
          (Ind{ind; lm // par,,(arg,,l)} **
            ((case_predicate_form is se)[0:=par] ** x[0:=par])).
      rewrite <- (glob_ind_eq_name sg_gs); trivial.

      elim (pp_ok H1); intros.
      rewrite H2.
      rewrite (glob_ind_eq_name sg_gs).
      constructor; trivial.

      Inversion_typ wf_triple.
      apply wf_def with s; auto.
      apply typ_conv_wf with t; auto.
      Inversion_typ H; trivial.

      apply var_wft; trivial.

      simpl in |- *.
      change_notation; apply cci_trans with (tyi ** (typr ** tybr)); auto.

      InvSignSat H2 ipattern:ty; case H3; intros.
      apply eq_term_cci_sub.
      apply eq_term_app; trivial.
      apply eq_term_app; trivial.
      generalize (inv_sum_l (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H7));
        intro.
      apply (inv_ind_a (cci_trans (eq_term_sym_cci_sub (redn_conv redtyi)) H9)).

      right; red in |- *; intros; apply b0.
      InvSignSat H2 ipattern:ty; case H3; intros.
      generalize (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H7); intro.
      generalize
        (cci_trans (eq_term_sym_cci_sub (redn_conv redtyi)) (inv_sum_l H9));
        intro subind.
      apply cci_trans with (1 := inv_sum_r (inv_sum_r H9)).
      apply cci_subst_eq_cci with x_ (i_param is:term).
      apply eq_term_sym.
      apply (inv_ind_p subind).

      generalize subind.
      rewrite (inv_ind_i subind); intro.
      elim
        eq_case_branch0
        with (1 := H6) (2 := subind0) (x1 := x_) (t0 := i_param is:term).
      intros.
      apply cci_trans with (1 := H11).
      apply incl_mark_sub_cci.
      apply (pp_least H1).
      split with x1; trivial.
      generalize H2.
      elim (inv_ind_i subind).
      intro.
      elim (inj_ind_is H12 sg_gs).
      elim (inj_ind_il H12 sg_gs).
      trivial.

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      Inversion_typ (subj_red_wf (forget_def _ _ _ _ H) H0).
      Inversion_typ H2.
      Inversion_typ H4.
      left with s4; trivial.

      apply ok_normalize.
      apply var_wft.
      (* cf wf_triple ci-dessous *)
      elim (pp_ok H1); intros.
      rewrite H2.
      apply
        wf_branch_vect with (i := ind) (a1 := arg) (se0 := se) (m := lm) (4 := H3);
        trivial.
      apply typ_conv_wf with typr; auto.
      Inversion_typ H.
      elim (subj_red_ty H4 H0) using inv_pair_sum; intros.
      elim H7 using inv_pair_sum; trivial.

      apply var_wft.
      apply wf_case_predicate with (1 := sg_gs) (2 := var_wft _ _ wf_ind').
      change (typed_sort se) in |- *.
      apply every_sort_typed. (*apply elim_sort_typed with (1 := a); trivial.*)

      (* preuve de wf_triple *)
      apply all_prod_sum.
      Inversion_typ wf_ind'.
      apply wf_var with s; trivial.

      apply all_prod_sum.
      Inversion_typ wf_ind'.
      apply wf_case_predicate with (1 := sg_gs) (2 := Wft_typed _ _ _ H2).
      change (typed_sort se) in |- *.
      apply every_sort_typed. (*apply elim_sort_typed with (1 := a); trivial.*)

      elim (pp_ok H1); intros.
      rewrite H2.
      apply
        wf_branch_vect with (i := ind) (a1 := arg) (se0 := se) (m := lm) (4 := H3);
        trivial.
      apply typ_conv_wf with typr; auto.
      Inversion_typ H.
      elim (subj_red_ty H4 H0) using inv_pair_sum; intros.
      elim H7 using inv_pair_sum; trivial.

      apply var_wft.
      apply wf_case_predicate with (1 := sg_gs) (2 := var_wft _ _ wf_ind').
      change (typed_sort se) in |- *.
      apply every_sort_typed. (*apply elim_sort_typed with (1 := a); trivial.*)

      InvSignSat H1 ipattern:ty; case H2; intros.
      generalize (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H6); intro.
      generalize
        (cci_trans (eq_term_sym_cci_sub (redn_conv redtyi)) (inv_sum_l H8));
        intro subind.
      generalize subind.
      rewrite (inv_ind_i subind); intro.
      elim
        eq_case_branch0
        with (1 := H5) (2 := subind0) (x0 := x_) (t0 := i_param is:term).
      generalize H1.
      elim (inv_ind_i subind).
      intro.
      elim (inj_ind_is sg_gs H9).
      elim (inj_ind_il sg_gs H9).
      intros.
      elim b0 with (mkvect x0).
      split with x0; trivial.

      exact (constr_in_ind_inj sg_gs).

      InvSignSat H1 ipattern:ty; case H2; intros.
      apply b0.
      generalize (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H6); intro.
      generalize
        (cci_trans (eq_term_sym_cci_sub (redn_conv redtyi)) (inv_sum_l H8));
        intro subind.
      generalize H1.
      elim (inv_ind_i subind).
      intro.
      elim (inj_ind_is H9 sg_gs).
      trivial.

      InvSignSat H1 ipattern:ty; case H2; intros.
      apply b0.
      generalize (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H6); intro.
      generalize
        (cci_trans (eq_term_sym_cci_sub (redn_conv redtyi)) (inv_sum_l H8));
        intro subind.
      generalize H1.
      elim (inv_ind_i subind).
      intro.
      elim (inj_ind_is H9 sg_gs).
      destruct b.
      apply elim_sort_sat with se0; trivial.
      apply lst_pr.
      change_notation; apply cci_trans with ((case_predicate_form is0 se0)[0:=par0]).
      apply (inv_sum_l (inv_sum_r H8)).

      apply eq_term_sym_cci_sub.
      elim (inj_ind_is H9 sg_gs).
      apply eq_term_subst_l.
      apply (inv_ind_p subind).

      auto.

      InvSignSat H1 ipattern:ty; case H2; intros.
      generalize (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H6); intro.
      generalize
        (cci_trans (eq_term_sym_cci_sub (redn_conv redtyi)) (inv_sum_l H8));
        intro subind.
      generalize H1.
      elim (inv_ind_i subind).
      intro.
      elim b0 with se.
      elim (inj_ind_is H9 sg_gs).
      apply cci_trans with (1 := inv_sum_l (inv_sum_r H8)).
      apply eq_term_sym_cci_sub.
      apply eq_term_subst_l.
      apply (inv_ind_p subind).

      Inversion_typ (subj_red_wf (forget_def _ _ _ _ H) H0).
      Inversion_typ H1.
      Inversion_typ H3.
      left with s0; trivial.

      Inversion_typ H.
      Inversion_typ (subject_reduction H0 H2).
      apply wf_var with s1.
      apply subject_reduction with (1 := redtyi); trivial.

      InvSignSat H1 ipattern:ty; case H2; intros.
      elim b0 with (GlobInd inds is); auto.
      generalize (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H6); intro.
      generalize
        (cci_trans (eq_term_sym_cci_sub (redn_conv redtyi)) (inv_sum_l H8));
        intro subind.
      rewrite (inv_ind_i subind).
      trivial.

      InvSignSat H1 ipattern:ty; case H2; intros.
      generalize (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H6); intro.
      elim b0 with (1 := inv_sum_l H8).

      Inversion_typ (subj_red_wf (forget_def _ _ _ _ H) H0).
      Inversion_typ H1.
      left with s1; trivial.

      InvSignSat H0 ipattern:ty; case H1; intros.
      generalize (cci_trans (eq_term_sym_cci_sub (redn_conv redt)) H5); intro.
      elim b0 with (1 := inv_sum_r H7).

      Inversion_typ (subj_red_wf (forget_def _ _ _ _ H) redt).
      Inversion_typ H0.
      left with s2; trivial.

      InvSignSat H0 ipattern:ty; case H1; intros.
      apply (b0 _ _ H5).

      eapply var_wft, forget_def; eassumption.

      generalize b0.
      unfold not in |- *.
      InvSignSat H0 ipattern:ty; eauto.

      generalize b0.
      unfold not in |- *.
      InvSignSat H0 ipattern:ty; eauto.
    Qed.

  End ICaseDecOk.
