(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require PTS.
Require Export Mmain.
Require Export SortV6.

Module CicCore.

  Definition name := ml_string.
  Definition x_ := dummy_string.
  Instance eq_dec_name : has_eq_dec name := eq_string.

  (* Number of marks that were not yet destructured *)
  Definition le_marks := nat.
  Definition patterns := list name.

  (* TODO: Struct does not need the field names (a verifier)
     *       Fix should keep the names of the fixpoints
     *)
  Inductive cci_op : Set :=
    | Mark : cci_op
    | Mt_mark : cci_op
    | Lmark : cci_op
    | Nil_mark : cci_op
    | Cons_mark : cci_op
    | App : cci_op
    | Proj1 : cci_op
    | Proj2 : cci_op
    | Struct : list name -> cci_op
    | Record : list name -> cci_op
    | Field : name -> cci_op
    | Const : name -> cci_op
    | MutInd : name -> le_marks -> cci_op
    | MutConstr : name -> cci_op
    | MutCase : bool -> patterns -> cci_op
    | Fixp : cci_op
  .

  Definition oper := cci_op.

  Instance eq_dec_oper : has_eq_dec oper.
  Proof.
    repeat (hnf; intros).
    decide equality; try apply eq_dec; decide equality.
  Defined.

  Definition sort := SortV6.srt_v6.

  Instance eq_dec_sort : has_eq_dec sort := v6_sort_dec.

  Definition string_of (x : name) : ml_string := x.

  Opaque name x_ sort.

  Include PTS.Make.

  Notation "x @ y" := (Cst1 App (x,,y)) (at level 20).
  Notation "'Fix' { f }" := (Cst1 Fixp f) (at level 0, f at level 0).
  Notation "'Cstr' { c // a }" := (Cst1 (MutConstr c) a) (at level 0).
  Notation "'Ind' { i ; m // a }" := (Cst1 (MutInd i m) a) (at level 0).
  Notation "'Ind' { i // a }" := Ind {i; 0 // a} (at level 0).
  Notation "'Match' t => pr 'with' b — pat => br 'end'" :=
    (Cst1 (MutCase b pat) (t,,(pr,,br))) (at level 0, pr at level 100).
  Notation "t ..L" := (Cst1 Proj1 t) (at level 1, left associativity).
  Notation "t ..R" := (Cst1 Proj2 t) (at level 1, left associativity).
  Notation "t ..( f )" := (Cst1 (Field f) t) (at level 1, left associativity).
  Notation "{< flds : ar := v >}" := (Cst1 (Struct flds) (ar,,v))
    (at level 0, flds at level 80).
  Notation "{< flds : ar >}" := (Cst1 (Record flds) ar)
    (at level 0, flds at level 80).

  Notation "m && lm" := (Cst1 Cons_mark (m,,lm))
    (at level 40, left associativity).
  Notation "##" := (Cst0 Mt_mark) (at level 0).
  Notation "[]" := (Cst0 Nil_mark) (at level 0).

End CicCore.

Export CicCore.
