(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export CicPts.

Set Implicit Arguments.
Unset Strict Implicit.

Section ICicMeta.

  Context {sg : sigma}.

  Ltac Inversion_cst T :=
    generalize T; intro INVTYP; InvTypHyp INVTYP; intros t_inv u_inv MEM;
     clear INVTYP; inversion_clear MEM; simpl in |- *;
     try clear t_inv || (generalize t_inv; clear t_inv);
     try clear u_inv || (generalize u_inv; clear u_inv).



  (* Derived and inversion rules related to cci operators *)

  Definition conv' e a b := a = b \/ e ⊦ a <= b /\ wf_type e b.
  Hint Unfold conv': core.

  Lemma conv'_sub_cci : forall e a b, conv' e a b -> e ⊦ a <= b.
  Proof.
    simple destruct 1; intros.
    elim H0; trivial.

    inversion_clear H0; trivial.
  Qed.


  Lemma mkvect_wf : forall e m l, e ⊦ m : mkvect l -> ⊦ (e; x_ : mkvect l).
  Proof.
    simple destruct l; simpl in |- *; intros.
    generalize (type_correctness _ _ _ H); intros.
    inversion_clear H0.
    apply wf_var with s; trivial.

    generalize (type_correctness _ _ _ H); intros.
    inversion_clear H0.
    apply wf_var with s; trivial.
  Qed.


  Lemma wf_mkvect_comp :
    forall e a l n, item a l n -> ⊦ (e; x_ : mkvect l) -> ⊦ (e; x_ : a).
  Proof.
    simple induction 1; simpl in |- *; intros; auto.
    Inversion_typ H0.
    Inversion_typ H1.
    apply wf_var with s1; trivial.

    apply H1.
    Inversion_typ H2.
    Inversion_typ H3.
    apply wf_var with s2; trivial.
  Qed.


  (* Inversion of pairs *)

  Lemma inv_pair_sum :
    forall (P : Prop) e m t a b,
      (e ⊦ m : a -> e ⊦ t : b -> P) -> e ⊦ (m,,t) : a ** b -> P.
  Proof.
    intros.
    Inversion_typ H0.
    generalize (type_correctness _ _ _ H0); intro.
    inversion_clear H5.
    Inversion_typ H6.
    apply H.
    apply Typ_conv with a0 s1; auto; simpl in |- *.
    apply (inv_sum_l H4).

    apply Typ_conv with b0 s2; auto; simpl in |- *.
    apply (inv_sum_r H4).
  Qed.


  (* Typing of projections *)
  Lemma typ_proj1 : forall e m a b, e ⊦ m : a ** b -> e ⊦ m..L : a.
  Proof.
    intros.
    apply Typ_cst1 with (a ** b); trivial.
    generalize (type_correctness _ _ _ H); intros.
    inversion_clear H0.
    Inversion_typ H1.
    apply wf_var with s1; trivial.

    generalize (type_correctness _ _ _ H); intros.
    inversion_clear H0.
    apply wf_var with s; trivial.

    constructor.
  Qed.

  Lemma typ_proj2 : forall e m a b, e ⊦ m : a ** b -> e ⊦ m..R : b.
  Proof.
    intros.
    apply Typ_cst1 with (a ** b); trivial.
    generalize (type_correctness _ _ _ H); intros.
    inversion_clear H0.
    Inversion_typ H1.
    apply wf_var with s2; trivial.

    generalize (type_correctness _ _ _ H); intros.
    inversion_clear H0.
    apply wf_var with s; trivial.

    constructor.
  Qed.


  Lemma proj_vect :
    forall e l l' n,
      trunc n l l' -> forall m, e ⊦ m : mkvect l -> e ⊦ m .^ n : mkvect l'.
  Proof.
    simple induction 1; simpl in |- *; intros.
    trivial.

    apply H1.
    apply typ_proj2 with (1 := H2).
  Qed.


  Lemma projl_vect :
    forall e m a l n, item a l n -> e ⊦ m : mkvect l -> e ⊦ (m .^ n) ..L : a.
  Proof.
    intros.
    elim item_trunc_cons with (1 := H); intros.
    apply typ_proj1 with (mkvect x).
    unfold mkvect in |- *; fold (mkvect (a :: x)) in |- *.
    apply proj_vect with l; trivial.
  Qed.


  (* Inversion of projections *)
  Lemma inv_projn :
    forall e n (P : Prop) t u,
      (forall v,
        e ⊦ t : v ->
          (forall l,
            e ⊦ mkvect l <= v -> exists2 l' : _, trunc n l l' & e ⊦ mkvect l' <= u) ->
          P) -> e ⊦ t .^ n : u -> P.
  Proof.
    simple induction n; simpl in |- *; intros.
    apply H with u; trivial; intros.
    split with l; auto.

    apply H with (2 := H1); intros.
    Inversion_cst H2; intros.
    apply H0 with (1 := H6).
    simple destruct l; simpl in |- *; intros.
    elim one_not_sum with (1 := H8).

    elim H3 with l0; intros.
    split with x; auto.

    apply cci_trans with (2 := H7).
    apply inv_sum_r with (1 := H8).
  Qed.



  Lemma sr_ctx_proj1 :
    forall e a b T,
      (forall U, e ⊦ a : U -> e ⊦ b : U) -> e ⊦ a..L : T -> e ⊦ b..L : T.
  Proof.
    intros.
    Inversion_cst H0; rename b0 into v; intros u wfp wfu ty_a le_T.
    apply typ_conv_wf with u; auto.
    apply typ_proj1 with v; auto.

    apply type_correctness with (1 := H0).
  Qed.



  Lemma typ_app0 :
    forall e x m n a b,
      e ⊦ m : Forall x : a, b ->
        e ⊦ n : a -> ⊦ (e; x_ : (Forall x : a, b) ** a) -> e ⊦ m @ n : b[0:=n].
  Proof.
    intros.
    Inversion_typ H1.
    Inversion_typ H2.
    Inversion_typ H3.
    apply Typ_cst1 with ((Forall x : a, b) ** a).
    apply wf_var with s4.
    fold (%%s4[0:=n]) in |- *.
    apply substitution with x a; trivial.

    apply wf_var with s; trivial.

    apply Typ_pair with s; auto.

    constructor.
  Qed.

(*
  Lemma typ_app'': (e:env)(x:name)(m,n,a,a',b,b':term)
             (conv' e [: {0:=n}b :] b')
           ->[~ e |- a' <= a ~]
           ->[~ e |- m : (x : a)b ~]
           ->[~ e |- n : a' ~]
           ->[~ |- e; x_: (n:a & (x:a)b) ~]
           ->[~ e |- (m n) : b' ~].
Proof.
Intros.
(Cut [~ e |- n : a ~]; Intros).
(Elim H; Intros).
Elim H5.
(Apply typ_app0 with f x a; Trivial).

Inversion_clear H5.
(Apply typ_conv_wf with [: {0:=n}b :]; Trivial).
(Apply typ_app0 with f x a; Trivial).

(Apply typ_conv_wf with a'; Trivial).
Inversion_typ H3.
Inversion_typ H4.
Inversion_typ H5.
(Left with s0; Trivial).
Save.
*)

  Lemma sr_ctx_app :
    forall e a b m T,
      (forall U, e ⊦ a : U -> e ⊦ b : U) -> e ⊦ a @ m : T -> e ⊦ b @ m : T.
  Proof.
    intros.
    Inversion_cst H0; rename a0 into u; rename b0 into v;
      intros wfp wfu typair le_T.
    elim typair using inv_pair_sum; intros.
    apply typ_conv_wf with (v[0:=m]); simpl in |- *; trivial.
    apply typ_app0 with x u; auto.
    apply type_correctness with (1 := H0).
  Qed.

  (* Inversion of an application judgement *)
  Lemma inv_typ_lam_prod :
    forall (P : Prop) e x y a m b u,
      (forall s1 s2 s3,
        e ⊦ b <= a ->
          e ⊦ b : %%s1 ->
            (e; y : b) ⊦ m : u -> (e; y : b) ⊦ u : %%s2 -> rl s1 s2 s3 -> P) ->
      e ⊦ Fun x : a => m : Forall y : b, u -> P.
  Proof.
    intros.
    generalize (type_correctness _ _ _ H0); intro.
    inversion_clear H1.
    Inversion_typ H2.
    simpl in H5.
    Inversion_typ H0.
    simpl in H8.
    apply H with s1 s2 s3; auto.
    apply (inv_prod_l H8).

    apply Typ_conv with u0 s2; simpl in |- *; auto.
    apply subtype_env with (e; x : a); auto.
    unfold add_var, rel_typ in |- *; simpl in |- *.
    constructor 1.
    split; simpl in |- *; auto.
    apply (inv_prod_l H8).

    apply wf_var with s1; trivial.

    apply inv_prod_r with (1 := H8).
  Qed.



  (* Using the fact that all pairs are allowed *)
  Lemma all_prod_sum :
    forall e a b, ⊦ (e; x_ : a) -> ⊦ (e; x_ : b) -> ⊦ (e; x_ : a ** b).
  Proof.
    intros.
    Inversion_typ H.
    Inversion_typ H0.
    elim all_prod with s s0; intros.
    apply wf_var with x.
    apply Typ_sum with s s0; trivial.
  Qed.



(*
  Lemma wf_mkvect: (e:env)[~ |- e ~]
  ->(l:(list term))
    ((n:nat)(x:term)(item x l n)->[~ |- e; x_:x ~])
  ->[~ |- e; x_:!mkvect l! ~].
Proof.
(Induction l; Simpl; Intros).
Apply wf_var with so.
(Apply typ_one_so; Trivial).

Apply all_prod_sum.
(Apply H1 with O; Auto).

Apply H0.
Auto.
Intros.
(Apply H1 with (S n); Auto).
Save.

*)



  (* Application *)
  Lemma typ_app :
    forall e x m n a b,
      e ⊦ m : Forall x : a, b -> e ⊦ n : a -> e ⊦ m @ n : b[0:=n].
  Proof.
    intros.
    apply typ_app0 with x a; auto.
    generalize (type_correctness _ _ _ H); intro.
    inversion_clear H1.
    apply all_prod_sum.
    apply wf_var with s; trivial.

    Inversion_typ H2.
    apply wf_var with s1; trivial.
  Qed.


  Lemma typ_app' :
    forall e x m n a a' b b',
      conv' e b[0:=n] b' ->
      e ⊦ a' <= a -> e ⊦ m : Forall x : a, b -> e ⊦ n : a' -> e ⊦ m @ n : b'.
  Proof.
    intros.
    cut (e ⊦ n : a); intros.
    elim H; intros.
    elim H4.
    apply typ_app with x a; trivial.

    inversion_clear H4.
    apply typ_conv_wf with (b[0:=n]); trivial.
    apply typ_app with x a; trivial.

    apply typ_conv_wf with a'; trivial.
    generalize (type_correctness _ _ _ H1); intro.
    inversion_clear H3.
    Inversion_typ H4.
    left with s1; trivial.
  Qed.

  Lemma typ_app2 :
    forall e m n1 n2 a1 a2 b tm,
      match tm with
        | Forall _ : tmtyp1, (Forall _ : tmtyp2, tmbody) =>
          conv' e (tmbody[1:=n1])[0:=n2] b
            /\ e ⊦ a1 <= tmtyp1
            /\ e ⊦ a2 <= tmtyp2[0:=n1]
        | _ => False
      end ->
      e ⊦ m : tm -> e ⊦ n1 : a1 -> e ⊦ n2 : a2 -> e ⊦ (m @ n1) @ n2 : b.
  Proof.
    simple destruct tm; try (intros; elim H).
    simple destruct t0; simpl in |- *; intros; elim H.
    destruct H as [? [? ?]]; intros ? [? ?].
    apply typ_app' with n0 (t1[0:=n1]) a2 (t2[1:=n1]); auto.
    apply typ_app' with n t a1 (Forall n0 : t1, t2); auto.
  Qed.


  Lemma typ_cons_mark :
    forall e m lm, e ⊦ m : Mark -> e ⊦ lm : Lmark -> e ⊦ m && lm : Lmark.
  Proof.
    intros.
    apply typ_cst1 with (Mark ** Lmark).
    simpl in |- *.
    red in |- *.
    constructor.

    apply inst_pair; auto.
    apply all_prod_sum.
    apply wf_var with so.
    apply typ_mark_so.
    apply (typ_wf _ _ _ H).

    apply wf_var with so.
    apply typ_lmark_so.
    apply (typ_wf _ _ _ H).

    apply wf_var with so.
    apply typ_lmark_so.
    apply (typ_wf _ _ _ H).
  Qed.


  Class sign_ok := {

    (* 1- The signature is injective *)
    ok_gs_inj :
      forall x g1 g2,
        sign_glob_spec sg x g1 -> sign_glob_spec sg x g2 -> g1 = g2;

    (* 2- Constants are well-typed *)
    ok_cst :
      forall x co,
        sign_glob_spec sg x (GlobConst co) ->
        ⊦ ((( ); x_ : c_par co); c_decl co);

    (* 3- Inductive types are ok *)
    ok_ind :
      forall ind inds is,
        sign_glob_spec sg ind (GlobInd inds is) ->
        ⊦ (((( ); x_ : i_param is); x_ : i_arg is); x_ : %%(i_kind is));

    (* 4- Constructors are ok *)
    ok_constr :
      forall x is c,
        sign_glob_spec sg x (GlobCstr is c) ->
        ((( ); x_ : i_param is); x_ : c_arg c) ⊦ c_inst c : ^(1)(i_arg is);

    (* 5- Strong normalization *)
    ok_normalize :
      forall e t, wf_type e t -> sn (ctxt redn_term) e t

  }.

  Section ICicMetaOk.

    Context {sgn_ok : sign_ok}.

    (* 1- *)

    Lemma inj_ind_is :
      forall x il il' is is',
        sign_glob_spec sg x (GlobInd il is) ->
        sign_glob_spec sg x (GlobInd il' is') -> is = is'.
    Proof.
      intros.
      generalize (ok_gs_inj H H0); intros.
      injection H1; trivial.
    Qed.

    Lemma inj_ind_il :
      forall x il il' is is',
        sign_glob_spec sg x (GlobInd il is) ->
        sign_glob_spec sg x (GlobInd il' is') -> il = il'.
    Proof.
      intros.
      generalize (ok_gs_inj H H0); intros.
      injection H1; trivial.
    Qed.


    Lemma constr_in_ind_inj :
      forall i inds is,
        sign_glob_spec sg i (GlobInd inds is) -> one_ind_unique is.
    Proof.
      red in |- *; intros.
      generalize (ok_gs_inj (glob_ind_cst H H0) (glob_ind_cst H H1)).
      intro.
      injection H2; trivial.
    Qed.

    (* 2- Constants are well-typed *)

    (* 3- Inductive types are ok *)

    Lemma ind_kind_ok :
      forall ind inds is,
        sign_glob_spec sg ind (GlobInd inds is) -> in_dom_rel ax (i_kind is).
    Proof.
      intros.
      Inversion_typ (ok_ind H).
      Inversion_typ H0.
      split with s2; trivial.
    Qed.

    Lemma ind_par_arg_one :
      forall ind inds is,
        sign_glob_spec sg ind (GlobInd inds is) ->
        ⊦ ((( ); x_ : i_param is); x_ : i_arg is).
    Proof.
      intros.
      apply (inv_wf _ _ (ok_ind H)).
    Qed.


    Lemma ind_par_arg_inst :
      forall ind inds is e p,
        sign_glob_spec sg ind (GlobInd inds is) ->
        e ⊦ p : i_param is ->
          ⊦ (e; x_ : i_param is ** ((i_arg is)[0:=p] ** Lmark)).
    Proof.
      intros.
      generalize (ind_par_arg_one H); intro.
      apply all_prod_sum.
      unfold add_var in |- *.
      apply thin_par_wf.
      apply (inv_wf _ _ H1).

      apply (typ_wf _ _ _ H0).

      apply all_prod_sum.
      apply subst_wf with (x := x_) (1 := H0).
      Inversion_typ H1.
      apply wf_var with s.
      apply thin_par_def; auto.
      apply (typ_wf _ _ _ H0).

      apply wf_var with so.
      apply typ_lmark_so.
      apply (typ_wf _ _ _ H0).
    Qed.


    (* 4- Constructors are ok *)


    (* 5- *)



    Transparent sort.
    Lemma rl_full : forall s1 s2, exists s3 : _, rl s1 s2 s3.
    Proof v6_full.
    Opaque sort.

    Lemma all_rules :
      forall e x a b, ⊦ ((e; x : a); x_ : b) -> ⊦ (e; x_ : Forall x : a, b).
    Proof.
      intros.
      Inversion_typ H.
      Inversion_typ (typ_wf _ _ _ H0).
      elim rl_full with s0 s; intros.
      apply wf_var with x0.
      apply Typ_prd with s0 s; trivial.
    Qed.




    (* derived typing rules involving sign_ok *)

    Lemma typ_induc0 :
      forall i inds is e p a l m,
        sign_glob_spec sg i (GlobInd inds is) ->
        e ⊦ p : i_param is ->
          e ⊦ a : (i_arg is)[0:=p] ->
            e ⊦ l : Lmark -> e ⊦ Ind{i; m // p,,(a,,l)} : %%(i_kind is).
    Proof.
      intros.
      apply typ_cst1 with (i_param is ** ((i_arg is)[0:=p] ** Lmark)).
      simpl in |- *; red in |- *.
      apply Mem_mutind with inds; trivial.

      apply inst_triple; trivial.
      apply ind_par_arg_inst with (1 := H); trivial.

      elim ind_kind_ok with (1 := H); intros.
      apply wf_var with x.
      constructor; auto.
      apply (typ_wf _ _ _ H0).
    Qed.


    Lemma typ_induc :
      forall i inds is e p a l m,
        sign_glob_spec sg i (GlobInd inds is) ->
        e ⊦ p : i_param is ->
          e ⊦ a : (i_arg is)[0:=p] ->
            e ⊦ l : Lmark -> ⊦ (e; x_ : Ind{i; m // p,,(a,,l)}).
    Proof.
      intros.
      apply wf_var with (i_kind is).
      apply typ_induc0 with (1 := H); trivial.
    Qed.



    (* pourquoi a-t-on besoin de sign_ok (via typ_induc0) ? *)
    Lemma typ_ind_cons_mark :
      forall e i m m' p a l x s,
        e ⊦ Ind{i; m // p,,(a,,l)} : %%s ->
          e ⊦ x : Mark -> e ⊦ Ind{i; m' // p,,(a,,x && l)} : %%s.
    Proof.
      intros.
      Inversion_cst H; intros.
      apply Typ_conv_srt with (%%(i_kind is)); auto.
      elim H4 using inv_pair_sum; intros.
      elim H7 using inv_pair_sum; intros.
      clear H7 H4.
      apply typ_induc0 with (1 := H1); trivial.
      apply typ_cons_mark; trivial.
    Qed.


    Lemma ty_constr_ok :
      forall e x p a is c,
        sign_glob_spec sg x (GlobCstr is c) ->
        e ⊦ p : i_param is ->
          e ⊦ a : (c_arg c)[0:=p] ->
            ⊦ (e; x_ : Ind{i_name is // p,,(((c_inst c)[1:=p])[0:=a],,[])}).
    Proof.
      intros.
      elim (glob_cst_ind H); intros.
      apply typ_induc with (1 := H2); trivial.
      rewrite <- (simpl_sl ^(1)a (i_arg is)).
      rewrite distr_subst.
      rewrite simpl_sl.
      apply substitution with x_ ((c_arg c)[0:=p]).
      apply substitution1 with x_ (i_param is:term); auto.
      apply thin_par_def2.
      apply ok_constr with (1 := H).

      apply typ_wf with (1 := H0).
      auto.

      apply typ_nil_mark.
      apply typ_wf with (1 := H0).
    Qed.


    Lemma typ_constructor :
      forall x is c e p a,
        sign_glob_spec sg x (GlobCstr is c) ->
        e ⊦ p : i_param is ->
          e ⊦ a : (c_arg c)[0:=p] ->
            e ⊦ Cstr{x // p,,a}
              : Ind{i_name is // p,,(((c_inst c)[1:=p])[0:=a],,[])}.
    Proof.
      intros.
      rewrite (glob_cstr_eq_name H).
      apply typ_cst1 with (i_param is ** (c_arg c)[0:=p]).
      simpl in |- *; red in |- *.
      constructor.
      rewrite <- (glob_cstr_eq_name H); trivial.

      apply inst_pair; trivial.
      apply all_prod_sum.
      Inversion_typ (typ_wf _ _ _ (ok_constr H)).
      unfold add_var in |- *.
      apply thin_par_wf.
      apply (typ_wf _ _ _ H2).

      apply (typ_wf _ _ _ H0).

      apply subst_wf with x_ (i_param is:term); trivial.
      apply thin_par_wf2.
      apply (typ_wf _ _ _ (ok_constr H)).

      apply (typ_wf _ _ _ H0).

      apply ty_constr_ok with (1 := H); trivial.
    Qed.


    (* 2 auxiliary lemmas for subject reduction iota_case *)
    Lemma inv_typ_constr_ind :
      forall (P : Prop) e c i p a p' ia l m,
        (forall inds is cd l',
          sign_glob_spec sg i (GlobInd inds is) ->
          constr_in_ind c is cd ->
          e ⊦ p : i_param is ->
            e ⊦ a : (c_arg cd)[0:=p'] ->
              eq_term e p p' ->
              eq_term e ((c_inst cd)[1:=p'])[0:=a] ia ->
              eq_term e l l' ->
              incl_mark e Ind{i // p',,(ia,,[])} Ind{i; m // p',,(ia,,l')} -> P) ->
        e ⊦ Cstr{c // p,,a} : Ind{i; m // p',,(ia,,l)} -> P.
    Proof.
      intros.
      Inversion_cst H0.
      intros.
      elim H4 using inv_pair_sum; intros.
      clear H4.
      generalize (inv_ind_p H5); intro.
      elim (glob_cst_ind H1); intros.
      elim inv_ind_m with (1 := H5).
      intros l1' eql' (l2', eql2', incl_mrk).
      cut (ind_marks e Ind{i // p',,(ia,,l1')} Ind{i; m // p',,(ia,,l2')}).
      clear incl_mrk; intro incl_mrk.
      2: inversion_clear incl_mrk; trivial.
      2: constructor.
      2: apply Incl_refl.
      inversion_clear incl_mrk.
      elim commut_red_incl with e 0 m l2' l1' []; trivial; intros.
      clear eql' H9.
      apply H with x is c0 x0; auto.
      elim (inv_ind_i H5); trivial.

      apply glob_ind_cst2; trivial.

      apply typ_conv_wf with ((c_arg c0)[0:=p]); auto; simpl in |- *.
      apply wf_type_subst with x_ (i_param is:term).
      Inversion_typ (typ_wf _ _ _ (ok_constr H1)).
      left with s.
      apply thin_par_def; trivial.
      apply typ_wf with (1 := H0).

      generalize (type_correctness _ _ _ H0); intro.
      inversion_clear H9.
      Inversion_cst H12; intros.
      elim H15 using inv_pair_sum; intros.
      generalize H9.
      elim inv_ind_i with (1 := H5); intro.
      elim (inj_ind_is H19 H8).
      trivial.

      unfold le_type; simpl.
      apply eq_term_sym_cci_sub.
      apply eq_term_subst_l; auto.

      apply eq_term_trans with (2 := inv_ind_a H5).
      rewrite comm_subst.
      rewrite comm_subst.
      apply eq_term_subst_l.
      auto.

      apply eq_term_trans with l2'; auto.

      constructor.
      constructor.
      trivial.

      elim CR with e [] l1'; intros; auto.
      inversion_clear H11.
      generalize H13.
      clear H13.
      elim red_normal with (1 := H10).
      intros.
      generalize H12.
      elim red_normal with (1 := H13); trivial.
      red in |- *; red in |- *; intros.
      inversion_clear H11.
      inversion H14.

      red in |- *; red in |- *; intros.
      inversion_clear H11.
      elim hn_subterm_normal with e [] u; trivial.
      simpl in |- *; trivial.
    Qed.



    Lemma match_branch :
      forall ind x pats n tybrs q mrk inds is c,
        sign_glob_spec sg ind (GlobInd inds is) ->
        match_pattern pats x n ->
        case_branch inds is q mrk pats tybrs ->
        constr_in_ind x is c ->
        exists2 arg : _,
        str_substp inds ^(1)mrk (c_arg c) arg & item (branch_ty c arg q) tybrs n.
    Proof.
      intros.
      generalize (for_all2_item_ex (first_item_is_item H0) H1).
      intros (ty_brn, itm_ty, (c_brn, sign_c)).
      elim (constr_in_ind_inj H H2 sign_c).
      intros (arg, str_arg, eq_arg).
      split with arg; trivial.
      elim eq_arg; trivial.
    Qed.


    (* Records *)

    Transparent sort.
    Lemma no_untyped : forall s, in_dom_rel ax s.
    Proof v6_not_topped.
    Opaque sort.

    Lemma all_sort_typed : forall e s, ⊦ e -> ⊦ (e; x_ : %%s).
    Proof.
      intros.
      elim no_untyped with s; intros.
      apply wf_var with x.
      constructor; trivial.
    Qed.


    Lemma typ_record :
      forall e sign,
        ⊦ (e; x_ : record_arity sign) ->
          list_unique (field_names sign) -> ⊦ (e; x_ : record_of sign).
    Proof.
      intros.
      Inversion_typ H.
      apply wf_var with s.
      unfold record_of in |- *.
      apply Typ_cst1 with (%%s); trivial.
      apply all_sort_typed with (1 := inv_wf _ _ H).

      apply all_sort_typed with (1 := inv_wf _ _ H).

      simpl in |- *.
      constructor; auto.
    Qed.

    Lemma typ_projr :
      forall e sign m ty x n,
        item ty (fields_type (expand_sign m sign) 0) n ->
        item x (field_names sign) n ->
        ⊦ (e; x_ : ty) -> e ⊦ m : record_of sign -> e ⊦ m..( x) : ty.
    Proof.
      intros.
      apply Typ_cst1 with (record_of sign); trivial.
      generalize (type_correctness _ _ _ H2); intro.
      inversion_clear H3.
      apply wf_var with s; trivial.

      simpl in |- *.
      red in |- *.
      apply Mem_field with n; auto.
    Qed.

  End ICicMetaOk.

End ICicMeta.

Ltac Inversion_cst T :=
  generalize T; intro INVTYP; InvTypHyp INVTYP; intros t_inv u_inv MEM;
   clear INVTYP; inversion_clear MEM; simpl in |- *;
   try clear t_inv || (generalize t_inv; clear t_inv);
   try clear u_inv || (generalize u_inv; clear u_inv).

Hint Unfold conv': core.
