(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export TypCase.
Require Export TypFix.
Require Export InvSubtyping.

Set Implicit Arguments.
Unset Strict Implicit.

Open Scope pts_scope.
Open Scope cci_scope.

(* Typing: definition of the signature of CCI *)

Ltac change_notation :=
  match goal with
    | |- context C [ (?e ⊦ ?a <= ?b)%pts ] => let r := context C [(e ⊦ a <= b)%cci] in change r
    | |- context C [ (?e ⊦ ?a <= ?b)%cci ] => let r := context C [(e ⊦ a <= b)%pts] in change r
  end.

Tactic Notation "change_notation" "in" hyp(H) :=
  match H with
    | context C [ (?e ⊦ ?a <= ?b)%pts ] => let r := context C [(e ⊦ a <= b)%cci] in change r in H
    | context C [ (?e ⊦ ?a <= ?b)%cci ] => let r := context C [(e ⊦ a <= b)%pts] in change r in H
  end.

(* For Cst0 objects : *)

Inductive mem_sign0 : cci_op -> term -> Prop :=
  | Mem0_mt_mark : mem_sign0 Mt_mark Mark
  | Mem0_mark : mem_sign0 Mark %%so
  | Mem0_nil_mark : mem_sign0 Nil_mark Lmark
  | Mem0_lmark : mem_sign0 Lmark %%so
.

Section ICicPts.

  Context {sg : sigma}.

  Lemma par_ctxt_mkvect :
    forall R e v1 v2,
      for_all2 (par_ctxt R e) v1 v2 -> par_ctxt R e (mkvect v1) (mkvect v2).
  Proof.
    simple induction 1; simpl in |- *; intros; auto.
  Qed.

  Lemma par_ctxt_mktuple :
    forall R e v1 v2,
      for_all2 (par_ctxt R e) v1 v2 -> par_ctxt R e (mktuple v1) (mktuple v2).
  Proof.
    simple induction 1; simpl in |- *; intros; auto.
  Qed.

  (* Besoin de la stabilite de la reduction par sous-typage (marques) *)

  Lemma red_fix_ty : forall e fd u,
    par_ctxt redn_term e (fix_ty fd) u ->
    exists2 fd' : _,
    u = fix_ty fd' &
    (forall b d1,
      par_ctxt redn_term (d1 :: e) (fix_scheme b fd) (fix_scheme b fd')).
  Proof.
    unfold fix_ty at 1 in |- *; intros.
    inversion_clear H.
    elim hn_subterm_normal with (2 := H0); simpl in |- *; trivial.

    inversion_clear H1.
    elim hn_subterm_normal with (2 := H); simpl in |- *; trivial.

    inversion_clear H.
    elim hn_subterm_normal with (2 := H1); simpl in |- *; trivial.

    inversion_clear H1.
    elim hn_subterm_normal with (2 := H); simpl in |- *; trivial.

    inversion_clear H3.
    elim hn_subterm_normal with (2 := H1); simpl in |- *; trivial.

    split with
      (mk_fi (fi_parname fd) m' (fi_argname fd) (fi_indty fd) m'1 m'2
        (fi_lem fd) n'2 n'0); auto.
    unfold fix_scheme in |- *; simpl in |- *; intros.
    apply Pctx_prd.
    apply par_ctxt_liftn with e; auto.
    exact R_lift_redn.

    unfold lift in |- *.
    apply Pctx_prd.
    apply Pctx_cst1.
    apply Pctx_pair.
    apply par_ctxt_lift_rec with (1 := R_lift_redn) (g := e) (d := d1) (2 := H);
      auto.

    apply Pctx_pair.
    apply par_ctxt_lift_rec with (1 := R_lift_redn) (g := e) (d := d1) (2 := H1);
      auto.

    apply Pctx_cst1.
    apply Pctx_pair; trivial.
    apply par_ctxt_lift_rec with (1 := R_lift_redn) (g := e) (d := d1) (2 := H4);
      auto.

    apply
      (par_ctxt_stable _ _ (R_stable_redn (R:=cci_subtype)))
      with
        ((d1 :: e; fi_parname fd : ^(1//0)(fi_party fd));
          fi_argname fd
          : Ind{fi_indty fd; fi_lem fd
            // ^(1//1)(fi_ipar fd),,(^(1//1)(fi_iarg fd),,^(1//1)(fi_marks fd))}).
    apply par_ctxt_lift_rec with (1 := R_lift_redn) (g := e) (d := d1) (2 := H2);
      auto.
    unfold add_var in |- *.
    apply ins_tl'; auto.
    apply ins_tl'; auto.

    unfold add_var in |- *.
    constructor.
    simpl in |- *.
    constructor; simpl in |- *; auto.
    case b.
    apply incl_mark_sub_cci.
    apply incl_rm_mark.

    apply incl_mark_sub_cci.
    apply incl_str_marks.
  Qed.

  Lemma red_fix_ty1 : forall e m fds t,
    par_ctxt redn_term e (fix_types fds) t ->
    exists2 fds' : _,
    t = fix_types fds' &
    (forall b,
      par_ctxt redn_term (e; m : Mark) (mkvect (map (fix_scheme b) fds))
      (mkvect (map (fix_scheme b) fds'))).
  Proof.
    unfold fix_types in |- *.
    simple induction fds; simpl in |- *; intros.
    split with (nil (A:=fix_info)); auto.
    simpl in |- *.
    symmetry  in |- *.
    apply red_normal with (ctxt redn_term) e.
    fold (red redn_term) in |- *.
    apply incl_par_red; trivial.

    apply one_normal.

    inversion_clear H0.
    elim hn_subterm_normal with (2 := H1); simpl in |- *; trivial.

    elim H with (1 := H2); intros.
    elim red_fix_ty with (1 := H1); intros.
    rewrite H4; rewrite H0.
    split with (x0 :: x); auto.
    simpl in |- *.
    intro.
    apply Pctx_sum; auto.
    unfold add_var in |- *; trivial.
  Qed.

  Lemma par_ctxt_fixs : forall e m fds fds',
    (forall b,
      par_ctxt redn_term (e; m : Mark) (mkvect (map (fix_scheme b) fds))
      (mkvect (map (fix_scheme b) fds'))) ->
    par_ctxt redn_term e (fix_bodies m fds) (fix_bodies m fds').
  Proof.
    intros.
    unfold fix_bodies in |- *.
    apply Pctx_prd; auto.
    apply Pctx_prd.
    trivial.

    trivial.
    apply par_ctxt_liftn with (e; m : Mark); auto.
    exact R_lift_redn.
  Qed.

  Lemma par_ctxt_cbr : forall e p q q' l pats tybr inds is,
    par_ctxt redn_term e q q' ->
    case_branch inds is q l pats tybr ->
    exists2 tybr' : _,
    case_branch inds is q' l pats tybr' &
    par_ctxt redn_term e (mkvect tybr)[0:=p] (mkvect tybr')[0:=p].
  Proof.
    simple induction 2; intros.
    split with (nil (A:=term)); auto.
    red in |- *; auto.

    inversion_clear H3.
    inversion_clear H1.
    inversion_clear H6.
    split with (branch_ty x1 x2 q' :: x0); auto.
    constructor; trivial.
    split with x1; trivial.
    split with x2; trivial.

    simpl in |- *.
    rewrite H7.
    apply Pctx_sum; trivial.
    unfold branch_ty in |- *.
    simpl in |- *.
    apply Pctx_prd; auto.
    apply Pctx_cst1; auto.
    apply Pctx_pair; auto.
    apply Pctx_cst1; auto.
    apply Pctx_pair; auto.
    repeat (rewrite simpl_subst; auto).
    apply par_ctxt_liftn with e; auto.
    exact R_lift_redn.
  Qed.

  (* Typing: definition of the signature of CCI *)

  (* For Cst1 objects : *)

  Inductive mem_sign (e : env) :
            term (* should be some (Cst1 op args) *)
         -> term (* type of args *)
         -> term (* type of the whole *)
         -> Prop :=
    | Mem_cons_mark :
        forall p, mem_sign e (Cst1 Cons_mark p) (Mark ** Lmark) Lmark
    | Mem_proj1 :      (* p..L = Cst1 proj1 p *)
        forall p a b, mem_sign e p..L (a**b) a
    | Mem_proj2 :
        forall p a b, mem_sign e p..R (a**b) b
    | Mem_app :        (* f@m = Cst1 app (f,,m) *)
        forall x f m a b, mem_sign e (f@m) ((Forall x:a, b)**a) b[0:=m]
    | Mem_cst :        (* every constant has a parameter (c_par c). *)
        forall x m c,
        sign_glob_spec sg x (GlobConst c) ->
        mem_sign e (Cst1 (Const x) m) (c_par c) (TYPE (c_decl c))[0:=m]
  (* For Records: *)
  (* 1) record types *)
    | Mem_record :     (* {<names:ar>} = Cst1 (Record names) ar *)
        forall (sign:rspec) m s,      (* rspec = list (name*term) *)
        list_unique (field_names sign) ->
        eq_term e m (record_arity sign) ->   (* eq_term : equality with conv *)
        mem_sign e {<field_names sign : m>} %%s %%s
  (* 2) record objects *)
    | Mem_struct :
        forall (md:mspec) s,          (* mspec = list (name*term*term) *)
        mem_sign e (struct_of md) (%%s ** mkvect (fields_type md 0))
          (record_of (rspec_of md))
  (* 3) record projection *)
    | Mem_field :      (* m..(x) = Cst1 (Field x) m *)
        forall n x m ty (sign:rspec),
        item x (field_names sign) n ->
        item ty (fields_type (expand_sign m sign) 0) n ->
        mem_sign e m..(x) (record_of sign) ty
  (* End Records *)
    | Mem_fix :        (* Fix {types,,bodies} = Cst1 Fixp (types,,bodies) *)
        forall m (fixdefs:list fix_info) bds s,
        mem_sign e (Fix {(fix_types fixdefs,,bds)})
          (%%s ** fix_bodies m fixdefs) (fix_types fixdefs)
    | Mem_case :       (* Match c=>pr with pats=>br end =
                          Cst1 (MutCase pats) (c,,(pr,,br)) *)
        forall inds is (pats:list name) p c br m t b,
        sign_glob_spec sg (i_name is) (GlobInd inds is) ->
        type_case inds is pats p c b m t ->
        mem_sign e (Match c => p with b — pats => br end) m t
    | Mem_mutind :     (* Ind{I;m// p,,(a,,l)}=Cst1 (MutInd m) (p,,(a,,l))*)
        forall ind inds is p a l m,
        sign_glob_spec sg ind (GlobInd inds is) ->
        mem_sign e Ind{ind; m // p,,(a,,l)}
          (i_param is ** ((i_arg is)[0:=p] ** Lmark))
          %%(i_kind is)
    | Mem_constr :     (* Cstr{C//p,,a} = Cst1 (MutConstr C) (p,,a) *)
        forall x p a is c,
        sign_glob_spec sg x (GlobCstr is c) ->
        mem_sign e Cstr{x // p,,a} (i_param is ** (c_arg c)[0:=p])
          Ind{i_name is // p,,(((c_inst c)[1:=p])[0:=a],,[])}
  .

  Definition mem_sign1 : sign_relation :=
    fun c e m t u => mem_sign e (Cst1 c m) t u.

  Lemma mem_sign1_stable : forall R e f m t u,
    rel_in_list (decl_rel (fun _ t1 t2 => t1 = t2) R) f e ->
    mem_sign e m t u -> mem_sign f m t u.
  Proof.
    intros.
    inversion_clear H0;
      try (constructor; trivial) || prolog [ Mem_field Mem_case Mem_mutind ] 5.
    apply (R_stable_eq_special (R:=R)) with (1 := H2); trivial.
  Qed.

  Lemma in_sign_lift_1 : forall c e f g d k m t u,
    ins_in_env g d k e f ->
    mem_sign1 c e m t u -> mem_sign1 c f ^(1//k)m ^(1//k)t ^(1//k)u.
  Proof.
    intros.
    inversion_clear H0; try (simpl in |- *; constructor; trivial; fail).
    4: unfold struct_of in |- *.
    simpl in |- *.
    rewrite distr_lift_subst.
    constructor.

    rewrite closed0_lift.
    rewrite distr_lift_subst.
    rewrite (closed_lift_rec 1 (TYPE (c_decl c0))).
    constructor; trivial.

    apply closed_weak with 1; auto with arith.

    simpl in |- *.
    rewrite <- record_names_lift with (n := 1) (k := k).
    constructor.
    rewrite record_names_lift; trivial.

    rewrite <- record_arity_lift.
    apply R_lift_eq with (2 := H); trivial.

    simpl in |- *.
    fold (fields_of md) in |- *.
    rewrite record_arity_lift.
    rewrite prod_arity_lift.
    rewrite mkvect_lift.
    fold (0 + k) in |- *.
    rewrite fields_type_lift; simpl in |- *.
    rewrite vals_of_lift with (k' := k).
    rewrite rspec_of_lift with (k' := k).
    rewrite fields_of_lift with (n := 1) (k := k) (k' := k).
    fold (struct_of (lift_msign 1 md k k)) in |- *.
    unfold fields_of in |- *.
    fold (record_of (rspec_of (lift_msign 1 md k k))) in |- *.
    constructor.

    rewrite record_of_lift.
    apply Mem_field with n.
    rewrite record_names_lift; trivial.

    rewrite <- expand_sign_lift.
    pattern k at 2 in |- *.
    replace k with (0 + k); auto.
    rewrite <- fields_type_lift.
    apply item_map' with u; trivial.

    unfold lift_rec in |- *; fold lift_rec in |- *.
    rewrite fix_types_lift.
    rewrite fix_bodies_lift.
    constructor.

    simpl in |- *.
    apply Mem_case with inds is; auto.
    apply type_case_lift; trivial.

    simpl in |- *.
    rewrite closed0_lift.
    rewrite distr_lift_subst.
    rewrite closed1_lift.
    apply Mem_mutind with inds; auto.

    simpl in |- *.
    rewrite closed0_lift.
    rewrite distr_lift_subst.
    rewrite closed1_lift.
    rewrite distr_lift_subst.
    rewrite distr_lift_subst1.
    rewrite closed2_lift.
    constructor; trivial.
  Qed.

  Lemma in_sign_subst_1 : forall c e f g d k s m t u,
    sub_in_env g s d k e f ->
    mem_sign1 c e m t u -> mem_sign1 c f m[k:=s] t[k:=s] u[k:=s].
  Proof.
    intros.
    inversion_clear H0; try (simpl in |- *; constructor; trivial; fail).
    4: unfold struct_of in |- *.
    simpl in |- *.
    rewrite distr_subst.
    constructor.

    rewrite closed0_subst.
    rewrite distr_subst.
    rewrite (closed_subst_rec s (TYPE (c_decl c0))).
    constructor; trivial.

    apply closed_weak with 1; auto with arith.

    simpl in |- *.
    rewrite <- record_names_subst with (s := s) (k := k).
    constructor.
    rewrite record_names_subst; trivial.

    rewrite <- record_arity_subst.
    apply R_subst_eq with (2 := H); trivial.

    simpl in |- *.
    fold (fields_of md) in |- *.
    rewrite record_arity_subst.
    rewrite prod_arity_subst.
    rewrite mkvect_subst.
    fold (0 + k) in |- *.
    rewrite fields_type_subst; simpl in |- *.
    rewrite vals_of_subst with (k' := k).
    rewrite rspec_of_subst with (k' := k).
    rewrite fields_of_subst with (s := s) (k := k) (k' := k).
    fold (struct_of (subst_msign s md k k)) in |- *.
    unfold fields_of in |- *.
    fold (record_of (rspec_of (subst_msign s md k k))) in |- *.
    constructor.

    rewrite record_of_subst.
    apply Mem_field with n.
    rewrite record_names_subst; trivial.

    rewrite <- expand_sign_subst.
    pattern k at 2 in |- *.
    replace k with (0 + k); auto.
    rewrite <- fields_type_subst.
    apply item_map' with u; trivial.

    unfold subst_rec in |- *; fold subst_rec in |- *.
    rewrite fix_types_subst.
    rewrite fix_bodies_subst.
    constructor.

    simpl in |- *.
    apply Mem_case with inds is; auto.
    apply type_case_subst; trivial.

    simpl in |- *.
    rewrite closed0_subst.
    rewrite distr_subst.
    rewrite closed1_subst.
    apply Mem_mutind with inds; auto.

    simpl in |- *.
    rewrite closed0_subst.
    rewrite distr_subst.
    rewrite closed1_subst.
    rewrite distr_subst.
    rewrite distr_subst1.
    rewrite closed2_subst.
    constructor; trivial.
  Qed.

  Lemma in_sign_stable_1 : forall c e f m t u,
    rel_in_list (rel_typ cci_subtype) e f -> mem_sign1 c f m t u -> mem_sign1 c e m t u.
  Proof.
    intros.
    inversion_clear H0;
      try (unfold struct_of in |- *; fold (struct_of md) in |- *);
        try (constructor; trivial) || (hnf; prolog [ Mem_field Mem_case Mem_mutind ] 5).
    apply (R_stable_eq (R:=cci_subtype)) with (2 := H); trivial.
  Qed.

  Instance Signature1 : InvariantSignature mem_sign1 cci_subtype := {
    in_sign_lift := in_sign_lift_1;
    in_sign_subst := in_sign_subst_1;
    in_sign_stable := in_sign_stable_1
  }.

  Lemma find_in_rspec : forall m rs t k n,
    item t (fields_type (expand_sign m rs) k) n -> exists p : _, item p rs n.
  Proof.
    simple induction rs; simpl in |- *; intros.
    inversion H.

    destruct a.
    inversion_clear H0.
    split with (n0, t0); auto.

    elim map_inv_item with (1 := H1); intros.
    elim H with (1 := H0); intros.
    split with x0; auto.
  Qed.

  Fixpoint struct_psubst (m : term) (flds : list name) (n : nat) {struct n} : list term :=
    match n, flds with
      | S k, x :: xl => struct_psubst m xl k ++ m..( x) :: nil
      | _, _ => nil
    end.

  Lemma map_struct_psubst : forall f m k flds,
    (forall m n, f m..( n) = (f m) ..( n)) ->
    map f (struct_psubst m flds k) = struct_psubst (f m) flds k.
  Proof.
    simple induction k; simpl in |- *; trivial.
    simple destruct flds; intros.
    trivial.

    rewrite map_app; simpl in |- *.
    rewrite H; auto.
    rewrite H0; trivial.
  Qed.

  Lemma par_ctxt_struct_psubst : forall R, R_lift R ->
    forall e m m' k flds, par_ctxt R e m m' ->
      for_all2 (par_ctxt R e) (struct_psubst m flds k) (struct_psubst m' flds k).
  Proof.
    simple induction k; simpl in |- *; intros; trivial.
    case flds; intros; trivial.
    apply for_all2_app; simpl in |- *; auto.
  Qed.

  Lemma record_psubst : forall rs m x ty t n k,
    item (x, ty) rs n ->
    item t (fields_type (expand_sign m rs) k) n ->
    t = psubst_rec (struct_psubst ^(k)m (field_names rs) n) ty 0.
  Proof.
    simple induction rs; simpl in |- *; intros.
    inversion H0.

    destruct a; simpl in |- *.
    destruct n.
    inversion H0.
    subst ty x l0.
    inversion H1.
    subst t l0; simpl in |- *.
    rewrite psubst_id.
    reflexivity.

    simpl in |- *.
    rewrite psubst_app; rewrite psubst_cons; rewrite psubst_id.
    simpl in |- *.
    inversion_clear H0.
    inversion_clear H1.
    elim map_inv_item with (1 := H0); intros.
    rewrite map_struct_psubst; simpl in |- *; auto.
    rewrite <- simpl_lift.
    rewrite <- (H _ _ _ _ _ _ H2 H1).
    trivial.
  Qed.

  Lemma par_ctxt_fields_type : forall e m m' sign,
    par_ctxt redn_term e m m' ->
    for_all2 (par_ctxt redn_term e) (fields_type (expand_sign m sign) 0)
    (fields_type (expand_sign m' sign) 0).
  Proof.
    intros.
    apply for_all2_ext; intros.
    elim find_in_rspec with (1 := H0); intros.
    destruct x0.
    rewrite (record_psubst H2 H0).
    rewrite (record_psubst H2 H1).
    apply par_ctxt_psubst with e; auto.
    exact R_lift_redn.

    apply par_ctxt_struct_psubst.
    exact R_lift_redn.

    repeat rewrite lift0.
    trivial.

    generalize 0.
    induction sign as [| a sign Hrecsign]; simpl in |- *; intros; auto.
    destruct a.
    simpl in |- *.
    repeat rewrite map_length; auto with arith.
  Qed.

  Axiom par_ctxt_struct :
    forall md e a b,
      par_ctxt redn_term e (prod_arity (rspec_of md)) a ->
      par_ctxt redn_term e (vals_of md) b ->
      exists2 md' : _,
      a = prod_arity (rspec_of md') /\ b = vals_of md' &
      par_ctxt redn_term e (mkvect (fields_type md 0))
      (mkvect (fields_type md' 0)) /\
      par_ctxt redn_term e (record_arity (rspec_of md))
      (record_arity (rspec_of md')) /\ fields_of md = fields_of md'.

  Lemma mem_sign_stable_val : forall c e m m' t u,
    par_ctxt redn_term e m m' ->
    mem_sign e (Cst1 c m) t u ->
    exists2 t' : _,
    par_ctxt redn_term e t t' &
    (exists2 u' : _,
      par_ctxt redn_term e u u' & mem_sign e (Cst1 c m') t' u').
  Proof.
    intros.
    generalize H.
    clear H.
    inversion_clear H0; unfold struct_of in |- *; intros.
    split with (Mark ** Lmark); trivial.
    split with (Lmark:term); trivial.
    constructor.

    split with (u ** b); auto.
    split with u; auto.
    constructor.

    split with (a ** u); auto.
    split with u; auto.
    constructor.

    (* App *)
    split with ((Forall x : a, b) ** a); auto.
    inversion_clear H.
    elim hn_subterm_normal with (2 := H0); simpl in |- *; trivial.

    split with (b[0:=n']).
    apply par_ctxt_subst_l with e; auto.
    exact R_lift_redn.

    constructor.

    (* Const *)
    split with (c_par c0:term); auto.
    split with ((TYPE (c_decl c0))[0:=m']).
    apply par_ctxt_subst_l with e; auto.
    exact R_lift_redn.

    constructor; trivial.

    (* Record *)
    split with (%%s); auto.
    split with (%%s); auto.
    constructor; trivial.
    apply eq_term_trans with m; trivial.
    apply eq_term_sym.
    apply redn_conv.
    apply incl_par_red; trivial.

    (* Struct *)
    inversion_clear H.
    elim hn_subterm_normal with (2 := H0); simpl in |- *; trivial.

    elim par_ctxt_struct with (1 := H0) (2 := H1); intros.
    inversion_clear H2.
    inversion_clear H4.
    split with (%%s ** mkvect (fields_type x 0)); auto.
    split with (record_of (rspec_of x)); auto.
    unfold record_of in |- *; fold (fields_of x) in |- *.
    elim H5; auto.

    inversion_clear H.
    rewrite H4; rewrite H6.
    rewrite H5.
    fold (struct_of x) in |- *.
    constructor.

    (* Field *)
    split with (record_of sign); auto.
    elim
      for_all2_item_ex
      with
        (1 := H1)
        (R := par_ctxt redn_term e)
        (l2 := fields_type (expand_sign m' sign) 0).
    intros.
    split with x0; trivial.
    apply Mem_field with n; trivial.

    apply par_ctxt_fields_type; trivial.

    (* Fix *)
    inversion_clear H.
    elim hn_subterm_normal with (2 := H0); simpl in |- *; trivial.

    elim red_fix_ty1 with (1 := H0) (m := m0); intros.
    rewrite H.
    split with (%%s ** fix_bodies m0 x).
    apply Pctx_sum; trivial.
    apply par_ctxt_fixs; trivial.

    split with m'0; trivial.
    rewrite H.
    constructor.

    inversion_clear H0.
    elim hn_subterm_normal with (2 := H2); simpl in |- *; trivial.

    inversion_clear H3.
    elim hn_subterm_normal with (2 := H0); simpl in |- *; trivial.

    (* Case *)
    inversion_clear H1.
    elim par_ctxt_cbr with (1 := H0) (2 := H6) (p := par); intros.
    split
      with
        (Ind{i_name is; m0 // par,,(arg,,l)} **
          ((case_predicate_form is se)[0:=par] ** (mkvect x)[0:=par])).
    auto.

    split with ((m'1 @ arg) @ m'0); auto.
    apply Mem_case with inds is; auto.
    constructor; auto.

    inversion_clear H0.
    elim hn_subterm_normal with (2 := H1); simpl in |- *; trivial.

    (* Ind *)
    inversion_clear H2.
    elim hn_subterm_normal with (2 := H0); simpl in |- *; trivial.

    split with (i_param is ** ((i_arg is)[0:=m'0] ** Lmark)); auto.
    apply Pctx_sum; auto.
    apply Pctx_sum; auto.
    apply par_ctxt_subst_l with e; auto.
    exact R_lift_redn.

    split with (%%(i_kind is)); auto.
    apply Mem_mutind with inds; trivial.

    (* Cstr *)
    inversion_clear H0.
    elim hn_subterm_normal with (2 := H1); simpl in |- *; trivial.

    split with (i_param is ** (c_arg c0)[0:=m'0]).
    apply Pctx_sum; auto.
    apply par_ctxt_subst_l with e; auto.
    exact R_lift_redn.

    split with Ind{i_name is // m'0,,(((c_inst c0)[1:=m'0])[0:=n'],,[])}.
    apply Pctx_cst1.
    apply Pctx_pair; auto.
    apply Pctx_pair; auto.
    apply par_ctxt_subst_l2 with e; auto.
    exact R_lift_redn.

    constructor; trivial.
  Qed.

  (* Defining the PTS *)

  Instance cci_pts : PTSO_spec := {
    axiom := ax;
    rule := rl;
    prd := pr;
    subrule := cci_subtyping_rule;
    in_sign0 := mem_sign0;
    in_sign1 := mem_sign1
  }.

  Lemma str_subst_eq_term :
    forall inds l m br, str_substp inds l m br ->
      forall e l', eq_term e l l' ->
        exists2 br' : _, str_substp inds l' m br' & eq_term e br br'.
  Proof.
    fix str_sub_rec 5 with
    (mark_rec (inds:_) (l:_) (s1:_) (s2:_) (y:mark_record inds l s1 s2)
      {struct y} :
      forall e l',
        eq_term e l l' ->
        exists2 s2' : _,
        mark_record inds l' s1 s2' &
        field_names s2' = field_names s2 /\
        eq_term e (record_arity s2) (record_arity s2')).
    simple destruct 1; intros.
    split with t; auto.
    constructor.

    exists Ind{x // p,,(a,,l')}.
    constructor; trivial.

    red in |- *; auto 6.

    elim str_sub_rec with inds (^(1)l0) b b' (e; x : a) (^(1)l'); trivial; intros.
    exists (Forall x : a, x0).
    constructor; trivial.

    red in |- *; auto.

    apply (Rlift0 _ R_lift_eq); trivial.

    elim str_sub_rec with inds l0 a a' e l'; trivial; intros.
    elim str_sub_rec with inds l0 b b' e l'; trivial; intros.
    exists (x ** x0).
    constructor; trivial.

    red in |- *; auto.

    elim mark_rec with inds l0 s1 s2 e l'; trivial; intros.
    exists (record_of x).
    constructor; trivial.

    unfold record_of in |- *.
    destruct H3.
    rewrite H3.
    red in |- *; auto.

    simple destruct 1; intros.
    exists (nil (A:=name * term)); auto.
    constructor.

    simpl in |- *.
    elim str_sub_rec with inds l0 ty1 ty2 e l'; trivial; intros.
    elim mark_rec with inds (^(1)l0) s0 s3 (e; x : ty2) (^(1)l'); trivial; intros.
    exists ((x, x0) :: x1).
    constructor; trivial.

    destruct H5.
    split; simpl in |- *.
    rewrite H5; trivial.

    red in |- *; auto.

    apply (Rlift0 _ R_lift_eq); trivial.
  Qed.

  Lemma str_subst_cci_sub :
    forall inds l u br, str_substp inds ^(1)l u br ->
      forall e i i' x p p' a a' l' t m m',
        e ⊦ Ind{i'; m' // p',,(a',,l')} <= Ind{i; m // p,,(a,,l)} ->
        exists2 br' : _, str_substp inds ^(1)l' u br' & (e; x : t) ⊦ br' <= br.
  Proof.
    intros.
    elim (inv_ind_m H0); intros.
    inversion_clear H2.
    elim str_subst_eq_term with (1 := H) (2 := Rlift0 _ R_lift_eq _ _ _ H3 x t).
    intros.
    elim
      str_subst_incl_seq
      with (1 := H2) (k := m) (k' := m') (e := (e; x : t)) (l' := ^(1)x0);
        intros.
    elim
      str_subst_eq_term
      with (1 := H6) (2 := Rlift0 _ R_lift_eq _ _ _ (eq_term_sym H1) x t).
    intros.
    split with x4; trivial.
    apply cci_trans with x3; auto.
    apply cci_trans with x2; auto.

    unfold lift in |- *.
    apply incl_seq_lift; trivial.
    inversion_clear H4.
    inversion_clear H6; trivial.

    apply Incl_refl.
  Qed.

  Lemma eq_case_branch0 :
    forall inds is pats q l' br, case_branch inds is q l' pats br ->
      forall e x t l p p' a a' m m',
        e ⊦ Ind{i_name is; m // p,,(a,,l)} <=
        Ind{i_name is; m' // p',,(a',,l')} ->
        exists2 lt : _,
        case_branch inds is q l pats lt & (e; x : t) ⊦ mkvect br <= mkvect lt.
  Proof.
    simple induction 1; intros.
    split with (nil (A:=term)); red in |- *.
    simple apply Fa2_nil.
    change (clos_refl_trans term (cci_le (e; x : t))
      (mkvect nil) (mkvect nil)); simple apply rt_refl.

    inversion_clear H0.
    inversion_clear H5.
    elim str_subst_cci_sub with (1 := H0) (2 := H3) (x := x0) (t := t); intros.
    elim H2 with (1 := H3) (x := x0) (t := t); intros.
    split with (branch_ty x1 x3 q :: x4).
    constructor; trivial.
    split with x1; trivial.
    split with x3; trivial.

    rewrite H6.
    simpl in |- *.
    apply cci_sub_sum; trivial.
    unfold branch_ty in |- *.
    auto.
  Qed.

  Lemma alpha_stab_typ :
    forall e f t T, alpha_stab f e -> ⊦ f -> e ⊦ t : T -> f ⊦ t : T.
  Proof.
    intros.
    apply reduction_env_rec with
      (decl_rel (fun _ t1 t2 => t1 = t2) (fun _ t1 t2 => t1 = t2)) e;
      trivial.
    unfold le_type.
    simpl in |- *.
    intros.
    inversion_clear H2.
    elim sub_ty0; auto.

    simpl in |- *; unfold mem_sign1 in |- *; intros.
    apply mem_sign1_stable with (1 := H2); trivial.

    simpl in |- *; intros.
    change le_type with cci_subtype.
    apply subtype_stable_special.
  Qed.

  (* --> Metatheory (!!! alpha_typ) *)
  Lemma alpha_stab_env : forall e f,
      rel_in_list (rel_typ (fun _ t1 t2 => t1 = t2)) f e -> ⊦ e -> ⊦ f.
  Proof.
    intros.
    generalize f H.
    clear H f.
    elim H0; intros.
    inversion_clear H.

    unfold add_var in H3.
    inversion_clear H3.
    generalize x0 H4.
    clear H4 x0.
    intros (x0, ty, bd).
    unfold rel_typ in |- *; simpl in |- *.
    intros (eqty, eqbd).
    rewrite eqty; rewrite eqbd.
    fold (e0; x0 : T) in |- *.
    apply wf_var with s; auto.

    fold (l1; x : T) in |- *.
    apply wf_var with s.
    apply alpha_stab_typ with e0; auto.

    unfold add_def in H4.
    inversion_clear H4.
    generalize x0 H5.
    clear H5 x0.
    intros (x0, ty, bd).
    unfold rel_typ in |- *; simpl in |- *.
    intros (eqty, eqbd).
    rewrite eqty; rewrite eqbd.
    fold (e0; x0 := M : T) in |- *.
    apply wf_def with s; auto.

    fold (l1; x := M : T) in |- *.
    apply wf_def with s.
    apply alpha_stab_typ with e0; auto.

    apply alpha_stab_typ with e0; auto.
  Qed.

  Lemma alpha_stab_typ' : forall e f t T,
    rel_in_list (rel_typ (fun _ t1 t2 => t1 = t2)) f e ->
    e ⊦ t : T -> f ⊦ t : T.
  Proof.
    intros.
    apply alpha_stab_typ with e; auto.
    apply alpha_stab_env with e; trivial.
    apply typ_wf with (1 := H0).
  Qed.

  Lemma alpha_stab_alpha_var : forall e x y t,
    rel_in_list (rel_typ (fun _ t1 t2 => t1 = t2)) (e; x : t) (e; y : t).
  Proof.
    unfold add_var in |- *; constructor.
    unfold rel_typ in |- *; simpl in |- *; auto.
  Qed.

  Lemma alpha_stab_alpha_def : forall e x y m t,
    rel_in_list (rel_typ (fun _ t1 t2 => t1 = t2)) (e; x := m : t)
    (e; y := m : t).
  Proof.
    unfold add_def in |- *; constructor.
    unfold rel_typ in |- *; simpl in |- *; auto.
  Qed.

  Hint Resolve alpha_stab_alpha_var alpha_stab_alpha_def: core.

  (* Definir les proprietes que doit verifier une signature bien formee:
      noms differents, types corrects, etc... *)

  Lemma typ_mark_so : forall e, ⊦ e -> e ⊦ Mark : %%so.
  Proof.
    intros.
    constructor; auto.
    elim so_typed; intros.
    apply wf_var with x; auto.
    constructor; auto.
  Qed.

  Lemma typ_lmark_so : forall e, ⊦ e -> e ⊦ Lmark : %%so.
  Proof.
    intros.
    constructor; auto.
    elim so_typed; intros.
    apply wf_var with x; auto.
    constructor; auto.
  Qed.

  Lemma typ_mt_mark : forall e, ⊦ e -> e ⊦ ## : Mark.
  Proof.
    intros.
    constructor; auto.
    apply wf_var with so.
    apply typ_mark_so; trivial.
    constructor.
  Qed.

  Lemma typ_nil_mark : forall e, ⊦ e -> e ⊦ [] : Lmark.
  Proof.
    intros.
    constructor; auto.
    apply wf_var with so.
    apply typ_lmark_so; trivial.
    constructor.
  Qed.

  Lemma typ_unit_one : forall e, ⊦ e -> e ⊦ unit : one.
  Proof typ_nil_mark.

  Lemma typ_one_so : forall e, ⊦ e -> e ⊦ one : %%so.
  Proof typ_lmark_so.


  Lemma so_wf : ⊦ (( ); x_ : %%so).
  Proof.
    elim so_typed; intros.
    apply wf_var with x.
    constructor; auto.
  Qed.

  Lemma mrk_wf : ⊦ (( ); x_ : Mark).
  Proof.
    apply wf_var with so.
    apply typ_mark_so.
    constructor.
  Qed.

  Lemma lmrk_wf : ⊦ (( ); x_ : Lmark).
  Proof.
    apply wf_var with so.
    apply typ_lmark_so.
    constructor.
  Qed.

End ICicPts.

Hint Resolve @alpha_stab_alpha_var @alpha_stab_alpha_def: core.
Existing Instance cci_pts.
