(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export SubjRed.
Require Export SubtypingDec.

Set Implicit Arguments.
Unset Strict Implicit.

Ltac InvSignSat H ty :=
  elim H; clear H; intros ty ISH (MEMSIGN, WFUU); simpl in ISH;
    generalize WFUU; clear WFUU; elim ISH; clear ISH;
      inversion_clear MEMSIGN.

Ltac Right_Neg :=
  intros NEG_HYP; right; red in |- *; intros; apply NEG_HYP; clear NEG_HYP.

  Section ICoreDecOk.

    Context {sg : sigma}.
    Context {sgn_ok : sign_ok}.

    Lemma whnf_cci :
      forall (e : env) (t : term),
        wf_type e t ->
        {u : term | red redn_term e t u &  head_normal redn_term e u}.
    Proof fun e t wft => whnf_cci_sn ok_gs_inj (ok_normalize wft).

    Lemma red_to_sort_dec :
      forall (e : env) (t : term),
        wf_type e t ->
        {s : sort | red redn_term e t %%s} + {(forall s : sort, ~ e ⊦ t <= %%s)}.
    Proof fun e t wft => red_to_sort_dec_sn ok_gs_inj (ok_normalize wft).

    Lemma red_to_prd_dec :
      forall (e : env) (t : term),
        wf_type e t ->
        {p : name * (term * term) |
          let (x, y) := p in let (a, b) := y in red redn_term e t (Forall x : a, b)} +
        {(forall (x : name) (a b : term), ~ e ⊦ t <= Forall x : a, b)}.
    Proof fun e t wft => red_to_prd_dec_sn ok_gs_inj (ok_normalize wft).

    Lemma red_to_sum_dec :
      forall (e : env) (t : term),
        wf_type e t ->
        {p : term * term | let (a, b) := p in red redn_term e t (a ** b)} +
        {(forall a b : term, ~ e ⊦ t <= a ** b)}.
    Proof fun e t wft => red_to_sum_dec_sn ok_gs_inj (ok_normalize wft).


    Lemma red_to_ind_dec :
      forall e t,
        wf_type e t ->
        {ii : _ |
          match ii with
            | (i, (p, (a, (m, l)))) => red redn_term e t Ind{i; m // p,,(a,,l)}
          end} + {(forall i x m, ~ e ⊦ t <= Ind{i; m // x})}.
    Proof.
      intros.
      elim whnf_cci with e t; trivial.
      intros hnft.
      generalize (is_ind_dec hnft).
      intros [(i, m, p, a, l)| not_ind] redt hnfht.
      left.
      split with (i, (p, (a, (m, l)))); trivial.

      RightStr err_not_ind; red in |- *; intros.
      cut (sub_hn_inv e hnft Ind{i; m // x}); intros.
      generalize not_ind; clear not_ind.
      generalize (subj_red_wft H redt).
      inversion_clear H1.
      intro.
      inversion_clear H1.
      Inversion_cst H3; auto.

      intro.
      inversion_clear H1.
      Inversion_cst H5; auto.

      apply inv_cumul_hn with t Ind{i; m // x}; auto.
    Qed.

    Lemma dest_cst_pair :
      forall e m t,
        ⊦ (e; x_ := m : t) ->
          {q : _ |
            match q with
              | ((v1, t1), (v2, t2)) =>
                m = (v1,,v2) /\
                red redn_term e t (t1 ** t2) /\
                  ⊦ (e; x_ := v1 : t1) /\ ⊦ (e; x_ := v2 : t2)
            end} +
          {~
            (exists v1 : _,
              (exists2 v2 : _,
                m = (v1,,v2) &
                (exists t1 : _, (exists t2 : _, e ⊦ t <= t1 ** t2))))}.
    Proof.
      intros.
      elim is_pair_dec with m.
      intros ((v1, v2), eqm).
      elim red_to_sum_dec with e t.
      intros ((t1, t2), redt).
      left.
      split with (v1, t1, (v2, t2)).
      split; auto.
      split; auto.
      Inversion_typ H.
      rewrite eqm in H0.
      elim (subj_red_ty H0 redt) using inv_pair_sum; intros.
      Inversion_typ (subject_reduction redt H1).
      split.
      apply wf_def with s1; trivial.

      apply wf_def with s2; trivial.

      right; red in |- *; intros (vv1, (vv2, eqmm, (t1, (t2, eqt)))).
      elim b with t1 t2; auto.

      apply var_wft.
      apply forget_def with (1 := H).

      right; red in |- *; intros (vv1, (vv2, eqmm, (t1, (t2, eqt)))).
      elim b with vv1 vv2; auto.
    Qed.





    Lemma eq_term_sort_dec :
      forall e t, wf_type e t -> cast_to_sort_dec e t.
    Proof.
      red in |- *; intros.
      elim red_to_sort_dec with e t; intros; trivial.
      left.
      inversion_clear a.
      split with x.
      simpl in |- *.
      fold le_sort in |- *.
      split; intros; auto.
      change_notation; simple apply eq_term_cci_sub; simple apply redn_conv; exact H0.
      change ((fun s3 s4 : sort => forall e0 : env, (e0 ⊦ %%s3 <= %%s4)%cci) x y).
      refine (@sub_sort_env_indep _ e _ _ _).
      (*apply sub_sort_env_indep with e.*)
      apply cci_trans with t; auto.

      right; red in |- *; intros.
      apply (b x); trivial.
    Qed.




    (* cf infer *)
    Transparent sort.
    Lemma infer_axiom :
      forall s, {sp : _ | ppal (ax s) le_sort sp} + {~ typed_sort s}.
    Proof.
      left.
      elim (v6_inf_axiom s); intros.
      split with x.
      split; intros.
      exact (pp_ok p).

      cut (univ x y).
      red in |- *; red in |- *; red in |- *; intros.
      hnf.
      elim H0; intros; auto.
      apply rt_step.
      right.
      constructor 1.
      left; auto.

      apply rt_trans with (%%y0); trivial.

      apply (pp_least p); trivial.
    Qed.


    Lemma infer_rule :
      forall s1 s2, ppal_dec (fun s3 => complete rl s1 s2 s3) le_sort.
    Proof.
      left.
      elim (v6_inf_rule s1 s2); intros.
      split with x.
      split; intros.
      split with s1; auto.
      hnf; intro.
      change_notation; apply cci_refl.

      split with s2; simpl in |- *; auto.
      intro; change_notation; apply cci_refl.

      cut (univ x y).
      red in |- *; red in |- *; red in |- *; intros.
      hnf; elim H0; intros; auto.
      apply rt_step.
      right.
      constructor 1.
      left; auto.

      apply rt_trans with (%%y0); trivial.

      inversion_clear H.
      inversion_clear H1.
      simpl in H0, H.
      apply q with (1 := H2); auto.
      red in |- *.
      apply inv_sort with ( ); simple apply H0.

      apply inv_sort with ( ); simple apply H.
    Qed.


    Lemma infer_prod :
      forall s1 s2,
        {s3 : _ | ppal (fun s => complete pr s1 s2 s) le_sort s3}.
    Proof.
      intros.
      elim (v6_inf_pair s1 s2); intros.
      split with x.
      split; intros.
      split with s1; auto.
      simpl in |- *; auto.
      intro; change_notation; apply cci_refl.

      split with s2; simpl in |- *; auto.
      intro; change_notation; apply cci_refl.

      cut (univ x y).
      red in |- *; red in |- *; red in |- *; intros.
      hnf; elim H0; intros; auto.
      apply rt_step.
      right.
      constructor 1.
      left; auto.

      apply rt_trans with (%%y0); trivial.

      inversion_clear H.
      inversion_clear H1.
      simpl in H0, H.
      apply q with (1 := H2); auto.
      red in |- *.
      apply inv_sort with ( ); simple apply H0.

      apply inv_sort with ( ); simple apply H.
    Qed.
    Opaque sort.

    Lemma wk_rl :
      forall e s1 b,
        (exists s : _, e ⊦ b : %%s) ->
        ppal_dec (weak_to_least_sort (complete rl) e s1 b)
        (cci_subtype e).
    Proof.
      left.
      split with b.
      split; intros.
      red in |- *; simpl in |- *.
      split; auto.
      change_notation; apply cci_refl.
      inversion_clear H.
      split with x; auto.
      elim rl_full with s1 x; intros.
      split with x0; auto.
      red in |- *.
      split with s1; simpl in |- *; auto.
      intro; change_notation; apply cci_refl.
      split with x; simpl in |- *; auto.
      intro; change_notation; apply cci_refl.

      inversion_clear H0; trivial.
    Qed.



    Lemma cci_subtype_dec :
      forall e u v,
        wf_type e u -> wf_type e v -> decide (e ⊦ u <= v).
    Proof.
      intros.
      apply subtype_dec.
      exact ok_gs_inj.
      apply ok_normalize; trivial.
      apply ok_normalize; trivial.
    Qed.



    Lemma cci_topsort_untyped :
      forall e s s' t, e ⊦ %%s <= t -> e ⊦ t : %%s' -> typed_sort s.
    Proof.
      intros.
      elim whnf_cci with e t; intros.
      cut (sub_hn_inv e %%s x); intros.
      generalize p.
      inversion_clear H1; intros.
      elim univ_hierarchy with (1 := H2); intros.
      split with x0; trivial.

      Inversion_typ (subject_reduction p0 H0).
      split with s2; trivial.

      apply inv_cumul_hn with (%%s) t; auto.

      left with s'; trivial.
    Qed.



    (* Typing constants is decidable *)

    Lemma sign_dec_const :
      forall c e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat (Const c) e m t u) (cci_subtype e).
    Proof.
      intros.
      elim search_const with c; intros.
      inversion_clear a.
      generalize (ok_cst H0); intro.
      elim subtype_dec with e t (c_par x:term); intros.
      left.
      split with ((TYPE (c_decl x))[0:=m]).
      split; intros.
      split with (c_par x:term).
      split; auto.
      unfold add_var in |- *; apply thin_par_wf; trivial.
      apply (inv_wf _ _ H1).

      apply (inv_wf _ _ H).

      split.
      constructor; trivial.

      cut (e ⊦ m : c_par x).
      generalize H1.
      case x; simpl in |- *; intros p [z [ty clty] [[bd clbd]| ]]; simpl in |- *;
        intros.
      change (⊦ ((( ); x_ : p); z := bd : ty)) in H2.
      simpl in clty, clbd.
      Inversion_typ H2.
      apply wf_var with s.
      fold (%%s[0:=m]) in |- *.
      apply substitution with x_ (p:term); trivial.
      apply thin_par_def; auto.
      apply (inv_wf _ _ H).

      change (⊦ ((( ); x_ : p); z : ty)) in H2.
      Inversion_typ H2.
      apply wf_var with s.
      fold (%%s[0:=m]) in |- *.
      apply substitution with x_ (p:term); trivial.
      apply thin_par_def; auto.
      apply (inv_wf _ _ H).

      apply typ_conv_wf with t; auto.
      Inversion_typ H; trivial.

      apply var_wft.
      unfold add_var in |- *.
      apply thin_par_wf.
      apply (inv_wf _ _ H1).

      apply (inv_wf _ _ H).

      inversion_clear H2.
      inversion_clear H4.
      inversion_clear H2.
      generalize (ok_gs_inj H0 H4); intro.
      injection H2; intro.
      elim H6; auto.

      right; red in |- *; intros.
      inversion_clear H2.
      inversion_clear H4.
      inversion_clear H3.
      generalize H4.
      inversion_clear H2.
      simpl in |- *.
      generalize (ok_gs_inj H0 H3); intro.
      injection H2; intro.
      elim H7; trivial.

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      Inversion_typ H.
      left with s; trivial.

      apply ok_normalize.
      Inversion_typ (inv_wf _ _ (ok_cst H0)).
      left with s.
      apply thin_par_def0; auto.
      apply (inv_wf _ _ H).

      right; red in |- *; intros.
      inversion_clear H0.
      inversion_clear H2.
      inversion_clear H0.
      elim b with c0; trivial.
    Qed.



    (* Typing projections is decidable *)
    (* Il existe une version utilisant whnf_cci au lieu de whnf_ty
       dans la revision 1.7 de SignDec. *)

    Lemma whnf_ty :
      forall e t,
        ⊦ (e; x_ : t) ->
          {u : _ | e ⊦ t <= u /\ e ⊦ u <= t &
            ⊦ (e; x_ : u) /\ head_normal redn_term e u}.
    Proof.
      intros e t wf_hyp.
      case whnf_cci with e t.
      apply var_wft; trivial.

      intros t' red_t hn_t'; exists t'.
      split.
      simple apply eq_term_cci_sub; simple apply redn_conv; exact red_t.
      simple apply eq_term_sym_cci_sub; simple apply redn_conv; exact red_t.
      split; trivial.
      Inversion_typ wf_hyp.
      apply wf_var with s.
      apply subject_reduction with (2 := H); trivial.
    Defined.

    Lemma sign_dec_pi1 :
      forall e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat Proj1 e m t u) (cci_subtype e).
    Proof.
      intros.
      elim whnf_ty with e t.
      intros hnft.
      generalize (is_sum_dec hnft); intros [(u, v)| not_sum] (leth, leht) (wfh, hh).
      left.
      split with u.
      split; intros.
      split with (u ** v); auto.
      split.
      constructor.

      Inversion_typ wfh.
      Inversion_typ H0.
      apply wf_var with s1; trivial.

      inversion_clear H0.
      inversion_clear H2.
      cut (e ⊦ u ** v <= x).
      inversion_clear H0.
      intro.
      apply (inv_sum_l H0).

      apply cci_trans with (1 := leht).
      inversion_clear H1; trivial.

      right; red in |- *; intros.
      inversion_clear H0.
      inversion_clear H1.
      generalize H0.
      inversion_clear H2.
      inversion_clear H1; intros.
      apply not_sum.
      apply conv_sum_eq with e hnft x b; auto.
      apply cci_trans with t; trivial.

      Inversion_typ H.
      apply wf_var with s; trivial.
    Qed.


    Lemma sign_dec_pi2 :
      forall e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat Proj2 e m t u) (cci_subtype e).
    Proof.
      intros.
      elim whnf_ty with e t.
      intros hnft.
      generalize (is_sum_dec hnft); intros [(u, v)| not_sum] (leth, leht) (wfh, hh).
      left.
      split with v.
      split; intros.
      split with (u ** v); auto.
      split.
      constructor.

      Inversion_typ wfh.
      Inversion_typ H0.
      apply wf_var with s2; trivial.

      inversion_clear H0.
      inversion_clear H2.
      cut (e ⊦ u ** v <= x).
      inversion_clear H0.
      intro.
      apply (inv_sum_r H0).

      apply cci_trans with (1 := leht).
      inversion_clear H1; trivial.

      right; red in |- *; intros.
      inversion_clear H0.
      inversion_clear H1.
      generalize H0.
      inversion_clear H2.
      inversion_clear H1; intros.
      apply not_sum.
      apply conv_sum_eq with e hnft a x; auto.
      apply cci_trans with t; trivial.

      Inversion_typ H.
      apply wf_var with s; trivial.
    Qed.


    (* Typing application is decidable *)

    Lemma sign_dec_app :
      forall e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat App e m t u) (cci_subtype e).
    Proof.
      intros.
      generalize (is_pair_dec m).
      intros [((f, a), eqp)| not_pair].
      rewrite eqp.
      elim whnf_cci with e t.
      intros ht.
      generalize (is_sum_dec ht).
      intros [(u, v)| not_sum] redt hnft.
      elim whnf_cci with e u.
      intros hu.
      generalize (is_prd_dec hu).
      intros [(x, dom, rng)| not_sum] redu hnfu.
      elim subtype_dec with e v dom; intros.
      left.
      split with (rng[0:=a]).
      split; intros.
      split with ((Forall x : dom, rng) ** dom).
      split.
      change_notation; apply cci_trans with (u ** v); auto.

      cut (⊦ (e; x_ : (Forall x : dom, rng) ** v)); intros.
      Inversion_typ H0.
      Inversion_typ H1.
      Inversion_typ H2.
      apply all_prod_sum.
      apply wf_var with s1; trivial.

      apply wf_var with s0; trivial.

      apply subj_red_wf with t.
      Inversion_typ H.
      apply wf_var with s; trivial.

      apply red_trans with (u ** v); auto.

      split.
      constructor.

      cut (⊦ (e; x_ : (Forall x : dom, rng) ** v)); intros.
      Inversion_typ H0.
      Inversion_typ H1.
      Inversion_typ H2.
      apply wf_var with s4.
      fold (%%s4[0:=a]) in |- *.
      apply substitution with x dom; auto.
      cut (e ⊦ (f,,a) : u ** v).
      intro.
      elim H10 using inv_pair_sum; intros.
      apply typ_conv_wf with v; auto.
      left with s0; trivial.

      apply subj_red_ty with t; auto.
      elim eqp.
      Inversion_typ H; trivial.

      apply subj_red_wf with t.
      Inversion_typ H.
      apply wf_var with s; trivial.

      apply red_trans with (1 := redt).
      auto.

      inversion_clear H0.
      inversion_clear H1.
      inversion_clear H2.
      generalize H0.
      inversion_clear H1; intros.
      apply cci_subst_var with x1 a1.
      apply inv_prod_r with x dom.
      apply cci_trans with u; auto.
      apply inv_sum_l with v a1.
      apply cci_trans with t; auto.

      right; red in |- *; intros.
      apply b.
      inversion_clear H0.
      inversion_clear H1.
      inversion_clear H2.
      generalize H0.
      inversion_clear H1; intros.
      cut (e ⊦ (Forall x : dom, rng) ** v <= (Forall x2 : a0, b0) ** a0); intros.
      apply cci_trans with a0.
      apply (inv_sum_r H2).

      apply (inv_prod_l (inv_sum_l H2)).

      apply cci_trans with (2 := H1).
      apply cci_trans with (u ** v); auto.

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      Inversion_typ H.
      generalize (subject_reduction redt H1); intro.
      Inversion_typ H2.
      left with s2; trivial.

      Inversion_typ H.
      generalize (subject_reduction redt H1); intro.
      Inversion_typ H2.
      generalize (subject_reduction redu H3); intro.
      Inversion_typ H7.
      apply ok_normalize.
      left with s0; trivial.

      right; red in |- *; intros.
      inversion_clear H0.
      inversion_clear H1.
      inversion_clear H2.
      generalize H0.
      inversion_clear H1; intros.
      apply not_sum.
      apply conv_prd_eq with e x1 u a0 b; trivial.
      apply inv_sum_l with v a0.
      apply cci_trans with (2 := H1).
      auto.

      Inversion_typ H.
      generalize (subject_reduction redt H1); intro.
      Inversion_typ H2.
      left with s1; trivial.

      right; red in |- *; intros.
      inversion_clear H0.
      inversion_clear H1.
      inversion_clear H2.
      generalize H0.
      inversion_clear H1; intros.
      apply not_sum.
      apply conv_sum_eq with (3 := H1); trivial.

      Inversion_typ H.
      left with s; trivial.

      right; red in |- *; intros.
      generalize not_pair.
      inversion_clear H0.
      inversion_clear H2.
      inversion_clear H0; intros.
      elim not_pair0 with f m0; trivial.
    Qed.


    Lemma sign_dec_cons :
      forall e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat Cons_mark e m t u) (cci_subtype e).
    Proof.
      intros.
      cut (⊦ (e; x_ : Mark ** Lmark)); intros.
      elim subtype_dec with e t (Mark ** Lmark); intros.
      left.
      split with (Lmark:term).
      split; intros.
      split with (Mark ** Lmark); auto.
      split.
      constructor.

      Inversion_typ H0.
      Inversion_typ H1.
      apply wf_var with s2; trivial.

      inversion_clear H1.
      inversion_clear H3.
      inversion_clear H1; trivial.

      right; red in |- *; intros.
      inversion_clear H1.
      inversion_clear H2.
      inversion_clear H3.
      generalize H1.
      inversion_clear H2; trivial.

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      Inversion_typ H.
      left with s; trivial.

      apply ok_normalize.
      apply var_wft.
      trivial.

      apply all_prod_sum.
      apply wf_var with so.
      apply typ_mark_so.
      apply (inv_wf _ _ H).

      apply wf_var with so.
      apply typ_lmark_so.
      apply (inv_wf _ _ H).
    Qed.


    Lemma sign0_dec :
      forall c,
        {u : _ | mem_sign0 c u &  ⊦ (( ); x_ : u)} +
        {(forall u, ⊦ (( ); x_ : u) -> ~ mem_sign0 c u)}.
    Proof.
      intro.
      (*Realizer
         Cases c of
         Mt_mark => (inleft ? (Cst0 Mark))
         | Mark => (inleft ? (Srt so))
         | Nil_mark => (inleft ? (Cst0 Lmark))
         | Lmark => (inleft ? (Srt so))
         | _ => (inright term)
         end.
         *)
      case c; try (right; red in |- *; intros; inversion_clear H0; fail).
      left.
      split with (%%so); try constructor.
      exact so_wf.

      left.
      split with (Mark:term); try constructor.
      exact mrk_wf.

      left.
      split with (%%so); try constructor.
      exact so_wf.

      left.
      split with (Lmark:term); try constructor.
      exact lmrk_wf.
    Qed.

    Lemma mem_sign0_unique :
      forall c u u', mem_sign0 c u -> mem_sign0 c u' -> u = u'.
    Proof.
      simple destruct 1; intros; inversion_clear H0; trivial.
    Qed.


    Lemma inf_sign0_dec :
      forall c,
        ppal_dec (fun u => mem_sign0 c u /\ ⊦ (( ); x_ : u)) (eq (A:=_)).
    Proof.
    (*Realizer sign0_dec.*)
      intro c.
      elim sign0_dec with c; intros.
      left.
      inversion_clear a.
      split with x; auto.
      split; auto.
      intros.
      inversion_clear H1.
      apply (mem_sign0_unique H H2).

      right; red in |- *; intros.
      inversion_clear H; auto.
      red in b.
      eauto.
    Qed.

  End ICoreDecOk.
