(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export LambdaRules.

Set Implicit Arguments.
Unset Strict Implicit.

Section IEqual.

  Context {sg : sigma}.


  Lemma alpha_sym : forall e x y, red alpha e x y -> red alpha e y x.
Proof.
simple induction 1; intros; auto.
red in |- *; red in |- *; apply rt_step.
elim H0; intros; auto.
constructor.
inversion_clear H1; constructor.

apply red_trans with y0; trivial.
Qed.

  Lemma delta_norm :
   forall x m a b e n,
   item_rel e n (mkdecl x a (Some m)) -> red delta e b ^(1//n)(b[n:=m]).
Proof.
simple induction b; simpl in |- *; intros; auto.
elim (lt_eq_lt_dec n0 n); intros; auto.
elim a0.
case n; intros.
inversion a1.

rewrite lift_ref_ge; simpl in |- *; auto with arith.

simple induction 1.
unfold lift in |- *.
rewrite simpl_lift_rec; simpl in |- *; auto with arith.
fold (^(S n0)m) in |- *.
red in |- *; red in |- *; apply rt_step.
apply Ctx_rule.
split with x a; trivial.

rewrite lift_ref_lt; auto.

apply red_red_prd; auto.
unfold add_var in |- *; auto.

apply red_red_lam; auto.
unfold add_var in |- *; auto.
Qed.

  Lemma delta_norm0 :
   forall e x m a b, red delta (e; x := m : a) b ^(1)(b[0:=m]).
Proof.
intros.
unfold add_def, lift in |- *.
apply delta_norm with x a; auto.
Qed.


  (* Reunion of all the reduction rules *)

  Definition red_term :=
    reunion_list
      (Cs_rl delta
         (Cs_rl (delta_glob sg)
            (Cs_rl alpha
               (Cs_rl proj1
                  (Cs_rl proj2
                     (Cs_rl projr
                        (Cs_rl beta (Cs_rl iota_fix (Cs_rl iota_case Mt_rl))))))))).

  Hint Unfold red_term: core.


  Lemma R_lift_red : R_lift red_term.
Proof.
unfold red_term in |- *.
apply reunion_list_lift.
auto 10.
Qed.

  Lemma R_subst_weak_red : R_subst_weak red_term.
Proof.
change
  (R_subst_weak
     (reunion delta
        (reunion_list
           (Cs_rl (delta_glob sg)
              (Cs_rl alpha
                 (Cs_rl proj1
                    (Cs_rl proj2
                       (Cs_rl projr
                          (Cs_rl beta
                             (Cs_rl iota_fix (Cs_rl iota_case Mt_rl)))))))))))
 in |- *.
apply reunion_subst_weak; auto.
apply weak_subst; auto 10.
Qed.

  Lemma R_stable_red : forall R, R_stable (rel_typ R) red_term.
Proof.
intros.
unfold red_term in |- *.
apply reunion_list_stable.
auto 10.
Qed.

  Lemma R_stable_red_special :
   forall R, R_stable R delta -> R_stable R red_term.
Proof.
intros.
unfold red_term in |- *.
apply reunion_list_stable.
auto 10.
Qed.



  (* Equality of terms in CIC *)

  Definition eq_term := conv red_term.

  Lemma eq_term_refl : forall e x, eq_term e x x.
Proof fun e : env => rst_refl _ (ctxt red_term e).

  Lemma eq_term_sym : forall e x y, eq_term e x y -> eq_term e y x.
Proof fun e : env => rst_sym _ (ctxt red_term e).

  Lemma eq_term_trans :
   forall e x y z, eq_term e x y -> eq_term e y z -> eq_term e x z.
Proof conv_trans red_term.

  Lemma eq_term_step : forall e a b, red_term e a b -> eq_term e a b.
Proof.
intros.
red in |- *; red in |- *; red in |- *; apply rst_step; apply Ctx_rule;
 trivial.
Qed.

  Lemma eq_term_exp : forall e a b, red_term e b a -> eq_term e a b.
Proof.
intros.
red in |- *; red in |- *; red in |- *; apply rst_sym; apply rst_step;
 apply Ctx_rule; trivial.
Qed.

  Hint Resolve eq_term_refl eq_term_step eq_term_exp: core.
  Hint Immediate eq_term_sym: core.

  (* Delta reduction is included in term equality *)

  Lemma delta_eq_term : forall e x y, red delta e x y -> eq_term e x y.
Proof.
red in |- *.
simple induction 1; intros; auto.
red in |- *; red in |- *; apply rst_step.
elim H0; intros; auto.
apply Ctx_rule.
red in |- *; simpl in |- *; auto.

apply conv_trans with y0; trivial.
Qed.

  Hint Resolve delta_eq_term: core.


  Lemma eq_term_app :
   forall e a a' b b',
   eq_term e a a' -> eq_term e b b' -> eq_term e (a @ b) (a' @ b').
Proof.
unfold eq_term in |- *; intros; auto.
Qed.

  Lemma eq_term_proj :
   forall e n a b, eq_term e a b -> eq_term e (a .^ n) (b .^ n).
Proof.
simple induction n; simpl in |- *; intros; auto.
apply H.
red in |- *.
apply conv_conv_cst; auto.
Qed.


  Hint Resolve eq_term_app eq_term_proj: core.


  (* Consistency with De Bruijn indexes *)

  Lemma R_lift_eq : R_lift eq_term.
Proof conv_lift _ R_lift_red.

  Lemma R_subst_eq : R_subst eq_term.
Proof conv_subst _ R_subst_weak_red.

  Lemma R_stable_eq : forall R, R_stable (rel_typ R) eq_term.
Proof fun R => stable_rst _ _ (ctxt_stable _ _ (R_stable_red (R:=R))).

  Lemma R_stable_eq_special :
   forall R, R_stable (decl_rel (fun _ t1 t2 => t1 = t2) R) eq_term. 
Proof
  fun R =>
  stable_rst _ _ (ctxt_stable _ _ (R_stable_red_special (stb_decl_rel (R:=R)))).


  (* eq_term is a subtyping rule *)
  Definition eq_subtyping_rule :=
    Build_Subtyping_rule _ R_lift_eq R_subst_eq (R_stable_eq (R:=eq_term))
      (fun e => Build_preorder _ _ (eq_term_refl e) (eq_term_trans (e:=e))).



  Lemma eq_term_subst_l :
   forall e a a' b, eq_term e a a' -> eq_term e b[0:=a] b[0:=a'].
Proof.
red in |- *; intros.
apply conv_subst_l with e; auto.
exact R_lift_red.
Qed.

  Lemma eq_term_subst :
   forall e x m t a b,
   eq_term (e; x := m : t) a b -> eq_term e a[0:=m] b[0:=m].
intros.
apply R_subst_eq with (mkdecl x t (Some m)) e (e; x := m : t); auto.
unfold add_def in |- *; constructor.
compute in |- *; reflexivity.
Qed.


  (* Reduction without alpha (for normalization) *)

  Definition redn_term :=
    reunion_list
      (Cs_rl delta
         (Cs_rl (delta_glob sg)
            (Cs_rl proj1
               (Cs_rl proj2
                  (Cs_rl projr
                     (Cs_rl beta (Cs_rl iota_fix (Cs_rl iota_case Mt_rl)))))))).
  Hint Unfold redn_term: core.

  (* Consistency with De Bruijn indexes *)
  Lemma R_lift_redn : R_lift redn_term.
Proof.
unfold redn_term in |- *.
apply reunion_list_lift.
auto 10.
Qed.

  Lemma R_subst_weak_redn : R_subst_weak redn_term.
Proof.
change
  (R_subst_weak
     (reunion delta
        (reunion_list
           (Cs_rl (delta_glob sg)
              (Cs_rl proj1
                 (Cs_rl proj2
                    (Cs_rl projr
                       (Cs_rl beta (Cs_rl iota_fix (Cs_rl iota_case Mt_rl))))))))))
 in |- *.
apply reunion_subst_weak; auto.
apply weak_subst; auto 10.
Qed.

  Lemma R_stable_redn : forall R, R_stable (rel_typ R) redn_term.
Proof.
intros.
unfold redn_term in |- *.
apply reunion_list_stable.
auto 10.
Qed.


  Lemma R_stable_redn_special :
   forall R, R_stable (decl_rel (fun _ t1 t2 => t1 = t2) R) redn_term.
Proof.
intros.
unfold redn_term in |- *.
apply reunion_list_stable.
constructor.
apply stb_decl_rel.

auto 10.
Qed.



  Lemma redn_red : forall e a b, ctxt redn_term e a b -> eq_term e a b.
Proof.
red in |- *; simple induction 1; intros; auto.
apply eq_term_step.
red in |- *; simpl in |- *.
inversion_clear H0; auto.
inversion_clear H1; auto 10.
Qed.

  Hint Resolve redn_red: core.

  Lemma redn_conv : forall e a b, red redn_term e a b -> eq_term e a b.
Proof.
simple induction 1; intros; auto.
apply eq_term_trans with y; auto.
Qed.

  Hint Resolve redn_conv: core.


  Lemma alpha_red_term : forall e a b, red alpha e a b -> red red_term e a b.
Proof.
simple induction 1; intros; auto.
red in |- *; red in |- *.
apply rt_step.
elim H0; intros; auto.
constructor.
right.
right; left; trivial.

apply red_trans with y; trivial.
Qed.

  Lemma redn_term_red_term :
   forall e a b, red redn_term e a b -> red red_term e a b.
Proof.
simple induction 1; intros; auto.
red in |- *; red in |- *.
apply rt_step.
elim H0; intros; auto.
constructor.
red in |- *.
simpl in |- *.
inversion_clear H1; auto.
inversion_clear H2; auto.
auto 10.

apply red_trans with y; trivial.
Qed.


  Lemma ctxt_red_term_or :
   forall e x y,
   ctxt red_term e x y -> ctxt redn_term e x y \/ ctxt alpha e x y.
Proof.
simple induction 1; intros.
inversion_clear H0; auto 20.
inversion_clear H1; auto 20.
inversion_clear H0; auto 20.

inversion_clear H1; auto.

inversion_clear H1; auto.

inversion_clear H1; auto.

inversion_clear H1; auto.

inversion_clear H1; auto.

inversion_clear H1; auto.

inversion_clear H1; auto.

inversion_clear H1; auto.

inversion_clear H1; auto.
Qed.

End IEqual.

Hint Unfold red_term: core.
Hint Resolve @eq_term_refl @eq_term_step @eq_term_exp: core.
Hint Immediate @eq_term_sym: core.
Hint Resolve @delta_eq_term: core.
Hint Resolve @eq_term_app @eq_term_proj: core.
Hint Unfold redn_term: core.
Hint Resolve @redn_red: core.
Hint Resolve @redn_conv: core.

