(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export ReductionDec.

Set Implicit Arguments.
Unset Strict Implicit.

Section IEqualDecA.

  Context {sg : sigma}.

  Section ConversionAlgorithm.

    Hypothesis sign_ok_glob_inj :
      forall x g1 g2,
        sign_glob_spec sg x g1 -> sign_glob_spec sg x g2 -> g1 = g2.

    Section FixBody.

      Variables (e : env) (x y : term).

      Hypothesis conv_rec :
        forall e' x' y',
          subterm eq_term (e', x') (e, x) ->
          subterm eq_term (e', y') (e, y) -> decide (eq_term e' x' y').

      Lemma convert_hn : decide (eq_hn_inv e x y).
      Proof.
        red in |- *.
        generalize conv_rec; clear conv_rec.
        case x.
        case y; intros; try (right; red in |- *; intros; inversion_clear H; fail).
        elim eq_dec_sort with s0 s; intros.
        left.
        elim a; auto.
        apply inv_eq_hn.
        apply eq_term_refl.
        apply hn_subterm_hn; unfold hn_subterm; exact I.
        apply hn_subterm_hn; unfold hn_subterm; exact I.

        right; red in |- *; intros.
        apply b.
        inversion_clear H; auto.

        case y; intros; try (right; red in |- *; intros; inversion_clear H; fail).
        elim eq_nat_dec with n0 n; intros.
        left.
        elim a; constructor.

        right; red in |- *; intros.
        apply b.
        inversion_clear H; auto.

        intros c.
        case y; intros; try (right; red in |- *; intros; inversion_clear H; fail).
        elim eq_dec_oper with c o; intros.
        left.
        elim a; auto.
        apply inv_eq_hn.
        apply eq_term_refl.
        apply hn_subterm_hn; unfold hn_subterm; exact I.
        apply hn_subterm_hn; unfold hn_subterm; exact I.

        right; red in |- *; intros.
        apply b.
        inversion_clear H; auto.

        intros c M.
        case y; intros; try (right; red in |- *; intros; inversion_clear H; fail).
        elim eq_dec_oper with c o; intros.
        elim a.
        elim conv_rec with e M t; intros; auto.
        left; constructor; trivial.

        right; red in |- *; intros.
        inversion_clear H; auto.

        right; red in |- *; intros.
        apply b.
        inversion_clear H; auto.

        intros n A M.
        case y; intros; try (right; red in |- *; intros; inversion_clear H; fail).
        elim conv_rec with e A t; intros; auto.
        elim conv_rec with (e; n : A) M t0; intros; auto.
        left; constructor; trivial.

        right; red in |- *; intros.
        inversion_clear H; auto.

        unfold add_var in |- *.
        constructor 2 with (mkdecl n A None); auto.
        split; auto.

        unfold add_var in |- *.
        constructor 2 with (mkdecl n0 t None); auto.
        split; auto.

        right; red in |- *; intros.
        inversion_clear H; auto.

        intros n A M.
        case y; intros; try (right; red in |- *; intros; inversion_clear H; fail).
        elim conv_rec with e A t; intros; auto.
        elim conv_rec with (e; n : A) M t0; intros; auto.
        left; constructor; trivial.

        right; red in |- *; intros.
        inversion_clear H; auto.

        unfold add_var in |- *.
        constructor 2 with (mkdecl n A None); auto.
        split; auto.

        unfold add_var in |- *.
        constructor 2 with (mkdecl n0 t None); auto.
        split; auto.

        right; red in |- *; intros.
        inversion_clear H; auto.

        intros M N.
        case y; intros; try (right; red in |- *; intros; inversion_clear H; fail).
        elim conv_rec with e M t; intros; auto.
        elim conv_rec with e N t0; intros; auto.
        left; constructor; trivial.

        right; red in |- *; intros.
        inversion_clear H; auto.

        right; red in |- *; intros.
        inversion_clear H; auto.

        intros M N.
        case y; intros; try (right; red in |- *; intros; inversion_clear H; fail).
        elim conv_rec with e M t; intros; auto.
        elim conv_rec with e N t0; intros; auto.
        left; constructor; trivial.

        right; red in |- *; intros.
        inversion_clear H; auto.

        right; red in |- *; intros.
        inversion_clear H; auto.
      Qed.

    End FixBody.

    Theorem convert : forall e x y,
      sn (ctxt redn_term) e x ->
      sn (ctxt redn_term) e y -> decide (eq_term e x y).
    Proof.
      intros.
      generalize y H0.
      clear H0 y.
      apply whnf_induc with (R' := eq_term) (5 := H).
      apply R_stable_redn.

      exact (whnf_cci_sn sign_ok_glob_inj).

      intros.
      elim H1 with y0; intros; trivial.
      left.
      apply eq_term_trans with y; auto.

      right; red in |- *; intros.
      apply b.
      apply eq_term_trans with x0; auto.
      apply eq_term_sym.
      auto.

      intros.
      elim (whnf_cci_sn sign_ok_glob_inj) with e0 y; intros; trivial.
      elim convert_hn with e0 x0 x1; intros.
      left.
      apply eq_term_trans with x1; auto.
      apply inv_eq_conv; exact a.
      apply eq_term_sym; auto.

      right; red in |- *; intros; apply b.
      apply inv_eq_hn_red with x0 y; auto.

      apply H1; trivial.
      apply subterm_sn with eq_term e0 x1; trivial.
      apply R_stable_redn.

      apply sn_red_sn with y; auto.
    Qed.

  End ConversionAlgorithm.

End IEqualDecA.
