(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export CicPts.

Delimit Scope sig_scope with sig.
Close Scope cci_scope.
Close Scope pts_scope.
Open Scope sig_scope.

(* We use Unicode notations here to avoid conflicts *)
Notation "s – e ⊢ t == u" := (@eq_term s e t u) (at level 50, t at next level) : sig_scope.
Notation "s – e ⊢ m : t" := (@typ (@cci_pts s) e m t)
  (at level 50, m at next level) : sig_scope.
Notation "s – ⊢  e" := (@wf (@cci_pts s) e) (at level 50) : sig_scope.
Notation "s – e ⊢ t <= u" := (@cci_subtype s e t u) (at level 50, t at next level) : sig_scope.

Set Implicit Arguments.
Unset Strict Implicit.


Inductive sig_item_ord : sig_item -> sig_item -> Prop :=
| Si_refl : forall si, sig_item_ord si si
| Si_abstr :
  forall ip,
    sig_item_ord (MutInductive (abstract_ind ip)) (MutInductive ip).

Hint Resolve Si_refl Si_abstr.

Definition incl_sig s1 s2 :=
  forall i, In i s1 -> exists2 i' : _, In i' s2 & sig_item_ord i i'.

Hint Unfold incl_sig: core.

Lemma incl_cons : forall (sg : sigma) si, incl_sig sg (si :: sg).
Proof.
  red in |- *; intros.
  exists i; auto.
Qed.

Lemma incl_ind_abstr :
  forall sg ip,
    incl_sig (MutInductive (abstract_ind ip) :: sg) (MutInductive ip :: sg).
Proof.
  red in |- *; intros sg ip si Hin.
  destruct Hin as [Hin|Hin].
  subst si.
  exists (MutInductive ip); auto.

  exists si; auto.
Qed.


  Section IInclSignOk.

  Variable s1 s2 : sigma.
  Variable incl_s1_s2 : incl_sig s1 s2.

  Lemma incl_gs_cst :
    forall x c,
      sign_glob_spec s1 x (GlobConst c) -> sign_glob_spec s2 x (GlobConst c).
  Proof.
    simple destruct 1; intros.
    split with x0; trivial.
    elim incl_s1_s2 with (1 := H0).
    inversion_clear H1; intros; trivial.
    generalize H1.
    inversion_clear H2; trivial.
  Qed.


  Lemma incl_gs_ind :
    forall x inds is,
      sign_glob_spec s1 x (GlobInd inds is) ->
      sign_glob_spec s2 x (GlobInd inds is) \/
      (exists2 is' : _, is = abs_is is' & sign_glob_spec s2 x (GlobInd inds is')).
  Proof.
    simple destruct 1; intros.
    elim incl_s1_s2 with (1 := H0); intros.
    generalize H1 H2.
    inversion_clear H3; intros.
    left.
    split with x1; trivial.

    right.
    inversion_clear H3.
    unfold abstract_ind in H5.
    elim map_inv with (1 := H5); intros.
    split with x2; trivial.
    replace (inds_of_pack (abstract_ind ip)) with (inds_of_pack ip).
    split with (MutInductive ip); trivial.
    rewrite H6.
    simpl in |- *.
    constructor; trivial.

    unfold inds_of_pack in |- *.
    elim ip; simpl in |- *; auto.
    intros.
    rewrite H7; trivial.
  Qed.


  Lemma incl_gs_cstr :
    forall x is c,
      sign_glob_spec s1 x (GlobCstr is c) -> sign_glob_spec s2 x (GlobCstr is c).
  Proof.
    simple destruct 1; intros.
    elim incl_s1_s2 with (1 := H0); intros.
    generalize H1 H2.
    inversion_clear H3; intros.
    split with x1; trivial.

    inversion_clear H3.
    elim abstract_no_cstr with (1 := H5) (2 := H6).
  Qed.




  Lemma eq_incl : forall e a b, s1 – e ⊢ a == b -> s2 – e ⊢ a == b.
  Proof.
    intros.
    red in |- *; red in |- *; red in |- *.
    elim H; intros; auto.
    apply rst_step.
    elim H0; intros; auto.
    apply Ctx_rule.
    unfold red_term in |- *; simpl in |- *.
    inversion_clear H1; auto.
    right.
    inversion_clear H2; auto.
    left.
    inversion_clear H1.
    split with c; auto.
    apply incl_gs_cst; trivial.

    apply rst_sym; trivial.

    apply rst_trans with y; trivial.
  Qed.


  Lemma sub_incl : forall e a b, s1 – e ⊢ a <= b -> s2 – e ⊢ a <= b.
  Proof.
    unfold cci_subtype, R_rt in |- *.
    simple induction 1; intros; auto.
    apply rt_step.
    inversion_clear H0; auto.
    left.
    apply eq_incl; trivial.

    apply rt_trans with y; trivial.
  Qed.



  Lemma incl_mem_sign :
    forall e m t u, @mem_sign s1 e m t u -> @mem_sign s2 e m t u.
  Proof.
    intros e m t u mems1.
    inversion_clear mems1; try (constructor; auto).
    apply incl_gs_cst; trivial.

    apply eq_incl; trivial.

    apply Mem_field with n; trivial.

    apply Mem_case with inds is; trivial.
    elim incl_gs_ind with (1 := H); intros; trivial.
    inversion_clear H1.
    rewrite H2 in H0.
    inversion_clear H0.
    inversion_clear H1.
    discriminate H0.

    elim incl_gs_ind with (1 := H); intros.
    apply Mem_mutind with inds; trivial.

    inversion_clear H0.
    rewrite H1; simpl in |- *.
    apply Mem_mutind with inds; trivial.

    apply incl_gs_cstr; trivial.
  Qed.


  (* Weakening of signature *)
  Lemma typ_incl : forall e m t, s1 – e ⊢ m : t -> s2 – e ⊢ m : t.
  Proof.
    intros e m t j.
    elim j using typ_mind with (P0 := fun e => s2 – ⊢ e); intros.
    constructor; auto.

    constructor; auto.

    constructor; auto.

    apply Typ_cst1 with t0; simpl in |- *; auto.
    apply incl_mem_sign with (1 := H5); trivial.

    apply Typ_lam with s; trivial.

    apply Typ_prd with s0 s3; trivial.

    apply Typ_sum with s0 s3; trivial.

    apply Typ_pair with s; trivial.

    apply Typ_conv with U s; trivial.
    simpl in |- *; apply sub_incl; trivial.

    apply Typ_conv_srt with U; trivial.
    simpl in |- *; apply sub_incl; trivial.

    constructor.

    apply wf_var with s; trivial.

    apply wf_def with s; trivial.
  Qed.

  Lemma wf_incl : forall e, s1 – ⊢ e -> s2 – ⊢ e.
  Proof.
    simple induction 1; intros.
    constructor.

    apply wf_var with s.
    apply typ_incl; trivial.

    apply wf_def with s.
    apply typ_incl; trivial.

    apply typ_incl; trivial.
  Qed.


  Lemma wft_incl :
    forall e t, @wf_type (@cci_pts s1) e t -> @wf_type (@cci_pts s2) e t.
  Proof.
    intros.
    inversion_clear H; auto.
    left with s.
    apply typ_incl; auto.
  Qed.

End IInclSignOk.
