(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Import CoreDec.

Set Implicit Arguments.
Unset Strict Implicit.

  Section IIndDecOk.

    Context {sg : sigma}.
    Context {sgn_ok : sign_ok}.

    (* Typing inductive types is decidable *)
    Lemma sign_dec_indt :
      forall ind l e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat (MutInd ind l) e m t u)
          (cci_subtype e).
    Proof.
      intros.
      elim is_pair_dec with m.
      2: right; red in |- *; intros.
      2: generalize b.
      2: inversion_clear H0.
      2: inversion_clear H2.
      2: inversion_clear H0; intros.
      2: apply (b0 p (a,,l0)); trivial.
      intros ((p, am), eqp).
      rewrite eqp.
      rewrite eqp in H.
      clear eqp.
      elim is_pair_dec with am.
      2: right; red in |- *; intros.
      2: generalize b.
      2: inversion_clear H0.
      2: inversion_clear H2.
      2: inversion_clear H0; intros.
      2: apply (b0 a l0); trivial.
      intros ((a, mrk), eqp).
      rewrite eqp.
      rewrite eqp in H.
      clear eqp.
      elim red_to_sum_dec with e t.
      2: right; red in |- *; intros.
      intros ((typar, typam), redt).
      elim red_to_sum_dec with e typam.
      2: right; red in |- *; intros.
      intros ((tyarg, tymrk), redtypam).
      cut (red redn_term e t (typar ** (tyarg ** tymrk))); intros.
      clear redtypam redt typam.

      (* Check that ind refers to an inductive type *)
      elim search_global with ind.
      2: right; red in |- *; intros.
      intros ([co| inds is| is c], sg_gs).
      right; red in |- *; intros.
      generalize sg_gs.
      inversion_clear H1.
      inversion_clear H3.
      inversion_clear H1.
      intro.
      generalize (ok_gs_inj sg_gs0 H3).
      discriminate.
      2: right; red in |- *; intros.
      2: generalize sg_gs.
      2: inversion_clear H1.
      2: inversion_clear H3.
      2: inversion_clear H1.
      2: intro.
      2: generalize (ok_gs_inj sg_gs0 H3).
      2: discriminate.
      (* OK *)
      elim subtype_dec with e typar (i_param is:term); intros.
      2: right; red in |- *; intros.
      cut (e ⊦ p : i_param is).
      intro wf_par.
      elim subtype_dec with e tyarg ((i_arg is)[0:=p]); intros.
      2: right; red in |- *; intros.
      cut (e ⊦ a : (i_arg is)[0:=p]).
      intro wf_arg.
      elim subtype_dec with e tymrk (Lmark:term); intros.
      2: right; red in |- *; intros.
      cut (e ⊦ mrk : Lmark).
      intro wf_mrk.
      left.
      split with (%%(i_kind is)).
      split; intros.
      split with (i_param is ** ((i_arg is)[0:=p] ** Lmark)).
      simpl in |- *; split.
      change_notation; apply cci_trans with (typar ** (tyarg ** tymrk)); auto.

      apply ind_par_arg_inst with (1 := sg_gs); trivial.

      split.
      simpl in |- *; red in |- *.
      apply Mem_mutind with inds; trivial.

      elim (ind_kind_ok sg_gs); intros.
      apply wf_var with x.
      constructor; trivial.
      apply (inv_wf _ _ H).

      inversion_clear H1.
      inversion_clear H3.
      inversion_clear H1.
      elim (inj_ind_is sg_gs H3); auto.

      apply Typ_conv with tymrk so; auto.
      cut (e ⊦ (p,,(a,,mrk)) : typar ** (tyarg ** tymrk)); intros.
      elim H1 using inv_pair_sum; intros.
      elim H3 using inv_pair_sum; intros.
      trivial.

      Inversion_typ H.
      apply subj_red_ty with (1 := H1); trivial.

      apply typ_lmark_so.
      apply (inv_wf _ _ H).

      (* Neg parts *)
      inversion_clear H1.
      inversion_clear H2.
      inversion_clear H3.
      generalize H1.
      inversion_clear H2; simpl in |- *; intros.
      apply b.
      apply
        (inv_sum_r (inv_sum_r (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H2))).

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      generalize (subj_red_wf (forget_def _ _ _ _ H) H0); intro.
      Inversion_typ H1.
      Inversion_typ H2.
      Inversion_typ H4.
      left with s4; trivial.

      apply ok_normalize.
      left with so.
      apply typ_lmark_so.
      apply (inv_wf _ _ H).

      apply typ_conv_wf with tyarg; simpl in |- *; trivial.
      cut (e ⊦ (p,,(a,,mrk)) : typar ** (tyarg ** tymrk)); intros.
      elim H1 using inv_pair_sum; intros.
      elim H3 using inv_pair_sum; intros.
      trivial.

      Inversion_typ H.
      apply subj_red_ty with (1 := H1); trivial.

      apply var_wft.
      Inversion_typ (ind_par_arg_inst sg_gs wf_par).
      Inversion_typ H1.
      Inversion_typ H3.
      apply wf_var with s0; trivial.

      inversion_clear H1.
      inversion_clear H2.
      inversion_clear H3.
      generalize H1.
      inversion_clear H2; simpl in |- *; intros.
      apply b.
      elim (inj_ind_is H3 sg_gs).
      apply
        (inv_sum_l (inv_sum_r (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H2))).

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      generalize (subj_red_wf (forget_def _ _ _ _ H) H0); intro.
      Inversion_typ H1.
      Inversion_typ H2.
      Inversion_typ H4.
      left with s0; trivial.

      apply ok_normalize.
      (* Bis: *)
      apply var_wft.
      Inversion_typ (ind_par_arg_inst sg_gs wf_par).
      Inversion_typ H1.
      Inversion_typ H3.
      apply wf_var with s0; trivial.

      apply typ_conv_wf with typar; trivial.
      cut (e ⊦ (p,,(a,,mrk)) : typar ** (tyarg ** tymrk)); intros.
      elim H1 using inv_pair_sum; intros.
      trivial.

      Inversion_typ H.
      apply subj_red_ty with (1 := H1); trivial.

      generalize (ind_par_arg_one sg_gs); intros.
      Inversion_typ H1.
      Inversion_typ (typ_wf _ _ _ H2).
      left with s0.
      apply thin_par_def0; trivial.
      apply (inv_wf _ _ H).

      inversion_clear H1.
      inversion_clear H2.
      inversion_clear H3.
      generalize H1.
      inversion_clear H2; simpl in |- *; intros.
      apply b.
      elim (inj_ind_is H3 sg_gs).
      apply (inv_sum_l (cci_trans (eq_term_sym_cci_sub (redn_conv H0)) H2)).

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      generalize (subj_red_wf (forget_def _ _ _ _ H) H0); intro.
      Inversion_typ H1.
      Inversion_typ H2.
      left with s1; trivial.

      apply ok_normalize.
      generalize (ind_par_arg_one sg_gs); intros.
      Inversion_typ H1.
      Inversion_typ (typ_wf _ _ _ H2).
      left with s0.
      apply thin_par_def0; trivial.
      apply (inv_wf _ _ H).

      inversion_clear H1.
      inversion_clear H3.
      inversion_clear H1.
      apply (b _ H3).

      apply red_trans with (typar ** typam); auto.

      inversion_clear H0.
      inversion_clear H1.
      inversion_clear H2.
      generalize H0.
      inversion_clear H1; simpl in |- *; intros.
      elim
        b
        with (1 := inv_sum_r (cci_trans (eq_term_sym_cci_sub (redn_conv redt)) H1)).

      Inversion_typ H.
      generalize (subject_reduction redt H1); intros.
      Inversion_typ H2.
      left with s2; trivial.

      inversion_clear H0.
      inversion_clear H1.
      inversion_clear H2.
      generalize H0.
      inversion_clear H1; simpl in |- *; intros.
      apply (b _ _ H1).

      apply var_wft.
      apply (forget_def _ _ _ _ H).
    Qed.



    (* Typing constructors is decidable *)

    Lemma sign_dec_cstr :
      forall c e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat (MutConstr c) e m t u)
          (cci_subtype e).
    Proof.
      intros.
      elim is_pair_dec with m.
      2: right; red in |- *; intros.
      2: inversion_clear H0.
      2: inversion_clear H2.
      2: generalize b.
      2: inversion_clear H0; intros.
      2: apply (b0 p a); trivial.
      intros ((u, v), eqp).
      rewrite eqp.
      rewrite eqp in H.
      clear eqp.
      elim search_global with c.
      2: right; red in |- *; intros.
      2: inversion_clear H0.
      2: inversion_clear H2.
      2: inversion_clear H0.
      2: apply (b (GlobCstr is c0)); trivial.
      intros ([co| inds is| is cs], sg_gs).
      right; red in |- *; intros.
      inversion_clear H0.
      inversion_clear H2.
      inversion_clear H0.
      generalize (ok_gs_inj sg_gs H2).
      discriminate.

      right; red in |- *; intros.
      inversion_clear H0.
      inversion_clear H2.
      inversion_clear H0.
      generalize (ok_gs_inj sg_gs H2).
      discriminate.

      elim whnf_cci with e t.
      intros ht.
      generalize (is_sum_dec ht).
      intros [(tpar, targ)| not_sum] redt hnft.
      elim subtype_dec with e tpar (i_param is:term); intros.
      cut (e ⊦ u : i_param is); intros.
      elim subtype_dec with e targ ((c_arg cs)[0:=u]); intros.
      left.
      split with Ind{i_name is // u,,(((c_inst cs)[1:=u])[0:=v],,[])}.
      split; intros.
      split with (i_param is ** (c_arg cs)[0:=u]).
      split.
      change_notation; apply cci_trans with (tpar ** targ); auto.

      generalize (typ_wf _ _ _ (ok_constr sg_gs)); intro.

      Inversion_typ H1.
      apply all_prod_sum.
      apply wf_thin1.
      apply typ_wf with (1 := H2).

      apply (inv_wf _ _ H).

      apply wf_var with s.
      fold (%%s[0:=u]) in |- *.
      apply substitution with x_ (i_param is:term).
      apply thin_par_def; trivial.
      apply (inv_wf _ _ H).

      Inversion_typ H.
      Inversion_typ H3.
      apply typ_conv_wf with a1; auto.
      apply var_wft.
      apply wf_thin1; auto.
      apply (typ_wf _ _ _ H2).

      apply (inv_wf _ _ H).

      simpl in |- *.
      apply cci_trans with (2 := a).
      apply inv_sum_l with b targ.
      apply cci_trans with (1 := H8); auto.

      split.
      constructor; trivial.

      apply ty_constr_ok with (1 := sg_gs); trivial.
      Inversion_typ H.
      elim (subj_red_ty H1 redt) using inv_pair_sum; intros.
      apply typ_conv_wf with targ; auto.
      apply var_wft.
      apply subst_wf with (x := x_) (1 := H0).
      Inversion_typ (typ_wf _ _ _ (ok_constr sg_gs)).
      apply wf_var with s0.
      apply thin_par_def; trivial.
      apply (inv_wf _ _ H).

      inversion_clear H1.
      inversion_clear H2.
      inversion_clear H3.
      generalize H1.
      inversion_clear H2.
      generalize (ok_gs_inj sg_gs H3); intro.
      injection H2; intros eqc eqis.
      elim eqc; elim eqis; auto.

      right; red in |- *; intros.
      inversion_clear H1.
      inversion_clear H2.
      inversion_clear H3.
      generalize H1.
      inversion_clear H2.
      generalize (ok_gs_inj sg_gs H3); intro.
      injection H2; intros eqc eqis.
      elim eqc; elim eqis; clear eqc eqis H3 H2 c0 is0.
      simpl in |- *; intros.
      apply b.
      apply inv_sum_r with tpar (i_param is:term).
      apply cci_trans with t; auto.

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      Inversion_typ H.
      generalize (subject_reduction redt H2); intro.
      Inversion_typ H3.
      left with s2; trivial.

      apply ok_normalize.
      apply wf_type_subst with x_ (i_param is:term); auto.
      generalize (typ_wf _ _ _ (ok_constr sg_gs)); intro.
      Inversion_typ H1.
      left with s.
      apply thin_par_def; trivial.
      apply (inv_wf _ _ H).

      Inversion_typ H.
      Inversion_typ H0.
      apply typ_conv_wf with a0; auto.
      generalize (typ_wf _ _ _ (ok_constr sg_gs)); intro.
      apply var_wft.
      apply wf_thin1; auto.
      apply (inv_wf _ _ H6).

      apply (inv_wf _ _ H).

      simpl in |- *.
      apply cci_trans with (2 := a).
      apply inv_sum_l with b targ.
      apply cci_trans with (1 := H5); auto.

      right; red in |- *; intros.
      inversion_clear H0.
      inversion_clear H1.
      inversion_clear H2.
      generalize H0.
      inversion_clear H1.
      generalize (ok_gs_inj sg_gs H2); intro.
      injection H1; intros eqc eqis.
      elim eqc; elim eqis; clear eqc eqis H2 H1 c0 is0.
      simpl in |- *; intros.
      apply b.
      apply inv_sum_l with targ ((c_arg cs)[0:=u]).
      apply cci_trans with t; auto.

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      Inversion_typ H.
      generalize (subject_reduction redt H1); intro.
      Inversion_typ H2.
      left with s1; trivial.

      apply ok_normalize.
      generalize (typ_wf _ _ _ (ok_constr sg_gs)); intro.
      apply var_wft.
      apply wf_thin1; auto.
      apply (inv_wf _ _ H0).

      apply (inv_wf _ _ H).

      right; red in |- *; intros.
      inversion_clear H0.
      inversion_clear H1.
      inversion_clear H2.
      generalize H0.
      inversion_clear H1; intros.
      apply not_sum.
      apply conv_sum_eq with (3 := H1); trivial.

      Inversion_typ H.
      left with s; trivial.
    Qed.

  End IIndDecOk.
