(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export InvReduction.

Set Implicit Arguments.
Unset Strict Implicit.

Section IInvEqual.

  Variable sg : sigma.

  (* Inversion of eq_term *)


  Lemma CR :
   forall e a b,
   eq_term e a b ->
   exists2 a' : _,
     red redn_term e a a' &
     (exists2 b' : _, red redn_term e b b' & red alpha e a' b').
Proof.
simple induction 1; intros.
elim ctxt_red_term_or with (1 := H0); intros.
split with y; auto.
split with y; auto.

split with x; auto.
split with y; auto.

split with x; auto.
split with x; auto.

inversion_clear H1.
inversion_clear H3.
split with x1; auto.
split with x0; auto.
apply alpha_sym; trivial.

inversion_clear H1.
inversion_clear H5.
inversion_clear H3.
inversion_clear H7.
elim confluent_redn with (1 := H1) (2 := H5); intros.
red in H7.
elim commut_alpha_redn with (1 := H9) (2 := H8); intros.
elim commut_alpha_redn with (1 := H7) (2 := alpha_sym H6); intros.
split with x6.
apply red_trans with x0; trivial.

split with x5.
apply red_trans with x3; trivial.

apply red_trans with x4; trivial.
apply alpha_sym; trivial.
Qed.


  Lemma confl_red_term : confluent (red red_term).
Proof.
red in |- *; intros.
red in |- *; intros.
red in H0.
elim CR with e x z; intros.
inversion_clear H2.
split with x1.
red in |- *.
apply red_trans with x0.
apply redn_term_red_term; trivial.

apply alpha_red_term; trivial.

apply redn_term_red_term; trivial.

apply eq_term_trans with y; auto.
red in |- *.
apply conv_sym.
apply red_conv; trivial.

red in |- *.
apply red_conv; trivial.
Qed.



  Lemma stb0_var :
   forall e x y t t' a b,
   eq_term e t t' -> eq_term (e; x : t) a b -> eq_term (e; y : t') a b.
Proof.
intros.
apply (R_stable_eq (R:=eq_term)) with (e; x : t); auto.
unfold add_var in |- *; constructor.
split; simpl in |- *; auto.
Qed.



  Inductive eq_hn_inv : env -> term -> term -> Prop :=
    | eqhi_srt : forall e s, eq_hn_inv e %%s %%s
    | eqhi_rel : forall e n, eq_hn_inv e #n #n
    | eqhi_cst0 : forall e c, eq_hn_inv e (Cst0 c) (Cst0 c)
    | eqhi_cst1 :
        forall e c x y, eq_term e x y -> eq_hn_inv e (Cst1 c x) (Cst1 c y)
    | eqhi_prd :
        forall e x y A A' B B',
        eq_term e A A' ->
        eq_term (e; x : A) B B' ->
        eq_hn_inv e (Forall x : A, B) (Forall y : A', B')
    | eqhi_lam :
        forall e x y A A' M M',
        eq_term e A A' ->
        eq_term (e; x : A) M M' ->
        eq_hn_inv e (Fun x : A => M) (Fun y : A' => M')
    | eqhi_sum :
        forall e A A' B B',
        eq_term e A A' -> eq_term e B B' -> eq_hn_inv e (A ** B) (A' ** B')
    | eqhi_pair :
        forall e M M' N N',
        eq_term e M M' -> eq_term e N N' -> eq_hn_inv e (M,,N) (M',,N').

  (* eq_hn_inv is an equivalence relation (modulo Leibniz and alpha): *)

  Lemma inv_eq_refl : forall e x, eq_hn_inv e x x.
Proof.
simple destruct x; constructor; auto.
Qed.

  Lemma inv_eq_sym : forall e x y, eq_hn_inv e x y -> eq_hn_inv e y x.
Proof.
simple induction 1; intros; constructor; auto.
apply stb0_var with x0 A; auto.

apply stb0_var with x0 A; auto.
Qed.


  Lemma inv_eq_trans :
   forall e x y z, eq_hn_inv e x y -> eq_hn_inv e y z -> eq_hn_inv e x z.
Proof.
simple destruct 1; intros.
inversion_clear H0; constructor.

inversion_clear H0; constructor.

inversion_clear H0; constructor.

inversion_clear H1; constructor.
apply eq_term_trans with y0; trivial.

inversion_clear H2; constructor.
apply eq_term_trans with A'; trivial.

apply eq_term_trans with B'; trivial.
apply stb0_var with y0 A'; trivial.
red in |- *; apply conv_sym; trivial.

inversion_clear H2; constructor.
apply eq_term_trans with A'; trivial.

apply eq_term_trans with M'; trivial.
apply stb0_var with y0 A'; trivial.
red in |- *; apply conv_sym; trivial.

inversion_clear H2; constructor.
apply eq_term_trans with A'; trivial.

apply eq_term_trans with B'; trivial.

inversion_clear H2; constructor.
apply eq_term_trans with M'; trivial.

apply eq_term_trans with N'; trivial.
Qed.



  (* eq_hn_inv is equivalent to eq_term on head normal terms *)

  Lemma inv_redn_eq : forall e x y, redn_term_hn_inv e x y -> eq_hn_inv e x y.
Proof.
simple destruct 1; constructor; auto.
Qed.


  Lemma inv_eq_conv : forall e x y, eq_hn_inv e x y -> eq_term e x y.
Proof.
red in |- *; simple destruct 1; intros; auto.
apply conv_trans with (Forall x0 : A', B'); auto.
red in |- *; red in |- *; apply rst_step; apply Ctx_rule; do 2 right; left.
constructor.

apply conv_trans with (Fun x0 : A' => M'); auto.
red in |- *; red in |- *; apply rst_step; apply Ctx_rule; do 2 right; left.
constructor.
Qed.

  Hint Resolve inv_eq_conv: core.


  Lemma inv_eq_refl_alpha : forall e x y, red alpha e x y -> eq_hn_inv e x y.
Proof.
simple induction 1.
simple induction 1; intros; try constructor; auto.
inversion_clear H1; constructor; auto.

intros.
apply inv_eq_refl.

intros.
apply inv_eq_trans with y0; trivial.
Qed.


  Lemma inv_eq_hn_red :
   forall e x x' y y',
   eq_term e x y ->
   red redn_term e x x' ->
   head_normal redn_term e x' ->
   red redn_term e y y' -> head_normal redn_term e y' -> eq_hn_inv e x' y'.
Proof.
intros.
elim CR with e x' y'; intros.
inversion_clear H5.
apply inv_eq_trans with x0.
apply inv_redn_eq.
apply inv_redn_hn; trivial.

apply inv_eq_trans with x1.
apply inv_eq_refl_alpha; trivial.

apply inv_eq_sym.
apply inv_redn_eq.
apply inv_redn_hn; trivial.

apply eq_term_trans with x.
apply eq_term_sym; auto.

apply eq_term_trans with y; auto.
Qed.

  Lemma inv_eq_hn :
   forall e x y,
   eq_term e x y ->
   head_normal redn_term e x -> head_normal redn_term e y -> eq_hn_inv e x y.
Proof.
intros.
apply inv_eq_hn_red with x y; auto.
Qed.
  Hint Resolve inv_eq_hn: core.



  Lemma inv_sum_eq_l :
   forall e a1 a2 b1 b2, eq_term e (a1 ** b1) (a2 ** b2) -> eq_term e a1 a2.
Proof.
intros.
cut (eq_hn_inv e (a1 ** b1) (a2 ** b2)); intros; auto.
inversion_clear H0; trivial.
Qed.

  Lemma inv_sum_eq_r :
   forall e a1 a2 b1 b2, eq_term e (a1 ** b1) (a2 ** b2) -> eq_term e b1 b2.
Proof.
intros.
cut (eq_hn_inv e (a1 ** b1) (a2 ** b2)); intros; auto.
inversion_clear H0; trivial.
Qed.


  Lemma inv_one_not_eq_sum : forall e a b, ~ eq_term e (a ** b) one.
Proof.
red in |- *; intros.
cut (eq_hn_inv e (a ** b) one); unfold one in |- *; intros; auto.
inversion_clear H0.
Qed.


  Lemma one_not_eq_sum : forall e a b, ~ eq_term e one (a ** b).
Proof.
red in |- *; intros.
cut (eq_hn_inv e one (a ** b)); unfold one in |- *; intros.
inversion_clear H0.

apply inv_eq_hn; auto.
Qed.


  Lemma eq_one_red :
   forall e t x,
   red redn_term e t x ->
   head_normal redn_term e x -> eq_term e t one -> x = one.
Proof.
intros.
cut (eq_hn_inv e x one); intros.
inversion_clear H2; trivial.

apply inv_eq_hn_red with t one; unfold one in |- *; auto.
Qed.


  Lemma conv_pair_eq :
   forall e t u a b,
   red redn_term e t u ->
   head_normal redn_term e u ->
   eq_term e t (a,,b) -> {p : term * term | let (x, y) := p in u = (x,,y)}.
Proof.
intros.
cut (eq_hn_inv e u (a,,b)).
case u; intros; auto; try (apply False_rec; inversion_clear H2; fail).
exists (t0, t1); trivial.

apply inv_eq_hn_red with t (a,,b); auto.
Qed.


  Lemma inv_prd_eq_l :
   forall e x y a1 a2 b1 b2,
   eq_term e (Forall x : a1, b1) (Forall y : a2, b2) -> eq_term e a1 a2.
Proof.
intros.
cut (eq_hn_inv e (Forall x : a1, b1) (Forall y : a2, b2)); intros; auto.
inversion_clear H0; trivial.
Qed.

  
  
  Lemma inv_prd_eq_r :
   forall e x y a1 a2 b1 b2,
   eq_term e (Forall x : a1, b1) (Forall y : a2, b2) ->
   eq_term (e; x : a1) b1 b2.
Proof.
intros.
cut (eq_hn_inv e (Forall x : a1, b1) (Forall y : a2, b2)); intros; auto.
inversion_clear H0; trivial.
Qed.


  Lemma inv_pair_r :
   forall e a1 a2 b1 b2, eq_term e (a1,,b1) (a2,,b2) -> eq_term e b1 b2.
Proof.
intros.
cut (eq_hn_inv e (a1,,b1) (a2,,b2)); intros; auto.
inversion_clear H0; trivial.
Qed.

  Lemma inv_pair_l :
   forall e a1 a2 b1 b2, eq_term e (a1,,b1) (a2,,b2) -> eq_term e a1 a2.
Proof.
intros.
cut (eq_hn_inv e (a1,,b1) (a2,,b2)); intros; auto.
inversion_clear H0; trivial.
Qed.


  Lemma eq_term_arities :
   forall s1 s2 e,
   eq_term e (record_arity s1) (record_arity s2) ->
   eq_term e (prod_arity s1) (prod_arity s2).
Proof.
simple induction s1; [ idtac | intros (x, ty) ];
 (simple destruct s2; [ idtac | intros (x', ty') ]); 
 simpl in |- *; intros; auto.
elim one_not_eq_sum with (1 := H).

elim one_not_eq_sum with (1 := eq_term_sym H0).

apply eq_term_trans with (Forall x : ty', prod_arity l0).
red in |- *; apply conv_conv_prd.
apply (inv_sum_eq_l H0).

apply H.
apply (inv_prd_eq_r (inv_sum_eq_r H0)).

red in |- *; apply red_conv; apply alpha_red_term.
red in |- *; red in |- *; apply rt_step; constructor.
constructor.
Qed.

End IInvEqual.

Implicit Arguments CR [ [sg] ].
Implicit Arguments eq_hn_inv [ [sg] ].
