(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export Subtyping.

Set Implicit Arguments.
Unset Strict Implicit.

Section IInvSubtyping.

  Context {sg : sigma}.


  (* Inversion lemmas about <= *)

  Lemma stb0_var_sub :
   forall e x y t t' a b,
   cci_subtype e t t' ->
   cci_subtype (e; x : t') a b -> cci_subtype (e; y : t) a b.
Proof.
intros.
apply (subtype_stable (R:=cci_subtype)) with (e; x : t'); auto.
unfold add_var in |- *; constructor.
split; simpl in |- *; auto.
Qed.


  Lemma commut_red_incl :
   forall e k k',
   commut term (fun x y => incl_seq k x k' y)
     (fun x y => red redn_term e y x).
Proof.
red in |- *.
simple induction 1; intros.
split with z; auto.
apply Incl_refl.

elim H1 with (1 := H4); intros.
elim H3 with (1 := H5); intros.
split with x1; trivial.
apply Incl_trans with m2 x0; trivial.

split with []; auto.
apply Incl_mt.

split with z; trivial.
apply Incl_str.

cut (redn_term_hn_inv e (m && ml) z); intros.
inversion_clear H1.
cut (redn_term_hn_inv e (m,,ml) y0); intros.
inversion_clear H1.
split with N'; trivial.
apply Incl_remove.

apply inv_redn_hn with (1 := H2); auto.

apply inv_redn_hn with (1 := H0); auto.

split with (## && z); auto.
apply Incl_add_mt.

cut (redn_term_hn_inv e (m && ml) z); intros.
inversion_clear H3.
cut (redn_term_hn_inv e (m,,ml) y0); intros.
inversion_clear H3.
elim H1 with (1 := H6); intros.
split with (M' && x0); auto.
apply Incl_deep_le.
trivial.

apply inv_redn_hn with (1 := H4); auto.

apply inv_redn_hn with (1 := H2); auto.

cut (redn_term_hn_inv e (m && ml) z); intros.
inversion_clear H3.
cut (redn_term_hn_inv e (m,,ml) y0); intros.
inversion_clear H3.
elim H1 with (1 := H6); intros.
split with (M' && x0); auto.
apply Incl_deep_lt; trivial.

apply inv_redn_hn with (1 := H4); auto.

apply inv_redn_hn with (1 := H2); auto.
Qed.

  Lemma commut_red_incl_tr :
   forall e k k',
   commut term (fun x y => incl_seq k y k' x)
     (fun x y => red redn_term e y x).
Proof.
red in |- *.
simple induction 1; intros.
split with z; auto.
apply Incl_refl.

elim H3 with (1 := H4); intros.
elim H1 with (1 := H5); intros.
split with x1; trivial.
apply Incl_trans with m2 x0; trivial.

elim red_normal with (1 := H0).
split with t; auto.
apply Incl_mt.

apply nil_mark_normal.

split with z; trivial.
apply Incl_str.

split with (m && z); auto.
apply Incl_remove.

cut (redn_term_hn_inv e (## && ml) z); intros.
inversion_clear H1.
cut (redn_term_hn_inv e (##,,ml) y0); intros.
inversion_clear H1.
split with N'; auto.
elim red_normal with (1 := H3).
apply Incl_add_mt.

apply mt_mark_normal.

apply inv_redn_hn with (1 := H2); auto.

apply inv_redn_hn with (1 := H0); auto.

cut (redn_term_hn_inv e (m && ml2) z); intros.
inversion_clear H3.
cut (redn_term_hn_inv e (m,,ml2) y0); intros.
inversion_clear H3.
elim H1 with (1 := H6); intros.
split with (M' && x0); auto.
apply Incl_deep_le; trivial.

apply inv_redn_hn with (1 := H4); auto.

apply inv_redn_hn with (1 := H2); auto.

cut (redn_term_hn_inv e (m && ml2) z); intros.
inversion_clear H3.
cut (redn_term_hn_inv e (m,,ml2) y0); intros.
inversion_clear H3.
elim H1 with (1 := H6); intros.
split with (M' && x0); auto.
apply Incl_deep_lt; trivial.

apply inv_redn_hn with (1 := H4); auto.

apply inv_redn_hn with (1 := H2); auto.
Qed.



  Inductive sub_hn_inv : env -> term -> term -> Prop :=
    | subhi_srt :
        forall e s s',
        clos_refl_trans _ incl_sort s s' -> sub_hn_inv e %%s %%s'
    | subhi_rel : forall e n, sub_hn_inv e #n #n
    | subhi_cst0 : forall e (c : cci_op), sub_hn_inv e c c
    | subhi_cst1 :
        forall e c a a',
        eq_term e a a' -> sub_hn_inv e (Cst1 c a) (Cst1 c a')
    | subhi_prd :
        forall e x y A A' B B',
        cci_subtype e A' A ->
        cci_subtype (e; y : A') B B' ->
        sub_hn_inv e (Forall x : A, B) (Forall y : A', B')
    | subhi_lam :
        forall e x y A A' M M',
        eq_term e A A' ->
        eq_term (e; y : A') M M' ->
        sub_hn_inv e (Fun x : A => M) (Fun y : A' => M')
    | subhi_sum :
        forall e A A' B B',
        cci_subtype e A A' ->
        cci_subtype e B B' -> sub_hn_inv e (A ** B) (A' ** B')
    | subhi_pair :
        forall e M M' N N',
        eq_term e M M' -> eq_term e N N' -> sub_hn_inv e (M,,N) (M',,N')
    | subhi_ind :
        forall e i x y l l' p a m m',
        incl_seq m l m' l' ->
        eq_term e x (p,,(a,,l)) ->
        eq_term e y (p,,(a,,l')) ->
        sub_hn_inv e Ind{i; m // x} Ind{i; m' // y}
    | subhi_rec :
        forall e s1 s2 x y,
        eq_term e x (record_arity s1) ->
        eq_term e y (record_arity s2) ->
        sub_record e s1 s2 ->
        sub_hn_inv e {<field_names s1 : x>} {<field_names s2 : y>}.


  Lemma inv_sub_cci : forall e x y, sub_hn_inv e x y -> cci_subtype e x y.
Proof.
simple induction 1; intros; auto.
elim H0; intros; auto.
apply cci_sub_step.
left; auto.

apply cci_trans with (%%y0); trivial.

apply eq_term_cci_sub.
red in |- *.
apply conv_conv_cst.
trivial.

apply cci_trans with (Forall y0 : A, B); auto.
apply eq_term_cci_sub.
red in |- *; apply red_conv; apply alpha_red_term.
red in |- *; red in |- *; apply rt_step; constructor.
constructor.

apply eq_term_cci_sub.
apply eq_term_trans with (Fun y0 : A => M); auto.
red in |- *; apply red_conv; apply alpha_red_term.
red in |- *; red in |- *; apply rt_step; constructor.
constructor.

apply eq_term_sym.
red in |- *; apply conv_conv_lam; auto.

apply eq_term_cci_sub.
red in |- *; apply conv_conv_pair; trivial.

apply cci_trans with Ind{i; m // p,,(a,,l)}; auto.
apply eq_term_cci_sub.
red in |- *; auto.

apply cci_trans with Ind{i; m' // p,,(a,,l')}; auto.
apply cci_sub_step.
right.
constructor.
trivial.

apply eq_term_sym_cci_sub.
red in |- *; auto.

apply cci_trans with (record_of s1).
apply eq_term_sym_cci_sub.
red in |- *; unfold record_of in |- *; auto.

apply cci_trans with (record_of s2).
apply record_sub_arity.
trivial.

apply eq_term_sym_cci_sub.
red in |- *; unfold record_of in |- *; auto.
Qed.



  Lemma eq_term_record_sub :
   forall s1 s2 e,
   field_names s1 = field_names s2 ->
   eq_term e (record_arity s1) (record_arity s2) -> sub_record e s1 s2.
Proof.
simple induction s1; simple destruct s2; simpl in |- *; intros.
constructor.

discriminate H.

discriminate H0.

generalize a p H0 H1; clear H0 H1 a p.
intros (x, ty) (x', ty') H0 H1.
injection H0; intros.
fold map field_names in H2.
simpl in H3.
elim H3.
constructor.
apply eq_term_cci_sub.
apply (inv_sum_eq_l H1).

apply H; auto.
apply (inv_prd_eq_r (inv_sum_eq_r H1)).
Qed.



  Lemma inv_cumul_trans :
   forall e x y z, sub_hn_inv e x y -> sub_hn_inv e y z -> sub_hn_inv e x z.
Proof.
simple induction 1; intros.
inversion_clear H1.
constructor.
apply rt_trans with s'; auto.

inversion_clear H0; constructor.

inversion_clear H0; constructor.

inversion_clear H1.
constructor.
apply eq_term_trans with a'; trivial.

apply subhi_ind with l l' p a0; auto.
apply eq_term_trans with a'; trivial.

apply subhi_rec; trivial.
apply eq_term_trans with a'; trivial.

inversion_clear H2; constructor.
apply cci_trans with A'; auto.

apply cci_trans with B'; auto.
apply stb0_var_sub with y0 A'; trivial.

inversion_clear H2; constructor.
apply eq_term_trans with A'; auto.

apply eq_term_trans with M'; auto.
apply stb0_var with y0 A'; trivial.

inversion_clear H2; constructor.
apply cci_trans with A'; auto.

apply cci_trans with B'; auto.

inversion_clear H2; constructor.
apply eq_term_trans with M'; auto.

apply eq_term_trans with N'; auto.

(* Ind *)
inversion_clear H3.
apply subhi_ind with l l' p a; trivial.
apply eq_term_trans with y0; auto.

(* Diagramme de commutation *)
elim CR with e0 l0 l'; intros.
inversion_clear H7.
elim commut_red_incl with (1 := H4) (2 := H3); intros.
elim commut_red_incl_tr with (1 := H0) (2 := H8); intros.
elim commut_alpha_incl with (1 := H10) (2 := H9); intros.
apply subhi_ind with x4 x5 p a.
apply Incl_trans with m' x2; trivial.

apply eq_term_trans with (1 := H1).
red in |- *.
apply conv_conv_pair.
red in |- *; red in |- *; trivial.

apply conv_conv_pair.
red in |- *; red in |- *; trivial.

apply redn_conv; trivial.

apply eq_term_trans with (1 := H6).
red in |- *.
apply conv_conv_pair.
apply (inv_pair_l (eq_term_trans (conv_sym _ _ _ _ H5) H2)).

apply conv_conv_pair.
apply (inv_pair_l (inv_pair_r (eq_term_trans (conv_sym _ _ _ _ H5) H2))).

fold eq_term in |- *.
apply eq_term_trans with x3.
apply redn_conv; trivial.

red in |- *.
apply red_conv.
apply alpha_red_term; trivial.

apply (inv_pair_r (inv_pair_r (eq_term_trans (conv_sym _ _ _ _ H5) H2))).

(* Record *)
inversion H3.
apply subhi_rec; trivial.
apply eq_term_trans with y0; auto.

apply subhi_rec; trivial.
apply sub_record_trans with (1 := H2).
apply sub_record_trans with (2 := H10).
apply eq_term_record_sub; auto.
apply eq_term_trans with y0; auto.
Qed.


  Lemma inv_sub_eq : forall e x y, eq_hn_inv e x y -> sub_hn_inv e x y.
Proof.
simple destruct 1; intros; constructor; auto.
apply eq_term_cci_sub.
apply stb0_var with x0 A; auto.

apply stb0_var with x0 A; auto.
Qed.


(* Les deux lemmes suivants font une recurrence sur clos_le de TypCtxt:
 * mauvais pour la modularite!.
 *)
  Lemma commut_red_hnf_cci_subtype :
   forall e,
   commut term (cci_subtype e)
     (fun x y => red redn_term e y x /\ head_normal redn_term e x).
red in |- *.
simple induction 1.
simple induction 1; intros.
inversion_clear H2.
elim CR with e z y0; intros; auto.
inversion_clear H5.
split with x2.
split; auto.
apply hn_alpha with x1; trivial.
apply hn_red_hn with z; auto.

apply cci_trans with x1; auto.
apply eq_term_cci_sub.
red in |- *.
apply red_conv.
apply alpha_red_term; trivial.

apply eq_term_trans with x0; auto.
apply eq_term_sym.
auto.

generalize H2.
case H1; intros.
generalize H4.
inversion_clear H3.
inversion_clear H5; intros.
inversion_clear H5.
split with (%%s2); auto.
apply cci_trans with (%%s1); auto.
apply cci_sub_step.
left; auto.

inversion_clear H5; intros.
split with Ind{x2; l2 // p,,(a,,m2)}; auto.
apply cci_trans with Ind{x2; l1 // p,,(a,,m1)}; auto.
inversion_clear H5; trivial.
auto.

apply cci_sub_step.
right; constructor; trivial.

split with z; auto.

inversion_clear H5.
split with (Forall x1 : a2, b2); auto.
apply cci_trans with (Forall x1 : a1, b1); auto.
apply cci_sub_prd.
red in |- *; red in |- *; apply rt_step; right; trivial.

red in |- *; red in |- *; apply rt_step; right; trivial.

inversion_clear H5.
split with (a2 ** b2); auto.
apply cci_trans with (a1 ** b1); auto.
apply cci_sub_sum.
red in |- *; red in |- *; apply rt_step; right; trivial.

red in |- *; red in |- *; apply rt_step; right; trivial.

inversion_clear H4.
split with (record_of s2); auto.
unfold record_of in |- *; auto.

apply cci_trans with (1 := eq_term_sym_cci_sub (redn_conv H5)).
apply record_sub_arity.
apply sub_record_intro; trivial.

intros.
split with z; auto.

intros.
elim H1 with (1 := H4); intros.
elim H3 with (1 := H5); intros.
split with x2; auto.
apply cci_trans with x1; trivial.
Qed.



  Lemma inv_cumul_hn :
   forall e x y,
   cci_subtype e x y ->
   forall x' y',
   red redn_term e x x' ->
   head_normal redn_term e x' ->
   red redn_term e y y' -> head_normal redn_term e y' -> sub_hn_inv e x' y'.
Proof.
simple induction 1.
simple destruct 1.
intros.
apply inv_sub_eq.
apply inv_eq_hn_red with x0 y0; trivial.

simple destruct 1.
simple destruct 1.
simple destruct 1; intros.
cut (redn_term_hn_inv e1 %%s1 x'); intros.
inversion_clear H9.
cut (redn_term_hn_inv e1 %%s2 y'); intros.
inversion_clear H9.
constructor; auto.

apply inv_redn_hn; auto.

apply inv_redn_hn; auto.

simple destruct 1; intros.
apply inv_cumul_trans with Ind{x2; l1 // p,,(a,,m1)}.
apply inv_sub_eq.
apply inv_eq_sym.
apply inv_redn_eq.
apply inv_redn_hn; auto.

apply inv_cumul_trans with Ind{x2; l2 // p,,(a,,m2)}.
apply subhi_ind with m1 m2 p a; auto.

apply inv_sub_eq.
apply inv_redn_eq.
apply inv_redn_hn; auto.

intros.
apply inv_sub_eq.
apply inv_eq_hn_red with x1 x1; auto.

intros.
apply inv_cumul_trans with (Forall x1 : a1, b1).
apply inv_sub_eq.
apply inv_eq_sym.
apply inv_redn_eq.
apply inv_redn_hn; auto.

apply inv_cumul_trans with (Forall x1 : a2, b2).
constructor; red in |- *; red in |- *; apply rt_step; right; auto.

apply inv_sub_eq.
apply inv_redn_eq.
apply inv_redn_hn; auto.

intros.
apply inv_cumul_trans with (a1 ** b1).
apply inv_sub_eq.
apply inv_eq_sym.
apply inv_redn_eq.
apply inv_redn_hn; auto.

apply inv_cumul_trans with (a2 ** b2).
constructor; auto; red in |- *; red in |- *; apply rt_step; right; auto.

apply inv_sub_eq.
apply inv_redn_eq.
apply inv_redn_hn; auto.

intros.
apply inv_cumul_trans with (record_of s1).
apply inv_sub_eq.
apply inv_eq_sym.
apply inv_redn_eq.
apply inv_redn_hn; unfold record_of in |- *; auto.

apply inv_cumul_trans with (record_of s2).
unfold record_of in |- *.
apply subhi_rec; trivial.
apply sub_record_intro; trivial.

apply inv_sub_eq.
apply inv_redn_eq.
apply inv_redn_hn; unfold record_of in |- *; auto.

intros.
apply inv_sub_eq.
apply inv_eq_hn_red with x0 x0; auto.

intros.
elim commut_red_hnf_cci_subtype with e y0 x0 x'; intros; auto.
inversion_clear H8.
apply inv_cumul_trans with x1; auto.
Qed.


  Lemma inv_sub_hn :
   forall e x y,
   cci_subtype e x y ->
   head_normal redn_term e x -> head_normal redn_term e y -> sub_hn_inv e x y.
Proof.
intros.
apply inv_cumul_hn with x y; auto.
Qed.
  Hint Resolve inv_sub_hn: core.



  Lemma one_not_sum : forall e a b, ~ cci_subtype e one (a ** b).
Proof.
red in |- *; intros.
cut (sub_hn_inv e one (a ** b)); intros.
inversion_clear H0.

apply inv_sub_hn; unfold one in |- *; auto.
Qed.


  Lemma conv_srt_eq :
   forall e t u s,
   red redn_term e t u ->
   head_normal redn_term e u ->
   cci_subtype e t %%s -> exists s' : _, %%s' = u.
Proof.
intros.
cut (sub_hn_inv e u %%s); intros.
inversion_clear H2.
split with s0; trivial.

apply inv_cumul_hn with t (%%s); auto.
Qed.


  Lemma conv_prd_eq :
   forall e x t u a b,
   red redn_term e t u ->
   head_normal redn_term e u -> cci_subtype e t (Forall x : a, b) -> is_prd u.
Proof.
intros.
cut (sub_hn_inv e u (Forall x : a, b)).
case u; intros; auto; try (apply False_rec; inversion_clear H2).

apply inv_cumul_hn with t (Forall x : a, b); auto.
Qed.


  Lemma conv_sum_eq :
   forall e t u a b,
   red redn_term e t u ->
   head_normal redn_term e u -> cci_subtype e t (a ** b) -> is_sum u.
Proof.
intros.
cut (sub_hn_inv e u (a ** b)).
case u; intros; auto; try (apply False_rec; inversion_clear H2).

apply inv_cumul_hn with t (a ** b); auto.
Qed.




  (* The inversion lemmas *)
  Lemma inv_sort :
   forall e s1 s2,
   cci_subtype e %%s1 %%s2 -> clos_refl_trans _ incl_sort s1 s2.
Proof.
intros.
cut (sub_hn_inv e %%s1 %%s2); intros; auto.
inversion_clear H0; trivial.
Qed.


  Lemma inv_sum_l :
   forall e a1 a2 b1 b2,
   cci_subtype e (a1 ** b1) (a2 ** b2) -> cci_subtype e a1 a2.
Proof.
intros.
cut (sub_hn_inv e (a1 ** b1) (a2 ** b2)); intros; auto.
inversion_clear H0; trivial.
Qed.

  Lemma inv_sum_r :
   forall e a1 a2 b1 b2,
   cci_subtype e (a1 ** b1) (a2 ** b2) -> cci_subtype e b1 b2.
Proof.
intros.
cut (sub_hn_inv e (a1 ** b1) (a2 ** b2)); intros; auto.
inversion_clear H0; trivial.
Qed.

  Lemma inv_prod_l :
   forall e x1 x2 a1 a2 b1 b2,
   cci_subtype e (Forall x1 : a1, b1) (Forall x2 : a2, b2) ->
   cci_subtype e a2 a1.
Proof.
intros.
cut (sub_hn_inv e (Forall x1 : a1, b1) (Forall x2 : a2, b2)); intros; auto.
inversion_clear H0; trivial.
Qed.

  Lemma inv_prod_r :
   forall e x1 x2 a1 a2 b1 b2,
   cci_subtype e (Forall x1 : a1, b1) (Forall x2 : a2, b2) ->
   cci_subtype (e; x2 : a2) b1 b2.
Proof.
intros.
cut (sub_hn_inv e (Forall x1 : a1, b1) (Forall x2 : a2, b2)); intros; auto.
inversion_clear H0; trivial.
Qed.


  Lemma inv_mkvect :
   forall e n l1 l2 t1 t2,
   cci_subtype e (mkvect l1) (mkvect l2) ->
   item t1 l1 n -> item t2 l2 n -> cci_subtype e t1 t2.
Proof.
simple induction n; intros.
generalize H.
inversion_clear H0.
inversion_clear H1; simpl in |- *; intros.
apply (inv_sum_l H0).

generalize H0.
inversion_clear H1.
inversion_clear H2; simpl in |- *; intros.
apply H with (1 := inv_sum_r H2) (2 := H3) (3 := H1).
Qed.


  Lemma sub_record_intro_arity :
   forall e s1 s2,
   field_names s1 = field_names s2 ->
   cci_subtype e (prod_arity s2) (prod_arity s1) -> sub_record e s1 s2.
Proof.
intros e s1.
generalize e; clear e.
elim s1; simple destruct s2; simpl in |- *; intros; try discriminate.
constructor.

destruct a; destruct p; simpl in H0, H1.
injection H0; intros.
subst n.
constructor.
apply inv_prod_l with (1 := H1).

apply H; auto.
apply inv_prod_r with (1 := H1).
Qed.

  Lemma inv_record_sub_arity :
   forall e s1 s2,
   cci_subtype e (record_of s1) (record_of s2) -> sub_record e s1 s2.
Proof.
intros.
cut (sub_hn_inv e (record_of s1) (record_of s2)); unfold record_of in |- *;
 intros; auto.
apply sub_record_intro_arity.
inversion_clear H0; trivial.
apply (same_names H3).

inversion_clear H0.
apply eq_term_sym_cci_sub.
apply eq_term_arities; trivial.

apply cci_trans with (1 := eq_term_cci_sub (eq_term_arities H2)).
apply cci_trans with (2 := eq_term_sym_cci_sub (eq_term_arities H1)).
apply (incl_arity H3).
Qed.


  (* inutilise pour l'instant *)
  Lemma inv_ind_i :
   forall e i1 i2 m1 m2 t1 t2,
   cci_subtype e Ind{i1; m1 // t1} Ind{i2; m2 // t2} -> i1 = i2.
Proof.
intros.
cut (sub_hn_inv e Ind{i1; m1 // t1} Ind{i2; m2 // t2}); intros; auto.
inversion_clear H0; trivial.
Qed.

  Lemma inv_ind_m :
   forall e i1 i2 m1 m2 p1 a1 l1 p2 a2 l2,
   cci_subtype e Ind{i1; m1 // p1,,(a1,,l1)}
     Ind{i2; m2 // p2,,(a2,,l2)} ->
   exists2 l1' : _,
     eq_term e l1 l1' &
     (exists2 l2' : _,
        eq_term e l2 l2' &
        incl_mark e Ind{i2; m1 // p2,,(a2,,l1')}
          Ind{i2; m2 // p2,,(a2,,l2')}).
Proof.
intros.
cut
 (sub_hn_inv e Ind{i1; m1 // p1,,(a1,,l1)} Ind{i2; m2 // p2,,(a2,,l2)});
 intros; auto.
inversion_clear H0.
split with l2.
apply (inv_pair_r (inv_pair_r H1)).

split with l2; auto.

split with l.
apply (inv_pair_r (inv_pair_r H2)).

split with l'.
apply (inv_pair_r (inv_pair_r H3)).

constructor.
constructor; trivial.
Qed.


(* NON! SI !! *)
  Lemma inv_ind_p :
   forall e i1 i2 m1 m2 p1 a1 l1 p2 a2 l2,
   cci_subtype e Ind{i1; m1 // p1,,(a1,,l1)}
     Ind{i2; m2 // p2,,(a2,,l2)} -> eq_term e p1 p2.
Proof.
intros.
cut
 (sub_hn_inv e Ind{i1; m1 // p1,,(a1,,l1)} Ind{i2; m2 // p2,,(a2,,l2)});
 intros; auto.
inversion_clear H0.
apply inv_pair_l with (1 := H1).

apply eq_term_trans with p.
apply inv_pair_l with (1 := H2).

red in |- *; apply conv_sym.
apply inv_pair_l with (1 := H3).
Qed.


  Lemma inv_ind_a :
   forall e i1 i2 m1 m2 p1 a1 l1 p2 a2 l2,
   cci_subtype e Ind{i1; m1 // p1,,(a1,,l1)}
     Ind{i2; m2 // p2,,(a2,,l2)} -> eq_term e a1 a2.
Proof.
intros.
cut
 (sub_hn_inv e Ind{i1; m1 // p1,,(a1,,l1)} Ind{i2; m2 // p2,,(a2,,l2)});
 intros; auto.
inversion_clear H0.
exact (inv_pair_l (inv_pair_r H1)).

apply eq_term_trans with a.
exact (inv_pair_l (inv_pair_r H2)).

red in |- *; apply conv_sym.
exact (inv_pair_l (inv_pair_r H3)).
Qed.




(* Si le sous-typage passe au types inductifs...
  Lemma inv_ind_inst:
    (e:env)(i1,i2:name)(m1,m2:le_marks)(p1,a1,l1,p2,a2,l2:term)
    (par:(cterm O))(arg:(cterm (1)))(knd:sort)(lc:(list constructor))
    (c:constructor)
      (cci_subtype e [: Ind{i1;m1|(p1,(a1,l1))} :]
                     [: Ind{i2;m2|(p2,(a2,l2))} :])
->(select_inddef_cstr sg i1 par arg knd lc)
->(In c lc)
->(arg:term)(eq_term e [: {0:=arg}{1:=p1}!c_inst c! :]
                       [: {0:=arg}{1:=p2}!c_inst c! :]).
*)

End IInvSubtyping.
