(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export TypCtxt.

Set Implicit Arguments.
Unset Strict Implicit.

Inductive is_ind : term -> Set :=
  Is_ind : forall i m p a l, is_ind Ind{i; m // p,,(a,,l)}.
Hint Resolve Is_ind: core.

Lemma is_ind_dec : forall t, is_ind t + {is_ind t -> False}.
Proof.
  refine (fun t =>
    match t with
      | Cst1 (MutInd i m) (p ,, a ,, l) => inleft _ _
      | _ => inright _ _
    end); try (intro eq_ind; inversion eq_ind; fail); [ ].
  auto; fail.
Defined.

Lemma is_untyped_ind_dec : forall t,
  {tr : _ | match tr with
              | ((ind, lm), u) => t = Ind{ind; lm // u}
            end
  } + {(forall ind lm u, t = Ind  {ind; lm // u} -> False)}.
Proof.
  refine (fun t =>
    match t as t0 return
      {tr : _ | match tr with
                  | ((ind, lm), u) => t0 = Ind{ind; lm // u}
                end
      } + {(forall ind lm u, t0 = Ind  {ind; lm // u} -> False)}
    with
      | Cst1 (MutInd i m) u => inleft _ _
      | _ => inright _ _
    end); try (intros; discriminate).
  exists (i, m, u); trivial; fail.
Defined.

Inductive no_marks : term -> Prop :=
  | Nomarks_empty : no_marks []
  | Nomarks_mt : forall t, no_marks t -> no_marks (## && t)
.

Lemma inv_no_marks : forall (P : term -> Prop) a b,
  (no_marks b -> P ##) -> no_marks (a && b) -> P a.
Proof.
  intros.
  inversion H0.
  auto.
Qed.

(* Mark system *)

Inductive incl_seq : le_marks -> term -> le_marks -> term -> Prop :=
  | Incl_refl : forall m t, incl_seq m t m t
  | Incl_trans : forall m1 m2 m3 t1 t2 t3,
      incl_seq m1 t1 m2 t2 -> incl_seq m2 t2 m3 t3 -> incl_seq m1 t1 m3 t3
  | Incl_mt : forall m t, incl_seq m t 0 []
  | Incl_str : forall k ml, incl_seq 0 ml k ml
  | Incl_remove : forall k m ml, incl_seq k (m && ml) (pred k) ml
  | Incl_add_mt : forall k ml, incl_seq (pred k) ml k (## && ml)
  | Incl_deep_le : forall lm lm2 m ml ml2,
      incl_seq lm ml lm2 ml2 ->
      incl_seq (S lm) (m && ml) (S lm2) (m && ml2)
  | Incl_deep_lt : forall m ml ml2,
      incl_seq 0 ml 0 ml2 -> incl_seq 0 (m && ml) 0 (m && ml2)
.

Lemma no_marks_incl : forall k t t', no_marks t' -> incl_seq k t 0 t'.
Proof.
  simple induction 1; intros.
  apply Incl_mt.

  apply Incl_trans with (pred 0) t0; trivial.
  apply Incl_add_mt.
Qed.

Lemma incl_no_marks :
  forall k k' t t', incl_seq k t k' t' -> no_marks t -> no_marks t'.
Proof.
  simple induction 1; intros; auto.
  constructor.

  elim H0 using inv_no_marks; trivial.

  constructor; trivial.

  elim H2 using inv_no_marks; intros.
  constructor; auto.

  elim H2 using inv_no_marks; intros.
  constructor; auto.
Qed.

Section Marks.

  Hint Constructors incl_seq.

  Lemma incl_seq_lift : forall n k la lb a b,
    incl_seq la a lb b -> incl_seq la ^(n//k)a lb ^(n//k)b.
  Proof.
    simple induction 1; simpl in |- *; intros; auto.
    apply Incl_trans with m2 (^(n//k)t2); trivial.
  Qed.

  Lemma incl_seq_subst : forall k la lb u a b,
    incl_seq la a lb b -> incl_seq la a[k:=u] lb b[k:=u].
  Proof.
    simple induction 1; simpl in |- *; intros; auto.
    apply Incl_trans with m2 (t2[k:=u]); trivial.
  Qed.

  Lemma incl_seq_OO :
    forall k k' l l', incl_seq k l k' l' -> incl_seq 0 l 0 l'.
  Proof.
    simple induction 1; intros; try constructor; trivial.
    apply Incl_trans with 0 t2; trivial.

    apply (Incl_add_mt 0 ml).
  Qed.

End Marks.

Inductive ind_marks : red_rule :=
  Mark_intro : forall e x l1 l2 p a m1 m2,
    incl_seq l1 m1 l2 m2 ->
    ind_marks e Ind{x; l1 // p,,(a,,m1)} Ind{x; l2 // p,,(a,,m2)}.

Lemma R_lift_ind_marks : R_lift ind_marks.
Proof.
  red in |- *; intros.
  inversion_clear H.
  simpl in |- *.
  constructor.
  apply incl_seq_lift; trivial.
Qed.

Lemma R_subst_ind_marks : R_subst ind_marks.
Proof.
  red in |- *; intros.
  inversion_clear H.
  simpl in |- *.
  constructor.
  apply incl_seq_subst; trivial.
Qed.

Lemma stable_ind_marks : forall R, R_stable R ind_marks.
Proof.
  red in |- *.
  simple destruct 1; constructor; auto.
Qed.

Definition incl_mark := clos_le ind_marks.

Hint Unfold incl_mark: core.

Lemma incl_mark_lift : R_lift incl_mark.
Proof R_lift_clos R_lift_ind_marks.

Lemma incl_mark_subst : R_subst incl_mark.
Proof R_subst_clos R_subst_ind_marks.

Lemma incl_mark_stable : forall R, R_stable R incl_mark.
Proof fun R : env -> decl -> decl -> Prop => stable_clos (stable_ind_marks (R:=R)).

Lemma mkvect_mark :
  forall (A : Set) e (f1 f2 : A -> term),
    (forall x, incl_mark e (f1 x) (f2 x)) ->
    forall l, incl_mark e (mkvect (map f1 l)) (mkvect (map f2 l)).
Proof.
  unfold incl_mark in |- *.
  simple induction l; simpl in |- *; intros; auto.
Qed.

(* about incl_mark or ind_marks ? *)

Lemma incl_mt_mark : forall e i m p a l,
    incl_mark e Ind{i; m // p,,(a,,l)} Ind{i // p,,(a,,[])}.
Proof.
  intros.
  red in |- *; apply Clos_rule; constructor.
  apply Incl_mt.
Qed.

Lemma incl_rm_mark : forall e i m m1 p a l,
  incl_mark e Ind{i; S m // p,,(a,,m1 && l)} Ind{i; m // p,,(a,,l)}.
Proof.
  intros.
  red in |- *; apply Clos_rule; constructor.
  apply (Incl_remove (S m) m1 l).
Qed.

(* used in fix_exp_incl (for SR fix), and... *)
Lemma incl_str_marks : forall e i m m1 p a l,
  incl_mark e Ind{i // p,,(a,,m1 && l)} Ind{i; m // p,,(a,,l)}.
Proof.
  intros.
  red in |- *; apply Clos_rule; constructor.
  apply Incl_trans with (pred 0) l.
  apply Incl_remove.

  apply Incl_str.
Qed.

(* used in SR fix (the fix can be applied to the body once the mark
   is erased. *)
Lemma incl_add_mt_mark : forall e i m p a l,
  incl_mark e Ind{i; m // p,,(a,,l)} Ind{i; S m // p,,(a,,## && l)}.
Proof.
  intros.
  red in |- *; apply Clos_rule; constructor.
  exact (Incl_add_mt (S m) l).
Qed.

(* used in SR case: the empty mark is the largest mark *)
Lemma incl_mt_highest : forall e i i' m p a p' a' l,
  incl_mark e Ind{i // p,,(a,,[])} Ind{i; m // p,,(a,,l)} ->
  incl_mark e Ind{i' // p',,(a',,[])} Ind{i' // p',,(a',,l)}.
Proof.
  intros.
  inversion_clear H; auto.
  constructor.
  inversion_clear H0.
  constructor; trivial.
  apply no_marks_incl.
  apply incl_no_marks with (1 := H).
  constructor.
Qed.
