(**************************************************************************)
(*                      Certified Extraction in Coq                       *)
(*                                                                        *)
(*                            Stéphane Glondu                             *)
(*                                                                        *)
(*                           © 2007-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

(***** More properties on CCI ****)

Require Export Typechecker.

Set Implicit Arguments.
Unset Strict Implicit.

Open Scope cci_scope.

Ltac invc H := inversion H; subst; clear H.
Ltac inv P :=
  match goal with
    | H : P _ |- _ => inversion H; subst; clear H
    | H : P _ _ |- _ => inversion H; subst; clear H
    | H : P _ _ _ |- _ => inversion H; subst; clear H
    | H : P _ _ _ _ |- _ => inversion H; subst; clear H
    | H : P _ _ _ _ _ |- _ => inversion H; subst; clear H
    | H : P _ _ _ _ _ _ |- _ => inversion H; subst; clear H
    | H : P _ _ _ _ _ _ _ |- _ => inversion H; subst; clear H
  end.
Ltac inv_closed :=
  repeat match goal with
    | H : closed _ (_ _) |- _ => inversion H; subst; clear H
    | H : closed _ (_ _ _) |- _ => inversion H; subst; clear H
    | H : closed _ (_ _ _ _) |- _ => inversion H; subst; clear H
  end.
Ltac inv_typ :=
  repeat match goal with
    | H : _ – _ ⊢ (_ _) : _ |- _ =>
      InvTypHyp H; intros;
      let x := type of H in assert (True -> x) by auto; clear H
    | H : _ – _ ⊢ (_ _ _) : _ |- _ =>
      InvTypHyp H; intros;
      let x := type of H in assert (True -> x) by auto; clear H
    | H : _ – _ ⊢ (_ _ _ _) : _ |- _ =>
      InvTypHyp H; intros;
      let x := type of H in assert (True -> x) by auto; clear H
    | H : in_sign1 _ _ _ (_ _) _ _ |- _ =>
      inversion H; subst;
      let x := type of H in assert (True -> x) by auto; clear H
    | H : in_sign1 _ _ _ (_ _ _) _ _ |- _ =>
      inversion H; subst;
      let x := type of H in assert (True -> x) by auto; clear H
    | H : in_sign1 _ _ _ (_ _ _ _) _ _ |- _ =>
      inversion H; subst;
      let x := type of H in assert (True -> x) by auto; clear H
  end.


(***** Termes sans constante globale *****)

Inductive gclosed : term -> Prop :=
  | gc_Srt : forall s, gclosed (Srt s)
  | gc_Rel : forall n, gclosed (Rel n)
  | gc_Cst0 : forall c, gclosed (Cst0 c)
  | gc_Cst1 : forall c m,
      gclosed m ->
      (forall x, c <> Const x) ->
      gclosed (Cst1 c m)
  | gc_Lam : forall x a m,
      gclosed a -> gclosed m -> gclosed (Lam x a m)
  | gc_Prd : forall x a m,
      gclosed a -> gclosed m -> gclosed (Prd x a m)
  | gc_sum : forall a b,
      gclosed a -> gclosed b -> gclosed (Sum a b)
  | gc_pair : forall m a,
      gclosed m -> gclosed a -> gclosed (Pair0 m a).


Ltac inv_gclosed :=
  repeat match goal with
    | H : gclosed (_ _) |- _ => inversion H; subst; clear H
    | H : gclosed (_ _ _) |- _ => inversion H; subst; clear H
    | H : gclosed (_ _ _ _) |- _ => inversion H; subst; clear H
  end.
Ltac auto_closed :=
  match goal with
    | |- closed _ _ => inv_closed; repeat econstructor; eauto
    | |- gclosed _ => inv_gclosed; repeat econstructor; eauto
    | _ => auto
  end.


Section CicProperties.

  Context {sg : sigma}.
  Context {Hsg : wf_sign sg}.
  Let sign_ok := wf_sign_ok Hsg.
  Existing Instance sign_ok.

Lemma typ_app_arg : forall e x t t' T T',
  sg – e ⊢ (Fun x : T' => t) @ t' : T ->
  sg – e ⊢ t' : T'.
Proof.
  intros.
  InvTypHyp H; intros.
  inv in_sign1.
  inversion H3 using inv_pair_sum; intros.
  InvTypHyp H0; intros.
  InvTypHyp H7; intros.
  simpl in *.
  eapply Typ_conv; eauto.
  simpl; change_notation; eapply inv_prod_l; change_notation; eauto.
Qed.

Lemma sort_is_normal : forall e s,
  head_normal redn_term e %%s.
Proof.
  intros e s t Ht u Htu.
  apply trans_rt1n in Ht.
  invc Ht.
  repeat (invc Htu; [inversion H | rename H into Htu ]); contradiction.
  clear H0 Htu.
  invc H.
  rename H0 into Htu.
  repeat (invc Htu; [inversion H | rename H into Htu ]); contradiction.
Qed.


(***** Stabilité par contexte des réductions *****)

Lemma rt_wctxt_stable : forall e t u,
  wctxt (R_rt (wctxt redn_term)) e t u ->
  R_rt (wctxt redn_term) e t u.
Proof.
  induction 1; auto;
  induction IHwctxt; eauto;
  unfold R_rt in *.
  apply rt_trans with (Cst1 c y); eauto.
  apply rt_trans with (Fun x : y => n); eauto.
  apply rt_trans with (Forall x : y, n); eauto.
  apply rt_trans with (y ** n); eauto.
  apply rt_trans with (t ** y); eauto.
  apply rt_trans with (y ,, n); eauto.
  apply rt_trans with (m ,, y); eauto.
Qed.

Lemma rt_wctxt_stable2 : forall e t u,
  R_rt (wctxt redn_term) e t u ->
  wctxt (R_rt (wctxt redn_term)) e t u.
Proof.
  induction 1.
  do 2 constructor; auto.
  auto.
  constructor.
  hnf; apply rt_trans with y; auto.
Qed.

Lemma t_wctxt_stable  : forall e t u,
  wctxt (R_t (wctxt redn_term)) e t u ->
  R_t (wctxt redn_term) e t u.
Proof.
  induction 1; auto;
  induction IHwctxt; eauto;
  unfold R_t in *; eauto.
  apply t_trans with (Cst1 c y); eauto.
  apply t_trans with (Fun x : y => n); eauto.
  apply t_trans with (Forall x : y, n); eauto.
  apply t_trans with (y ** n); eauto.
  apply t_trans with (t ** y); eauto.
  apply t_trans with (y ,, n); eauto.
  apply t_trans with (m ,, y); eauto.
Qed.

Lemma t_wctxt_stable2 : forall e t u,
  R_t (wctxt redn_term) e t u ->
  wctxt (R_t (wctxt redn_term)) e t u.
Proof.
  induction 1.
  do 2 constructor; auto.
  constructor.
  hnf; apply t_trans with y; auto.
Qed.


Section TypingProperties.

Lemma closed_normal_ind_form :
  forall e t T, sg – e ⊢ t : T ->
    forall I n p a l,
    closed 0 t -> gclosed t ->
    normal (wctxt redn_term) e t ->
    sg – e ⊢ T <= Ind  {I; n // p,, a,, l} ->
    exists C, exists a', t = Cstr {C // p,, a'}.
Proof.
  intros e t T Ht; elim Ht; clear e t T Ht.
  intros e s1 s2 He Hs I n p a l Hc Hg Hn Hle.
    elimtype False.
    admit (* Hle *).
  intros e v t He Hv I n p a l Hc Hg Hn Hle.
    inversion Hc; elimtype False; omega.
  intros e c u He Hu Hsign I n p a l Hc Hg Hn Hle.
    elimtype False.
    inversion Hsign; subst; clear Hsign.
    admit (* Hle *).
    admit (* Hle *).
    admit (* Hle *).
    admit (* Hle *).
  intros e c m t u Hu Ht Hm IH Hsign I n p a l Hc Hg Hn Hle.
    inversion Hsign; subst; clear Hsign.
    elimtype False; admit (* Hle *).
    admit (* needs inversion of sg – e ⊢ t : a ** b *).
    admit (* needs inversion of sg – e ⊢ t : a ** b *).
    admit (* needs inversion of sg – e ⊢ t : Forall x : a, b *).
    inversion Hg; subst; edestruct H2; reflexivity (* buggy gclosed *).
    elimtype False; admit (* Hle *).
    elimtype False; admit (* Hle *).
    elimtype False; admit (* needs inversion of sg – e ⊢ t : record_of sign *).
    elimtype False; admit (* Hle *).
    elimtype False; admit (* Hn? H4? *).
    elimtype False; admit (* Hle *).
    (* p0 ~= p thanks to Hle *)
    replace p0 with p by admit.
    repeat esplit; fail.
  intros e x m t u s Ht IHt Hu IHs I n p a l Hc Hg Hn Hle.
    elimtype False; admit (* Hle *).
  intros e x m b s1 s2 s3 Ha IHa Hb IHb Hs I n p a l Hc Hg Hn Hle.
    elimtype False; admit (* Hle *).
  intros e m b s1 s2 s3 Ha IHa Hb IHb Hs I n p a l Hc Hg Hn Hle.
    elimtype False; admit (* Hle *).
  intros e m t c b s Ha IHa Hb IHb Hab IH I n p a l Hc Hg Hm Hle.
    elimtype False; admit (* Hle *).
  intros e t U V s HU IHU HleU HV IHV I n p a l Hc Hg Hm Hle.
    refine (let HH := IHU I n p a l Hc Hg Hm _ in _).
    apply cci_trans with V; assumption.
    destruct HH as [C [a' HCa']].
    repeat eexists; eassumption.
  intros e t U s Ht IH HleU I n p a l Hc Hg Hn Hle.
    elimtype False; admit (* Hle *).
Qed.

Lemma closed_ind_form_cstr :
  forall (P : Prop) e t I n p a l,
    closed 0 t -> gclosed t ->
    (forall is inds c C a',
      sign_glob_spec sg I (GlobInd inds is) ->
      sign_glob_spec sg C (GlobCstr is c) ->
      R_rt (wctxt redn_term) e t Cstr  {C // p,, a'} ->
      P) ->
    sg – e ⊢ t : Ind  {I; n // p,, a,, l} -> P.
Proof.


(*
 1. définir les termes "bloqués" (variables, constantes non définies,
    et dans les contextes de réduction)
 2. types bloqués habités par termes bloqués
 3. dans un contexte, les types produits (resp. inductifs,
    constructeurs de types sauf sortes) sont habités soit par termes
    bloqués ou lambda (resp. constructeurs, etc.) simultanément
 4. indépendamment, les habitants d'une sorte sont les sortes,
    inductifs, etc.
*)
Admitted.

Lemma closed_ind_form : forall e t I n p a l,
  closed 0 t -> gclosed t ->
  sg – e ⊢ t : Ind{I ; n // p,, a,, l} ->
  exists C, exists a',
  R_rt (wctxt redn_term) e t Cstr{C // p,, a'}.
Proof.
  intros e t I n p a l Hc Hg Ht.
  inversion Ht using closed_ind_form_cstr.
  assumption.
  assumption.
  intros; repeat esplit; eassumption.
Qed.

Lemma fix_types_projn_arg :
  forall e x fixdefs,
  sg – e ⊢ x : fix_types fixdefs ->
  forall n t1 t2 T,
  sg – e ⊢ (x .^ n)..L @ t1 @ t2 : T ->
  exists I, exists n, exists p, exists a, exists l,
  sg – e ⊢ t2 : Ind{I ; n // p,, a,, l}.
Proof.
  unfold fix_types.
  intros e x fixdefs Hx n t1 t2 T Hxn.
  destruct (list_item fixdefs n) as [ [f Hf] | Ha ].
  assert (Hproj : sg – e ⊢ (x .^ n)..L : fix_ty f)
    by eauto using projl_vect, item_map.
  set (y := (x .^ n)..L) in *.
  clearbody y; clear - Hxn Hproj.
  unfold fix_ty in *.
  admit.
  elimtype False.
  admit.
Qed.


Lemma typ_fix_arg : forall e n t t1 t2 T,
  sg – e ⊢ ((Fix {t} .^ n) ..L @ t1) @ t2 : T ->
  exists I, exists n, exists p, exists a, exists l,
  sg – e ⊢ t2 : Ind{I ; n // p,, a,, l}.
Proof.
  intros.
  InvTypHyp H; intros; rename H into Htruc; clear H1 H2.
  inversion H0; subst; clear H0.
  inversion H3 using inv_pair_sum; intros; clear H3.
  rename H0 into Ht2.
  InvTypHyp H; intros; clear H H1 H2.
  inversion H0; subst; clear H0.
  inversion H3 using inv_pair_sum; intros; clear H3.
  rename H0 into Ht1.
  InvTypHyp H; intros; clear H H1 H2.
  inversion H3 using inv_projn; intros; clear H3.
  InvTypHyp H; intros; clear H H3 H7.
  inversion H2; subst; rename H2 into Htruc2.
  inversion H8 using inv_pair_sum; intros.
  inversion H8.

  simple eapply fix_types_projn_arg.
  simple eapply Typ_cst1.
  simple eapply Wf_cons_var.
  simple eapply typ_wf; eassumption.
  eassumption.
  simple eapply Wf_cons_var.
  simple eapply typ_wf; eassumption.
  eassumption.
  eassumption.
  eassumption.
  eassumption.

  simple eapply fix_types_projn_arg.
  simple eapply Typ_cst1.
  simple eapply Wf_cons_var.
  simple eapply typ_wf; eassumption.
  eassumption.
  simple eapply Wf_cons_var.
  simple eapply typ_wf; eassumption.
  eassumption.
  eassumption.
  eassumption.
  eassumption.
Qed.

(* lemme d'inversion *)


Lemma subjred_wctxt : rule_sound (wctxt redn_term).
Proof.
  hnf; intros.
  assert (H100 : red redn_term e x y)
   by (apply rt_step; apply wctxt_ctxt; auto).
  assert (H101 := subject_reduction H100).
  auto.
Qed.

Lemma subjred_rt_wctxt : forall e t u T,
  sg – e ⊢ t : T -> R_rt (wctxt redn_term) e t u ->
  sg – e ⊢ u : T.
Proof.
  induction 2; eauto using subjred_wctxt.
Qed.

Lemma subjred_t_wctxt : forall e t u T,
  sg – e ⊢ t : T -> R_t (wctxt redn_term) e t u ->
  sg – e ⊢ u : T.
Proof.
  induction 2; eauto using subjred_wctxt.
Qed.

Lemma gclosed_stable_wctxt : forall e t u,
  gclosed t -> wctxt redn_term e t u ->
  gclosed u.
Admitted. (* facile / si t contient une sorte, alors tous les t' la contiennent -> adapter aux variables *)

Lemma gclosed_stable_rt_wctxt : forall e t u,
  gclosed t -> R_rt (wctxt redn_term) e t u ->
  gclosed u.
Proof.
  induction 2; eauto using gclosed_stable_wctxt.
Qed.

Lemma gclosed_stable_t_wctxt : forall e t u,
  gclosed t -> R_t (wctxt redn_term) e t u ->
  gclosed u.
Proof.
  induction 2; eauto using gclosed_stable_wctxt.
Qed.

Lemma closed_stable_wctxt : forall e t u n,
  closed n t -> wctxt redn_term e t u ->
  closed n u.
Admitted.

Lemma closed_stable_rt_wctxt : forall e t u n,
  closed n t -> R_rt (wctxt redn_term) e t u ->
  closed n u.
Proof.
  induction 2; eauto using closed_stable_wctxt.
Qed.

Lemma closed_stable_t_wctxt : forall e t u n,
  closed n t -> R_t (wctxt redn_term) e t u ->
  closed n u.
Proof.
  induction 2; eauto using closed_stable_wctxt.
Qed.

End TypingProperties.


Section SubtypingProperties.

Lemma hn_ind_inv_left : forall e u n a I P,
  wf_type e u ->
  head_normal redn_term e u ->
  (forall n' p' a' l', u = Ind { I ; n' // p',, a',, l' } -> P) ->
  sg – e ⊢ u <= Ind { I ; n // a } ->
  P.
Proof.
  intros e u.
  generalize (is_ind_dec u).
  intros [ [i m p a' l] | not_ind ]; intros n a I P Htu Hnu Hinv Hle.
  replace I with i in Hinv; eauto using inv_ind_i.
  elimtype False.
  assert (sub_hn_inv e u Ind { I ; n // a })
    by (eapply inv_cumul_hn; eauto).
  inv sub_hn_inv; inv wf_type; inv_typ; inv in_sign1; auto.
Qed.

Lemma hn_ind_inv_right : forall e u n a I (P : Prop),
  wf_type e u ->
  head_normal redn_term e u ->
  (forall n' p' a' l', u = Ind { I ; n' // p',, a',, l' } -> P) ->
  sg – e ⊢ Ind { I ; n // a } <= u ->
  P.
Proof.
  intros e u.
  generalize (is_ind_dec u).
  intros [ [i m p a' l] | not_ind ]; intros n a I P Htu Hnu Hinv Hle.
  eapply Hinv.
  replace i with I; eauto using inv_ind_i.
  elimtype False.
  assert (sub_hn_inv e Ind {I ; n // a } u)
    by (eapply inv_cumul_hn; eauto).
  inv sub_hn_inv; inv wf_type; inv_typ; inv in_sign1; auto.
Qed.

Lemma Prop_is_minimal : forall e s,
  sg – e ⊢ %%s <= %%(Sprop Neg) -> s = Sprop Neg.
Proof.
  intros.
  assert (H100 := inv_sort H).
  apply trans_rtn1 in H100.
  remember (Sprop Neg) as s'.
  revert Heqs' H.
  induction H100; intros; subst; auto.
  inv incl_sort.
  auto.
Qed.

End SubtypingProperties.


(***** Normalisation forte *****)

Lemma wctxt_is_strongly_normalizing : forall e t T,
  sg – e ⊢ t : T ->
  sn (wctxt redn_term) e t.
Proof.
  intros e t T Htyp.
  assert (Ht := cci_is_strongly_normalizing_axiom Hsg Htyp).
  induction Ht as [ t _ Hrec ].
  constructor; intros u Htu.
  apply Hrec.
  apply wctxt_ctxt; auto.
  hnf in Htu.
  eapply subjred_wctxt; eauto.
Qed.

End CicProperties.

(* Élimine un sous-typage contradictoire *)
Ltac contradict_le :=
  match goal with
    H : (?e ⊦ ?t <= ?u)%pts, sg : sigma |- _ =>
      change (sg – e ⊢ t <= u) in H;
      let H100 := fresh in
      assert (H100 : sub_hn_inv e t u)
        by (eapply inv_cumul_hn; eauto);
      invc H100;
      fail
  end.
