(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export CoreDec.
Require Export SignCorrectness.

Set Implicit Arguments.
Unset Strict Implicit.

  Section IRecordDecOk.

    Context {sg : sigma}.
    Context {sgn_ok : sign_ok}.

    Lemma red_to_one :
      forall e t,
        ⊦ (e; x_ : t) -> {red redn_term e t one} + {~ eq_term e t one}.
    Proof.
      (*
       Realizer [e,t](term_eq (whnf_cci e t) (Cst0 one)).
       *)
      intros.
      elim whnf_cci with e t; intros.
      elim eq_dec_term with x (one:term); intros.
      left.
      elim a; trivial.

      right; red in |- *; intros.
      apply b.
      apply eq_one_red with (1 := p); auto.

      apply var_wft; trivial.
    Qed.


    Lemma dest_one_field :
      forall e t,
        ⊦ (e; x_ : t) ->
          {tr : _ |
            match tr with
              | (x, (u, v)) =>
                exists2 u' : _,
                eq_term e u u' & red redn_term e t (u ** (Forall x : u', v))
            end} + {(forall x u v, ~ eq_term e t (u ** (Forall x : u, v)))}.
    Proof.
      (*Realizer [e,t]
        Cases (red_to_sum_dec e t) of
          (inleft(u,a)) =>
            Cases (red_to_prd_dec e a) of
              (inleft(x,(u',v))) =>
                if (convert e u u') then (inleft ? (x,(u,v)))
                else (inright ?)
            | _ => (inright ?)
            end
        | _ => (inright ?)
        end.*)
      intros e t wfet.
      assert (wft := var_wft _ _ wfet).
      case red_to_sum_dec with e t; trivial.
      intros ((u, a), red_t).
      assert (wfea : ⊦ (e; x_ : a)).
      Inversion_typ (subj_red_wf wfet red_t).
      Inversion_typ H.
      apply wf_var with (1 := H1).

      case red_to_prd_dec with e a.
      Inversion_typ wfea.
      left with s; trivial.

      intros ((x, (u', v)), red_a).
      assert (wf_prd := subj_red_wf wfea red_a).
      case (convert ok_gs_inj) with e u u'.
      apply ok_normalize.
      apply var_wft.
      Inversion_typ (subj_red_wf wfet red_t).
      Inversion_typ H.
      apply wf_var with (1 := H0).

      apply ok_normalize.
      Inversion_typ wf_prd.
      Inversion_typ H.
      left with s1; trivial.

      intros conv_u_u'.
      left; exists (x, (u, v)).
      exists u'; trivial.
      apply red_trans with (u ** a); auto.

      intros nconv_u_u'.
      right; red in |- *; intros y u'' v'' conv_t; apply nconv_u_u'.
      assert (conv_sum := eq_term_trans (eq_term_sym (redn_conv red_t)) conv_t).
      apply eq_term_trans with (1 := inv_sum_eq_l conv_sum).
      assert
        (conv_prd :=
          eq_term_trans (eq_term_sym (inv_sum_eq_r conv_sum)) (redn_conv red_a)).
      apply (inv_prd_eq_l conv_prd).

      intros not_prod.
      right; red in |- *; intros y u' v' conv_t; apply not_prod with y u' v'.
      apply eq_term_cci_sub.
      assert (conv_prd := eq_term_trans (eq_term_sym (redn_conv red_t)) conv_t).
      apply (inv_sum_eq_r conv_prd).

      intros not_sum.
      right; red in |- *; intros x u v conv_t;
        apply not_sum with u (Forall x : u, v).
      simple apply eq_term_cci_sub; exact conv_t.
    Defined.

    Lemma dest_all_fields :
      forall flds e m,
        ⊦ (e; x_ : m) ->
          {x : _ | eq_term e m (record_arity x) /\ ⊦ (e; x_ : record_arity x) &
            flds = field_names x} +
          {~
            (exists2 sign : _,
              eq_term e m (record_arity sign) & flds = field_names sign)}.
    Proof.
      (*Realizer Fix dest_flds{ dest_flds [l:?]: ? :=
        [e;m]Cases l of
          nil => if (red_to_one e m) then (inleft ? (nil ?)::rspec)
                 else (inright ?)
        | (cons a l') =>
            Cases (dest_one_field e m) of
              (inleft(x,(u,v))) =>
            	  Cases (dest_flds l' (add_var e x u) v) of
                  (inleft sign) => (inleft ? (cons ? (a,u) sign))
                | _ => (inright ?)
                end
            | inright => (inright ?)
            end
        end}.*)
      fix dest_flds 1.
      intros flds e m wfem.
      case flds.
      case red_to_one with e m; trivial.
      intros red_m.
      left; exists (nil:rspec); simpl in |- *; trivial.
      split; auto.
      apply wf_var with so.
      apply typ_one_so.
      apply (inv_wf _ _ wfem).

      intros neq_one.
      right; red in |- *; intros ([| rs l], eq_m, eq_f); apply neq_one; trivial;
        discriminate.

      intros a flds'.
      case dest_one_field with e m; trivial.
      intros ((x, (u, v)), fld_eqn).
      case dest_flds with flds' (e; x : u) v.
      case fld_eqn; intros u' eq_u red_m.
      assert (wf_sum := subj_red_wf wfem red_m).
      Inversion_typ wf_sum.
      Inversion_typ H.
      Inversion_typ H1.
      apply wf_var with s4.
      apply subtype_env with (1 := H5); auto.
      unfold add_var in |- *.
      constructor.
      split; simpl in |- *; auto.
      change_notation; simple apply eq_term_cci_sub; exact eq_u.

      exact (wf_var _ x _ _ H0).

      intros (sign, eq_v, eq_fld).
      left; exists ((a, u) :: sign).
      case fld_eqn; intros u' eq_u red_m.
      case eq_v; clear eq_v; intros eq_v wf_ar.
      simpl in |- *; split.
      red in |- *.
      apply conv_trans with (1 := redn_conv red_m).
      apply conv_conv_sum; auto.
      apply conv_trans with (Forall x : u, v); auto.
      apply conv_trans with (Forall x : u, record_arity sign); auto.
      apply red_conv.
      apply alpha_red_term.
      repeat constructor.

      apply all_prod_sum.
      Inversion_typ (inv_wf _ _ wf_ar).
      apply wf_var with s; trivial.

      apply all_rules; trivial.
      Inversion_typ wf_ar.
      apply wf_var with s.
      apply subtype_env with (1 := H); auto.
      unfold add_var in |- *.
      constructor.
      split; simpl in |- *; auto.
      change_notation; auto.

      Inversion_typ (inv_wf _ _ wf_ar).
      apply wf_var with s0; trivial.

      rewrite eq_fld; trivial.

      intros not_field_tail.
      right; red in |- *; intros ([| (y, t) sign], eq_m, eq_flds);
        apply not_field_tail; try discriminate.
      case fld_eqn; intros u' eq_u red_m.
      exists sign.
      simpl in eq_m.
      assert
        (eq_sum :
          eq_term e (u ** (Forall x : u, v)) (t ** (Forall y : t, record_arity sign))).
      apply
        eq_term_trans with (2 := eq_term_trans (eq_term_sym (redn_conv red_m)) eq_m);
        red in |- *; auto.

      exact (inv_prd_eq_r (inv_sum_eq_r eq_sum)).

      injection eq_flds; trivial.

      intros not_field; red in not_field.
      right; red in |- *; intros ([| (z, ty) sign], eq_m, eq_flds);
        try discriminate.
      apply not_field with z ty (record_arity sign); auto.
    Defined.

    Lemma red_to_record :
      forall e t,
        ⊦ (e; x_ : t) ->
          {sign : _ | eq_term e t (record_of sign) /\ ⊦ (e; x_ : record_of sign)} +
          {(forall sign, ~ e ⊦ t <= record_of sign)}.
    Proof.
      intros.
      elim whnf_cci with e t.
      intros hnft.
      elim (is_record hnft);
        [ intros ((f, arg), eqty) | RightStr err_not_record; red in |- *; intros ].
      rewrite eqty; clear eqty hnft; intros.
      elim dest_all_fields with f e arg;
        [ intros (sign, (eqarg, wfarg), eqf)
          | RightStr err_not_record; red in |- *; intros
          | idtac ].
      left.
      split with sign.
      split.
      rewrite eqf in p; clear eqf q f.
      apply eq_term_trans with (1 := redn_conv p).
      red in |- *; unfold record_of in |- *; auto.

      apply typ_record.
      trivial.

      Inversion_typ H.
      rewrite eqf in p.
      Inversion_cst (subject_reduction p H0); trivial.

      Inversion_typ H.
      apply b.
      Inversion_cst (subject_reduction p H1); intros.
      split with sign0; trivial.

      Inversion_typ H.
      Inversion_cst (subject_reduction p H0); intros.
      apply wf_var with s0; trivial.

      Inversion_typ H.
      cut (sub_hn_inv e hnft (record_of sign)); unfold record_of in |- *; intros.
      generalize b.
      inversion_clear H2.
      intros.
      elim b0 with (field_names sign, a); trivial.

      intros.
      elim b0 with (field_names s1, x); trivial.

      apply inv_sub_hn; auto.
      apply cci_trans with (2 := H0); auto.

      apply var_wft; trivial.
    Qed.



    Lemma sign_dec_record :
      forall flds e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat (Record flds) e m t u)
          (cci_subtype e).
    Proof.
      (*
      Realizer [flds:(list name); e:env; m,t:term]
        Cases (red_to_sort_dec e t) of
          (inleft s) =>
            Cases (dest_all_fields flds e m) of
              (inleft sign) =>
                if (list_unique_dec eq_name flds) then (inleft ? (Srt s))
                else (inright_str term err_field_name_clash)
            | _ => (inright_str term err_no_record_arity)
        | inright => (inright term)
        end.
      *)
      intros.
      elim red_to_sort_dec with e t; intros.
      inversion_clear a.
      elim dest_all_fields with (flds := flds) (e := e) (m := m); intros.
      elim (list_unique_dec (@eq_dec name _)) with flds; intros.
      left.
      split with (%%x).
      inversion_clear a.
      rewrite H2.
      split; intros.
      split with (%%x); simpl in |- *.
      split; auto.
      change_notation; simple apply eq_term_cci_sub; simple apply redn_conv; exact H0.
      apply (subj_red_wf (forget_def _ _ _ _ H) H0).

      split.
      constructor; trivial.
      elim H2; trivial.

      elim H1; trivial.

      apply (subj_red_wf (forget_def _ _ _ _ H) H0).

      inversion_clear H3.
      simpl in H4.
      inversion_clear H5.
      generalize H4.
      inversion_clear H3; intros.
      inversion_clear H3.
      apply cci_trans with (2 := H8); auto.

      RightStr err_field_name_clash; red in |- *; intros.
      apply b.
      InvSignSat H1 ipattern:ty; trivial.

      RightStr err_no_record_arity; red in |- *; intros.
      apply b.
      InvSignSat H1 ipattern:ty; intros.
      split with sign; trivial.

      apply wf_var with x.
      Inversion_typ H.
      apply subj_red_ty with (1 := H1); trivial.

      right; red in |- *; intros.
      InvSignSat H0 ipattern:ty; intros.
      elim b with s; trivial.

      eapply var_wft, forget_def; eassumption.
    Qed.




    Lemma coerce_to_field_types :
      forall e md t,
        ⊦ (e; x_ : prod_arity (rspec_of md)) ->
          e ⊦ vals_of md : t ->
            decide
            (e ⊦ t <= mkvect (fields_type md 0) /\
              ⊦ (e; x_ : mkvect (fields_type md 0))).
    Proof.
      (*
      Realizer Fix cft { cft [e:env; md:mspec]: term->bool :=
        [t]Case md of
          nil => (subtype_dec e t (Cst0 one))
        | (cons ((x,ty),v) l) =>
            Cases (red_to_sum_dec e t) of
              (inleft (a,b)) =>
      	  if (subtype_dec e a ty) then (cft e (subst_m2 v md O) b)
                else false
            | _ => false
        end}.
      *)
      intros e md.
      elim md using list_length_rec; intros.
      generalize l H0 H1 H2; clear H0 H1 H2 l; intros [| ((x, ty), v) sign];
        simpl in |- *; intros.
      elim subtype_dec with e t (one:term); intros; auto.
      right; red in |- *; intros; apply b.
      inversion H3; trivial.

      eapply ok_gs_inj; eassumption.
      apply ok_normalize.
      apply (type_correctness _ _ _ H2).

      apply ok_normalize.
      apply var_wft; trivial.

      fold rspec_of in H1.
      elim red_to_sum_dec with e t;
        [ intros ((a, b), redt) | right; red in |- *; intros | idtac ].
      elim subtype_dec with e a ty; intros; try solve [ eapply ok_gs_inj; eassumption ].
      elim H with (n := length sign) (l := subst_msign2 v sign 0) (t := b); intros.
      left.
      inversion_clear a1.
      rewrite subst2_fields.
      rewrite lift0.
      split.
      apply cci_trans with (1 := eq_term_cci_sub (redn_conv redt)).
      auto.

      apply all_prod_sum; auto.
      Inversion_typ H1.
      Inversion_typ H5.
      apply wf_var with s1; trivial.

      right; red in |- *; intros.
      apply b0.
      inversion_clear H3.
      split.
      generalize (inv_sum_r (cci_trans (eq_term_sym_cci_sub (redn_conv redt)) H4)).
      rewrite subst2_fields; trivial.
      rewrite lift0.
      trivial.

      Inversion_typ H5.
      Inversion_typ H3.
      rewrite <- subst2_fields.
      generalize H7.
      rewrite lift0; intros.
      apply wf_var with s2; trivial.

      elim H0.
      auto.

      apply length_subst2.

      rewrite rspec_subst2.
      rewrite <- prod_arity_subst.
      Inversion_typ H1.
      Inversion_typ H3.
      apply wf_var with s2.
      fold (%%s2[0:=v]) in |- *.
      apply substitution with x ty.
      trivial.

      unfold vals_of in H2; simpl in H2.
      Inversion_typ H2.
      apply Typ_conv with a1 s1; trivial.
      simpl in |- *.
      change_notation; apply cci_trans with a; trivial.
      apply (inv_sum_l (cci_trans H11 (eq_term_cci_sub (redn_conv redt)))).

      rewrite vals_subst2.
      unfold vals_of in H2; simpl in H2.
      Inversion_typ H2.
      apply typ_conv_wf with b0; trivial.
      generalize (type_correctness _ _ _ (subj_red_ty H2 redt)); intros.
      inversion_clear H7.
      Inversion_typ H8.
      left with s2; trivial.

      simpl in |- *.
      apply (inv_sum_r (cci_trans H6 (eq_term_cci_sub (redn_conv redt)))).

      right; red in |- *; intros.
      apply b0.
      inversion_clear H3.
      apply (inv_sum_l (cci_trans (eq_term_sym_cci_sub (redn_conv redt)) H4)).

      generalize (type_correctness _ _ _ (subj_red_ty H2 redt)); intros.
      inversion_clear H3.
      Inversion_typ H4.
      apply ok_normalize.
      left with s1; trivial.

      apply ok_normalize.
      Inversion_typ H1.
      Inversion_typ H3.
      left with s1; trivial.

      inversion_clear H3; eauto.
      red in b; eapply b; change_notation; eauto.

      apply (type_correctness _ _ _ H2).
    Qed.



    Lemma sign_dec_struct :
      forall flds e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat (Struct flds) e m t u)
          (cci_subtype e).
    Proof.
      (*
      Realizer [flds:(list name); e:env; m,t:term]
        Cases (dest_cst_pair e m t) of
          (inleft ((ar,kar),(vec,tyvec))) =>
            Cases (red_to_sort_dec e kar) of
              (inleft s) =>
                Cases (is_prod_arity flds ar) of
                  (inleft sign) =>
                    Cases (collect_values sign vec) of
                      (inleft md) =>
                        if (coerce_to_field_types e md tyvec O) then
      		    if (list_unique_dec ? eq_name flds) then
                            (inleft ? (record_of (rspec_of md)))
                          else (inright term)
                        else (inright term)
                    | _ => (inright term)
                    end
                | _ => (inright term)
                end
            | _ => (inright term)
        | _ => (inright term)
        end.
      *)
      intros.
      elim dest_cst_pair with e m t;
        [ intros (((ar, kar), (vec, tyvec)), (eq_m, (red_t, (wf_ar, wf_vec))))
          | Right_Neg
          | trivial ].
      rewrite eq_m; clear eq_m H m.
      elim red_to_sort_dec with e kar;
        [ intros (s, le_s) | right; red in |- *; intros | idtac ].
      elim is_prod_arity with flds ar;
        [ intros (sign, (eq_flds, eq_ar)) | right; red in |- *; intros ].
      rewrite eq_flds; clear eq_flds flds.
      rewrite eq_ar; rewrite eq_ar in wf_ar; clear eq_ar ar.
      elim collect_values with sign vec;
        [ intros (md, (eq_sign, eq_vec)) | right; red in |- *; intros ].
      rewrite eq_sign; rewrite eq_sign in wf_ar; clear eq_sign sign.
      rewrite eq_vec; rewrite eq_vec in wf_vec; clear eq_vec vec.
      elim coerce_to_field_types with e md tyvec;
        [ intros (le_tyvec, wf_tyvec) | Right_Neg | idtac | idtac ].
      elim (list_unique_dec (@eq_dec name _)) with (fields_of md);
        [ intros flds_unique | Right_Neg ].
      left.
      split with (record_of (rspec_of md)).
      cut (e ⊦ vals_of md : mkvect (fields_type md 0)); intros.
      Inversion_typ wf_ar.
      generalize (Typ_conv_srt _ _ _ _ H0 (eq_term_cci_sub (redn_conv le_s))); intros.
      generalize (wf_def _ x_ _ _ _ H2 (subject_reduction le_s H1));
        clear H2 H1 H0 s0 wf_ar; intro wf_ar.
      split; intros.
      split with (%%s ** mkvect (fields_type md 0)); simpl in |- *.
      split.
      change_notation; apply cci_trans with (kar ** tyvec); auto.

      apply all_prod_sum.
      apply (forget_def _ _ _ _ wf_ar).

      apply (mkvect_wf H).

      split.
      red in |- *.
      fold (fields_of md) (struct_of md) in |- *.
      constructor.

      apply typ_record; trivial.
      apply record_arity_ok.
      apply wf_var with s.
      Inversion_typ wf_ar; trivial.

      generalize (refl_equal (fields_of md)); unfold fields_of at 2 in |- *.
      generalize (refl_equal (prod_arity (rspec_of md)));
        unfold rspec_of at 1 in |- *.
      InvSignSat H0 ipattern:ty; unfold struct_of in |- *.
      intros.
      unfold fields_of in |- *.
      apply record_sub_arity.
      elim (rspec_inj H2).
      unfold rspec_of in |- *.
      apply sub_record_refl.

      apply typ_conv_wf with tyvec; auto.
      Inversion_typ wf_vec; trivial.

      apply var_wft.
      trivial.

      unfold fields_of in |- *.
      InvSignSat H ipattern:ty; unfold record_of, struct_of in |- *; intros.
      unfold fields_of in |- *.
      Inversion_typ WFUU.
      Inversion_cst H1; trivial.

      generalize (mspec_inj (s1:=md)).
      InvSignSat H ipattern:ty; unfold struct_of in |- *; intros.
      rewrite H1 with md0; trivial.
      split.
      apply (inv_sum_r (cci_trans (eq_term_sym_cci_sub (redn_conv red_t)) H)).

      Inversion_typ H0.
      Inversion_typ H2.
      apply wf_var with s3; trivial.

      apply wf_var with s.
      apply Typ_conv_srt with kar; simpl in |- *; auto.
      2: change_notation; simple apply eq_term_cci_sub; simple apply redn_conv; exact le_s.
      Inversion_typ wf_ar; trivial.

      Inversion_typ wf_vec; trivial.

      generalize b (rspec_inj (s1:=sign)).
      InvSignSat H ipattern:ty; unfold struct_of in |- *; intros.
      elim b0 with md; auto.

      generalize b.
      InvSignSat H ipattern:ty; unfold struct_of in |- *; intros.
      elim b0 with (rspec_of md); auto.

      InvSignSat H ipattern:ty; unfold struct_of in |- *; intros.
      elim b with s.
      apply (inv_sum_l (cci_trans (eq_term_sym_cci_sub (redn_conv red_t)) H)).

      eapply var_wft, forget_def; eassumption.

      InvSignSat H0 ipattern:ty; unfold struct_of in |- *; intros.
      prolog [ ex_intro ex_intro2 refl_equal ] 10.
    Qed.





    Lemma search_field :
      forall f m sign k,
        {fty : _ |
          exists ty : _,
            (exists2 n : _,
              item (f, ty) sign n & item fty (fields_type (expand_sign m sign) k) n)} +
        {~ In f (field_names sign)}.
    Proof.
      (*
      Realizer Fix ftype { ftype [f:name; m:term; sign:rspec]: nat->(sumor term) :=
        [k]Cases sign of
             nil => (inright term)
           | (cons (x,ty) l) =>
             	 if (eq_name f x) then (inleft ? ty)
               else
                 Cases (ftype f m l (S k)) of
                   (inleft u) => (inleft ? [: {0:=^(k)m.x}u :])
                 | _ => (inright term)
                 end
           end}.
      *)
      simple induction sign.
      right; red in |- *; intros.
      inversion H.

      intros (x, ty) l Hrec k.
      elim (eq_dec f x); intros.
      left.
      elim a.
      split with ty.
      split with ty.
      split with 0; simpl in |- *; auto.

      generalize (Hrec (S k)); intros [(u, is_fty)| no_fld].
      left.
      split with (u[0:=^(k)m..( x)]).
      generalize is_fty; clear is_fty.
      intros (tyn, (n, itm_sign, itm_fld)).
      split with tyn.
      split with (S n); auto.
      simpl in |- *.
      constructor.
      apply item_map' with u; auto.

      right; red in |- *; intros.
      generalize b no_fld.
      inversion_clear H; auto.
    Qed.


    Lemma sign_dec_field :
      forall f e m t,
        ⊦ (e; x_ := m : t) ->
          ppal_dec (fun u => in_sign_sat (Field f) e m t u) (cci_subtype e).
    Proof.
      (*Realizer [f,e,m,t]
        Cases (red_to_record e t) of
          (inleft sign) =>
            Cases (search_field f m sign O) of
              (inleft ty) => (inleft ? ty)
            | _ => (inright_str term err_unknown_field)
            end
        | _ => (inright term)
        end.*)
      intros f e m t wfemt.
      case red_to_record with e t.
      apply (forget_def _ _ _ _ wfemt).

      intros (rs, (eq_rec, wf_rs)).
      case (search_field f m rs 0).
      intros (fty, is_fty).
      left; exists fty.
      generalize is_fty; clear is_fty; intros (ty, (n, itm_sign, itm_fld)).
      split; intros.
      split with (record_of rs); simpl in |- *.
      split; auto.
      change_notation; simple apply eq_term_cci_sub; exact eq_rec.

      split.
      red in |- *.
      apply Mem_field with n; trivial.
      unfold field_names in |- *.
      apply item_map' with (2 := itm_sign); auto.

      apply wf_mkvect_comp with (1 := itm_fld).
      apply wf_fields.
      apply typ_conv_wf with t; auto.
      Inversion_typ wfemt; trivial.

      apply var_wft; trivial.

      simpl in |- *; auto.
      change_notation; simple apply eq_term_cci_sub; exact eq_rec.

      InvSignSat H ipattern:ty'; intros.
      generalize (item_map (fst (B:=_)) itm_sign).
      simpl in |- *.
      fold field_names in |- *.
      assert (sub_rec : sub_record e rs sign).
      apply inv_record_sub_arity.
      apply cci_trans with t; auto.

      rewrite (same_names sub_rec).
      intro.
      generalize H0.
      replace n0 with n; intros.
      apply inv_mkvect with (2 := itm_fld) (3 := H4).
      apply sub_arity_field; trivial.

      assert (fld_unique : list_unique (field_names sign)).
      elim (same_names sub_rec).
      Inversion_typ wf_rs.
      unfold record_of in H4.
      Inversion_cst H4; trivial.

      apply
        first_item_unique
        with
          (1 := list_unique_first H fld_unique)
          (2 := list_unique_first H3 fld_unique).

      intros no_such_field.
      right; red in |- *; intros u in_sgn.
      InvSignSat in_sgn ipattern:ty; intros.
      apply no_such_field.
      assert (sub_rec : sub_record e rs sign).
      apply inv_record_sub_arity.
      apply cci_trans with t; auto.

      rewrite (same_names sub_rec).
      elim H; auto.

      intros not_record.
      right; red in |- *; intros u in_sgn.
      InvSignSat in_sgn ipattern:ty; intros.
      elim not_record with sign; trivial.
    Defined.

  End IRecordDecOk.
