(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export Vect.

Set Implicit Arguments.
Unset Strict Implicit.

Lemma is_record : forall t,
  dec_find (fun p : _ * _ => let (flds, arg) := p in t = {<flds : arg>}).
Proof.
  refine (fun t =>
    match t as t0 return
      dec_find (fun p : _ * _ => let (flds, arg) := p in t0 = {<flds : arg>})
    with
      | Cst1 (Record flds) arg => inleft _ _
      | _ => inright _ _
    end); try (red; intros (fld, ar) eq_record; discriminate); [ ].
  exists (flds, arg); reflexivity.
Defined.

Definition rspec := list (name * term).

Definition field_names : rspec -> list name := map (fst (B:=_)).

Fixpoint record_arity (l : rspec) : term :=
  match l with
    | nil => one
    | (x, t) :: l' => t ** (Forall x : t, record_arity l')
  end.

Definition record_of (sign : rspec) : term :=
  {<field_names sign : record_arity sign>}.

Fixpoint prod_arity (l : rspec) : term :=
  match l with
    | nil => one
    | (x, t) :: l' => Forall x : t, prod_arity l'
  end.

Lemma rspec_inj : forall s1 s2, prod_arity s1 = prod_arity s2 -> s1 = s2.
Proof.
  induction s1 as [ | [x t] ]; destruct s2 as [ | [y u] ]; simpl; intro He.
  reflexivity.
  discriminate.
  discriminate.
  injection He; intros; subst y u; f_equal; auto; fail.
Qed.

Lemma rspec_record_inj :
  forall s1 s2, record_arity s1 = record_arity s2 -> s1 = s2.
Proof.
  induction s1 as [ | [x t] ]; destruct s2 as [ | [y u] ]; simpl; intro He.
  reflexivity.
  discriminate.
  discriminate.
  injection He; intros; subst y u; f_equal; auto; fail.
Qed.

Lemma inj_record_of : forall s1 s2, record_of s1 = record_of s2 -> s1 = s2.
Proof.
  unfold record_of.
  intros s1 s2 He; injection He; intros Hs1 Hs2.
  apply rspec_inj.
  rewrite (rspec_record_inj Hs1); reflexivity.
Qed.


Lemma is_record_arity : forall l t,
  dec_find (fun sign => l = field_names sign /\ t = record_arity sign).
Proof.
  simple induction l; intros.
  elim (t == one); intros.
  left.
  split with (nil (A:=name * term)); auto.

  right; red in |- *; intros.
  apply b.
  inversion_clear H.
  generalize H0 H1.
  generalize x.
  intros [| (yy, ty) sign]; simpl in |- *; intros; auto.
  discriminate H2.

  elim (is_sum_dec t); intros.
  elim a0; intros ll rr.
  elim (is_prd_dec rr); intros.
  elim a1; intros x aa bb.
  elim (a == x); intros.
  elim (ll == aa); intros.
  elim H with bb; intros.
  left.
  inversion_clear a4.
  split with ((a, ll) :: x0).
  inversion_clear H0.
  rewrite H1; rewrite H2.
  split; trivial.
  elim a3; elim a2; simpl in |- *; auto.

  right; red in |- *.
  intros [| (xx, ty) ll0]; simpl in |- *; intros.
  inversion_clear H0.
  discriminate H1.

  inversion_clear H0.
  elim b with ll0.
  inversion H1; inversion H2; auto.

  right; red in |- *.
  intros [| (xx, ty) ll0]; simpl in |- *; intros.
  inversion_clear H0.
  discriminate H1.

  apply b.
  inversion_clear H0.
  inversion H2; trivial.

  right; red in |- *.
  intros [| (xx, ty) ll0]; simpl in |- *; intros.
  inversion_clear H0.
  discriminate H1.

  inversion_clear H0.
  apply b.
  inversion H2.
  injection H1; trivial.

  right; red in |- *.
  intros [| (xx, ty) ll0]; simpl in |- *; intros.
  inversion_clear H0.
  discriminate H1.

  inversion_clear H0.
  generalize b.
  inversion H2; auto.

  right; red in |- *.
  intros [| (xx, ty) ll0]; simpl in |- *; intros.
  inversion_clear H0.
  discriminate H1.

  inversion_clear H0.
  generalize b.
  inversion H2; auto.
Qed.

Definition mspec := list (name * term * term).
Definition rspec_of : mspec -> rspec := map (fst (B:=_)).
Definition fields_of m := field_names (rspec_of m).

Definition vals_of (m : mspec) : term := mktuple (map (snd (B:=_)) m).

Definition struct_of md :=
  {<fields_of md : prod_arity (rspec_of md) := vals_of md>}.

Lemma mspec_inj : forall s1 s2,
  prod_arity (rspec_of s1) = prod_arity (rspec_of s2) ->
  vals_of s1 = vals_of s2 -> s1 = s2.
Proof.
  simple induction s1; simple destruct s2; simpl in |- *; trivial.
  intros.
  discriminate H0.

  intros.
  discriminate H1.

  generalize a; intros ((x, ty), v) ((x', ty'), v'); simpl in |- *; intros.
  unfold vals_of in H1.
  simpl in H1.
  injection H0; intros.
  injection H1; intros.
  elim H4; elim H3; elim H6.
  elim H with l0; trivial.
Qed.

Fixpoint fields_type (m : mspec) (k : nat) {struct m} : list term :=
  match m with
    | nil => nil
    | ((x, ty), v) :: l =>
      ty :: map (fun a : term => a[0:=^(k)v]) (fields_type l (S k))
  end.

Definition expand_sign (m : term) : rspec -> mspec :=
  map (fun p => (p, m..( fst p))).

(* Lifting and subtitution... *)

Fixpoint lift_rsign (n : nat) (l : rspec) (k : nat) {struct l} : rspec :=
  match l with
    | nil => nil
    | (x, t) :: l' => (x, ^(n//k)t) :: lift_rsign n l' (S k)
  end.

Fixpoint lift_msign (n : nat) (l : mspec) (k k' : nat) {struct l} : mspec :=
  match l with
    | nil => nil
    | ((x, t), v) :: l' =>
        (x, ^(n//k)t, ^(n//k')v) :: lift_msign n l' (S k) k'
  end.

Lemma record_arity_lift :
  forall n l k, ^(n//k)(record_arity l) = record_arity (lift_rsign n l k).
Proof.
  simple induction l; simpl in |- *; intros; trivial.
  case a; simpl in |- *; intros.
  rewrite H; trivial.
Qed.

Lemma prod_arity_lift :
  forall n l k, ^(n//k)(prod_arity l) = prod_arity (lift_rsign n l k).
Proof.
  simple induction l; simpl in |- *; intros; trivial.
  case a; simpl in |- *; intros.
  rewrite H; trivial.
Qed.

Lemma record_names_lift :
  forall n l k, field_names (lift_rsign n l k) = field_names l.
Proof.
  unfold field_names in |- *; simple induction l; simpl in |- *; intros;
    trivial.
  case a; simpl in |- *; intros.
  rewrite H; trivial.
Qed.

Lemma record_of_lift :
  forall (n : nat) (s : rspec) (k : nat),
    ^(n//k)(record_of s) = record_of (lift_rsign n s k).
Proof.
  intros; unfold record_of in |- *; simpl in |- *.
  rewrite record_names_lift.
  rewrite record_arity_lift.
  reflexivity.
Qed.

Lemma fields_type_lift : forall n md k k',
  map (fun t : term => ^(n//k' + k)t) (fields_type md k') =
  fields_type (lift_msign n md (k' + k) k) k'.
Proof.
  simple induction md; simpl in |- *; intros; trivial.
  generalize a; intros ((x, ty), v); simpl in |- *.
  rewrite map_map.
  replace (S (k' + k)) with (S k' + k); trivial.
  rewrite <- H.
  rewrite map_map.
  apply f_equal with (f := cons ^(n//k' + k)ty).
  apply map_ext; intros.
  rewrite distr_lift_subst.
  rewrite <- permute_lift.
  trivial.
Qed.

Lemma rspec_of_lift : forall n md k k',
  lift_rsign n (rspec_of md) k = rspec_of (lift_msign n md k k').
Proof.
  simple induction md; simpl in |- *; intros; auto.
  generalize a; intros ((x, ty), v); simpl in |- *.
  unfold rspec_of in |- *; simpl in |- *; fold rspec_of in |- *.
  rewrite <- H; trivial.
Qed.

Lemma vals_of_lift :
  forall n md k k', ^(n//k)(vals_of md) = vals_of (lift_msign n md k' k).
Proof.
  unfold vals_of in |- *.
  simple induction md; simpl in |- *; intros; auto.
  generalize a; intros ((x, ty), v); simpl in |- *.
  rewrite <- H; trivial.
Qed.

Lemma fields_of_lift :
  forall n md k k', fields_of md = fields_of (lift_msign n md k k').
Proof.
  unfold fields_of in |- *.
  intros.
  rewrite <- rspec_of_lift.
  rewrite record_names_lift; trivial.
Qed.

Lemma expand_sign_lift : forall n m sign k k',
  lift_msign n (expand_sign m sign) k k' =
  expand_sign ^(n//k')m (lift_rsign n sign k).
Proof.
  unfold expand_sign in |- *.
  simple induction sign; intros; auto.
  generalize a; intros (x, ty); simpl in |- *.
  rewrite H; trivial.
Qed.

Fixpoint subst_rsign (s : term) (l : rspec) (k : nat) {struct l} : rspec :=
  match l with
    | nil => nil
    | (x, t) :: l' => (x, t[k:=s]) :: subst_rsign s l' (S k)
  end.

Fixpoint subst_msign (s : term) (l : mspec) (k k' : nat) {struct l} : mspec :=
  match l with
    | nil => nil
    | ((x, t), v) :: l' =>
      (x, t[k:=s], v[k':=s]) :: subst_msign s l' (S k) k'
  end.

Lemma record_arity_subst :
  forall s l k, (record_arity l)[k:=s] = record_arity (subst_rsign s l k).
Proof.
  simple induction l; simpl in |- *; intros; trivial.
  case a; intros.
  simpl in |- *.
  rewrite H; trivial.
Qed.

Lemma prod_arity_subst :
  forall s l k, (prod_arity l)[k:=s] = prod_arity (subst_rsign s l k).
Proof.
  simple induction l; simpl in |- *; intros; trivial.
  case a; intros.
  simpl in |- *.
  rewrite H; trivial.
Qed.

Lemma record_names_subst :
  forall s l k, field_names (subst_rsign s l k) = field_names l.
Proof.
  unfold field_names in |- *; simple induction l; simpl in |- *; intros;
    trivial.
  case a; simpl in |- *; intros.
  rewrite H; trivial.
Qed.

Lemma record_of_subst :
  forall m s k, (record_of s)[k:=m] = record_of (subst_rsign m s k).
Proof.
  intros; unfold record_of in |- *; simpl in |- *.
  rewrite record_names_subst.
  rewrite record_arity_subst.
  reflexivity.
Qed.

Lemma fields_type_subst : forall s md k k',
  map (fun t : term => t[k' + k:=s]) (fields_type md k') =
  fields_type (subst_msign s md (k' + k) k) k'.
Proof.
  simple induction md; simpl in |- *; intros; trivial.
  generalize a; intros ((x, ty), v); simpl in |- *.
  rewrite map_map.
  replace (S (k' + k)) with (S k' + k); trivial.
  rewrite <- H.
  rewrite map_map.
  apply f_equal with (f := cons ty[k' + k:=s]).
  apply map_ext; intros.
  rewrite distr_subst.
  rewrite <- commut_lift_subst.
  trivial.
Qed.

Lemma rspec_of_subst : forall s md k k',
  subst_rsign s (rspec_of md) k = rspec_of (subst_msign s md k k').
Proof.
  simple induction md; simpl in |- *; intros; auto.
  generalize a; intros ((x, ty), v); simpl in |- *.
  unfold rspec_of in |- *; simpl in |- *; fold rspec_of in |- *.
  rewrite <- H; trivial.
Qed.

Lemma vals_of_subst :
  forall s md k k', (vals_of md)[k:=s] = vals_of (subst_msign s md k' k).
Proof.
  unfold vals_of in |- *.
  simple induction md; simpl in |- *; intros; auto.
  generalize a; intros ((x, ty), v); simpl in |- *.
  rewrite <- H; trivial.
Qed.

Lemma fields_of_subst :
  forall s md k k', fields_of md = fields_of (subst_msign s md k k').
Proof.
  unfold fields_of in |- *.
  intros.
  rewrite <- rspec_of_subst.
  rewrite record_names_subst; trivial.
Qed.

Lemma expand_sign_subst : forall s m sign k k',
  subst_msign s (expand_sign m sign) k k' =
  expand_sign m[k':=s] (subst_rsign s sign k).
Proof.
  unfold expand_sign in |- *.
  simple induction sign; intros; auto.
  generalize a; intros (x, ty); simpl in |- *.
  rewrite H; trivial.
Qed.

Lemma subst_fields_expand : forall s m sign k,
  map (fun a : term => a[k:=s]) (fields_type (expand_sign m sign) (S k)) =
  fields_type (expand_sign m (subst_rsign s sign k)) k.
Proof.
  simple induction sign.
  simpl in |- *; trivial.

  intros (x, ty) l Hrec k.
  simpl in |- *.
  apply f_equal with (f := cons ty[k:=s]).
  rewrite map_map.
  fold (expand_sign m (subst_rsign s l (S k))) (expand_sign m l) in |- *.
  elim Hrec with (S k).
  rewrite map_map.
  apply map_ext; intros.
  rewrite distr_subst.
  rewrite simpl_subst; auto.
Qed.

Lemma subst_length :
  forall s sign k, length (subst_rsign s sign k) = length sign.
Proof.
  simple induction sign; simpl in |- *; intros; auto with arith.
  case a; simpl in |- *; auto with arith.
Qed.

Fixpoint subst_msign2 (s : term) (l : mspec) (k : nat) {struct l} : mspec :=
  match l with
    | nil => nil
    | ((x, t), v) :: l' => (x, t[k:=s], v) :: subst_msign2 s l' (S k)
  end.

Lemma subst2_fields : forall s md k,
  map (fun a : term => a[k:=s]) (fields_type md (S k)) =
  fields_type (subst_msign2 s md k) k.
Proof.
  simple induction md.
  simpl in |- *; trivial.

  intros ((x, ty), v) l Hrec k.
  simpl in |- *.
  apply f_equal with (f := cons ty[k:=s]).
  rewrite map_map.
  elim Hrec with (S k).
  rewrite map_map.
  apply map_ext; intros.
  rewrite distr_subst.
  rewrite simpl_subst; auto.
Qed.

Lemma length_subst2 :
  forall v sign k, length (subst_msign2 v sign k) = length sign.
Proof.
  simple induction sign; simpl in |- *; auto with arith.
  intros ((x, ty), w); simpl in |- *; intros; auto with arith.
Qed.

Lemma rspec_subst2 : forall s md k,
  rspec_of (subst_msign2 s md k) = subst_rsign s (rspec_of md) k.
Proof.
  simple induction md; simpl in |- *; auto.
  intros ((x, ty), v); simpl in |- *; intros.
  unfold rspec_of in |- *; simpl in |- *.
  fold rspec_of in |- *.
  rewrite H; trivial.
Qed.

Lemma vals_subst2 :
  forall s md k, vals_of (subst_msign2 s md k) = vals_of md.
Proof.
  unfold vals_of in |- *.
  simple induction md; simpl in |- *; auto.
  intros ((x, ty), v); simpl in |- *; intros.
  rewrite H; trivial.
Qed.

Lemma exp_subst2 : forall s m sign k,
  subst_msign2 s (expand_sign m sign) k =
  expand_sign m (subst_rsign s sign k).
Proof.
  simple induction sign.
  simpl in |- *; trivial.

  intros (x, ty) l Hrec k.
  simpl in |- *.
  fold (expand_sign m l) in |- *.
  rewrite Hrec; trivial.
Qed.

Lemma is_prod_arity : forall l t,
    dec_find (fun sign => l = field_names sign /\ t = prod_arity sign).
Proof.
  (*
  Realizer Fix is_pa { is_pa [f:(list name)]: term->(sumor rspec) :=
    [t]Cases f of
      nil => if (term_eq t (Cst0 one)) then (inleft ? (nil ?))
             else (inright rspec)
    | (cons x l) =>
        Cases (is_prd_dec t) of
          (inleft (Is_prd y a b)) =>
              Cases (is_pa l b) of
                (inleft sign) =>
  	  if (eq_name x y) then (inleft ? (cons ? (x,a) sign))
            else (inright rspec)
              | _ => (inright rspec)
              end
        | _ => (inright rspec)
        end
    end}.
  *)
  simple induction l; intros.
  elim (t == one); intros.
  left.
  split with (nil (A:=name * term)); auto.

  right; red in |- *; intros.
  apply b.
  inversion_clear H.
  generalize H0 H1.
  generalize x.
  intros [| (yy, ty) sign]; simpl in |- *; intros; auto.
  discriminate H2.

  elim (is_prd_dec t); intros.
  elim a0; intros x aa bb.
  elim (a == x); intros.
  elim H with bb; intros.
  left.
  inversion_clear a2.
  split with ((a, aa) :: x0).
  inversion_clear H0.
  rewrite H1; rewrite H2.
  split; trivial.
  simpl in |- *.
  elim a1; trivial.

  right; red in |- *.
  intros [| (xx, ty) ll]; simpl in |- *; intros.
  inversion_clear H0.
  discriminate H1.

  inversion_clear H0.
  elim b with ll.
  injection H1; injection H2; trivial.
  intros.
  rewrite H5; rewrite H0; trivial.
  auto.

  right; red in |- *.
  intros [| (xx, ty) ll]; simpl in |- *; intros.
  inversion_clear H0.
  discriminate H1.

  inversion_clear H0.
  apply b.
  injection H2; intros.
  rewrite H4.
  injection H1; trivial.

  right; red in |- *.
  intros [| (xx, ty) ll0]; simpl in |- *; intros.
  inversion_clear H0.
  discriminate H1.

  inversion_clear H0.
  apply b.
  rewrite H2; trivial.
Qed.

Lemma collect_values :
  forall sign t, dec_find (fun md => sign = rspec_of md /\ t = vals_of md).
Proof.
  (*Realizer Fix is_va { is_va [r:?]: ? :=
    [t]Cases r of
      nil => if (term_eq t (Cst0 unit)) then (inleft ? (nil ?))
             else (inright mspec)
    | (cons x l) =>
        Cases (is_pair_dec t) of
          (inleft (a,b)) =>
              Cases (is_va l b) of
                (inleft md) => (inleft ? (cons ? (x,a) md))
              | _ => (inright mspec)
              end
        | _ => (inright mspec)
        end
    end}.*)
  fix is_va 1.
  intros sign t.
  case sign.
  case (t == unit).
  intro eq_unit.
  left; exists (nil (A:=name * term * term)); simpl in |- *; auto.

  intros neq_unit.
  unfold vals_of in |- *.
  right; red in |- *; intros [| ((xx, ty), vv) ll]; simpl in |- *;
    intros (eq_nil, eq_t); discriminate || auto.

  intros fs l.
  case (is_pair_dec t).
  intros ((a, b), t_pair); subst t.
  case (is_va l b).
  intros (md, (eq_l, eq_b)); subst l b.
  left; exists ((fs, a) :: md); unfold rspec_of in |- *; simpl in |- *; auto.

  unfold vals_of in |- *.
  intros not_spec; red in not_spec.
  right; red in |- *; intros [| ((xx, ty), vv) ll]; simpl in |- *;
    intros (eq_rs, eq_v);
      discriminate || (injection eq_rs; injection eq_v; intros; subst l b; eauto).

  intros not_pair; red in not_pair.
  unfold vals_of in |- *.
  right; red in |- *; intros [| ((xx, ty), vv) ll]; simpl in |- *;
    intros (eq_nil, eq_t); discriminate || eauto.
Defined.
