(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export InvSubtyping.
Require Import MlPrint.

Set Implicit Arguments.
Unset Strict Implicit.

Section IReductionDecA.

  Context {sg : sigma}.

  Section Whnf_function.

    (* Auxiliary functions *)

    Lemma delta_db : forall e n, dec_find (fun t => delta e #n t).
    Proof.
      fix delta_db 1.
      intros [| dcl e].
      right; red in |- *; intros t red_t.
      inversion_clear red_t.
      inversion H.

      case dcl; intros x ty obd [| n].
      case obd.
      intros bd.
      left; split with (^(1)bd).
      eapply delta_intro; eauto.

      right; red in |- *; intros t red_t; inversion_clear red_t.
      inversion_clear H.

      case (delta_db e n).
      intros (t, red_t).
      left; exists (^(1)t).
      inversion_clear red_t.
      rewrite <- simpl_lift.
      eapply delta_intro; eauto.

      intros no_red; right; red in |- *; intros t red_t.
      inversion_clear red_t.
      inversion_clear H.
      red in no_red; eapply no_red; eapply delta_intro; red in |- *; eauto.
    Defined.

    Hypothesis sign_ok_glob_inj :
      forall x g1 g2,
        sign_glob_spec sg x g1 -> sign_glob_spec sg x g2 -> g1 = g2.

    Lemma delta_cst :
      forall e c u, dec_find (fun t => delta_glob sg e (Cst1 (Const c) u) t).
    Proof.
      intros e c u.
      elim (search_const c).
      intros (co, srch_co).
      generalize (refl_equal (BODY (c_decl co))).
      pattern (BODY (c_decl co)) at 2 in |- *.
      case (BODY (c_decl co)).
      intros v eq_b.
      left; exists (v[0:=u]).
      eapply deltag_intro; eauto.

      intro no_body.
      right; red in |- *; intros t red_t.
      inversion_clear red_t.
      assert (same_g := sign_ok_glob_inj srch_co H).
      injection same_g; intro same_co.
      rewrite <- same_co in H0.
      rewrite no_body in H0; discriminate.

      intro not_found; red in not_found.
      right; red in |- *; intros t red_t.
      inversion_clear red_t.
      eauto.
    Defined.

    Lemma reduce_fix :
      forall (e : env) (t : term),
        sn (ctxt redn_term) e t ->
        (forall f u,
          clos_refl_trans _ (ord_norm1 redn_term redn_term) (f, u) (e, t) ->
          {v : term | red redn_term f u v &  head_normal redn_term f v}) ->
        {n : _ &
          {rfix : _ |
            exists2 efx : _, red redn_term e t (efx .^ n) & fix_free efx rfix}} +
        {ht : _ | red redn_term e t ht &  hn_mutfix e ht}.
    Proof.
      intros e t sn_t.
      elim (sn_ord_norm _ _ (R_stable_redn (R:=redn_term)) _ _ sn_t) using Acc2_rec.
      clear e t sn_t; intros e t acc_t reduce_fix whnf_cci.
      elim whnf_cci with e t; auto.
      intros ht red_t hn_t.
      destruct ht as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ right; exists (%%s)
        | right; exists (#n)
        | right; exists (Cst0 c)
        | idtac
        | right; exists (Forall n : t0, t1)
        | right; exists (Fun n : t0 => t1)
        | right; exists (t0 ** t1)
        | right; exists (t0,,t1)
        ]; auto 10.
      destruct c;
        [ right; exists (Cst1 Mark t0)
        | right; exists (Cst1 Mt_mark t0)
        | right; exists (Cst1 Lmark t0)
        | right; exists (Cst1 Nil_mark t0)
        | right; exists (Cst1 Cons_mark t0)
        | right; exists (Cst1 App t0)
        | right; exists t0..L
        | idtac
        | right; exists (Cst1 (Struct l) t0)
        | right; exists {<l : t0>}
        | right; exists t0..( n)
        | right; exists (Cst1 (Const n) t0)
        | right; exists Ind{n; l // t0}
        | right; exists Cstr{n // t0}
        | right; exists (Cst1 (MutCase b p) t0)
        | idtac
        ]; auto 10.
      (*..R *)
      elim reduce_fix with e t0; eauto.
      intros (n, (rfx, reducible)).
      left; exists (S n); exists rfx.
      destruct reducible.
      exists x; trivial.
      apply red_trans with t0..R; auto.
      simpl in |- *; replace (x..R .^ n) with (x .^ n) ..R; auto.

      generalize x; elim n; simpl in |- *; auto.

      intros (ht, red_ht, hn_ht).
      right; exists ht..R; auto.
      apply red_trans with t0..R; auto.

      split; trivial.
      eapply hn_red_hn; eauto.

      intros.
      apply whnf_cci.
      apply rt_trans with (e, t0); eauto 10.
      apply clos_t_clos_rt.
      fold (ord_norm redn_term redn_term) in |- *.
      eauto 10.

      (* fix *)
      rename t0 into fxp.
      elim whnf_cci with e fxp; eauto 20.
      intros fxph red_fxp hn_fx.
      assert (red_tfxp : red redn_term e t (Fix {fxph})).
      apply red_trans with (Fix {fxp}); auto.

      destruct fxph as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ right; exists (Fix {%%s})
        | right; exists (Fix {#n})
        | right; exists (Fix {Cst0 c})
        | right; exists (Fix {Cst1 c t0})
        | right; exists (Fix {Forall n : t0, t1})
        | right; exists (Fix {Fun n : t0 => t1})
        | right; exists (Fix {t0 ** t1})
        | idtac
        ]; auto 10.
      left; exists 0; simpl in |- *.
      exists ((t1 @ ##) @ (Fix {(t0,,t1)})).
      exists (Fix {(t0,,t1)}); auto.

      apply clos_t_clos_rt.
      fold (ord_norm redn_term redn_term) in |- *.
      eauto 10.
    Defined.

    Lemma red_red_proj :
      forall (R : red_rule) e n a a', red R e a a' -> red R e (a .^ n) (a' .^ n).
    Proof.
      simple induction n; simpl in |- *; intros; auto.
    Qed.

    Hint Resolve red_red_proj: core.

    Lemma whnf_cci_sn : forall (e : env) (t : term),
      sn (ctxt redn_term) e t ->
      {u : term | red redn_term e t u &  head_normal redn_term e u}.
    Proof.
      intros e t sn_t.
      elim (sn_ord_norm _ _ (R_stable_redn (R:=redn_term)) _ _ sn_t) using Acc2_rec.
      clear e t sn_t; intros e t acc_t whnf_cci_sn.
      destruct t as [s| n| c| c t| n t t0| n t t0| t t0| t t0];
        [ exists (%%s); auto
        | idtac
        | exists (Cst0 c); auto
        | idtac
        | exists (Forall n : t, t0); auto
        | exists (Fun n : t => t0); auto
        | exists (t ** t0); auto
        | exists (t,,t0); auto
        ].

      (* Case of Rel *)
      case (delta_db e n).
      intros (v, red_v).
      elim whnf_cci_sn with e v; auto 20.
      intros v' red_v' hn_v'.
      exists v'; trivial.
      apply red_trans with v; auto 20.

      intros no_red.
      exists (#n); auto.

      (* Case of Cst1 *)
      destruct c;
        [ exists (Cst1 Mark t)
        | exists (Cst1 Mt_mark t)
        | exists (Cst1 Lmark t)
        | exists (Cst1 Nil_mark t)
        | exists (Cst1 Cons_mark t)
        | idtac
        | idtac
        | idtac
        | exists (Cst1 (Struct l) t)
        | exists {<l : t>}
        | idtac
        | idtac
        | exists Ind{n; l // t}
        | exists Cstr{n // t}
        | idtac
        | exists (Fix {t})
        ]; auto.

      (* Case of (Cst1 App _) *)
      elim whnf_cci_sn with e t; eauto 20.
      intros ht red_t hn_t.
      destruct ht as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists (Cst1 App %%s)
        | exists (Cst1 App #n)
        | exists (Cst1 App (Cst0 c))
        | exists (Cst1 App (Cst1 c t0))
        | exists (Cst1 App (Forall n : t0, t1))
        | exists (Cst1 App (Fun n : t0 => t1))
        | exists (Cst1 App (t0 ** t1))
        | idtac
        ]; auto 10.
      rename t0 into u; rename t1 into v.
      elim whnf_cci_sn with e u; eauto 20.
      intros hu red_u hn_hu.
      assert (red_tu : red redn_term e (Cst1 App t) (hu @ v)).
      apply red_trans with (u @ v); auto 10.

      destruct hu as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists (%%s @ v)
        | exists (#n @ v)
        | exists (Cst0 c @ v)
        | idtac
        | exists ((Forall n : t0, t1) @ v)
        | idtac
        | exists ((t0 ** t1) @ v)
        | exists ((t0,,t1) @ v)
        ]; auto 10.
      destruct c;
        [ exists (Cst1 Mark t0 @ v)
        | exists (Cst1 Mt_mark t0 @ v)
        | exists (Cst1 Lmark t0 @ v)
        | exists (Cst1 Nil_mark t0 @ v)
        | exists (Cst1 Cons_mark t0 @ v)
        | idtac
        | exists (t0..L @ v)
        | exists (t0..R @ v)
        | exists (Cst1 (Struct l) t0 @ v)
        | exists ({<l : t0>} @ v)
        | exists (t0..( n) @ v)
        | exists (Cst1 (Const n) t0 @ v)
        | exists (Ind{n; l // t0} @ v)
        | exists (Cstr{n // t0} @ v)
        | exists (Cst1 (MutCase b p) t0 @ v)
        | exists ((Fix {t0}) @ v)
        ]; auto 10.
      rename t0 into pr.
      elim whnf_cci_sn with e pr; eauto 20.
      intros hpr red_pr hn_pr.
      assert (red_tpr : red redn_term e (Cst1 App t) (Cst1 App hpr @ v)).
      apply red_trans with (Cst1 App pr @ v); auto 10.

      destruct hpr as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists (Cst1 App %%s @ v)
        | exists (Cst1 App #n @ v)
        | exists (Cst1 App (Cst0 c) @ v)
        | exists (Cst1 App (Cst1 c t0) @ v)
        | exists (Cst1 App (Forall n : t0, t1) @ v)
        | exists (Cst1 App (Fun n : t0 => t1) @ v)
        | exists (Cst1 App (t0 ** t1) @ v)
        | idtac
        ]; auto 10.
      rename t0 into a; rename t1 into par.
      elim whnf_cci_sn with e v; eauto 20.
      intros hv red_v hn_v.
      assert (red_tc : red redn_term e (Cst1 App t) ((a @ par) @ hv)).
      apply red_trans with (1 := red_tpr); auto 10.

      assert (hn_apar : head_normal redn_term e (a @ par)).
      apply hn_red_hn with (Cst1 App pr); auto 10.
      destruct hv as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists ((a @ par) @ %%s)
        | exists ((a @ par) @ #n)
        | exists ((a @ par) @ Cst0 c)
        | idtac
        | exists ((a @ par) @ (Forall n : t0, t1))
        | exists ((a @ par) @ (Fun n : t0 => t1))
        | exists ((a @ par) @ (t0 ** t1))
        | exists ((a @ par) @ (t0,,t1))
        ]; auto 20.
      destruct c;
        [ exists ((a @ par) @ Cst1 Mark t0)
        | exists ((a @ par) @ Cst1 Mt_mark t0)
        | exists ((a @ par) @ Cst1 Lmark t0)
        | exists ((a @ par) @ Cst1 Nil_mark t0)
        | exists ((a @ par) @ Cst1 Cons_mark t0)
        | exists ((a @ par) @ Cst1 App t0)
        | exists ((a @ par) @ t0..L)
        | exists ((a @ par) @ t0..R)
        | exists ((a @ par) @ Cst1 (Struct l) t0)
        | exists ((a @ par) @ {<l : t0>})
        | exists ((a @ par) @ t0..( n))
        | exists ((a @ par) @ Cst1 (Const n) t0)
        | exists ((a @ par) @ Ind{n; l // t0})
        | idtac
        | exists ((a @ par) @ Cst1 (MutCase b p) t0)
        | exists ((a @ par) @ (Fix {t0}))
        ]; auto 10.
      rename t0 into carg; rename n into cn.
      elim whnf_cci_sn with e a; eauto 20.
      intros ha red_a hn_a.
      assert
        (red_ta : red redn_term e (Cst1 App t) ((ha @ par) @ Cstr{cn // carg})).
      apply red_trans with ((a @ par) @ Cstr{cn // carg}); auto 10.

      destruct ha as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists ((%%s @ par) @ Cstr{cn // carg})
        | exists ((#n @ par) @ Cstr{cn // carg})
        | exists ((Cst0 c @ par) @ Cstr{cn // carg})
        | idtac
        | exists (((Forall n : t0, t1) @ par) @ Cstr{cn // carg})
        | idtac
        | exists (((t0 ** t1) @ par) @ Cstr{cn // carg})
        | exists (((t0,,t1) @ par) @ Cstr{cn // carg})
        ]; auto 20.
      (* Cst1 *)
      assert (hn_hd : head_normal redn_term e (Cst1 c t0 @ par)).
      apply hn_red_hn with (a @ par); auto.

      destruct c;
        [ exists ((Cst1 Mark t0 @ par) @ Cstr{cn // carg})
        | exists ((Cst1 Mt_mark t0 @ par) @ Cstr{cn // carg})
        | exists ((Cst1 Lmark t0 @ par) @ Cstr{cn // carg})
        | exists ((Cst1 Nil_mark t0 @ par) @ Cstr{cn // carg})
        | exists ((Cst1 Cons_mark t0 @ par) @ Cstr{cn // carg})
        | exists ((Cst1 App t0 @ par) @ Cstr{cn // carg})
        | idtac
        | exists ((t0..R @ par) @ Cstr{cn // carg})
        | exists ((Cst1 (Struct l) t0 @ par) @ Cstr{cn // carg})
        | exists (({<l : t0>} @ par) @ Cstr{cn // carg})
        | exists ((t0..( n) @ par) @ Cstr{cn // carg})
        | exists ((Cst1 (Const n) t0 @ par) @ Cstr{cn // carg})
        | exists ((Ind{n; l // t0} @ par) @ Cstr{cn // carg})
        | exists ((Cstr{n // t0} @ par) @ Cstr{cn // carg})
        | exists ((Cst1 (MutCase b p) t0 @ par) @ Cstr{cn // carg})
        | exists (((Fix {t0}) @ par) @ Cstr{cn // carg})
        ]; auto 20.
      rename t0 into fx.
      elim reduce_fix with e fx.
      intros (n, (rfix, reducible)).
      elim whnf_cci_sn with e (((rfix .^ n) ..L @ par) @ Cstr{cn // carg}).
      intros hnf red_hnf id_hnf.
      exists hnf; trivial.
      destruct reducible.
      apply red_trans with ((fx..L @ par) @ Cstr{cn // carg}); auto 10.
      apply red_trans with (((x .^ n) ..L @ par) @ Cstr{cn // carg}); auto 20.
      apply red_trans with (((rfix .^ n) ..L @ par) @ Cstr{cn // carg}); auto 20.

      destruct reducible.
      apply ord_norm_step with (((x .^ n) ..L @ par) @ Cstr{cn // carg}); auto 20.
      apply red_trans with (1 := red_ta); auto 20.

      intros (rfx, red_rfx, hnf_fix).
      exists ((rfx..L @ par) @ Cstr{cn // carg}); auto 20.
      apply red_trans with (1 := red_ta); auto 20.

      apply hn_subterm_hn.
      split; auto.
      split; auto.
      apply hn_red_hn with (fx..L @ par); auto 10.

      split; auto.
      split; auto.
      split; auto.
      apply hn_red_hn with fx..L; auto.

      apply subterm_sn with redn_term e fx..L; auto.
      apply R_stable_redn.

      apply sn_red_sn with (1 := red_a).
      apply ord_norm_sn with redn_term.
      apply Acc_inv with (e, Cst1 App t); trivial.
      red in |- *; apply t_trans with (e, pr);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
      red in |- *; apply t_trans with (e, u);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
      red in |- *; apply t_trans with (e, t);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.

      intros.
      apply whnf_cci_sn.
      red in |- *; apply step_clos_trans with (e, fx); trivial.
      apply t_trans with (e, a); fold (ord_norm redn_term redn_term) in |- *;
        eauto 10.
      red in |- *; apply t_trans with (e, pr);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
      red in |- *; apply t_trans with (e, u);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
      red in |- *; apply t_trans with (e, t);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.

      (* Abs *)
      assert (abs : head_normal redn_term e ((Fun n : t0 => t1) @ par)).
      apply hn_red_hn with (a @ par); auto 10.
      elim abs with ((Fun n : t0 => t1) @ par) (t1[0:=par]); auto 20.

      red in |- *; apply t_trans with (e, pr);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
      red in |- *; apply t_trans with (e, u);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
      red in |- *; apply t_trans with (e, t);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.

      red in |- *; apply t_trans with (e, t);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.

      red in |- *; apply t_trans with (e, u);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
      red in |- *; apply t_trans with (e, t);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.

      (* Case of (fun x:ty => m) @ v *)
      rename t0 into ty; rename t1 into m.
      elim whnf_cci_sn with e (m[0:=v]); auto 10.
      intros ht red_ht hn_ht.
      exists ht; trivial.
      apply red_trans with ((Fun n : ty => m) @ v); auto 10.
      apply red_trans with (m[0:=v]); auto 20.

      apply ord_norm_step with ((Fun n : ty => m) @ v); auto 20.

      red in |- *; apply t_trans with (e, u,,v);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.

      (* Case of t..L *)
      elim whnf_cci_sn with e t; eauto 20.
      intros v red_v hn_v.
      destruct v as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists (%%s) ..L
        | exists (#n) ..L
        | exists (Cst0 c) ..L
        | exists (Cst1 c t0) ..L
        | exists (Forall n : t0, t1) ..L
        | exists (Fun n : t0 => t1) ..L
        | exists (t0 ** t1) ..L
        | idtac
        ]; auto 10.
      elim whnf_cci_sn with e t0; eauto 20.
      intros u red_t0 hn_u.
      exists u; trivial.
      apply red_trans with (t0,,t1) ..L; auto.
      apply red_trans with t0; auto 20.

      (* Case of t..R *)
      elim whnf_cci_sn with e t; eauto 20.
      intros v red_v hn_v.
      destruct v as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists (%%s) ..R
        | exists (#n) ..R
        | exists (Cst0 c) ..R
        | exists (Cst1 c t0) ..R
        | exists (Forall n : t0, t1) ..R
        | exists (Fun n : t0 => t1) ..R
        | exists (t0 ** t1) ..R
        | idtac
        ]; auto 10.
      elim whnf_cci_sn with e t1; eauto 20.
      intros u red_t1 hn_u.
      exists u; trivial.
      apply red_trans with (t0,,t1) ..R; auto.
      apply red_trans with t1; auto 20.


      (* Case of t.(n) *)
      elim whnf_cci_sn with e t; eauto 20.
      intros v red_v hn_v.
      destruct v as [s| n0| c| c t0| n0 t0 t1| n0 t0 t1| t0 t1| t0 t1];
        [ exists (%%s) ..( n)
        | exists (#n0) ..( n)
        | exists (Cst0 c) ..( n)
        | idtac
        | exists (Forall n0 : t0, t1) ..( n)
        | exists (Fun n0 : t0 => t1) ..( n)
        | exists (t0 ** t1) ..( n)
        | exists (t0,,t1) ..( n)
        ]; auto 10.
      destruct c;
        [ exists (Cst1 Mark t0) ..( n)
        | exists (Cst1 Mt_mark t0) ..( n)
        | exists (Cst1 Lmark t0) ..( n)
        | exists (Cst1 Nil_mark t0) ..( n)
        | exists (Cst1 Cons_mark t0) ..( n)
        | exists (Cst1 App t0) ..( n)
        | exists t0..L..( n)
        | exists t0..R..( n)
        | idtac
        | exists {<l : t0>} ..( n)
        | exists t0..( n0) ..( n)
        | exists (Cst1 (Const n0) t0) ..( n)
        | exists Ind{n0; l // t0} ..( n)
        | exists Cstr{n0 // t0} ..( n)
        | exists (Cst1 (MutCase b p) t0) ..( n)
        | exists (Fix {t0}) ..( n)
        ]; auto 10.
      elim whnf_cci_sn with e t0; eauto 20.
      intros w red_w hn_w.
      assert (red_tw : red redn_term e t..( n) (Cst1 (Struct l) w) ..( n)).
      apply red_trans with (Cst1 (Struct l) t0) ..( n); auto.
      destruct w as [s| n0| c| c t1| n0 t1 t2| n0 t1 t2| t1 t2| t1 t2];
        [ exists (Cst1 (Struct l) %%s) ..( n)
        | exists (Cst1 (Struct l) #n0) ..( n)
        | exists (Cst1 (Struct l) (Cst0 c)) ..( n)
        | exists (Cst1 (Struct l) (Cst1 c t1)) ..( n)
        | exists (Cst1 (Struct l) (Forall n0 : t1, t2)) ..( n)
        | exists (Cst1 (Struct l) (Fun n0 : t1 => t2)) ..( n)
        | exists (Cst1 (Struct l) (t1 ** t2)) ..( n)
        | idtac
        ]; auto 10.
      case (list_index eq_dec n l).
      intros (rnk, is_rnk).
      elim whnf_cci_sn with e (t2 .^ rnk) ..L; eauto 20.
      intros hf red_hf hn_hf.
      exists hf; trivial.
      apply red_trans with {<l : t1 := t2>} ..( n); auto.
      apply red_trans with (t2 .^ rnk) ..L; auto 20.

      intros no_fld.
      exists {<l : t1 := t2>} ..( n); auto 10.

      red in |- *; apply t_trans with (e, Cst1 (Struct l) t0);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.

      (* Case of (Cst1 (Const _) _) *)
      case (delta_cst e n t).
      intros (v, red_v).
      elim whnf_cci_sn with e v; eauto 20.
      intros v' red_v' hn_v'.
      exists v'; trivial.
      apply red_trans with v; auto 20.

      intros no_red.
      exists (Cst1 (Const n) t); auto.

      (* Case of (Cst1 (MutCase _) _) *)
      elim whnf_cci_sn with e t; eauto 20.
      intros v red_v hn_v.
      destruct v as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists (Cst1 (MutCase b p) %%s)
        | exists (Cst1 (MutCase b p) #n)
        | exists (Cst1 (MutCase b p) (Cst0 c))
        | exists (Cst1 (MutCase b p) (Cst1 c t0))
        | exists (Cst1 (MutCase b p) (Forall n : t0, t1))
        | exists (Cst1 (MutCase b p) (Fun n : t0 => t1))
        | exists (Cst1 (MutCase b p) (t0 ** t1))
        | idtac
        ]; auto 10.
      rename t0 into ct; rename t1 into pr.
      elim whnf_cci_sn with e pr; eauto 20.
      intros w red_w hn_w.
      assert
        (red_tw : red redn_term e (Cst1 (MutCase b p) t) (Cst1 (MutCase b p) (ct,,w))).
      apply red_trans with (Cst1 (MutCase b p) (ct,,pr)); auto 10.

      destruct w as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists (Cst1 (MutCase b p) (ct,,%%s))
        | exists (Cst1 (MutCase b p) (ct,,#n))
        | exists (Cst1 (MutCase b p) (ct,,Cst0 c))
        | exists (Cst1 (MutCase b p) (ct,,Cst1 c t0))
        | exists (Cst1 (MutCase b p) (ct,,Forall n : t0, t1))
        | exists (Cst1 (MutCase b p) (ct,,Fun n : t0 => t1))
        | exists (Cst1 (MutCase b p) (ct,,t0 ** t1))
        | idtac
        ]; auto 10.
      rename t0 into prd; rename t1 into br.
      elim whnf_cci_sn with e ct; eauto 20.
      intros uc red_uc hn_uc.
      assert
        (red_tuc :
          red redn_term e (Cst1 (MutCase b p) t) Match uc => prd with b — p => br end).
      apply red_trans with Match ct => prd with b — p => br end; auto 10.
      destruct uc as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists Match %%s => prd with b — p => br end
        | exists Match #n => prd with b — p => br end
        | exists Match Cst0 c => prd with b — p => br end
        | idtac
        | exists Match Forall n : t0, t1 => prd with b — p => br end
        | exists Match Fun n : t0 => t1 => prd with b — p => br end
        | exists Match t0 ** t1 => prd with b — p => br end
        | exists Match t0,,t1 => prd with b — p => br end
        ];  auto 10.
      destruct c;
        [ exists Match Cst1 Mark t0 => prd with b — p => br end
        | exists Match Cst1 Mt_mark t0 => prd with b — p => br end
        | exists Match Cst1 Lmark t0 => prd with b — p => br end
        | exists Match Cst1 Nil_mark t0 => prd with b — p => br end
        | exists Match Cst1 Cons_mark t0 => prd with b — p => br end
        | exists Match Cst1 App t0 => prd with b — p => br end
        | exists Match t0..L => prd with b — p => br end
        | exists Match t0..R => prd with b — p => br end
        | exists Match Cst1 (Struct l) t0 => prd with b — p => br end
        | exists Match {<l : t0>} => prd with b — p => br end
        | exists Match t0..( n) => prd with b — p => br end
        | exists Match Cst1 (Const n) t0 => prd with b — p => br end
        | exists Match Ind{n; l // t0} => prd with b — p => br end
        | idtac
        | exists Match Cst1 (MutCase b0 p0) t0 => prd with b — p => br end
        | exists Match Fix {t0} => prd with b — p => br end
        ]; auto 10.
      rename n into cn; rename t0 into cpr.
      elim whnf_cci_sn with e cpr; auto 10.
      intros chp red_chp hn_chp.
      assert
        (red_tchp :
          red redn_term e (Cst1 (MutCase b p) t)
            Match Cstr{cn // chp} => prd with b — p => br end).
      apply red_trans with Match Cstr{cn // cpr} => prd with b — p => br end; auto 10.

      destruct chp as [s| n| c| c t0| n t0 t1| n t0 t1| t0 t1| t0 t1];
        [ exists Match Cstr{cn // %%s} => prd with b — p => br end
        | exists Match Cstr{cn // #n} => prd with b — p => br end
        | exists Match Cstr{cn // Cst0 c} => prd with b — p => br end
        | exists Match Cstr{cn // Cst1 c t0} => prd with b — p => br end
        | exists Match Cstr{cn // Forall n : t0, t1} => prd with b — p => br end
        | exists Match Cstr{cn // Fun n : t0 => t1} => prd with b — p => br end
        | exists Match Cstr{cn // t0 ** t1} => prd with b — p => br end
        | idtac
        ]; auto 10.
      rename t0 into par; rename t1 into arg.
      elim (list_index eq_dec cn p).
      intros (rnk, mtch).
      elim whnf_cci_sn with e ((br .^ rnk) ..L @ arg); auto 20.
      intros hf red_hf hn_hf.
      exists hf; trivial.
      apply red_trans with Match Cstr{cn // par,,arg} => prd with b — p => br end;
      auto 10.
      apply red_trans with ((br .^ rnk) ..L @ arg); auto 20.

      apply ord_norm_step with Match Cstr{cn // par,,arg} => prd with b — p => br end;
      auto 20.

      intro no_match.
      cut (forall n, ~ match_pattern p cn n).
      exists Match Cstr{cn // par,,arg} => prd with b — p => br end; auto 10.

      red in |- *; intros n fitm_n; apply no_match; elim fitm_n; auto.

      red in |- *; apply t_trans with (e, Cstr{cn // cpr});
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
      red in |- *; apply t_trans with (e, Cstr{cn // cpr},,pr);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
      apply ord_norm_sub with (Cst1 (MutCase b p) (Cstr{cn // cpr},,pr)); eauto 20.
      apply red_trans with (Cst1 (MutCase b p) (ct,,pr)); auto 20.

      red in |- *; apply t_trans with (e, ct,,pr);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.

      red in |- *; apply t_trans with (e, ct,,pr);
        fold (ord_norm redn_term redn_term) in |- *; eauto 10.
    Qed.

    (* Reduction towards the type constructors: sort, product, sum and
       inductives *)

    Lemma red_to_sort_dec_sn : forall e t,
      sn (ctxt redn_term) e t ->
      {s : _ | red redn_term e t %%s} + {(forall s, ~ cci_subtype e t %%s)}.
    Proof.
      intros.
      elim whnf_cci_sn with e t; trivial.
      intros hnft.
      case hnft; intros.
      left.
      split with s; trivial.

      RightStr err_not_type; red in |- *; intros.
      elim conv_srt_eq with (1 := p) (2 := q) (3 := H0); intros.
      discriminate H1.

      RightStr err_not_type; red in |- *; intros.
      elim conv_srt_eq with (1 := p) (2 := q) (3 := H0); intros.
      discriminate H1.

      RightStr err_not_type; red in |- *; intros.
      elim conv_srt_eq with (1 := p) (2 := q) (3 := H0); intros.
      discriminate H1.

      RightStr err_not_type; red in |- *; intros.
      elim conv_srt_eq with (1 := p) (2 := q) (3 := H0); intros.
      discriminate H1.

      RightStr err_not_type; red in |- *; intros.
      elim conv_srt_eq with (1 := p) (2 := q) (3 := H0); intros.
      discriminate H1.

      RightStr err_not_type; red in |- *; intros.
      elim conv_srt_eq with (1 := p) (2 := q) (3 := H0); intros.
      discriminate H1.

      RightStr err_not_type; red in |- *; intros.
      elim conv_srt_eq with (1 := p) (2 := q) (3 := H0); intros.
      discriminate H1.
    Qed.

    Lemma red_to_sum_dec_sn : forall e t,
      sn (ctxt redn_term) e t ->
      {p : term * term | let (a, b) := p in red redn_term e t (a ** b)} +
      {(forall a b, ~ cci_subtype e t (a ** b))}.
    Proof.
      intros.
      elim whnf_cci_sn with e t; trivial.
      intros hnft.
      generalize (is_sum_dec hnft); intros [(u, v)| not_sum] redt hnfht.
      left.
      split with (u, v); trivial.

      RightStr err_not_sum; red in |- *; intros.
      apply not_sum.
      apply conv_sum_eq with (3 := H0); trivial.
    Qed.

    Lemma red_to_pair_dec_sn : forall e t,
      sn (ctxt redn_term) e t ->
      {p : term * term | let (a, b) := p in red redn_term e t (a,,b)} +
      {(forall a b, ~ eq_term e t (a,,b))}.
    Proof.
      intros.
      elim whnf_cci_sn with e t; trivial.
      intros hnft.
      generalize (is_pair_dec hnft);
        intros [((u, v), is_pair)| not_pair] redt hnfht.
      left.
      subst hnft.
      split with (u, v); trivial.

      right; red in |- *; intros.
      elim conv_pair_eq with (1 := redt) (2 := hnfht) (3 := H0); intros.
      destruct x; red in not_pair; eapply not_pair; eauto.
    Qed.

    Lemma red_to_prd_dec_sn : forall e t,
      sn (ctxt redn_term) e t ->
      {p : _ |
        match p with
          | (x, (a, b)) => red redn_term e t (Forall x : a, b)
        end
      } + {(forall x a b, ~ cci_subtype e t (Forall x : a, b))}.
    Proof.
      intros.
      elim whnf_cci_sn with e t; trivial.
      intros hnft.
      generalize (is_prd_dec hnft); intros [(x, a, b)| not_prd] redt hnfht.
      left.
      split with (x, (a, b)); trivial.

      RightStr err_not_prod; red in |- *; intros.
      apply not_prd.
      apply conv_prd_eq with (3 := H0); trivial.
    Qed.

    Lemma red_to_ind_dec_sn : forall e t,
      sn (ctxt redn_term) e t ->
      {ii : _ |
        match ii with
          | (i, (p, (a, (m, l)))) => red redn_term e t Ind{i; m // p,,(a,,l)}
        end
      } + {(forall i p a l m, ~ cci_subtype e t Ind{i; m // p,,(a,,l)})}.
    Proof.
      intros.
      elim whnf_cci_sn with e t; trivial.
      intros hnft.
      generalize (is_untyped_ind_dec hnft).
      intros [(((i, m), tr), eq_h)| not_ind] redt hnfht.
      subst hnft.
      elim red_to_pair_dec_sn with e tr.
      intros ((p, pr0), red_tr).
      elim red_to_pair_dec_sn with e pr0.
      intros ((a, l), red_pr).
      left.
      split with (i, (p, (a, (m, l)))); trivial.

      apply red_trans with Ind{i; m // p,,pr0}; auto 10.
      apply red_trans with Ind{i; m // tr}; auto 10.

      intros not_pair.
      right; red in |- *; intros.
      assert (sub_hn_inv e Ind{i; m // tr} Ind{i0; m0 // p0,,(a,,l)}).
      apply inv_cumul_hn with Ind{i; m // tr} Ind{i0; m0 // p0,,(a,,l)}; auto.
      apply cci_trans with t; auto 10.

      inversion_clear H1.
      elim not_pair with a l.
      assert (eq_term e (p,,pr0) (p0,,(a,,l))).
      apply eq_term_trans with tr; auto.
      apply eq_term_sym; auto.

      eapply inv_pair_r; eauto.

      elim not_pair with a0 l0.
      assert (eq_term e (p,,pr0) (p1,,(a0,,l0))).
      apply eq_term_trans with tr; auto.
      apply eq_term_sym; auto.

      eapply inv_pair_r; eauto.

      apply subterm_sn with redn_term e (p,,pr0); auto.
      apply R_stable_redn.

      apply sn_red_sn with tr; auto.
      apply subterm_sn with redn_term e Ind{i; m // tr}; auto.
      apply R_stable_redn.

      apply sn_red_sn with t; auto.

      intros not_pair.
      right; red in |- *; intros.
      assert (sub_hn_inv e Ind{i; m // tr} Ind{i0; m0 // p,,(a,,l)}).
      apply inv_cumul_hn with Ind{i; m // tr} Ind{i0; m0 // p,,(a,,l)}; auto.
      apply cci_trans with t; auto 10.

      inversion_clear H1.
      red in not_pair; eauto.

      red in not_pair; eauto.

      apply subterm_sn with redn_term e Ind{i; m // tr}; auto.
      apply R_stable_redn.

      apply sn_red_sn with t; auto.

      right; red in |- *; intros.
      assert (sub_hn_inv e hnft Ind{i; m // p,,(a,,l)}).
      apply inv_cumul_hn with hnft Ind{i; m // p,,(a,,l)}; auto.
      apply cci_trans with t; auto 10.

      inversion H1.
      subst c; eauto.

      eauto.
    Qed.

  End Whnf_function.

End IReductionDecA.
