(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export WfSign.
Require Export SignDec.

Set Implicit Arguments.
Unset Strict Implicit.

  Definition err_name_clash x :=
    app_str name_string (app_str (string_of x) err_already_defined).

  Section DecideInductive.

    Variable sg : sigma.
    Hypothesis wfs : wf_sign sg.

    Let wf_ok := wf_sign_ok wfs.


    Lemma wf_arity_dec : forall ip, decide (wf_arity sg ip).
    Proof.
      intros.
      elim dec_for_all with (l := ip) (P := fun is => wf_one_ind_arity sg is);
        intros.
      (* Ici, tous les inductifs sont ok individuellement. *)
      elim (list_unique_dec eq_dec (inds_of_pack ip)); intros.
      left.
      split; trivial.

      RightStr err_ind_name_clash; red in |- *; intros; apply b.
      exact (wf_ar_free H).

      (* Un des types inductifs est mal forme *)
      right; red in |- *; intros; apply b; intros.
      apply (wf_ar_ok H H0).

      (* FIXME: should be inferred by type classes *)
      assert (algos := @CCI_algos sg wf_ok).

      (* Verification d'un inductif. *)
      elim decide_wf with (((nil; x_ : i_param x); x_ : i_arg x); x_ : %%(i_kind x));
        intros.
      elim search_global with (i_name x); intros.
      RightStr (err_name_clash (i_name x)); red in |- *; intros.
      inversion_clear a0.
      elim (wf_one_ind_free H0) with x0; trivial.

      left.
      split; trivial.

      RightStr err_ind_ill_formed; red in |- *; intros; apply b.
      apply (wf_one_ind_ok H0).
    Qed.


    Lemma wf_constructor_dec :
     forall inds is c,
     sg – ⊢ (((( ); x_ : i_param is); x_ : i_arg is); x_ : %%(i_kind is)) ->
     decide (wf_constructor sg inds is c).
    Proof.
      (*
      Realizer [inds:(list name); is:one_ind; c:constructor]
        [ e1 = (add_var (nil ?) x_ (i_param is)) ]
        [ e2 = (add_var e1 x_ (c_arg c)) ]
        Cases (search_global sg (c_name c)) of
          inright =>
            if (DEC_TYPWF e1 (c_arg c) (Srt (i_ckind is))) then
              if (DEC_TYPWF e2 (c_inst c) (lift (1) (i_arg is))) then
                (constr_pos_dec inds c)
              else false
            else false
        | _ => false
        end.
      *)
      intros.
      elim search_global with (c_name c); intros.
      RightStr (err_name_clash (c_name c)); red in |- *; intros.
      inversion_clear a.
      elim (wf_cstr_free H0) with x; trivial.

      (* FIXME: should be inferred by type classes *)
      assert (algos := @CCI_algos sg wf_ok).

      elim chk_wk with (( ); x_ : i_param is) (c_arg c:term) (%%(i_ckind is));
       intros.
      elim
       chk_wk
        with
          ((nil; x_ : i_param is); x_ : c_arg c)
          (c_inst c:term)
          (^(1)(i_arg is)); intros.
      elim constr_pos_dec with inds c; intros.
      left.
      split; trivial.

      right; red in |- *; intros.
      apply b0.
      apply (wf_cstr_pos H0).

      RightStr err_bad_instance; red in |- *; intros; apply b0.
      apply (wf_cstr_inst H0).

      apply wf_var with (i_ckind is); trivial.

      apply wf_type_thinning with (e := (( ); x_ : ct (i_param is))); auto.
      Inversion_typ (inv_wf _ _ H).
      left with s; trivial.

      apply wf_var with (i_ckind is); trivial.

      RightStr err_bad_cstr_args; red in |- *; intros.
      apply b0.
      apply (wf_cstr_ok H0).

      apply (inv_wf _ _ (inv_wf _ _ H)).

      constructor.
    Qed.


    Lemma wf_constr_one_dec :
      forall inds is,
        sg – ⊢ (((( ); x_ : i_param is); x_ : i_arg is); x_ : %%(i_kind is)) ->
          decide (forall x c, constr_in_ind x is c -> wf_constructor sg inds is c).
    Proof.
      (*
      Realizer [inds:(list name); is:one_ind]
        Cases (i_constrs is) of
          (Some lc) => (dec_for_all ? (wf_constructor_dec inds is) lc)
        | None => true
        end.
      *)
      intros.
      generalize (refl_equal (i_constrs is)).
      pattern (i_constrs is) at 2 in |- *.
      case (i_constrs is); intros.
      elim dec_for_all with (P := wf_constructor sg inds is) (l := l).
      left; intros.
      apply a.
      inversion_clear H1.
      rewrite H0 in H2.
      injection H2; intros.
      rewrite H1; trivial.

      right; red in |- *; intros.
      apply b; intros.
      apply H1 with (c_name x).
      split with l; trivial.

      intros.
      apply wf_constructor_dec; trivial.

      left; intros.
      inversion_clear H1.
      rewrite H0 in H2.
      discriminate H2.
    Qed.


    Lemma wf_constr_dec :
      forall ip : ind_pack,
        (forall is,
          In is ip ->
          sg – ⊢ (((( ); x_ : i_param is); x_ : i_arg is); x_ : %%(i_kind is))) ->
        decide (wf_constr sg ip).
    Proof.
      intros.
      elim
        dec_for_all
        with
          (P := fun is =>
            sg – ⊢
              (((( ); x_ : i_param is); x_ : i_arg is); x_ : %%(i_kind is)) ->
              forall x c,
                constr_in_ind x is c -> wf_constructor sg (inds_of_pack ip) is c)
          (l := ip); intros.
      elim (list_unique_dec (@eq_dec name _)) with (cstr_ind_pack ip); intros.
      left.
      split; trivial; intros.
      apply a with x; trivial.
      auto.

      RightStr err_cstr_name_clash; red in |- *; intros; apply b.
      exact (wf_ind_free H0).

      right; red in |- *; intros; apply b; intros.
      apply (wf_ind_ok H0 H1 H3).

      elim wf_constr_one_dec with (inds_of_pack ip) x; intros; auto.
      right; red in |- *; intros.
      apply b.
      eauto.
    Qed.

  End DecideInductive.

  Section DecideWfSignature.

    Variable sg : sigma.
    Hypothesis wfs : wf_sign sg.

    Let wf_ok := wf_sign_ok wfs.


    Lemma wf_def_dec : forall co, decide (wf_cst sg co).
    Proof.
      (*
      Realizer [co:constant_obj]
        Cases (search_global sg (NAME (c_decl co))) of
          (inleft _) => false
        | inright =>
            (DEC_WF (cons ? (c_decl co)
                    (cons ? (mkdecl x_ (c_par co) (None ?))
                    (nil ?))))
        end.
      *)
      intros.
      elim search_global with (NAME (c_decl co)); intros.
      RightStr (err_name_clash (NAME (c_decl co))); red in |- *; intros.
      inversion_clear a.
      elim (wf_cst_free H) with x; trivial.

      (* FIXME: should be inferred by type classes *)
      assert (algos := @CCI_algos sg wf_ok).

      elim decide_wf with ((( ); x_ : c_par co); c_decl co); intros.
      left.
      split; trivial.

      RightStr err_constant_ill_typed; red in |- *; intros.
      apply b0.
      apply (wf_cst_ok H).
    Qed.


    Lemma wf_sign_ext : forall si, decide (wf_sign (si :: sg)).
    Proof.
      (*
      Realizer
        [si:sig_item]Cases si of
          (Constant co) => (wf_def_dec co)
        | (MutInductive ip) =>
            if (wf_arity_dec ip)
            then (wf_constr_dec (cons ? (MutInductive (abstract_ind ip)) sg) ip)
            else false
        end.
      *)
      intros [ip| ic].
      elim wf_arity_dec with sg ip; intros; trivial.
      elim wf_constr_dec with (MutInductive (abstract_ind ip) :: sg) ip; intros;
       trivial.
      left.
      apply wfs_ind; trivial.

      right; red in |- *; intros; apply b.
      inversion_clear H; trivial.

      apply wfs_ind; trivial.
      split; intros.
      unfold abstract_ind in H.
      elim map_inv with (1 := H); intros.
      rewrite H1.
      split; intros.
      simpl in |- *.
      apply (wf_one_ind_ok (wf_ar_ok a H0)).

      simpl in |- *.
      apply (wf_one_ind_free (wf_ar_ok a H0)).

      replace (inds_of_pack (abstract_ind ip)) with (inds_of_pack ip).
      exact (wf_ar_free a).

      unfold inds_of_pack in |- *.
      elim ip; simpl in |- *; intros; auto.
      rewrite H; trivial.

      split; intros.
      elim abstract_no_cstr with (1 := H) (2 := H0).

      replace (cstr_ind_pack (abstract_ind ip)) with (nil (A:=name)).
      red in |- *; intros.
      inversion H.

      unfold cstr_ind_pack, abstract_ind in |- *.
      elim ip; simpl in |- *; intros; auto.  admit.

      right; red in |- *; intros; apply b.
      inversion_clear H; trivial.

      elim wf_def_dec with ic; intros.
      left.
      constructor; trivial.

      right; red in |- *; intros; apply b.
      inversion_clear H; trivial.
    Qed.


  End DecideWfSignature.

  Lemma wf_sign_dec : forall sg, decide (wf_sign sg).
  Proof.
    (*Realizer Fix wf_sign_dec {wf_sign_dec [sg:?]: bool :=
      Cases sg of
        nil => true
      | (cons si s) => if (wf_sign_dec s) then (wf_sign_ext s si) else false
      end}.*)
    fix wf_sign_dec 1.
    intros sg.
    case sg.
    left; constructor.

    intros si s.
    case (wf_sign_dec s).
    intros s_ok.
    case (wf_sign_ext s_ok si).
    intros si_ok; left; auto.

    intros si_nok; right; trivial.

    intros s_nok.
    right; red in |- *; intros sg_ok; inversion sg_ok; auto.
  Defined.
