(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export CicMeta.

Set Implicit Arguments.
Unset Strict Implicit.

Section ISignCorrectness.

  Context {sg : sigma}.
  Context {sgn_ok : sign_ok}.

(*
  Lemma le_ind_rev: (m:nat)(P:nat->Prop)
      (P m)
    ->((n:nat)(lt n m)->(P (S n))->(P n))
    ->(n:nat)(le n m)->(P n).
Proof.
Intros.
Generalize H H0.
(Elim H1; Intros; Auto).
Save.
*)

  (* Used in RecordDec.v *)
  Lemma record_arity_ok :
    forall sign e,
      ⊦ (e; x_ : prod_arity sign) -> ⊦ (e; x_ : record_arity sign).
  Proof.
    simple induction sign; simpl in |- *; trivial.
    intros (x, ty); simpl in |- *; intros.
    Inversion_typ H0.
    Inversion_typ H1.
    apply all_prod_sum.
    apply wf_var with s1; trivial.

    apply all_rules.
    apply H.
    apply wf_var with s2; trivial.
  Qed.


  Lemma prod_arity_ok :
    forall sign e,
      ⊦ (e; x_ : record_arity sign) -> ⊦ (e; x_ : prod_arity sign).
  Proof.
    simple induction sign; simpl in |- *; trivial.
    intros (x, ty); simpl in |- *; intros.
    Inversion_typ H0.
    Inversion_typ H1.
    Inversion_typ H3.
    apply all_rules.
    apply H.
    apply wf_var with s4; trivial.
  Qed.




  Inductive contract (m : term) (s2 : rspec) : nat -> rspec -> Prop :=
  | Cont_O : contract m s2 0 s2
  | Cont_S :
    forall k s1 x ty,
      contract m s2 k ((x, ty) :: s1) ->
      contract m s2 (S k) (subst_rsign m..( x) s1 0).


  Lemma contract_field :
    forall m k s1 s2,
      contract m s1 k s2 ->
      forall ty n,
        item ty (fields_type (expand_sign m s2) 0) n ->
        item ty (fields_type (expand_sign m s1) 0) (k + n).
  Proof.
    simple induction 1; simpl in |- *; intros; trivial.
    rewrite plus_n_Sm.
    apply H1.
    fold (expand_sign m s0) in |- *.
    rewrite lift0.
    rewrite subst_fields_expand; auto.
  Qed.

  Lemma contract_name :
    forall m k s1 s2,
      contract m s1 k s2 ->
      forall x n, item x (field_names s2) n -> item x (field_names s1) (k + n).
  Proof.
    simple induction 1; simpl in |- *; intros; trivial.
    rewrite plus_n_Sm.
    apply H1.
    generalize H2.
    rewrite record_names_subst.
    unfold field_names in |- *; simpl in |- *; auto.
  Qed.


  Lemma wf_fields_rec :
    forall sign e m,
      e ⊦ m : {<field_names sign : record_arity sign>} ->
        forall sign' k,
          contract m sign k sign' ->
          ⊦ (e; x_ : prod_arity sign') ->
            ⊦ (e; x_ : mkvect (fields_type (expand_sign m sign') 0)).
  Proof.
    intros sign e m H sign'.
    elim sign' using list_length_ind; intros.
    generalize l H1 H2 H3; clear H1 H2 H3 l; intros [| (x, ty) sign0];
      simpl in |- *; intros eqlgth tym.
    trivial.

    intros.
    Inversion_typ H3.
    Inversion_typ H1.
    clear H6 H5 s3.
    apply all_prod_sum.
    apply wf_var with s1; trivial.

    generalize (Cont_S tym).
    fold map (expand_sign m sign0) in |- *.
    simpl in |- *; intros.
    rewrite subst_fields_expand.
    rewrite lift0.
    apply H0 with (pred m0) (S k); auto.
    elim eqlgth; simpl in |- *; auto with arith.

    elim eqlgth; simpl in |- *.
    apply subst_length.

    rewrite <- prod_arity_subst.
    apply subst_wf with x ty.
    apply typ_projr with (n := k) (sign1 := sign); trivial.
    replace k with (k + 0); auto with arith.
    apply contract_field with (1 := tym); simpl in |- *; auto.

    replace k with (k + 0); auto with arith.
    apply contract_name with (1 := tym); simpl in |- *; auto.
    unfold field_names in |- *; simpl in |- *; auto.

    apply wf_var with s1; trivial.

    apply wf_var with s2; trivial.
  Qed.

  Lemma wf_fields :
    forall sign e m,
      e ⊦ m : {<field_names sign : record_arity sign>} ->
        ⊦ (e; x_ : mkvect (fields_type (expand_sign m sign) 0)).
  Proof.
    intros.
    apply wf_fields_rec with sign 0; trivial; try constructor.
    apply prod_arity_ok.
    generalize (type_correctness _ _ _ H); intro.
    inversion_clear H0.
    Inversion_cst H1; intros.
    apply wf_var with s0; trivial.
  Qed.


  (* Used in FixDec.v *)

  Lemma typ_fix_corr :
    forall e b m fd,
      ⊦ (e; x_ : fix_ty fd) -> ⊦ ((e; m : Mark); x_ : fix_scheme b fd).
  Proof.
    unfold fix_ty, fix_scheme in |- *; intros.
    Inversion_typ H.
    clear H.
    Inversion_typ H0.
    clear H3 H0 s.
    apply wf_var with s3.
    apply Typ_prd' with s1 s2; auto.
    replace (%%s1) with (^(1)%%s1); trivial.
    apply thinning with e; auto.
    apply wf_var with so.
    apply typ_mark_so.
    apply typ_wf with (1 := H).

    intro wf_par.
    Inversion_typ H1.
    clear H H1 H2 s3.
    apply Typ_conv_srt with (%%s5); auto.
    apply Typ_prd' with s0 s4; auto.
    clear H5 H4 H3 s5 s4 s2.
    apply typ_ind_cons_mark with (fi_lem fd).
    change
      (((e; m : Mark); fi_parname fd : ^(1)(fi_party fd))
        ⊦ ^(1//1)Ind{fi_indty fd; fi_lem fd
          // fi_ipar fd,,(fi_iarg fd,,fi_marks fd)} :
        ^(1//1)%%s0) in |- *.
    apply typ_weak with e (mkdecl m Mark None) (e; fi_parname fd : fi_party fd);
      trivial.
    fold (e; m : Mark) in |- *.
    apply (typ_wf _ _ _ wf_par).

    unfold lift in |- *.
    apply ins_in_env_var.
    unfold add_var in |- *; constructor.

    constructor.
    apply wf_var_lift.
    apply (typ_wf _ _ _ wf_par).

    apply (typ_wf _ _ _ H0).

    split with (mkdecl m Mark None); auto.
    unfold add_var in |- *; auto.

    intro wf_ind'.
    apply
      subtype_env
      with
        (((e; m : Mark); fi_parname fd : ^(1)(fi_party fd));
          fi_argname fd
          : ^(1//1)Ind{fi_indty fd; fi_lem fd
            // fi_ipar fd,,(fi_iarg fd,,fi_marks fd)}).
    fold (^(1//2)%%s4) in |- *.
    apply
      typ_weak
      with
        e
        (mkdecl m Mark None)
        ((e; fi_parname fd : fi_party fd); fi_argname fd
          : Ind{fi_indty fd; fi_lem fd
            // fi_ipar fd,,(fi_iarg fd,,fi_marks fd)});
        trivial.
    fold (e; m : Mark) in |- *.
    apply (typ_wf _ _ _ wf_par).

    apply ins_in_env_var.
    unfold lift in |- *.
    apply ins_in_env_var.
    unfold add_var in |- *; constructor.

    unfold add_var in |- *.
    constructor.
    split; simpl in |- *; trivial.
    change_notation.
    apply incl_mark_sub_cci.
    case b.
    apply incl_rm_mark.

    apply incl_str_marks.

    apply wf_var with s0; trivial.


    change_notation; generalize (sub_sort_env_indep H5); intro H; simple apply H.
  Qed.

  (* unused! *)
  Lemma typ_fix_correctness :
    forall e m fds,
      ⊦ (e; x_ : fix_types fds) -> ⊦ (e; x_ : fix_bodies m fds).
  Proof.
    intros.
    unfold fix_bodies in |- *.
    apply all_rules.
    apply all_rules.
    apply wf_var_lift.
    generalize H.
    elim fds; simpl in |- *; intros.
    apply wf_var with so.
    apply typ_one_so.
    apply wf_var with so.
    apply typ_mark_so.
    apply (inv_wf _ _ H).

    Inversion_typ H1.
    unfold fix_types in H2.
    simpl in H2.
    Inversion_typ H2.
    apply all_prod_sum; auto.
    apply typ_fix_corr.
    apply wf_var with s1; trivial.

    apply H0.
    apply wf_var with s2; trivial.

    generalize H.
    elim fds; simpl in |- *; intros.
    apply wf_var with so.
    apply typ_one_so.
    apply wf_var with so.
    apply typ_mark_so.
    apply (inv_wf _ _ H).

    Inversion_typ H1.
    unfold fix_types in H2.
    simpl in H2.
    Inversion_typ H2.
    apply all_prod_sum; auto.
    apply typ_fix_corr.
    apply wf_var with s1; trivial.

    apply H0.
    apply wf_var with s2; trivial.
  Qed.



  Lemma ty_case_ok :
    forall e inds is pats p c br b t u,
      sign_glob_spec sg (i_name is) (GlobInd inds is) ->
      type_case inds is pats p c b t u ->
      ⊦ (e; x_ := (c,,(p,,br)) : t) -> ⊦ (e; x_ : u).
  Proof.
    intros e inds is pats p c br b t u sign_ind
      (m, par, arg, l, tybr, se, fpat, elim_srt, case_br) wf_tr.
    clear fpat elim_srt case_br t u pats.
    Inversion_typ wf_tr.
    elim H using inv_pair_sum; intros.
    elim H2 using inv_pair_sum; intros.
    Inversion_typ H0.
    clear H8 H7 H6 s3 s2 H4 H2 H wf_tr.
    Inversion_cst H5; intros.
    elim H6 using inv_pair_sum; intros.
    elim H9 using inv_pair_sum; intros.
    clear H11 H9 H8 H7 H6 H4 H2 H5 s1 H0 s.
    apply wf_var with se.
    apply typ_app2 with (2 := H3) (3 := H10) (4 := H1); simpl in |- *; split; [ |split]; auto.
    elim (inj_ind_is sign_ind H); auto.

    rewrite simpl_sl.
    rewrite lift0.
    apply incl_mark_sub_cci.
    apply incl_mt_mark.
  Qed.


(*

  Lemma wf_case_predicate2: (ind:name)(inds:(list name))(is:one_ind)
    (sign_glob_spec sg ind (GlobInd inds is))
  ->[~ |- [];
          x_:!i_param is!;
          x_:!i_arg is!;
          x_:Ind{i_name is | (#1, (#0, []))} ~].
Proof.
Intros.
Elim glob_ind_eq_name with 1:=H.
(Inversion_typ (ind_sign_ok H); Intros).
Apply typ_induc with 1:=H.
Constructor.
Apply (typ_wf H0).

Unfold add_var.
(Split with (mkdecl x_ (i_param is) (None ?)); Simpl; Auto).
Unfold lift.
(Rewrite closed0_lift; Trivial).

Constructor.
Apply (typ_wf H0).

Unfold add_var.
(Split with (mkdecl x_ (i_arg is) (None ?)); Simpl; Auto).
Apply id_closed1.

Apply typ_nil_mark.
Apply (typ_wf H0).
Save.


  Lemma wf_ind_wf_par: (ind:name)(inds:(list name))(is:one_ind)
      (e:env)(m:le_marks)(p,a,l:term)
      (sign_glob_spec sg ind (GlobInd inds is))
    ->(wf_type cci_pts e [: Ind{ind;m|(p,(a,l))} :])
    ->[~ e |- p : !i_param is! ~].
Proof.
Intros.
Inversion_clear H0.
(Inversion_cst H1; Intros).
(Elim H4 using inv_pair_sum; Intros).
(Elim (inj_ind_is H0 H); Trivial).
Save.


Intros.
Change [~ |- e;
             x_:{0:=p}!ct (i_arg is)!;
             x_:{1:=p}Ind{i_name is | (#1, (#0, []))} ~].
Apply subst_wf2 with x_ (i_param is)::term.
Apply wf_ind_wf_par with 1:=H 2:=H0.




Intros.
Apply wf_var with (i_kind is).
Change [~ |- e; x_:{0:=p}!ct (i_arg is)!
              ; x_:{1:=p}Ind{i_name is | (#1, (#0, []))} ~].

           : {1:=p}%%!i_kind is! ~].
Apply substitution1 with x_ (i_param is)::term.
Apply thin_par_def2.


*)

  Lemma wf_case_predicate2 :
    forall e p ind inds is,
      sign_glob_spec sg ind (GlobInd inds is) ->
      e ⊦ p : i_param is ->
        ⊦ ((e; x_ : (i_arg is)[0:=p]); x_ : Ind{i_name is // ^(1)p,,(#0,,[])}).
  Proof.
    intros.
    cut (⊦ (e; x_ : (i_arg is)[0:=p])); intros.
    elim (glob_ind_eq_name H).
    apply typ_induc with (1 := H).
    replace (i_param is:term) with (^(1)(i_param is)).
    apply thinning with e; auto.

    unfold lift in |- *.
    rewrite closed0_lift.
    trivial.

    constructor; trivial.
    unfold add_var in |- *.
    split with (mkdecl x_ (i_arg is)[0:=p] None); auto.
    simpl in |- *.
    unfold lift in |- *.
    rewrite distr_lift_subst.
    rewrite closed1_lift.
    trivial.

    apply typ_nil_mark; trivial.

    apply subst_wf with x_ (i_param is:term).
    trivial.

    apply thin_par_wf2.
    apply ind_par_arg_one with (1 := H).

    apply (typ_wf _ _ _ H0).
  Qed.

  Lemma wf_case_predicate1 :
    forall e p a l m ind inds is,
      sign_glob_spec sg ind (GlobInd inds is) ->
      wf_type e Ind{ind; m // p,,(a,,l)} ->
      ⊦ ((e; x_ : (i_arg is)[0:=p]); x_ : Ind{i_name is // ^(1)p,,(#0,,[])}).
  Proof.
    intros.
    apply wf_case_predicate2 with (1 := H).
    inversion_clear H0.
    Inversion_cst H1.
    intros.
    elim H4 using inv_pair_sum; intros.
    elim (inj_ind_is H0 H).
    trivial.
  Qed.

  Lemma wf_case_predicate :
    forall e p a l m se ind inds is,
      sign_glob_spec sg ind (GlobInd inds is) ->
      wf_type e Ind{ind; m // p,,(a,,l)} ->
      in_dom_rel ax se -> ⊦ (e; x_ : (case_predicate_form is se)[0:=p]).
  Proof.
    intros.
    unfold case_predicate_form in |- *; simpl in |- *.
    apply all_rules.
    apply all_rules.
    elim H1; intros.
    apply wf_var with x.
    constructor; trivial.
    exact (wf_case_predicate1 H H0).
  Qed.



  Lemma str_substp_ok :
    forall inds l b b',
      str_substp inds l b b' ->
      forall e, e ⊦ l : Lmark -> forall s, e ⊦ b : %%s -> e ⊦ b' : %%s.

  Proof.
    fix str_sub_ok 5
      with (mark_rec_ok (inds:_) (l:_) (s1:_) (s2:_) (y:mark_record inds l s1 s2)
        {struct y} :
        forall e,
          e ⊦ l : Lmark ->
            forall s, e ⊦ record_arity s1 : %%s -> e ⊦ record_arity s2 : %%s).
    simple destruct 1; intros.
    trivial.

    Inversion_cst H2; intros.
    elim H6 using inv_pair_sum; intros.
    elim H9 using inv_pair_sum; intros.
    clear H6 H9.
    apply Typ_conv_srt with (%%(i_kind is)); auto.
    apply typ_induc0 with (1 := H3); auto.

    Inversion_typ H2.
    apply Typ_conv_srt with (%%s3); auto.
    apply Typ_prd with s1 s2; auto.
    apply str_sub_ok with (1 := H0); auto.
    replace (Lmark:term) with (^(1)Lmark); auto.
    apply thinning with e; auto.
    apply (typ_wf _ _ _ H4).

    Inversion_typ H3.
    apply Typ_conv_srt with (%%s3); auto.
    apply Typ_sum with s1 s2; eauto.

    unfold record_of in H2; unfold record_of in |- *.
    Inversion_typ H2.
    apply Typ_conv_srt with u; trivial.
    apply Typ_cst1 with t; trivial.
    generalize H6.
    inversion_clear H3; intros.
    apply mark_rec_ok with (1 := H0); auto.

    red in |- *; simpl in |- *; red in |- *.
    inversion H3.
    constructor; auto.

    rewrite <- (mark_record_names H0).
    rewrite <- H8; trivial.

    simple destruct 1; simpl in |- *; trivial; intros.
    Inversion_typ H2.
    apply Typ_conv_srt with (%%s6); trivial.
    apply Typ_sum with s4 s5; auto.
    apply str_sub_ok with (1 := H); trivial.

    Inversion_typ H4.
    apply Typ_conv_srt with (%%s9); auto.
    apply Typ_prd with s7 s8; trivial.
    apply str_sub_ok with (1 := H); trivial.

    apply mark_rec_ok with (1 := H0); trivial.
    replace (Lmark:term) with (^(1)Lmark); auto.
    apply thinning with e; auto.
    apply wf_var with s7.
    apply str_sub_ok with (1 := H); trivial.

    apply subtype_env with (e; x : ty1); auto.
    unfold add_var in |- *; auto.
    constructor.
    split; simpl in |- *; auto.
    assert (incl_m := str_sub_smaller H); auto.
    change_notation; simple apply incl_mark_sub_cci; simple apply incl_m.

    apply wf_var with s7.
    apply str_sub_ok with (1 := H); trivial.
  Qed.


  Lemma str_substp_wf :
    forall inds e l b b',
      str_substp inds l b b' ->
      e ⊦ l : Lmark -> ⊦ (e; x_ : b) -> ⊦ (e; x_ : b').
  Proof.
    intros.
    Inversion_typ H1.
    apply wf_var with s.
    apply str_substp_ok with (1 := H) (2 := H0); auto.
  Qed.



  (* Used in CaseDec.v *)

  Lemma wf_branch :
    forall e inds is p l q se pat c arg,
      sign_glob_spec sg pat (GlobCstr is c) ->
      e ⊦ l : Lmark ->
        e ⊦ p : i_param is ->
          e ⊦ q : (case_predicate_form is se)[0:=p] ->
            str_substp inds ^(1)l (c_arg c) arg ->
            ⊦ (e; x_ : (branch_ty c arg q)[0:=p]).
  Proof.
    intros.
    unfold branch_ty in |- *.
    simpl in |- *.
    apply all_rules.
    apply wf_var with se.
    rewrite simpl_subst; auto.
    cut (⊦ (e; x_ : arg[0:=p])); intros.
    apply
      typ_app2
      with
        ((i_arg is)[0:=^(1)p])
        Ind{i_name is // ^(1)p,,(((c_inst c)[1:=^(1)p])[0:=#0],,[])}
        ((case_predicate_form is se)[0:=^(1)p]); simpl in |- *;
          auto.
    rewrite simpl_sl.
    rewrite lift0.
    replace (c_inst c:term) with (^(1//2)(c_inst c)).
    rewrite subst_id.
    rewrite closed2_lift.
    auto.

    rewrite closed2_lift; trivial.

    change ((e; x_ : arg[0:=p]) ⊦ ^(1)q : (case_predicate_form is se)[0:=^(1)p])
      in |- *.
    rewrite <- case_predicate_form_lift with (n := 1) (k := 0).
    unfold lift at 2 in |- *.
    rewrite <- distr_lift_subst.
    fold (^(1)((case_predicate_form is se)[0:=p])) in |- *.
    apply thinning with e; auto.

    rewrite <- subst_id.
    rewrite closed2_lift.
    rewrite <- (simpl_sl ^(1)#0 (i_arg is)).
    rewrite <- comm_subst.
    apply subtype_env with (e; x_ : (c_arg c)[0:=p]); auto.
    apply substitution with x_ (^(1)((c_arg c)[0:=p])).
    unfold lift at 1 in |- *.
    rewrite distr_lift_subst.
    rewrite closed1_lift.
    apply substitution1 with x_ (^(1)(i_param is)).
    unfold lift at 1 in |- *.
    rewrite closed0_lift.
    apply thin_par_def2; trivial.
    apply (ok_constr H).

    apply subst_wf with x_ (i_param is:term); trivial.
    apply thin_par_wf2; auto.
    apply (typ_wf _ _ _ (ok_constr H)).

    apply (inv_wf _ _ H4).

    apply thinning with e; auto.
    apply subst_wf with x_ (i_param is:term); trivial.
    apply thin_par_wf2; auto.
    apply (typ_wf _ _ _ (ok_constr H)).

    apply (inv_wf _ _ H4).

    constructor.
    apply subst_wf with x_ (i_param is:term); trivial.
    apply thin_par_wf2; auto.
    apply (typ_wf _ _ _ (ok_constr H)).

    apply (inv_wf _ _ H4).

    split with (mkdecl x_ (c_arg c)[0:=p] None); auto.
    unfold add_var in |- *; auto.

    unfold add_var in |- *; constructor.
    constructor; simpl in |- *; auto.
    change_notation; apply cci_subst_var with x_ p.
    apply incl_mark_sub_cci.
    apply str_sub_smaller with (1 := H3).

    apply typ_constructor; trivial.
    replace (i_param is:term) with (^(1)(i_param is)).
    elim glob_cstr_eq_name with (1 := H).
    trivial.

    unfold lift in |- *.
    rewrite closed0_lift; trivial.

    replace (i_param is:term) with (^(1)(i_param is)).
    apply thinning with e; auto.

    unfold lift in |- *.
    rewrite closed0_lift; trivial.

    apply typ_conv_wf with (^(1)(arg[0:=p])); simpl in |- *.
    constructor; auto.
    split with (mkdecl x_ arg[0:=p] None); auto.
    unfold add_var in |- *; auto.

    apply var_wft.
    replace ((c_arg c)[0:=^(1)p]) with (^(1)((c_arg c)[0:=p])).
    apply wf_var_lift; trivial.
    apply subst_wf with x_ (i_param is:term); auto.
    apply thin_par_wf2.
    apply (typ_wf _ _ _ (ok_constr H)).

    apply (typ_wf _ _ _ H0).

    unfold lift in |- *.
    rewrite distr_lift_subst.
    rewrite closed1_lift; auto.

    replace ((c_arg c)[0:=^(1)p]) with (^(1)((c_arg c)[0:=p])).
    unfold add_var in |- *.
    change_notation; apply cci_thin.
    apply cci_subst_var with x_ p.
    apply incl_mark_sub_cci.
    apply str_sub_smaller with (1 := H3).

    unfold lift in |- *.
    rewrite distr_lift_subst.
    rewrite closed1_lift; auto.

    apply subst_wf with x_ (i_param is:term); auto.
    apply str_substp_wf with (1 := H3); trivial.
    replace (Lmark:term) with (^(1)Lmark); auto.
    apply thinning with e; auto.
    unfold add_var in |- *.
    apply thin_par_wf.
    apply (inv_wf _ _ (typ_wf _ _ _ (ok_constr H))).

    apply (typ_wf _ _ _ H0).

    apply thin_par_wf2.
    apply (typ_wf _ _ _ (ok_constr H)).

    apply (typ_wf _ _ _ H0).
  Qed.

  Lemma wf_branch_vect :
    forall e i inds is p a l q se m pats tybr,
      sign_glob_spec sg i (GlobInd inds is) ->
      ⊦ (e; x_ : Ind{i; m // p,,(a,,l)}) ->
        e ⊦ q : (case_predicate_form is se)[0:=p] ->
          case_branch inds is q l pats tybr -> ⊦ (e; x_ : (mkvect tybr)[0:=p]).
  Proof.
    simple induction 4; simpl in |- *; intros.
    apply wf_var with so.
    apply typ_one_so.
    apply (typ_wf _ _ _ H1).

    apply all_prod_sum; trivial.
    clear H5 H4 l1 l2.
    inversion_clear H3.
    inversion_clear H5.
    rewrite H6.
    clear H6 y.
    Inversion_typ H0.
    Inversion_cst H5.
    intros.
    elim H9 using inv_pair_sum; intros.
    elim H12 using inv_pair_sum; intros.
    clear H12 H13 H10 H9 H8 H7.
    apply wf_branch with (pat := x) (4 := H1) (5 := H3); trivial.
    apply glob_ind_cst with (1 := H).
    trivial.

    elim (inj_ind_is H6 H); trivial.
  Qed.

End ISignCorrectness.
