(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export FixDec.
Require Export IndDec.
Require Export CaseDec.
Require Export RecordDec.

Set Implicit Arguments.
Unset Strict Implicit.

  Section ISignDecOk.

    Context {sg : sigma}.
    Context {MPTSO_algos : PTSO_algos}.
    Context {sgn_ok : sign_ok}.

    Lemma sign_dec :
      forall e c m t,
        e ⊦ m : t ->
          ppal_dec (fun u => in_sign_sat c e m t u) (cci_subtype e).
    Proof.
      intros e c m t wf_m.
      case (sorted_type infer_axiom) with e t.
      eapply typ_wf; eauto.

      apply type_correctness with (1 := wf_m).

      intros t_sorted.
      assert (wfemt : ⊦ (e; x_ := m : t)).
      case t_sorted; intros s ty_t.
      apply wf_def with s; trivial.

      clear t_sorted wf_m.
      refine
        match
          c as c
            return
            (ppal_dec (fun u : term => in_sign_sat c e m t u)
              (cci_subtype e))
          with
          | Const x => sign_dec_const x wfemt
          | Proj1 => sign_dec_pi1 wfemt
          | Proj2 => sign_dec_pi2 wfemt
          | App => sign_dec_app wfemt
          | Cons_mark => sign_dec_cons wfemt
          | Record flds => sign_dec_record flds wfemt
          | Struct flds => sign_dec_struct flds wfemt
          | Field f => sign_dec_field f wfemt
          | Fixp => sign_dec_fixp wfemt
          | MutInd ind l => sign_dec_indt ind l wfemt
          | MutConstr x => sign_dec_cstr x wfemt
          | MutCase b pats => sign_dec_case pats b wfemt
          | _ => _
        end;
        try (right; red in |- *; intros; InvSignSat ipattern:H ipattern:ty; fail).

      intros top_s.
      right; red in |- *; intros.
      apply top_s.
      assert (wft := type_correctness _ _ _ wf_m).
      case H; intros t' (le_t, wf_t') _; simpl in le_t.
      generalize le_t; clear le_t.
      inversion_clear wft; intros.
      exists s; trivial.

      Inversion_typ wf_t'.
      case cci_topsort_untyped with (1 := le_t) (2 := H0).
      intros s' ax_s; exists s'; constructor; auto; apply (typ_wf _ _ _ wf_m).
    Defined.

    Lemma cci_prod :
      forall e x a b1 b2,
        (e; x : a) ⊦ b1 <= b2 -> e ⊦ Forall x : a, b1 <= Forall x : a, b2.
    Proof.
      intro e; intro x; intro a; intro b1; intro b2;
        intro H; simple apply cci_sub_prd; auto.
    Qed.

    Lemma cci_sum :
      forall e a1 a2 b1 b2,
        e ⊦ a1 <= a2 -> e ⊦ b1 <= b2 -> e ⊦ a1 ** b1 <= a2 ** b2.
    Proof.
      intro e; intro a1; intro a2; intro b1; intro b2;
        intro H; intro H0; simple apply cci_sub_sum; auto.
    Qed.

    Instance CCI_algos : PTSO_algos := {
      ppa_infer_axiom := infer_axiom;
      ppa_infer_sign0 := inf_sign0_dec;
      ppa_infer_sign1 := sign_dec;
      ppa_least_sort := eq_term_sort_dec;
      ppa_infer_rule := infer_rule;
      ppa_weak_rule := wk_rl;
      ppa_le_type_prod := cci_prod;
      ppa_infer_prod := infer_prod;
      ppa_le_type_sum := cci_sum;
      ppa_topsort_untyped := cci_topsort_untyped;
      ppa_le_type_dec := cci_subtype_dec
    }.

  End ISignDecOk.

  Existing Instance CCI_algos.

  Hint Unfold hn_subterm hn_pair: core.
  Hint Resolve @hn_subterm_hn: core.
  Hint Resolve @cci_sub_prd @cci_sub_sum @eq_term_app: core.
  Hint Resolve @delta_eq_term @incl_mark_sub_cci: core.

  Hint Resolve @alpha_stab_alpha_var @alpha_stab_alpha_def: core.
  Hint Unfold conv': core.
