(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export CicMeta.

(* Subject-reduction property for the rules defined in LambdaRules.v *)

Set Implicit Arguments.
Unset Strict Implicit.

Section ISubjRed.

  Context {sg : sigma}.

  (* Subject Reduction of Fixpoint expansion *)

  Lemma sr_fix_free :
    forall e ty fixs T,
      e ⊦ (Fix {(ty,,fixs)}) : T -> e ⊦ (fixs @ ##) @ (Fix {(ty,,fixs)}) : T.
  Proof.
    intros e ty fixs T j.
    Inversion_cst j; rename fixdefs into fds; intros wfp wfu ty_fixb le_T.
    apply typ_app2 with Mark (fix_types fds) (fix_bodies m fds); [split; [ |split] | | | ]; simpl in |- *;
      auto.
    right; split.
    apply cci_trans with (2 := le_T).
    apply incl_mark_sub_cci.
    rewrite commut_lift_subst0.
    rewrite simpl_sl.
    apply fix_body_incl_types.

    apply type_correctness with (1 := j).

    apply incl_mark_sub_cci.
    apply fix_exp_incl.

    elim ty_fixb using inv_pair_sum; trivial.

    apply typ_mt_mark.
    apply typ_wf with (1 := ty_fixb).

    apply Typ_cst1 with (3 := ty_fixb); trivial.
    constructor.
  Qed.


  Lemma sr_iota_fix : rule_sound iota_fix.
  Proof.
    red in |- *.
    simple induction 1; intros.
    apply sr_ctx_app with (2 := H1); intros.
    apply sr_ctx_app with (2 := H2); intros.
    apply sr_ctx_proj1 with (2 := H3).
    clear H3 H2 H1 U0 U T.
    inversion_clear H0.
    intros.
    elim H0 using inv_projn.
    intros v ty_fix le_vect.
    Inversion_cst ty_fix; rename fixdefs into fds; unfold fix_types at 3 in |- *;
      intros wfp wfu ty_fixb le_v.
    elim le_vect with (1 := le_v); intros fixs_tl trnc le_U1.
    apply typ_conv_wf with (mkvect fixs_tl); auto.
    apply proj_vect with (map fix_ty fds); trivial.
    apply sr_fix_free.
    fold (fix_types fds) in |- *.
    apply Typ_cst1 with (3 := ty_fixb); trivial.
    constructor.

    apply type_correctness with (1 := H0).
  Qed.


  Section ISubjRedOk.

    Context {sgn_ok : sign_ok}.


    Lemma sr_iota_case : rule_sound iota_case.
    Proof.
      red in |- *.
      simple destruct 1; intros.
      clear H x y.
      Inversion_cst H1; generalize t_inv u_inv inds is H H2;
        clear t_inv u_inv inds is H H2;
          intros t u isc indc sign_ind
            (m, p', a', l, tybr, se, full_pat, elim_sortO, cbr) wfp wfu ty_case_arg le_T.
      elim ty_case_arg using inv_pair_sum.
      intros ty_cstr typair.
      elim typair using inv_pair_sum.
      intros ty_pred ty_br.
      elim ty_cstr using inv_typ_constr_ind.
      intros inds is cs l' sign_is sign_cs ty_p ty_a eq_p eq_ia eq_mrk incl_mrk.
      rewrite (inj_ind_il sign_ind sign_is) in cbr.
      rewrite (inj_ind_is sign_ind sign_is) in cbr.
      clear ty_p ty_pred typair ty_cstr ty_case_arg wfp elim_sortO full_pat sign_ind
        se t u.
      generalize (match_branch sign_is H0 cbr sign_cs).
      intros (br_arg, str_br_arg, itm_tybr).
      unfold branch_ty in itm_tybr.
      generalize ty_br.
      rewrite mkvect_subst.
      clear ty_br; intro.
      generalize (projl_vect (item_map (fun t => t[0:=p']) itm_tybr) ty_br);
        simpl in |- *; intro.
      apply typ_app' with (3 := H) (4 := ty_a).
      right; simpl in |- *; split.
      rewrite simpl_subst; auto.
      repeat rewrite simpl_sl.
      rewrite lift0.
      apply cci_trans with (2 := le_T).
      apply eq_term_cci_sub.
      apply eq_term_app; auto.
      elim (glob_cstr_eq_name (glob_ind_cst sign_is sign_cs)); trivial.
      red in |- *; auto.

      apply type_correctness with (1 := H1).

      apply cci_subst_var with x_ p'.
      elim
        str_subst_eq_term
        with (1 := str_br_arg) (2 := Rlift0 _ R_lift_eq _ _ _ eq_mrk x_ p');
          intros; auto.
      apply cci_trans with x; auto.
      apply incl_mark_sub_cci.
      apply (incl_arg_str_branch incl_mrk H2).
    Qed.


    (* Subject Reduction alpha *)

    (* Modifier les conv_conv_* pour qu'elles prennent en compte l'alpha *)

    Lemma sr_alpha : rule_sound alpha.
    Proof.
      red in |- *.
      simple induction 1; intros.
      Inversion_typ H0.
      simpl in H3.
      apply typ_conv_wf with (Forall y0 : a, u).
      Inversion_typ H2.
      simpl in H7.
      apply Typ_lam with s3.
      apply alpha_stab_typ' with (e; x0 : a); trivial.
      apply Typ_prd with s1 s2; auto.
      apply alpha_stab_typ' with (e; x0 : a); trivial.

      apply type_correctness with (1 := H0); auto.

      simpl in |- *.
      apply cci_trans with (2 := H3).
      apply eq_term_cci_sub.
      apply eq_term_step.
      do 2 right; left; constructor.

      Inversion_typ H0.
      simpl in H4.
      apply typ_conv_wf with (%%s3); simpl in |- *; auto.
      apply Typ_prd with s1 s2; auto.
      apply alpha_stab_typ' with (e; x0 : a); trivial.

      apply type_correctness with (1 := H0); auto.
    Qed.



    Lemma sr_beta : rule_sound beta.
    Proof.
      red in |- *.
      simple destruct 1; intros.
      Inversion_cst H0; rename a into dom; rename b into rng; rename x1 into z;
        intros wfp wfu typair le_T.
      elim typair using inv_pair_sum.
      intros ty_lam ty_n.
      elim ty_lam using inv_typ_lam_prod.
      intros s1 s2 s3 le_ty ty_a ty_m ty_b rule0.
      apply typ_conv_wf with (rng[0:=n]); simpl in |- *; trivial.
      apply substitution with z dom; trivial.

      apply type_correctness with (1 := H0).
    Qed.


    Lemma sr_proj1 : rule_sound proj1.
    Proof.
      red in |- *.
      simple destruct 1; intros.
      Inversion_cst H0.
      intros.
      elim H3 using inv_pair_sum; intros.
      apply typ_conv_wf with u_inv; auto; simpl in |- *.
      apply type_correctness with (1 := H0).
    Qed.



    Lemma sr_proj2 : rule_sound proj2.
    Proof.
      red in |- *.
      simple destruct 1; intros.
      Inversion_cst H0.
      intros.
      elim H3 using inv_pair_sum; intros.
      apply typ_conv_wf with u_inv; auto; simpl in |- *.
      apply type_correctness with (1 := H0).
    Qed.



    Lemma sr_projr : rule_sound projr.
    Proof.
      red in |- *.
      simple destruct 1; intros.
      Inversion_cst H1; rename u_inv into uu; rename n0 into fnum;
        rename H2 into itf; rename H3 into itval; intros wfrec wfuu tystr le_T.
      generalize H0 itval.
      clear H0 itval wfuu wfrec H y x.
      Inversion_cst tystr.
      intros.
      clear H tystr.
      fold (struct_of md) in itval.
      generalize (inv_record_sub_arity H3); clear H3; intro subrecord.
      elim H2 using inv_pair_sum; clear H2; intros ty_ar ty_body.
      replace n with fnum.
      elim smaller_record_field_type with (2 := itval) (1 := subrecord); intros.
      elim mkvect_expand_ok with (4 := H2) (g := e) (e0 := e); intros; auto.
      apply typ_conv_wf with x1; auto.
      apply projl_vect with (1 := H5); trivial.

      apply type_correctness with (1 := H1).

      simpl in |- *.

      change (@cci_subtype sg e x1 T).
      apply cci_trans with x; trivial.
      apply cci_trans with uu; trivial.

      Inversion_typ H0.
      unfold record_of in H3.
      unfold fields_of in |- *.
      Inversion_cst H3; trivial.

      apply first_item_unique with (1 := H4).
      apply list_unique_first.
      unfold fields_of in |- *; rewrite (same_names subrecord); trivial.

      Inversion_typ H0.
      unfold record_of in H.
      unfold fields_of in |- *.
      Inversion_cst H; trivial.
    Qed.



    Lemma sr_delta : rule_sound delta.
    Proof.
      red in |- *.
      simple destruct 1; intros.
      Inversion_typ H1.
      apply typ_conv_wf with (^(S n)(TYPE d)); simpl in |- *; trivial.
      elim inj_item_rel with (1 := H0) (2 := H2).
      simpl in |- *.
      elim item_trunc with (1 := H2); intros.
      apply thinning with x1; auto.

      apply wf_definition with (1 := H0); auto.
      apply typ_wf with (1 := H1).

      apply typ_wf with (1 := H1).

      apply type_correctness with (1 := H1); auto.
    Qed.


    Lemma sr_delta_glob : rule_sound (delta_glob sg).
    Proof.
      red in |- *.
      simple destruct 1; intros.
      Inversion_cst H2.
      intros.
      generalize (ok_gs_inj H0 H3); intros.
      injection H8; intros.
      rewrite H9 in H1.
      apply typ_conv_wf with ((TYPE (c_decl c0))[0:=m]); simpl in |- *; auto.
      apply substitution with x_ (c_par c0:term); auto.
      generalize c0 H1 (ok_cst H3).
      intros [cpar [nm [ty clt] [[b clb]| ]]]; simpl; intros.
      injection H10.
      simple induction 1.
      apply thin_par_def.
      Inversion_typ H11; trivial.

      apply typ_wf with (1 := H2).

      discriminate H10.

      apply type_correctness with (1 := H2).
    Qed.



    Lemma subj_red : rule_sound red_term.
    Proof.
      change
        (rule_sound
          (reunion delta
            (reunion (delta_glob sg)
              (reunion alpha
                (reunion proj1
                  (reunion proj2
                    (reunion projr
                      (reunion beta
                        (reunion iota_fix
                          (reunion iota_case (fun _ _ _ => False)))))))))))
        in |- *.
      repeat apply reunion_sound.
      exact sr_delta.

      exact sr_delta_glob.

      exact sr_alpha.

      exact sr_proj1.

      exact sr_proj2.

      exact sr_projr.

      exact sr_beta.

      exact sr_iota_fix.

      exact sr_iota_case.

      red in |- *; intros.
      elim H.
    Qed.

    Lemma subj_red_redn : rule_sound redn_term.
    Proof.
      apply incl_sound with (1 := subj_red); intros.
      red in |- *; simpl in |- *.
      inversion_clear H; auto.
      inversion_clear H0; auto 10.
    Qed.


  (* --> Soundness *)
(*
  Lemma incl_red_delta_stab: (e:env)
    (inclusion ? (red_stab (ctxt redn_term) e)
      (stab_delta eq_term cci_subtype e)).
Proof.
(Red; Intros).
Inversion_clear H.
(Split; Simpl).
Apply eq_term_sym_cci_sub.
(Apply redn_red; Trivial).

(Case b; Auto).

(Split; Simpl).
Auto.

(Red; Apply conv_sym).
(Apply redn_red; Trivial).
Save.


  Lemma stab_redn_sub: (R_stable (red_stab (ctxt redn_term)) cci_subtype).
Proof.
Apply stable_incl with 1:=stb_eq_cci.
Exact incl_red_delta_stab.
Save.
*)

    Theorem subject_reduction : rule_sound (red redn_term).
    Proof.
      unfold red in |- *.
      apply clos_refl_trans_sound.
      apply ctxt_sound; simpl in |- *; intros.
      change (@cci_subtype sg e y x).
      apply eq_term_sym_cci_sub.
      apply redn_red; trivial.

      change (@cci_subtype sg e x y).
      apply eq_term_cci_sub.
      apply redn_red; trivial.

      unfold mem_sign1 in |- *.
      apply mem_sign_stable_val with m; trivial.

      exact subj_red_redn.
    Qed.


    Lemma subj_red_wf :
      forall e t u, ⊦ (e; x_ : t) -> red redn_term e t u -> ⊦ (e; x_ : u).
    Proof.
      intros.
      Inversion_typ H.
      apply wf_var with s.
      apply subject_reduction with (2 := H1); trivial.
    Qed.

    Lemma subj_red_wft :
      forall e t u,
        wf_type e t -> red redn_term e t u -> wf_type e u.
    Proof.
      simple destruct 1; intros.
      left with s.
      apply subject_reduction with (2 := H0); trivial.

      elim red_normal with (1 := H0).
      right.

      red in |- *; red in |- *; intros.
      inversion_clear H1.
      elim hn_subterm_normal with (2 := H2); simpl in |- *; trivial.
    Qed.

    Lemma subj_red_ty :
      forall e m t u, e ⊦ m : t -> red redn_term e t u -> e ⊦ m : u.
    Proof.
      intros e m t u j.
      generalize (type_correctness _ _ _ j) j.
      clear j.
      intros [ty s jt| s] j redty.
      apply Typ_conv with ty s; simpl in |- *; auto.
      change (@cci_subtype sg e ty u).
      simple apply eq_term_cci_sub; simple apply redn_conv; exact redty.
      apply subject_reduction with (2 := jt); trivial.

      elim red_normal with (1 := redty); trivial.
      red in |- *; red in |- *; intros.
      inversion_clear H.
      elim hn_subterm_normal with (2 := H0); simpl in |- *; trivial.
    Qed.

  End ISubjRedOk.

End ISubjRed.
