(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export Cumulativity.
Require Export InvEqual.

Section ISubtyping.

  Context {sg : sigma}.

Set Implicit Arguments.
Unset Strict Implicit.

  Definition cci_step := reunion cumul ind_marks.
  Definition cci_le := reunion eq_term (clos_le cci_step).
  Hint Unfold cci_le cci_step: core.

  Definition cci_subtype := R_rt cci_le.

  Let le_sort_subtype (s1 s2 : sort) : Prop := forall e, cci_subtype e %%s1 %%s2.
  Hint Unfold le_sort_subtype : core.


  Lemma cci_refl : forall e x, cci_subtype e x x.
Proof fun e => rt_refl _ (cci_le e).
Hint Resolve cci_refl: core.

  Lemma cci_trans :
   forall e x y z,
   cci_subtype e x y -> cci_subtype e y z -> cci_subtype e x z.
Proof fun e => rt_trans _ (cci_le e).


  Lemma eq_term_cci_sub : forall e a b, eq_term e a b -> cci_subtype e a b.
Proof.
intros.
red in |- *; red in |- *; apply rt_step; left; trivial.
Qed.

  Lemma eq_term_sym_cci_sub :
   forall e a b, eq_term e b a -> cci_subtype e a b.
Proof.
intros.
red in |- *; red in |- *; apply rt_step; left; red in |- *; apply conv_sym;
 trivial.
Qed.


  Lemma cci_sub_step : forall e a b, cci_step e a b -> cci_subtype e a b.
Proof.
intros.
red in |- *; red in |- *; apply rt_step; right; apply Clos_rule; trivial.
Qed.

  Hint Resolve eq_term_cci_sub eq_term_sym_cci_sub cci_sub_step: core.

(* TODO: alpha-conversion *)
  Lemma cci_sub_prd :
   forall e x a1 b1 a2 b2,
   cci_subtype e a2 a1 ->
   cci_subtype (e; x : a2) b1 b2 ->
   cci_subtype e (Forall x : a1, b1) (Forall x : a2, b2).
Proof.
intros.
red in |- *; red in |- *; apply rt_trans with (Forall x : a2, b1).
elim H; intros; auto.
apply rt_step.
case H1; intros; auto.
left.
red in |- *; auto.

right; auto.

apply rt_trans with (1 := H4) (2 := H2).

elim H0; intros; auto.
apply rt_step.
case H1; intros; auto.
left.
red in |- *; auto.

right; auto.

apply rt_trans with (1 := H2) (2 := H4).
Qed.


  Lemma cci_sub_sum :
   forall e a1 b1 a2 b2,
   cci_subtype e a1 a2 ->
   cci_subtype e b1 b2 -> cci_subtype e (a1 ** b1) (a2 ** b2).
Proof.
red in |- *; red in |- *; intros.
apply rt_trans with (a2 ** b1).
elim H; intros; auto.
apply rt_step.
case H1; intros.
left; red in |- *; auto.

right; auto.

apply rt_trans with (1 := H2) (2 := H4).

elim H0; intros; auto.
apply rt_step.
case H1; intros.
left; red in |- *; auto.

right; auto.

apply rt_trans with (1 := H2) (2 := H4).
Qed.


  Hint Resolve cci_sub_prd cci_sub_sum: core.



  (* Mark weakening is included in the subtyping *)

  Lemma incl_mark_sub_cci :
   forall e x y, incl_mark e x y -> cci_subtype e x y.
Proof.
intros.
red in |- *; red in |- *; apply rt_step; right.
apply incl_clos_le with ind_marks; auto.
Qed.

  Hint Resolve incl_mark_sub_cci: core.


  (* Consistency with De Bruijn indexes *)
  Lemma subtype_lift : R_lift cci_subtype.
Proof
  R_rt_lift _
    (reunion_lift _ _ R_lift_eq
       (R_lift_clos (reunion_lift _ _ R_lift_cumul R_lift_ind_marks))).



  Lemma subtype_subst : R_subst cci_subtype.
Proof
  R_rt_subst _
    (weak_subst _
       (reunion_subst _ _ R_subst_eq
          (R_subst_clos (reunion_subst _ _ R_subst_cumul R_subst_ind_marks)))).

  Lemma subtype_stable : forall R, R_stable (rel_typ R) cci_subtype.
Proof
  fun R =>
  stable_rt _ _
    (reunion_stable _ _ _ (R_stable_eq (R:=R))
       (stable_clos
          (reunion_stable _ _ _ (stable_cumul (R:=rel_typ R))
             (stable_ind_marks (R:=rel_typ R))))).

  Lemma subtype_stable_special :
   forall R,
   R_stable (decl_rel (fun (_ : env) (t1 t2 : term) => t1 = t2) R)
     cci_subtype.
Proof
  fun R =>
  let stb := decl_rel (fun (_ : env) (t1 t2 : term) => t1 = t2) R in
  stable_rt _ _
    (reunion_stable _ _ _ (R_stable_eq_special (R:=R))
       (stable_clos
          (reunion_stable _ _ _ (stable_cumul (R:=stb)) (stable_ind_marks (R:=stb))))).

  (* cci_subtype is a subtyping rule *)
  Definition cci_subtyping_rule :=
    Build_Subtyping_rule _ subtype_lift subtype_subst
      (subtype_stable (R:=cci_subtype))
      (fun e => clos_rt_is_preorder _ (cci_le e)).




  Lemma cci_thin :
   forall e d a b, cci_subtype e a b -> cci_subtype (d :: e) ^(1)a ^(1)b.
Proof.
intros.
unfold lift in |- *.
apply subtype_lift with e e d; auto.
Qed.

  Lemma cci_subst :
   forall e x m t a b,
   cci_subtype (e; x := m : t) a b -> cci_subtype e a[0:=m] b[0:=m].
Proof.
intros.
apply subtype_subst with (mkdecl x t (Some m)) e (e; x := m : t); auto.
unfold add_def in |- *; constructor.
compute in |- *; reflexivity.
Qed.


  Lemma cci_subst_var :
   forall e x m t a b,
   cci_subtype (e; x : t) a b -> cci_subtype e a[0:=m] b[0:=m].
Proof.
intros.
apply subtype_subst with (mkdecl x t None) e (e; x : t); auto.
unfold add_var in |- *; constructor.
compute in |- *; trivial.
Qed.


  Lemma cci_subst_eq_cci :
   forall e x m t u a b,
   eq_term e t u ->
   cci_subtype (e; x : m) a b -> cci_subtype e a[0:=t] b[0:=u].
intros.
apply cci_trans with (a[0:=u]).
apply eq_term_cci_sub.
apply eq_term_subst_l; trivial.

apply cci_subst_var with (1 := H0).
Qed.




  Lemma sub_sort_env_indep :
   forall e s1 s2, cci_subtype e %%s1 %%s2 -> le_sort_subtype s1 s2.
Proof.
intros.
cut (cci_subtype nil %%s1 %%s2); intros.
red in |- *.
simple induction e0; intros; auto.
exact (cci_thin a H1).

generalize H.
clear H.
elim e; trivial.
intros (x, t, [b| ]) f lef lee.
apply lef.
exact (cci_subst lee).

apply lef.
exact (cci_subst_var t lee).
Qed.

(*
Lemma cci_sub_inv: (e:env)(x:name)(m,t,a,b:term)
    [~ sg | e |- {0:=m}a <= {0:=m}b ~]
      ->[~ sg | e; x:=m:t |- a <= b ~].
Proof.
Intros.
Apply cci_trans with [: ^(1){0:=m}a :].
Apply eq_term_cci_sub.
Apply delta_eq_term.
Apply delta_norm0.

Apply cci_trans with [: ^(1){0:=m}b :].
Unfold add_def.
(Apply cci_thin; Trivial).

Apply eq_term_cci_sub.
Apply eq_term_sym.
Apply delta_eq_term.
Apply delta_norm0.
Save.
*)



Ltac SplitRule Hyp HR :=
  elim Hyp; clear Hyp; [ intro HR; left | intro Hyp; right ].


  Lemma stb_eq_eq : R_stable (decl_rel eq_term cci_subtype) eq_term.
Proof.
red in |- *.
unfold eq_term in |- *.
simple induction 1; intros; auto.
generalize f H1.
clear H1 f.
elim H0; intros; auto.
inversion_clear H1.
inversion_clear H3.
elim rel_list_item with (1 := H1) (2 := H2); intros.
apply delta_eq_term.
red in |- *; red in |- *; apply rt_step; constructor.
split with x0 t1; auto.

elim item_trunc with (1 := H1); intros.
elim H3 with (1 := H4); intros.
elim H5.
clear H5.
intros (xx, ty, [bd| ]); intros.
inversion_clear H5.
simpl in sub_ty0, eq_body0.
apply eq_term_trans with (^(S n)bd).
apply delta_eq_term.
red in |- *; red in |- *; constructor; constructor.
split with xx ty; auto.

apply (iter_R_lift _ R_lift_eq) with x1; auto.

inversion_clear H5.
simpl in eq_body0.
elim eq_body0.

apply eq_term_step.
red in |- *; simpl in |- *.
right.
repeat SplitRule H3 ipattern:rule_hyp.
apply
 (stable_delta_glob (sg:=sg) (R:=decl_rel eq_term cci_subtype))
  with (1 := rule_hyp) (2 := H2).

apply
 (stable_alpha (R:=decl_rel eq_term cci_subtype))
  with (1 := rule_hyp) (2 := H2).

apply
 (stable_proj1 (R:=decl_rel eq_term cci_subtype))
  with (1 := rule_hyp) (2 := H2).

apply
 (stable_proj2 (R:=decl_rel eq_term cci_subtype))
  with (1 := rule_hyp) (2 := H2).

apply
 (stable_projr (R:=decl_rel eq_term cci_subtype))
  with (1 := rule_hyp) (2 := H2).

apply
 (stable_beta (R:=decl_rel eq_term cci_subtype))
  with (1 := rule_hyp) (2 := H2).

apply
 (stable_iota_fix (R:=decl_rel eq_term cci_subtype))
  with (1 := rule_hyp) (2 := H2).

apply
 (stable_iota_case (R:=decl_rel eq_term cci_subtype))
  with (1 := rule_hyp) (2 := H2).

trivial.

apply conv_sym; auto.

apply conv_trans with y; auto.
Qed.


  Lemma stb_eq_cci : R_stable (decl_rel eq_term cci_subtype) cci_subtype.
Proof.
unfold cci_subtype at 2 in |- *.
apply stable_rt.
unfold cci_le in |- *.
apply reunion_stable.
exact stb_eq_eq.

apply stable_clos.
unfold cci_step in |- *.
apply reunion_stable.
apply stable_cumul.

apply stable_ind_marks.
Qed.



  (* Record subtyping *)
  Inductive sub_record_step : env -> rspec -> rspec -> Prop :=
    | Subrec1_hd :
        forall e a b x r,
        cci_le e a b -> sub_record_step e ((x, a) :: r) ((x, b) :: r)
    | Subrec1_tl :
        forall e x ty r1 r2,
        sub_record_step (e; x : ty) r1 r2 ->
        sub_record_step e ((x, ty) :: r1) ((x, ty) :: r2).

  Lemma sub_record_step_names :
   forall e r1 r2, sub_record_step e r1 r2 -> field_names r1 = field_names r2.
simple induction 1; simpl in |- *; auto.
simple destruct 2; trivial.
Qed.



  Inductive sub_record : env -> rspec -> rspec -> Prop :=
    | Subrec_nil : forall e, sub_record e nil nil
    | Subrec_cons :
        forall e x a1 a2 s1 s2,
        cci_subtype e a1 a2 ->
        sub_record (e; x : a1) s1 s2 ->
        sub_record e ((x, a1) :: s1) ((x, a2) :: s2).


  Lemma sub_record_intro :
   forall e s1 s2, clos_le_record cci_step e s1 s2 -> sub_record e s1 s2.
Proof.
simple induction 1; constructor; auto.
red in |- *; red in |- *; apply rt_step; auto.
Qed.

  Lemma sub_record_refl : forall sign e, sub_record e sign sign.
Proof.
simple induction sign.
constructor.

intros (x, ty); intros; simpl in |- *.
constructor; auto.
Qed.

(*
Lemma sub_record_step_sub:
  (e,r1,r2:?)(sub_record_step e r1 r2) -> (sub_record e r1 r2).
Induction 1;Constructor 2; Auto.
Red;Red;Auto.

Apply sub_record_refl.
Save.
*)


Lemma sub_record_sub_record_step :
 forall e r1 r2,
 sub_record e r1 r2 -> clos_refl_trans _ (sub_record_step e) r1 r2.
simple induction 1; intros.
apply rt_refl.

apply rt_trans with ((x, a1) :: s2).
elim H2; intros; auto.
apply rt_step.
constructor; trivial.

apply rt_trans with (1 := H4); trivial.

elim H0; intros; auto.
apply rt_step; constructor; trivial.

apply rt_trans with (1 := H4); trivial.
Qed.


  Lemma sub_record_stable :
   forall e s1 s2,
   sub_record e s1 s2 ->
   forall f, rel_in_list (rel_typ cci_subtype) f e -> sub_record f s1 s2.
Proof.
simple induction 1.
constructor.

clear e s1 s2 H.
intros e x a1 a2 s1 s2 le_a1_a2 sub_s1_s2 sub_rec f le_f_e.
constructor.
apply (subtype_stable le_a1_a2 le_f_e).

apply sub_rec; auto.
Qed.


  Lemma sub_record_trans :
   forall e s1 s2,
   sub_record e s1 s2 -> forall s3, sub_record e s2 s3 -> sub_record e s1 s3.
simple induction 1; intros.
inversion_clear H0.
constructor.

inversion_clear H3.
constructor; auto.
apply cci_trans with a2; trivial.

apply H2.
apply sub_record_stable with (1 := H5).
auto.
unfold add_var in |- *.
constructor.
red in |- *.
simpl in |- *.
auto.
Qed.

  Lemma same_names :
   forall e s1 s2, sub_record e s1 s2 -> field_names s1 = field_names s2.
Proof.
unfold field_names in |- *; simple induction 1; simpl in |- *; intros;
 trivial.
elim H2; trivial.
Qed.

  Lemma incl_arity :
   forall e s1 s2,
   sub_record e s1 s2 -> cci_subtype e (prod_arity s2) (prod_arity s1).
Proof.
simple induction 1; simpl in |- *; intros; auto.
Qed.


  Lemma incl_arity_step :
   forall e s1 s2,
   sub_record_step e s1 s2 ->
   eq_term e (record_arity s1) (record_arity s2) \/
   clos_le_record cci_step e s1 s2.
simple induction 1; simpl in |- *; intros; auto.
destruct H0; auto.
left; red in |- *; auto.

destruct H1; auto.
left; red in |- *; auto.
Qed.

  Lemma record_sub_arity :
   forall e s1 s2,
   sub_record e s1 s2 -> cci_subtype e (record_of s1) (record_of s2).
intros.
assert (sr_st := sub_record_sub_record_step H).
elim sr_st; intros; auto.
red in |- *; red in |- *; apply rt_step.
elim (incl_arity_step H0); intros.
left.
unfold record_of in |- *.
rewrite (sub_record_step_names H0).
red in |- *; auto.

right.
apply Clos_record; trivial.

apply cci_trans with (1 := H1); trivial.
Qed.


  Lemma sub_arity_field :
   forall m s1 s2 e,
   sub_record e s1 s2 ->
   forall k,
   cci_subtype e (mkvect (fields_type (expand_sign m s1) k))
     (mkvect (fields_type (expand_sign m s2) k)).
Proof.
simple induction 1; simpl in |- *; intros.
auto.

apply cci_sub_sum; trivial.
rewrite <- mkvect_subst.
rewrite <- mkvect_subst.
apply cci_subst_var with x a1.
trivial.
Qed.

(* Plutot utiliser le resultat precedent... *)
  Lemma smaller_record_field_type :
   forall m e s2 s1,
   sub_record e s1 s2 ->
   forall n k x,
   item x (fields_type (expand_sign m s2) k) n ->
   exists2 y : _,
     cci_subtype e y x & item y (fields_type (expand_sign m s1) k) n.
Proof.
simple induction 1; simpl in |- *; intros.
inversion H0.

inversion_clear H3.
split with a1; auto.

fold (expand_sign m s0) in |- *.
fold (expand_sign m s3) in H4.
elim map_inv_item with (1 := H4); intros.
rewrite H5.
elim H2 with (1 := H3); intros.
split with (x2[0:=^(k)m..( x)]).
apply cci_subst_var with x a1; trivial.

constructor.
apply item_map' with x2; auto.
Qed.

(* ressemble au resultat precedent *)
  Lemma mkvect_expand_ok :
   forall (g : env) mc,
   list_unique (fields_of mc) ->
   forall md e k,
   trunc k e g ->
   trunc k mc md ->
   forall n tf,
   item tf (fields_type (expand_sign (struct_of mc) (rspec_of md)) k) n ->
   exists2 tf' : _, cci_subtype e tf' tf & item tf' (fields_type md k) n.
Proof.
simple induction md; intros.
simpl in H2.
inversion H2.

generalize a H2 H3; clear H2 H3 a; intros ((x, ty), v) H2.
simpl in |- *.
fold rspec_of in |- *.
fold (expand_sign (struct_of mc)) in |- *.
intros.
inversion_clear H3.
split with ty; auto.

elim map_inv_item with (1 := H4); intros.
clear H4.
elim H0 with (3 := H3) (e := (e; x : ty)); auto; intros.
split with (x1[0:=^(k)v]).
rewrite H5.
apply cci_subst_eq_cci with x ty; auto.
apply (iter_R_lift _ R_lift_eq) with g; trivial.
apply eq_term_sym.
apply eq_term_trans with (vals_of mc .^ k) ..L.
apply redn_red.
constructor.
unfold struct_of in |- *.
do 4 right; left; constructor.
apply list_unique_first; trivial.
unfold fields_of, field_names in |- *.
fold (fst (x, ty)) in |- *.
apply item_map.
unfold rspec_of in |- *.
fold (fst (x, ty, v)) in |- *.
apply item_map; trivial.
apply (trunc_it H2).

unfold vals_of in |- *.
elimtype (item v (map (snd (B:=_)) mc) k); simpl in |- *; intros; auto.
apply redn_red.
constructor.
do 2 right; left; constructor.

apply eq_term_trans with (2 := H8).
red in |- *.
apply conv_conv_cst.
fold eq_term in |- *.
apply eq_term_proj.
apply redn_red.
constructor.
do 3 right; left; constructor.

fold (snd (x, ty, v)) in |- *.
apply item_map.
apply (trunc_it H2).

constructor; apply item_map' with x1; auto.

apply trunc_inv_r with (1 := H2).
Qed.




(*
  Record sub_arity [e:env; s1,s2:rspec]: Prop :=
    { same_names: (field_names s1)=(field_names s2);
      incl_arity: (cci_subtype e (prod_arity s2) (prod_arity s1)) }.
*)

End ISubtyping.

Hint Unfold cci_le cci_step: core.
Hint Resolve @cci_refl: core.
Hint Resolve @eq_term_cci_sub @eq_term_sym_cci_sub @cci_sub_step: core.
Hint Resolve @cci_sub_prd @cci_sub_sum: core.
Hint Resolve @incl_mark_sub_cci: core.

Delimit Scope cci_scope with cci.

Notation "e ⊦ t <= u" := (cci_subtype e t u) (at level 50, t at next level) : cci_scope.
