(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export EqualDec.

Set Implicit Arguments.
Unset Strict Implicit.

Section ISubtypingDec.

  Context {sg : sigma}.

  Hypothesis sign_ok_glob_inj : forall x g1 g2,
    sign_glob_spec sg x g1 -> sign_glob_spec sg x g2 -> g1 = g2.

  Axiom unsafe_sn : forall e t, sn (ctxt redn_term) e t.
  Axiom unsafe_acc : forall A x, @Acc A (fun x y => True) x.

  Lemma sn_ctxt : forall e t u,
    red redn_term e t u ->
    sn (ctxt redn_term) e t ->
    sn (ctxt redn_term) e u.
  Proof.
    induction 1; intro Hsn.
    destruct Hsn as [Hsn].
    apply Hsn; assumption.
    assumption.
    auto.
  Defined.

  Definition seq (R R' : red_rule) :=
    fun e x y => exists z, R e x z /\ R' e y z.

  Lemma subtype_dec :
    forall e a b,
      sn (ctxt redn_term) e a ->
      sn (ctxt redn_term) e b -> decide (cci_subtype e a b).
  Proof.
    intros e a b Ha Hb.
    assert (H := unsafe_acc (a, b)).
    clear - H sign_ok_glob_inj; revert e; elim H; clear - sign_ok_glob_inj.
    intros [t u] H subtype_dec e.
    refine (
      let (t', Htt', Ht') := @whnf_cci_sn sg sign_ok_glob_inj e t _ in
      let (u', Huu', Hu') := @whnf_cci_sn sg sign_ok_glob_inj e u _ in
      let true := left _ _ in let false := right _ _ in
      let Ht := _ in let Hu := _ in
      match t' with
        | Srt s1 =>
          match u' with
            | Srt s2 => if univ_v6_dec s1 s2 then true else false
            | _ => false
          end
        | Sum a1 b1 =>
          match is_sum_dec u' with
            | inleft (Is_sum a2 b2) =>
              if subtype_dec (a1, a2) _ e then
                if subtype_dec (b1, b2) _ e then true else false
              else
                false
            | inright _ => false
          end
        | Prd x t1 m1 => _
          match is_prd_dec u' with
            | inleft (Is_prd y t2 m2) =>
              if subtype_dec (t2, t1) _ e then
                if subtype_dec (m1, m2) _ (add_var e y t2) then true else false
              else
                false
            | inright _ => false
          end
        | Cst1 (Record f1) arg1 =>
          match is_record u' with
            | inleft (exist (f2, arg2) _) =>
              if f1 == f2 then
                _
              else
                false
            | _ => false
          end
        | Cst1 (MutInd i1 m1) (p1,,a1,,l1) =>
          match is_ind_dec u' with
            | inleft (Is_ind i2 m2 p2 a2 l2) =>
              if i1 == i2 then
                if @convert sg sign_ok_glob_inj e (p1,,a1) (p2,,a2) _ _ then true else false
              else
                false
            | _ => false
          end
        | _ => if @convert sg sign_ok_glob_inj e t' u' Ht Hu then true else false
      end
    ); try (apply unsafe_sn || exact I).
    admit.
    admit.
    clear - arg1 arg2 subtype_dec sign_ok_glob_inj.
    revert e.
    assert (H := unsafe_acc (arg1, arg2)).
    elim H; clear - subtype_dec sign_ok_glob_inj.
    intros [t u] H sub_record_rec e.
    refine (
      let (t', Htt', Ht') := @whnf_cci_sn sg sign_ok_glob_inj e t _ in
      let (u', Huu', Hu') := @whnf_cci_sn sg sign_ok_glob_inj e u _ in
      let true := left _ _ in let false := right _ _ in
      let Ht := _ in let Hu := _ in
      match t', u' with
        | Sum a1 p1, Sum a2 p2 =>
          let (p1', Hp1', Hp1) := @whnf_cci_sn sg sign_ok_glob_inj e p1 _ in
          let (p2', Hp2', Hp2) := @whnf_cci_sn sg sign_ok_glob_inj e p2 _ in
          match p1', p2' with
            | Prd x _ r1, Prd _ _ r2 =>
              if subtype_dec (a1, a2) _ e then
                if sub_record_rec (r1, r2) _ (add_var e x a1) then true else false
              else
                false
            | _, _ => false
          end
        | _, _ => if @convert sg sign_ok_glob_inj e t' u' Ht Hu then true else false
      end
    ); try (apply unsafe_sn || exact I).
    admit.
    admit.
    trivial.
  Qed.

End ISubtypingDec.
