(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export Signature.
Require Export Marks.

Set Implicit Arguments.
Unset Strict Implicit.

Open Scope nat_scope.

  Transparent sort.
  Definition elim_sort : sort -> sort -> sort -> Prop := elim_v6.
  Opaque sort.


Section TypCase.

  Definition branch_ty c arg q :=
    Forall x_ : arg, (^(2)q @ c_inst c) @ Cstr{c_name c // #1,,#0}.

  Variable inds : list name.
  Variable is : one_ind.

  (* Adding the substructure annotation in the type of the constructor
     argument. The most marked term is not stable by substitution! *)


  Inductive str_substp : term -> term -> term -> Prop :=
    | Strm_refl : forall l t, str_substp l t t
    | Strm_ind :
        forall p a l x,
        In x inds ->
        str_substp l Ind{x // p,,(a,,[])} Ind{x // p,,(a,,l)}
    | Strm_prd :
        forall a b b' l x,
        str_substp ^(1)l b b' ->
        str_substp l (Forall x : a, b) (Forall x : a, b')
    | Strm_sum :
        forall a a' b b' l,
        str_substp l a a' ->
        str_substp l b b' -> str_substp l (a ** b) (a' ** b')
    | Strm_record :
        forall l s1 s2,
        mark_record l s1 s2 -> str_substp l (record_of s1) (record_of s2)
with mark_record : term -> rspec -> rspec -> Prop :=
  | Ma_rec_nil : forall l, mark_record l nil nil
  | Ma_rec_cons :
      forall x l ty1 ty2 s1 s2,
      str_substp l ty1 ty2 ->
      mark_record ^(1)l s1 s2 ->
      mark_record l ((x, ty1) :: s1) ((x, ty2) :: s2).


  Scheme str_substp_mind := Minimality for str_substp
    Sort Prop
    with mark_record_mind := Minimality for mark_record 
    Sort Prop.

(* Schema de recurrence sur les types:


  Lemma mark_branch_arg: (e:env)(l,a:term)
    { arg:term | (ppal [arg](str_substp l a arg) (incl_mark e) arg) }.
Proof.
(Fix   mbr 3
  with mrec 3:(e:env; l,a:term)
          {sp:rspec | ((sign:rspec) a=(record_arity sign)
               ->(ppal [s:rspec](mark_record l sign s)
                   [s1,s2:rspec](incl_mark e (prod_arity s2) (prod_arity s1))
                   sp))}; Intros).
Intros.
(Elim a using typ_ctxt_rec
  with Pr:=[sign:rspec]
            {sp:rspec |
               (ppal [s:rspec](mark_record l sign s)
                 [s1,s2:rspec]
                  (incl_mark e (prod_arity s2) (prod_arity s1)) sp)};
 Intros).




  Lemma typ_ctxt_rec: (P:term->Set)(Pr:rspec->Set)
      ((t:term)((is_sum t)->False)->((is_prd t)->False)->((is_ind t)->False)
          ->((sign:rspec)~t=(record_of sign))
       ->(P t))
    ->((a,b:term)(P a)->(P b)->(P [: a*b :]))
    ->((x:name)(a,b:term)(P a)->(P b)->(P [: (x:a)b :]))
    ->((i:name)(m:le_marks)(p,a,l:term)(P p)->(P a)->(P l)
                 ->(P [: Ind{i;m|(p,(a,l))} :]))
    ->((sign:rspec)(Pr sign)->(P (record_of sign)))
    ->(Pr (nil ?))
    ->((x:name)(ty:term)(sign:rspec)(P ty)->(Pr sign)->(Pr (cons (x,ty) sign)))
    ->(t:term)(P t).
Intros P Pr H H0 H1 H2 H3 H4 H5.
Fix t 1
 with typ_ctxt_record 1:(a:term; sign:rspec)
                         a=(record_arity sign)->(Pr sign).
(Destruct t; Intros; Try (Apply H; Try Red; Intros; Inversion H6; Fail)).
(Elim is_ind_op' with c; Intros).
Inversion_clear y.
Inversion_clear H6.
(Elim is_pair_dec with t1; ([ Intros ((a, b), eqt1) | Intros ])).
(Elim is_pair_dec with b; ([ Intros ((ba, bb), eqb) | Intros ])).
Replace (Cst1 c t1) with [: Ind{x; x0 | (a, (ba, bb))} :].
(Apply H2; Trivial).

Rewrite H7.
Rewrite eqt1.
(Rewrite eqb; Reflexivity).

(Apply H; Try Red; Intros; Inversion H6).
Elim y with a0 l.
Rewrite eqt1 in H10.
(Injection H10; Trivial).
Auto.

Unfold record_of in H9.
Unfold record_of in H10.
Rewrite eqt1 in H10.
Generalize H10.
(Generalize sign; Intros [ | (x, ty) s]; Intros; Discriminate H8).

<Your Tactic Text here>
<Your Tactic Text here>
Apply H1.
(Apply t; Trivial).

(Apply t; Trivial).



(Destruct a; Intros;
 Try (Split with (nil name*term); Intros [ | (x, ty) ll]; Intros;
      Discriminate H6)).
*)


  Lemma mark_record_names :
   forall l s1 s2, mark_record l s1 s2 -> field_names s1 = field_names s2.
Proof.
unfold field_names in |- *.
simple induction 1; simpl in |- *; intros; auto.
elim H2; trivial.
Qed.

  Lemma mark_record_refl : forall s l, mark_record l s s.
Proof.
simple induction s; intros.
constructor.

case a; intros.
constructor; trivial.
constructor.
Qed.



  Lemma str_sub_lift :
   forall n l a a',
   str_substp l a a' -> forall k, str_substp ^(n//k)l ^(n//k)a ^(n//k)a'.
Proof.
intros n l a a' H.
elim H using
 str_substp_mind
  with
    (P0 := fun l s1 s2 =>
           forall k,
           mark_record ^(n//k)l (lift_rsign n s1 k) (lift_rsign n s2 k));
 intros; try (simpl in |- *; constructor; auto).
rewrite permute_lift; trivial.

rewrite record_of_lift.
rewrite record_of_lift.
constructor; trivial.

rewrite permute_lift; trivial.
Qed.

  Lemma str_sub_subst :
   forall m l a a',
   str_substp l a a' -> forall k, str_substp l[k:=m] a[k:=m] a'[k:=m].
Proof.
intros m l a a' H.
elim H using
 str_substp_mind
  with
    (P0 := fun l s1 s2 =>
           forall k,
           mark_record l[k:=m] (subst_rsign m s1 k) (subst_rsign m s2 k));
 intros; try (simpl in |- *; constructor; auto).
rewrite <- commut_lift_subst; trivial.

rewrite record_of_subst.
rewrite record_of_subst.
constructor; trivial.

rewrite <- commut_lift_subst; trivial.
Qed.


  Lemma str_sub_smaller :
   forall l b b', str_substp l b b' -> forall e, incl_mark e b' b.
Proof.
intros l b b' H.
elim H using
 str_substp_mind
  with
    (P0 := fun (l : term) s1 s2 => forall e, clos_le_record ind_marks e s2 s1);
 unfold lift, incl_mark in |- *; intros; auto.
apply incl_mt_mark.
Qed.



(* Mark the type of the argument of a branch *)
(* TODO: utiliser is_ind, cf positive_dec
 * Pb: condition de garde
 *)
  Definition str_substp_support a :=
    match a with
    | Forall _ : _, _ => False
    | _ ** _ => False
    | Ind {ind // (_,,(_,,[]))} => False
    | {<_ : _>} => False
    | _ => True
    end.

  Lemma str_substp_triv :
   forall e l a,
   str_substp_support a ->
   {arg : _ | ppal (fun arg => str_substp l a arg) (incl_mark e) arg}.
Proof.
intros e l a strs.
split with a.
split; intros.
constructor.

generalize strs.
clear strs.
case H; intros; try case strs.
auto.
Qed.


  Lemma mark_branch_arg :
   forall e l a,
   {arg : _ | ppal (fun arg => str_substp l a arg) (incl_mark e) arg}.
Proof.
fix mbr 3
 with (mrec (e:_) (l:_) (a:_) {struct a} :
       {sp : _ |
       forall sign,
       a = record_arity sign ->
       ppal (fun s => mark_record l sign s)
         (fun s1 s2 => clos_le_record ind_marks e s1 s2) sp}); 
 intros.
(* Astuce, mais bugs en cascade, et prog extrait pas beau *)
refine
 match
   a as x
   return
     {arg : _ | ppal (fun arg0 => str_substp l x arg0) (incl_mark e) arg}
 with
 | Forall x : t, u => _
 | t ** u => _
 | Cst1 _ t => _
 | x => str_substp_triv e l (a:=x) I
 end.
refine
 match
   o as c
   return
     {arg : _ |
     ppal (fun arg0 => str_substp l (Cst1 c t) arg0) (incl_mark e) arg}
 with
 | Record f => _
 | MutInd ind m => _
 | cc => str_substp_triv e l (a:=Cst1 cc t) I
 end.
elim is_record_arity with f t; intros.
inversion_clear a0.
elim mrec with e l t; intros.
split with (record_of x0).
inversion_clear H.
generalize (p _ H1); intros.
rewrite H0; rewrite H1; clear p H0 H1 f t.
fold (record_of x) in |- *.
split; intros.
constructor.
apply (pp_ok H).

generalize (rspec_record_inj (s1:=x)).
unfold record_of in H0.
inversion_clear H0;
 [ fold (record_of x) in |- * | unfold record_of at 1 in |- * ];
 intros.
red in |- *.
apply Clos_record.
apply (pp_least H).
apply mark_record_refl.

red in |- *.
fold (record_of x0) in |- *.
apply Clos_record.
apply (pp_least H).
rewrite H0 with s1; trivial.

split with {<f : t>}.
split.
constructor.

intros.
generalize b.
inversion_clear H; unfold record_of in |- *; intros; auto.
elim b0 with s1; auto.

case m; intros.
2: exact (str_substp_triv e l (a:=Ind{ind; S n // t}) I).
refine
 match
   t as x
   return
     {arg : _ |
     ppal (fun arg0 => str_substp l Ind{ind // x} arg0) (incl_mark e) arg}
 with
 | (u,,v) => _
 | x => str_substp_triv e l (a:=Ind{ind // x}) I
 end.
refine
 match
   v as x
   return
     {arg : _ |
     ppal (fun arg0 => str_substp l Ind{ind // u,,x} arg0) 
       (incl_mark e) arg}
 with
 | (_,,ll) => _
 | x => str_substp_triv e l (a:=Ind{ind // u,,x}) I
 end.
refine
 match
   ll as x
   return
     {arg : _ |
     ppal (fun arg0 => str_substp l Ind{ind // u,,(t0,,x)} arg0)
       (incl_mark e) arg}
 with
 | [] => _
 | x => str_substp_triv e l (a:=Ind{ind // u,,(t0,,x)}) I
 end.
elim (list_index (@eq_dec name _)) with ind inds; intros.
cut (In ind inds); intros.
split with Ind{ind // u,,(t0,,l)}.
split; intros.
constructor; trivial.

inversion_clear H0; auto.
apply incl_mt_mark.

inversion_clear a0; auto.
elim H; intros; auto.

split with Ind{ind // u,,(t0,,[])}.
split; intros.
constructor.

inversion_clear H.
auto.

elim b; trivial.

elim (mbr (e; x : t) ^(1)l u); intros.
split with (Forall x : t, x0).
split; intros.
constructor.
apply (pp_ok p).

inversion_clear H.
red in |- *.
apply Clos_prd; auto.
apply (pp_least p).
constructor.

red in |- *; apply Clos_prd; auto.
apply (pp_least p).
trivial.

elim (mbr e l t); intros.
elim (mbr e l u); intros.
split with (x ** x0).
split; intros.
constructor.
apply (pp_ok p).

apply (pp_ok p0).

inversion_clear H.
red in |- *.
apply Clos_sum.
apply (pp_least p).
constructor.

apply (pp_least p0).
constructor.

red in |- *.
apply Clos_sum.
apply (pp_least p); trivial.

apply (pp_least p0); trivial.

(* mark_record *)
case a; intros;
 try
  (split with (nil (A:=name * term)); intros [| (x, ty) ll]; intros;
    discriminate H).
split with (nil (A:=name * term)); intros [| (x, ty) ll]; simpl in |- *;
 intros.
split.
constructor.

intros.
inversion_clear H0; simpl in |- *; auto.

discriminate H.

case t0; intros;
 try
  (split with (nil (A:=name * term)); intros [| (x, ty) ll]; intros;
    discriminate H).
elim (mbr e l t); intros.
elim (mrec (e; n : x) ^(1)l t2); intros.
split with ((n, x) :: x0).
intros [| (x1, ty) ll]; simpl in |- *; intros.
discriminate H.

injection H; intros.
generalize (p0 _ H0).
elim H2.
elim H3.
clear H3 H2 H1 H0 H p0; intros.
split; intros.
constructor.
apply (pp_ok p).

apply (pp_ok H).

inversion_clear H0; simpl in |- *.
apply Clos_field.
apply (pp_least p).
trivial.

apply (pp_least H).
trivial.
Qed.




  (* Ensures that the case reduction is sound *)

  Lemma incl_arg_str_branch0 :
   forall i m l' br carg,
   str_substp l' carg br ->
   forall e p a l,
   l' = ^(1)l ->
   incl_mark e Ind{i // p,,(a,,[])} Ind{i; m // p,,(a,,l)} ->
   forall x t, incl_mark (e; x : t) carg br.
Proof.
intros i m l' br carg H.
elim H using
 str_substp_mind
  with
    (P0 := fun l' s1 s2 =>
           forall e p a l,
           l' = ^(1)l ->
           incl_mark e Ind{i // p,,(a,,[])} Ind{i; m // p,,(a,,l)} ->
           forall x t, clos_le_record ind_marks (e; x : t) s1 s2); 
 intros; auto.
rewrite H1.
apply incl_mt_highest with i m (^(1)p0) (^(1)a0).
change
  (incl_mark (e; x0 : t) ^(1)Ind{i // p0,,(a0,,[])}
     ^(1)Ind{i; m // p0,,(a0,,l0)}) in |- *.
apply (Rlift0 _ incl_mark_lift); trivial.

red in |- *.
apply Clos_prd; auto.
apply H1 with (^(1)p) (^(1)a0) l; trivial.
rewrite H2.
change
  (incl_mark (e; x0 : t) ^(1)Ind{i // p,,(a0,,[])}
     ^(1)Ind{i; m // p,,(a0,,l0)}) in |- *.
apply (Rlift0 _ incl_mark_lift); trivial.

red in |- *.
apply Clos_sum.
apply H1 with p a0 l0; auto.

apply H3 with p a0 l0; auto.

red in |- *.
apply Clos_record.
apply H1 with p a l0; auto.

simpl in |- *.
apply Clos_field.
apply H1 with p a l0; auto.

apply H3 with (^(1)p) (^(1)a) l; trivial.
rewrite H4.
change
  (incl_mark (e; x0 : t) ^(1)Ind{i // p,,(a,,[])}
     ^(1)Ind{i; m // p,,(a,,l0)}) in |- *.
apply (Rlift0 _ incl_mark_lift); trivial.
Qed.

  Lemma incl_arg_str_branch :
   forall e i m l p a br carg,
   incl_mark e Ind{i // p,,(a,,[])} Ind{i; m // p,,(a,,l)} ->
   str_substp ^(1)l carg br -> forall x t, incl_mark (e; x : t) carg br.
Proof.
intros.
apply incl_arg_str_branch0 with (1 := H0) (3 := H); trivial.
Qed.


  Lemma str_subst_incl_seq :
   forall k k' l m br,
   str_substp l m br ->
   forall e l',
   incl_seq k' l' k l ->
   exists2 br' : _, str_substp l' m br' & incl_mark e br' br.
Proof.
intros k k' l m br H.
elim H using
 str_substp_mind
  with
    (P0 := fun l s1 s2 =>
           forall e l',
           incl_seq k' l' k l ->
           exists2 s2' : _,
             mark_record l' s1 s2' & clos_le_record ind_marks e s2' s2);
 intros.
split with t; auto.
constructor.

split with Ind{x // p,,(a,,l')}; auto.
constructor.
trivial.

constructor.
constructor.
apply (incl_seq_OO H1).

elim H1 with (e; x : a) (^(1)l'); intros.
split with (Forall x : a, x0); auto.
constructor.
trivial.

unfold lift in |- *.
apply incl_seq_lift; trivial.

elim H1 with e l'; trivial; intros.
elim H3 with e l'; trivial; intros.
split with (x ** x0); auto.
constructor; trivial.

elim H1 with e l'; trivial; intros.
split with (record_of x).
constructor; trivial.

red in |- *.
apply Clos_record; trivial.

split with (nil (A:=name * term)); simpl in |- *; auto.
constructor.

elim H1 with (e := e) (1 := H4); intros.
elim H3 with (e; x : x0) (^(1)l'); intros.
split with ((x, x0) :: x1); simpl in |- *.
constructor; trivial.

apply Clos_field; trivial.

unfold lift in |- *.
apply incl_seq_lift; trivial.
Qed.


  (* Given lc the list of constructor of an inductive type, and a list
     of patterns, builds the list of the constructor corresponding to
     each branch *)

  Definition case_branch q l (pats : patterns) lt :=
    for_all2
      (fun pat t =>
       exists2 c : _,
         constr_in_ind pat is c &
         (exists2 arg : _,
            str_substp ^(1)l (c_arg c) arg & t = branch_ty c arg q)) pats lt.
(* TODO: on ne doit pas pouvoir typer un case avec 0 branches avec un type
   inductif abstrait... *)


  Lemma case_branch_lift :
   forall q l pats lt n,
   case_branch q l pats lt ->
   forall k,
   case_branch ^(n//k)q ^(n//k)l pats (map (fun t => ^(n//S k)t) lt).
Proof.
unfold case_branch, branch_ty, lift in |- *.
intros q l pats lt n cbr_ty.
elim cbr_ty; simpl in |- *; intros; auto.
constructor; auto.
inversion_clear H.
split with x0; trivial.
inversion_clear H3.
split with (^(n//S k)x1).
replace (c_arg x0:term) with (^(n//S k)(c_arg x0)).
rewrite permute_lift_rec; simpl in |- *; auto with arith.
apply str_sub_lift; trivial.

rewrite closed1_lift; auto with arith.

rewrite H4; simpl in |- *.
rewrite closed2_lift.
fold (2 + k) in |- *.
rewrite <- permute_lift_rec; auto with arith.
Qed.


  Lemma case_branch_subst :
   forall q m l pats lt,
   case_branch q l pats lt ->
   forall k, case_branch q[k:=m] l[k:=m] pats (map (fun a => a[S k:=m]) lt).
Proof.
unfold case_branch, branch_ty, lift in |- *.
intros q m l pats lt cbr_ty.
elim cbr_ty; simpl in |- *; intros; auto.
constructor; auto.
inversion_clear H.
split with x0; trivial.
inversion_clear H3.
split with (x1[S k:=m]).
replace (c_arg x0:term) with ((c_arg x0)[S k:=m]).
rewrite commut_lift_subst_rec; simpl in |- *; auto with arith.
apply str_sub_subst; trivial.

rewrite closed1_subst; auto with arith.

rewrite H4; simpl in |- *.
rewrite closed2_subst.
fold (2 + k) in |- *.
rewrite <- commut_lift_subst_rec; auto with arith.
Qed.


  Definition err_no_such_constructor x :=
    app_str (app_str name_string (string_of x)) err_not_constructor.


  Lemma mark_branch_type :
   forall e q l pat,
   one_ind_unique is ->
   ppal_dec
     (fun t =>
      exists2 c : _,
        constr_in_ind pat is c &
        (exists2 arg : _,
           str_substp ^(1)l (c_arg c) arg & t = branch_ty c arg q))
     (fun x y => incl_mark e y x).
Proof.
intros.
elim (search_constr_ind pat is); intros.
inversion_clear a.
elim (mark_branch_arg e ^(1)l (c_arg x)); intros.
left.
split with (branch_ty x x0 q).
split; intros.
split with x; trivial.
split with x0; auto.
apply (pp_ok p).

elim H1.
intros x' cx'.
replace x' with x; intros.
inversion_clear H2.
rewrite H4.
unfold branch_ty in |- *.
red in |- *.
apply Clos_prd; trivial.
apply (pp_least p).
trivial.

red in H.
apply H with pat; trivial.

RightStr (err_no_such_constructor pat); red in |- *; intros.
inversion_clear H0.
elim b with x0; trivial.
Qed.


  (* Inference of the largest type for the branch vector *)
  Lemma mark_branch_types :
   forall e q l pats,
   one_ind_unique is ->
   ppal_dec
     (fun t => exists2 lt : _, t = mkvect lt & case_branch q l pats lt)
     (fun x y => incl_mark e y x).
Proof.
intros.
elim pats; intros.
left.
split with one.
split; intros.
split with (nil (A:=term)); trivial.
red in |- *.
auto.

inversion_clear H0.
rewrite H1.
inversion_clear H2; auto.

elim (mark_branch_type e q l a H); intros.
inversion_clear a0.
elim H0; intros.
inversion_clear a0.
left.
split with (x ** x0).
split; intros.
elim (pp_ok H1); intros.
elim (pp_ok H2); intros.
clear H1 H2.
elim H4; intros.
rewrite H2.
rewrite H5.
split with (branch_ty x1 x3 q :: x2); simpl in |- *; auto.
constructor.
split with x1; trivial.
elim H2; trivial.

trivial.

inversion_clear H3.
rewrite H4.
inversion_clear H5.
simpl in |- *.
red in |- *.
apply Clos_sum.
apply (pp_least H1).
trivial.

apply (pp_least H2).
split with l2; trivial.

right; red in |- *; intros.
inversion_clear H2.
inversion_clear H4.
elim b with (mkvect l2).
split with l2; auto.

right; red in |- *; intros.
inversion_clear H1.
inversion_clear H3.
elim (b y).
trivial.
Qed.



  (* Exhaustivity of patterns *)
  Definition full_patt (pats : patterns) : Prop :=
    exists2 lc : _,
      i_constrs is = Some lc & (forall c, In c lc -> In (c_name c) pats).


  (* Case predicate *)
  Definition case_predicate_form se :=
    Forall x_ : i_arg is,
    (Forall x_ : Ind{i_name is // #1,,(#0,,[])}, %%se).

  Lemma case_predicate_form_lift :
   forall n k se, ^(n//S k)(case_predicate_form se) = case_predicate_form se.
Proof.
intros.
unfold case_predicate_form in |- *; simpl in |- *.
rewrite closed1_lift; trivial.
Qed.

  Lemma case_predicate_form_subst :
   forall m k se, (case_predicate_form se)[S k:=m] = case_predicate_form se.
Proof.
intros.
unfold case_predicate_form in |- *; simpl in |- *.
rewrite closed1_subst; trivial.
Qed.

  Inductive type_case (pats : patterns) (q c : term) (b : bool) :
  term -> term -> Prop :=
      Ty_case :
        forall m par arg l tybr se,
        full_patt pats ->
        (if b
          then elim_sort (i_ckind is) (i_kind is) se
          else List.length pats = O \/ List.length pats = 1 /\ i_ckind is = so)
        ->
        case_branch q l pats tybr ->
        type_case pats q c b
          (Ind{i_name is; m // par,,(arg,,l)} **
           ((case_predicate_form se)[0:=par] ** (mkvect tybr)[0:=par]))
          ((q @ arg) @ c).


  Lemma type_case_lift :
   forall pats n k p c b m t,
   type_case pats p c b m t ->
   type_case pats ^(n//k)p ^(n//k)c b ^(n//k)m ^(n//k)t.
Proof.
simple destruct 1; intros.
unfold lift_rec in |- *; fold lift_rec in |- *.
rewrite distr_lift_subst.
rewrite case_predicate_form_lift.
rewrite distr_lift_subst.
rewrite mkvect_lift.
constructor; trivial.
apply case_branch_lift; trivial.
Qed.

  Lemma type_case_subst :
   forall pats k s p c b m t,
   type_case pats p c b m t -> type_case pats p[k:=s] c[k:=s] b m[k:=s] t[k:=s].
Proof.
simple destruct 1; intros.
unfold subst_rec in |- *; fold subst_rec in |- *.
rewrite distr_subst.
rewrite case_predicate_form_subst.
rewrite distr_subst.
rewrite mkvect_subst.
constructor; trivial.
apply case_branch_subst; trivial.
Qed.


  (* Decidability stuffs *)

  Lemma full_pat_dec_cstr :
   forall pats lc, decide (forall c, In c lc -> In (c_name c) pats).
Proof.
simple induction lc; intros.
left; intros.
inversion H.

elim (list_index eq_dec (c_name a) pats); intros.
elim H; intros.
left; intros.
inversion_clear H0.
inversion_clear a0.
elim H0; intros; auto.

left; subst; reflexivity.
auto.

right; red in |- *; intros.
apply b; intros.
auto.

right; red in |- *; intros.
apply b.
auto.
Qed.


  Lemma full_patt_dec : forall pats, decide (full_patt pats).
Proof.
unfold full_patt in |- *.
intros.
case (i_constrs is); intros.
elim full_pat_dec_cstr with pats l; intros.
left.
split with l; trivial.

right; red in |- *; intros; apply b.
inversion_clear H.
injection H0; intros.
apply H1.
elim H; trivial.

right; red in |- *; intros.
inversion_clear H.
discriminate H0.
Qed.


End TypCase.
