(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export Records.

Set Implicit Arguments.
Unset Strict Implicit.

Section ContextClosureTypes.


  Variable stab_rel : env -> decl -> decl -> Prop.
  Variable R : red_rule.


  Inductive clos_le : red_rule :=
    | Clos_rule : forall e x y, R e x y -> clos_le e x y
    | Clos_refl : forall e x, clos_le e x x
    | Clos_prd :
        forall e x a1 a2 b1 b2,
        clos_le e a2 a1 ->
        clos_le (e; x : a2) b1 b2 ->
        clos_le e (Forall x : a1, b1) (Forall x : a2, b2)
    | Clos_sum :
        forall e a1 a2 b1 b2,
        clos_le e a1 a2 -> clos_le e b1 b2 -> clos_le e (a1 ** b1) (a2 ** b2)
    | Clos_record :
        forall e s1 s2,
        clos_le_record e s1 s2 -> clos_le e (record_of s1) (record_of s2)
with clos_le_record : env -> rspec -> rspec -> Prop :=
  | Clos_mt : forall e : env, clos_le_record e nil nil
  | Clos_field :
      forall e x a1 a2 s1 s2,
      clos_le e a1 a2 ->
      clos_le_record (e; x : a1) s1 s2 ->
      clos_le_record e ((x, a1) :: s1) ((x, a2) :: s2).

  Scheme clos_le_mind := Minimality for clos_le
    Sort Prop
    with clos_le_record_mind := Minimality for clos_le_record 
    Sort Prop.

  Lemma clos_le_record_refl : forall rs e, clos_le_record e rs rs.
Proof.
simple induction rs.
constructor.

simple destruct a; constructor; trivial; apply Clos_refl.
Qed.

  Hint Resolve Clos_rule Clos_refl Clos_prd Clos_sum Clos_record Clos_mt
    Clos_field clos_le_record_refl: core.



  Hypothesis R_lifts : R_lift R.
  Hypothesis R_substs : R_subst R.
  Hypothesis stable : R_stable stab_rel R.

  Lemma R_lift_clos : R_lift clos_le.
Proof.
red in |- *; intros g e d t u cl_t_u.
elim cl_t_u using
 clos_le_mind
  with
    (P0 := fun e s1 s2 =>
           forall k f,
           ins_in_env g d k e f ->
           clos_le_record f (lift_rsign 1 s1 k) (lift_rsign 1 s2 k));
 (simpl in |- *; intros; eauto; fail) || intros.
repeat rewrite record_of_lift; eauto.
Qed.

  Lemma R_subst_clos : R_subst clos_le.
Proof.
red in |- *; intros s d g e t u cl_t_u.
elim cl_t_u using
 clos_le_mind
  with
    (P0 := fun e s1 s2 =>
           forall k f,
           sub_in_env g s d k e f ->
           clos_le_record f (subst_rsign s s1 k) (subst_rsign s s2 k));
 (simpl in |- *; intros; eauto; fail) || intros.
repeat rewrite record_of_subst; eauto.
Qed.

  Lemma stable_clos : R_stable stab_rel clos_le.
Proof.
red in |- *; intros e t u cl_t_u.
elim cl_t_u using
 clos_le_mind
  with
    (P0 := fun e s1 s2 =>
           forall f, rel_in_list stab_rel f e -> clos_le_record f s1 s2);
 eauto.
Qed.

End ContextClosureTypes.


  Hint Resolve Clos_rule Clos_refl Clos_prd Clos_sum Clos_record Clos_mt
    Clos_field clos_le_record_refl R_lift_clos R_subst_clos stable_clos: core.

Section Monotonicity.

  Variable R1 R2 : red_rule.
  Hypothesis incl : forall e x y, R1 e x y -> R2 e x y.

  Lemma incl_clos_le : forall e x y, clos_le R1 e x y -> clos_le R2 e x y.
Proof.
intros e x y cl_x_y.
elim cl_x_y using
 clos_le_mind with (P0 := fun e s1 s2 => clos_le_record R2 e s1 s2); 
 eauto.
Qed.

End Monotonicity.
