(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export SignCheck.

Set Implicit Arguments.
Unset Strict Implicit.

(* Error messages stuffs *)
Definition msg_if_fail_sumor (A:Set) (B:Prop) s (p : A + {B}) :=
  match p with
  | inleft a => inleft B a
  | inright b => inright_str A s b
  end.
Definition msg_if_fail_sumbool A B s (p : {A} + {B}) :=
  match p with
  | left a => left B a
  | right b => right_str A s b
  end.
Ltac PrintIfFail s :=
  apply msg_if_fail_sumor with (1 := s) ||
    apply msg_if_fail_sumbool with (1 := s).


  (* The real interface *)

  Definition genv := {sg : _ | wf_sign sg}.

  Definition genv_intro (*: (sg:?)(wf_sign sg)->genv*)  sg wfs :=
    exist (fun sg => wf_sign sg) sg wfs:genv.

  Instance sig_genv (g : genv) : sigma := let (sg, _) := g in sg.

  Definition wfs_genv g : wf_sign (sig_genv g) :=
    match g return (wf_sign (sig_genv g)) with
    | exist _ wfg => wfg
    end.

  (* The empty signature *)
  Definition nil_genv := genv_intro wfs_nil.

  Definition gtyp g M T := sig_genv g – ( ) ⊢ M : T.
  Definition gwft g T := @wf_type (@cci_pts (sig_genv g)) nil T.

  Definition cci_typ_ppal g t := @infer_typ_ppal (@cci_pts (sig_genv g)) nil t.

  Definition cci_sort_ppal g x t :=
    @infer_sort_ppal (@cci_pts (sig_genv g)) nil x t.

  Definition cci_sign_extend g d :=
    {g' : _ | sig_genv g' = d :: sig_genv g} + {~ wf_sign (d :: sig_genv g)}.

  Unset Elimination Schemes.

  Record CCI_TC : Set :=
    {cci_inf_ppal_type : forall g t, cci_typ_ppal g t;
     cci_inf_ppal_sort : forall g x t, cci_sort_ppal g x t;
     cci_chk_typ : forall g t T, decide (gtyp g t T);
     cci_chk_decl : forall g d, cci_sign_extend g d;
     cci_chk_wk : forall g t T, gwft g T -> decide (gtyp g t T);
     cci_chk_wft : forall g T, decide (gwft g T)}.

  Instance TC : forall (g : genv), @PTSO_TC (@cci_pts (sig_genv g)).
  Proof.
    intros [sg Hsg].
    refine (@IPTSO_TC _ _).
    apply @CCI_algos.
    apply wf_sign_ok.
    exact Hsg.
  Defined.


  Theorem cci_kernel : CCI_TC.
  Proof.
    split.
    red in |- *; intros.
    apply inf_ppal_type.
    apply Wf_nil.

    red in |- *; intros.
    apply inf_ppal_sort.
    apply Wf_nil.

    unfold gtyp in |- *; intros.
    apply chk_typ.
    apply Wf_nil.

    (*
    Realizer [g:genv; d:sig_item]
      if (wf_sign_ext (sig_genv g) d)
      then (inleft ? (existS ? ? (cons ? d (sig_genv g)) (env_genv g)))
      else (inright genv).
    *)
    red in |- *; intros.
    elim wf_sign_ext with (sig_genv g) d; intros.
    left.
    split with (genv_intro a); reflexivity.

    right; trivial.

    apply wfs_genv.

    unfold gtyp in |- *; intros.
    apply chk_wk; trivial.

    unfold gwft in |- *; intros.
    apply chk_wft; trivial.
  Qed.


  Definition infer_type := cci_inf_ppal_type cci_kernel.
  Definition check_type := cci_chk_typ cci_kernel.
  Definition enter_decl := cci_chk_decl cci_kernel.
