(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

(*Require Import MinclSign.*)
Require Export InclSign.
Require Export CicMeta.

Set Implicit Arguments.
Unset Strict Implicit.

(* Definition of a well-formed signature *)

Unset Elimination Schemes.

(* Constant *)
Record wf_cst (sg : sigma) (c : constant_obj) : Prop :=
  {wf_cst_ok : sg – ⊢ ((( ); x_ : c_par c); c_decl c);
    wf_cst_free : forall g, ~ sign_glob_spec sg (NAME (c_decl c)) g}.

(* Abstract inductive type *)
Record wf_one_ind_arity (sg : sigma) (is : one_ind) : Prop :=
  {wf_one_ind_ok :
    sg – ⊢ (((( ); x_ : i_param is); x_ : i_arg is); x_ : %%(i_kind is));
    wf_one_ind_free : forall g, ~ sign_glob_spec sg (i_name is) g}.

Record wf_arity (sg : sigma) (ip : ind_pack) : Prop :=
  {wf_ar_ok : forall is, In is ip -> wf_one_ind_arity sg is;
    wf_ar_free : list_unique (inds_of_pack ip)}.

Set Elimination Schemes.

(* Inductive type with constructors *)

Lemma is_ind_op :
  forall c,
    {i : _ &  {m : _ | c = MutInd i m}} + {(forall i m, c <> MutInd i m)}.
Proof.
  simple destruct c; intros; try (right; intros; discriminate).
  left.
  exists n.
  exists l; trivial.
Qed.


Inductive inds_free (inds : list name) : term -> Prop :=
  | Ind_srt : forall s, inds_free inds %%s
  | Ind_rel : forall n, inds_free inds #n
  | Ind_cst0 : forall c : cci_op, inds_free inds c
  | Ind_cst1 :
      forall c a,
      inds_free inds a ->
      (forall i m, In i inds -> c <> MutInd i m) ->
      inds_free inds (Cst1 c a)
  | Ind_abs :
      forall v a b,
      inds_free inds a ->
      inds_free inds b -> inds_free inds (Fun v : a => b)
  | Ind_prd :
      forall v a b,
      inds_free inds a ->
      inds_free inds b -> inds_free inds (Forall v : a, b)
  | Ind_pair :
      forall a b,
      inds_free inds a -> inds_free inds b -> inds_free inds (a,,b)
  | Ind_sum :
      forall a b,
      inds_free inds a -> inds_free inds b -> inds_free inds (a ** b).

Lemma inds_free_dec : forall inds m, decide (inds_free inds m).
Proof.
  (*Realizer [inds] Fix indf_rec { indf_rec [m:?]: bool :=
    Cases m of
      (Cst1 o t) =>
        if (indf_rec t)
        then Cases (is_ind_op o) of
          (inleft (existS x l)) =>
                        Cases (list_index ? eq_name x inds) of
                          (inleft _) => false
                        | _ => true
                        end
        | inright => true
        end
        else false
    | (Lam _ a b) => if (indf_rec a) then (indf_rec b) else false
    | (Prd _ a b) => if (indf_rec a) then (indf_rec b) else false
    | (Pair a b) => if (indf_rec a) then (indf_rec b) else false
    | (Sum a b) => if (indf_rec a) then (indf_rec b) else false
    | _ => true
    end}.*)
  fix indf_rec 2.
  intros inds m.
  case m; try (left; constructor; trivial; fail).
  intros o t.
  case (indf_rec inds t).
  intros t_ok.
  case (is_ind_op o).
  intros (x, (l, o_ind)); subst o.
  case (list_index (@eq_dec name _)) with x inds.
  intros (n, x_in_inds).
  right; red in |- *; intros free_ind; inversion_clear free_ind.
  apply H0 with x l; trivial; elim x_in_inds; auto.

  intros x_not_in_inds.
  left; constructor; auto.
  red in |- *; intros i l' in_inds eq_ind; apply x_not_in_inds;
   injection eq_ind; intros; subst x; trivial.

  intros o_not_ind.
  left; constructor; auto.

  intros inds_in_t.
  right; red in |- *; intros m_ok; inversion m_ok; auto.

  (* Prod *)
  intros x a b.
  case (indf_rec inds a);
   [ intros a_ok; case (indf_rec inds b);
      [ intros b_ok; left; constructor; trivial | intros b_nok ]
   | intros a_nok ];
   try (right; red in |- *; intros m_ok; inversion_clear m_ok; auto).

  (* Abs*)
  intros x a b.
  case (indf_rec inds a);
   [ intros a_ok; case (indf_rec inds b);
      [ intros b_ok; left; constructor; trivial | intros b_nok ]
   | intros a_nok ];
   try (right; red in |- *; intros m_ok; inversion_clear m_ok; auto).

  (* Sum *)
  intros a b.
  case (indf_rec inds a);
   [ intros a_ok; case (indf_rec inds b);
      [ intros b_ok; left; constructor; trivial | intros b_nok ]
   | intros a_nok ];
   try (right; red in |- *; intros m_ok; inversion_clear m_ok; auto).

  (* Pair *)
  intros a b.
  case (indf_rec inds a);
   [ intros a_ok; case (indf_rec inds b);
      [ intros b_ok; left; constructor; trivial | intros b_nok ]
   | intros a_nok ];
   try (right; red in |- *; intros m_ok; inversion_clear m_ok; auto).
Defined.


Inductive positive_orig (inds : list name) : nat -> term -> Prop :=
| Pos_noccur : forall k m, inds_free inds m -> positive_orig inds k m
| Pos_ind :
  forall x m a k,
    In x inds ->
    inds_free inds a -> positive_orig inds k Ind{x; m // #k,,(a,,[])}
| Pos_sum :
  forall k a b,
    positive_orig inds k a -> positive_orig inds k b -> positive_orig inds k (a ** b)
| Pos_prd :
  forall k x a b,
    inds_free inds a ->
    positive_orig inds (S k) b -> positive_orig inds k (Forall x : a, b).

Inductive positive_rspec (inds : list name) : nat -> rspec -> Prop :=
| Pos_nil : forall k, positive_rspec inds k nil
| Pos_cons : forall k n t l, positive_orig inds k t ->
  positive_rspec inds (S k) l -> positive_rspec inds k ((n,t)::l).

Inductive positive (inds : list name) : nat -> term -> Prop :=
| Pos2_compat : forall k m, positive_orig inds k m -> positive inds k m
| Pos2_record : forall k sign, positive_rspec inds k sign ->
  positive inds k {< field_names sign:record_arity sign >}.


  Lemma positive_orig_dec : forall inds m k, decide (positive_orig inds k m).
  Proof.
    (*
    Realizer [inds:(list name)] Fix pos_dec {pos_dec [m:term]: nat->bool :=
      [k]Cases m of
           (Prd _ a b) => if (inds_free_dec inds a) then (pos_dec b (S k))
                          else false
         | (Sum a b) => if (pos_dec a k) then (pos_dec b k) else false
         | (Cst1 _ _) =>
             Cases (is_ind_dec m) of
               (inleft (Is_ind i l p a lm)) =>
                 Cases (list_index ? eq_name i inds) of
                   (inleft _) => if (term_eq p [: #k :]) then
                                   if (term_eq lm (Cst0 Nil_mark))
                                   then (inds_free_dec inds a)
                                   else (inds_free_dec inds m)
                                 else (inds_free_dec inds m)
                 | inright => (inds_free_dec inds m)
                 end
             | _ => (inds_free_dec inds m)
             end
         | _ => (inds_free_dec inds m)
         end}.
    *)
    simple induction m; intros; try (left; repeat constructor; trivial; fail).
    elim (is_ind_dec (Cst1 o t)); intros.
    elim a; clear a; intros i l p a lm.
    elim (list_index eq_dec i inds); intros.
    inversion_clear a0.
    elim (eq_dec p #k); intros.
    rewrite a0.
    elim (eq_dec lm []); intros.
    rewrite a1.
    elim inds_free_dec with inds a; intros.
    left; constructor 2; trivial.
    elim (first_item_is_item H0); auto.

    RightStr err_nested_rec_type; red in |- *; intros; apply b.
    inversion_clear H1; trivial.
    inversion_clear H2.
    inversion_clear H1.
    inversion_clear H4; trivial.

    RightStr err_rec_type_marked; red in |- *; intros; apply b.
    inversion_clear H1; trivial.
    inversion_clear H2.
    elim H3 with i l; trivial.
    elim H0; auto.

    RightStr err_rec_type_bad_param; red in |- *; intros; apply b.
    inversion_clear H1; trivial.
    inversion_clear H2.
    elim H3 with i l; trivial.
    elim H0; auto.

    elim inds_free_dec with inds Ind{i; l // p,,(a,,lm)}; intros.
    left; constructor; trivial.

    RightStr err_nested_rec_type; red in |- *; intros.
    generalize b b0; unfold not in |- *.
    inversion_clear H0; auto.

    elim inds_free_dec with inds (Cst1 o t); intros.
    left; constructor; trivial.

    RightStr err_negative_occ; red in |- *; intros.
    generalize b b0; unfold not in |- *.
    inversion_clear H0; intros; auto.

    elim inds_free_dec with inds t; intros.
    elim H0 with (S k); intros.
    left.
    constructor 4; trivial.

    right; red in |- *; intros.
    apply b.
    inversion_clear H1; trivial.
    constructor.
    inversion_clear H2; trivial.

    RightStr err_negative_occ; red in |- *; intros; apply b.
    inversion_clear H1; trivial.
    inversion_clear H2; trivial.

    elim inds_free_dec with inds (Fun n : t => t0); intros.
    left; constructor; trivial.

    RightStr err_negative_occ; red in |- *; intros; apply b.
    inversion_clear H1; trivial.

    elim H with k; intros.
    elim H0 with k; intros.
    left; constructor 3; trivial.

    right; red in |- *; intros; apply b.
    inversion_clear H1; trivial.
    constructor.
    inversion_clear H2; trivial.

    right; red in |- *; intros; apply b.
    inversion_clear H1; trivial.
    constructor.
    inversion_clear H2; trivial.

    elim inds_free_dec with inds (t,,t0); intros.
    left; constructor; trivial.

    RightStr err_negative_occ; red in |- *; intros; apply b.
    inversion_clear H1; trivial.
  Qed.

  Lemma positive_rspec_dec : forall inds sign k, decide (positive_rspec inds k sign).
  Proof.
    induction sign.
    left; constructor; auto.
    intros.
    destruct a.
    destruct (positive_orig_dec inds t k).
    destruct (IHsign (S k)).
    left; constructor; auto.
    right; contradict n0. inversion n0; auto.
    right; contradict n0. inversion n0; auto.
  Qed.

  Lemma positive_dec : forall inds m k, decide (positive inds k m).
  Proof.
    intros.
    destruct (positive_orig_dec inds m k).
    left; constructor; auto.
    destruct (is_record m) as [[[names ar] H0]|H0].
    subst.
    destruct (is_record_arity names ar) as [[sign [H1 H2]]|H1].
    subst.
    destruct (positive_rspec_dec inds sign k); auto.
    left; constructor 2; auto.
    right; intro.
    inversion H; auto.
    subst.
    rewrite (rspec_record_inj H3) in H2; auto.
    right; intro.
    inversion H; auto.
    apply (H1 sign); auto.
    right; intro.
    inversion H; auto.
    apply (H0 (field_names sign, record_arity sign)); auto.
  Qed.


  Definition constr_pos inds c :=
    positive inds 0 (c_arg c) /\ inds_free inds (c_inst c).

  Lemma constr_pos_dec : forall inds c, decide (constr_pos inds c).
  Proof.
    (*
    Realizer [inds:(list name); c:constructor]
      if (inds_free_dec inds (c_inst c)) then (positive_dec inds (c_arg c) O)
      else false.
    *)
    intros.
    elim inds_free_dec with inds (c_inst c:term); intros.
    elim positive_dec with inds (c_arg c:term) 0; intros.
    left.
    split; trivial.

    right; red in |- *; intros; apply b.
    inversion_clear H; trivial.

    RightStr err_nested_rec_type; red in |- *; intros; apply b.
    inversion_clear H; trivial.
  Qed.

  Unset Elimination Schemes.

  Record wf_constructor (sg : sigma) (inds : list name)
    (is : one_ind) (c : constructor) : Prop :=
    {wf_cstr_ok : sg – (( ); x_ : i_param is) ⊢ c_arg c : %%(i_ckind is);
     wf_cstr_inst :
      sg – ((( ); x_ : i_param is); x_ : c_arg c) ⊢
      c_inst c : ^(1)(i_arg is);
     wf_cstr_pos : constr_pos inds c;
     wf_cstr_free : forall g, ~ sign_glob_spec sg (c_name c) g}.


  Record wf_constr (sg : sigma) (ip : ind_pack) : Prop :=
    {wf_ind_ok :
      forall x is c,
      In is ip ->
      constr_in_ind x is c -> wf_constructor sg (inds_of_pack ip) is c;
     wf_ind_free : list_unique (cstr_ind_pack ip)}.

  Set Elimination Schemes.

  (* The whole signature *)
  Inductive wf_sign : sigma -> Prop :=
    | wfs_nil : wf_sign nil
    | wfs_cst :
        forall sg c, wf_sign sg -> wf_cst sg c -> wf_sign (Constant c :: sg)
    | wfs_ind :
        forall sg ip,
        wf_sign sg ->
        wf_arity sg ip ->
        wf_constr (MutInductive (abstract_ind ip) :: sg) ip ->
        wf_sign (MutInductive ip :: sg).


  (* The ultimate axiom: the type system is strongly normalizing *)
  Axiom
    cci_is_strongly_normalizing_axiom :
    forall sg e t T,
      wf_sign sg -> sg – e ⊢ t : T -> sn (ctxt redn_term) e t.


  Lemma types_are_strongly_normalizing :
    forall sg e t,
      wf_sign sg -> wf_type e t -> sn (ctxt redn_term) e t.
  Proof.
    clear.
    intros sg e t Hsg [ T s Ht | s ].
    eapply cci_is_strongly_normalizing_axiom; eauto.
    (* We have to prove that sorts are sn *)
    constructor; intros u Htu; hnf in *.
    inversion Htu; subst; clear Htu.
    repeat (inversion H; subst; clear H; [ inversion H0 | rename H0 into H ]).
    inversion H.
  Qed.

  Section WfSignIsOk.

    Variable sg : sigma.
    Hypothesis wfs_sg : wf_sign sg.


    (* These lemmas entail subject reduction delta_glob *)
    Lemma sign_ok_glob_inj :
      forall x g1 g2,
        sign_glob_spec sg x g1 -> sign_glob_spec sg x g2 -> g1 = g2.
    Proof.
      intro x.
      elim wfs_sg; intros.
      inversion H.
      inversion H1.

      clear sg wfs_sg; elim incl_inv_cst with (1 := H2); intros.
      elim incl_inv_cst with (1 := H3); intros.
      auto.

      elim (wf_cst_free H1) with g1.
      rewrite H5 in H3.
      inversion_clear H3.
      generalize H4.
      inversion_clear H7; trivial.

      elim incl_inv_cst with (1 := H3); intros.
      elim (wf_cst_free H1) with g2.
      rewrite H4 in H2.
      inversion_clear H2.
      generalize H5.
      inversion_clear H7; trivial.

      rewrite H4; rewrite H5; trivial.

      clear sg wfs_sg; elim inv_cons_induc with (1 := H3); intros;
        elim inv_cons_induc with (1 := H4); intros; auto.
      inversion_clear H6.
      inversion_clear H8.
      rewrite H6 in H4.
      elim (wf_one_ind_free (wf_ar_ok H1 H7)) with g1.
      elim (@glob_ind_eq_name (MutInductive ip :: sg0)) with (1 := H4); trivial.

      inversion_clear H6.
      rewrite H9 in H4.
      elim (wf_cstr_free (wf_ind_ok H2 H7 H8)) with g1.
      apply glob_spec_cons.
      elim (@glob_cstr_eq_name (MutInductive ip :: sg0))  with (1 := H4); trivial.

      inversion_clear H5.
      inversion_clear H8.
      rewrite H5 in H3.
      elim (wf_one_ind_free (wf_ar_ok H1 H7)) with g2.
      elim (@glob_ind_eq_name (MutInductive ip :: sg0)) with (1 := H3); trivial.

      inversion_clear H5.
      rewrite H9 in H3.
      elim (wf_cstr_free (wf_ind_ok H2 H7 H8)) with g2.
      apply glob_spec_cons.
      elim (@glob_cstr_eq_name (MutInductive ip :: sg0)) with (1 := H3); trivial.

      inversion_clear H5.
      inversion_clear H6.
      inversion_clear H8; inversion_clear H9.
      rewrite H6; rewrite H8.
      elim (unique_map (wf_ar_free H1)) with (1 := H7) (2 := H5); trivial.
      rewrite H6 in H3.
      rewrite H8 in H4.
      elim (@glob_ind_eq_name (MutInductive ip :: sg0)) with (1 := H3); trivial.
      elim (@glob_ind_eq_name (MutInductive ip :: sg0)) with (1 := H4); trivial.

      inversion_clear H8.
      rewrite H6 in H3.
      rewrite H10 in H4.
      elim inv_incl_ind with (1 := H3); intros.
      elim (wf_cstr_free (wf_ind_ok H2 H5 H9)) with (GlobInd (inds_of_pack ip) x0).
      elim (@glob_cstr_eq_name (MutInductive ip :: sg0)) with (1 := H4); trivial.
      apply glob_spec_cons; trivial.

      elim
        (wf_cstr_free (wf_ind_ok H2 H5 H9))
        with (GlobInd (inds_of_pack (abstract_ind ip)) (abs_is x0)).
      split with (MutInductive (abstract_ind ip)); auto.
      elim (@glob_cstr_eq_name (MutInductive ip :: sg0)) with (1 := H4); trivial.
      rewrite (glob_ind_eq_name H3); trivial.
      replace (i_name x0) with (i_name (abs_is x0)); auto.
      constructor; trivial.
      unfold abstract_ind in |- *.
      apply in_map; assumption.

      inversion_clear H6.
      rewrite H10 in H3.
      rewrite H8 in H4.
      elim inv_incl_ind with (1 := H4); intros.
      elim (wf_cstr_free (wf_ind_ok H2 H7 H9)) with (GlobInd (inds_of_pack ip) x1).
      elim (@glob_cstr_eq_name (MutInductive ip :: sg0)) with (1 := H3); trivial.
      apply glob_spec_cons; trivial.

      elim
        (wf_cstr_free (wf_ind_ok H2 H7 H9))
        with (GlobInd (inds_of_pack (abstract_ind ip)) (abs_is x1)).
      split with (MutInductive (abstract_ind ip)); auto.
      elim (@glob_cstr_eq_name (MutInductive ip :: sg0)) with (1 := H3); trivial.
      rewrite (glob_ind_eq_name H4); trivial.
      replace (i_name x1) with (i_name (abs_is x1)); auto.
      constructor; trivial.
      unfold abstract_ind in |- *.
      apply in_map; assumption.

      inversion_clear H6.
      inversion_clear H8.
      rewrite H10.
      rewrite H11.
      apply unique_constr_names with (2 := H7) (3 := H5) (4 := H9) (5 := H6).
      exact (wf_ind_free H2).
    Qed.


    Theorem wf_sign_ok : sign_ok.
    Proof.
      split.
      exact sign_ok_glob_inj.

      intros x co.
      elim wfs_sg; intros.
      inversion_clear H.
      inversion H0.

      apply wf_incl with sg0; auto.
      apply incl_cons.

      clear sg wfs_sg; elim inv_incl_cst with (1 := H2); intros; auto.
      elim H3.
      apply wf_cst_ok; trivial.

      apply wf_incl with sg0.
      apply incl_cons.

      apply H0.
      apply inv_incl_cst_tl with (1 := H3).

      intros inds ind is.
      elim wfs_sg; intros.
      inversion_clear H.
      inversion H0.

      apply wf_incl with sg0.
      apply incl_cons.

      clear sg wfs_sg; elim incl_inv_cst with (1 := H2); intros; auto.
      discriminate H3.

      apply wf_incl with sg0.
      apply incl_cons.

      clear sg wfs_sg; elim inv_incl_ind with (1 := H3); intros; auto.
      apply (wf_one_ind_ok (wf_ar_ok H1 H4)).

      intros x is c.
      elim wfs_sg; intros.
      inversion_clear H.
      inversion H0.

      apply typ_incl with sg0.
      apply incl_cons.

      clear sg wfs_sg; elim incl_inv_cst with (1 := H2); intros; auto.
      discriminate H3.

      clear sg wfs_sg; elim inv_incl_cstr with (1 := H3); intros; auto.
      apply typ_incl with sg0; auto.
      apply incl_cons.

      inversion_clear H4.
      apply typ_incl with (MutInductive (abstract_ind ip) :: sg0).
      apply incl_ind_abstr.

      apply (wf_cstr_inst (wf_ind_ok H2 H5 H6)).

      intros; apply types_are_strongly_normalizing; trivial.
    Qed.

  End WfSignIsOk.
