(**************************************************************************)
(*                      Certified Extraction in Coq                       *)
(*                                                                        *)
(*                            Stéphane Glondu                             *)
(*                                                                        *)
(*                           © 2007-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Export Extraction.Relational.

Set Implicit Arguments.
Unset Strict Implicit.

Section ExtractionFunction.

Context {g : genv}.
Let sg := sig_genv g.
Existing Instance sg.


Definition discard : forall e, sg – ⊢ e -> forall t,
  decide (exists T, sg – e ⊢ t : T /\ sg – e ⊢ T : %%so).
Proof.
  intros e He t.
  destruct (inf_ppal_type _ t He) as [(Hppal, ?)|Ha]; subst.
  destruct Hppal.
  generalize (pp_ok ppal_ty0); intro.
  simpl in *.
  destruct (chk_typ _ TYP0 %%so He) as [HTYP|Ha].

  left; eauto.

  right; intro Ha1; apply Ha.
  destruct Ha1 as (T, (HVAL, HT)).
  apply (pp_least ppal_ty0) in HVAL; simpl in *.
  hnf in ppal_ty0.
  clear - HVAL HT H.
  revert TYP0 T HVAL VAL0 HT H.
  induction 1; intros; auto.
  admit.
  admit.
  right; intro Ha1; destruct Ha1 as (T, (Ht, HT)).
  exact (Ha T Ht).
Defined.

Ltac elim_discard :=
  match goal with
    | extract_term : env -> term -> _ -> sig _,
      He : wf _ |- sig (fun _ => Pe _ ?t _ ) =>
      destruct (discard He t) as [Hd|Hd];
        [ clear extract_term; (* trouble auto *)
          exists ##; destruct Hd as (T', (?, ?));
          apply Pe_Prop with T'; auto
        | ]
  end.

Ltac trivial_extract :=
  match goal with
    | H : ex (fun _ => typ _ (Cst1 ?c ?t) _) |- _ =>
      intros; elimtype False; destruct H as (?, H100);
      InvTypHyp H100; intros; inv in_sign1; fail
    | |- sig (fun _ => Pe _ _ _) =>
      exists ##; auto; fail
    | H : ex (fun _ => typ _ (Cst1 ?c ?t) _),
      extract_term : env -> term -> _ -> sig _
      |- sig (fun _ => Pe ?e (Cst1 ?c ?t) _) =>
      assert (Ht : exists T, sg – e ⊢ t : T)
        by (destruct H as (T, H); inv_typ; eauto);
      destruct (extract_term _ _ Ht) as (t', Ht');
      clear extract_term;
      exists (Cst1 c t'); econstructor; eauto; fail
  end.

(* En mettant le T dans les forall au lieu du exists,
   il faut utiliser la procédure de décision. *)

Definition extract_term : forall e t,
  (exists T, sg – e ⊢ t : T) -> { t' | Pe e t t' }.
Proof with auto.
  fix 2; intros e t H.
  assert (He : sg – ⊢ e)
    by (destruct H as (?, HT); eapply typ_wf; eauto).
  destruct t; elim_discard; try trivial_extract.

  (* Rel *)
  exists #n...

  (* Cst1 *)
  destruct o; try trivial_extract.

    (* Cons_mark *)
    destruct Hd.
    clear; admit.

    (* App *)
    destruct t; try trivial_extract.
    assert (Ht1 : exists T1, sg – e ⊢ t1 : T1)
      by (destruct H as (T, H); inv_typ; eauto).
    assert (Ht2 : exists T2, sg – e ⊢ t2 : T2)
      by (destruct H as (T, H); inv_typ; eauto).
    destruct (extract_term _ _ Ht1) as (t1', Ht1').
    destruct (extract_term _ _ Ht2) as (t2', Ht2').
    exists (t1' @ t2')...

    (* Struct *)
    destruct t; try trivial_extract.
    assert (Ht2 : exists T2, sg – e ⊢ t2 : T2)
      by (destruct H as (T, H); inv_typ; eauto).
    destruct (extract_term _ _ Ht2) as (t2', Ht2').
    exists {<l : ## := t2'>}...

    (* MutConstr *)
    destruct t as [ | | | | | | |p a]; try trivial_extract; [ ].
    assert (Hp : exists P, sg – e ⊢ p : P)
      by (destruct H as (T, H); inv_typ; eauto).
    assert (Ha : exists A, sg – e ⊢ a : A)
      by (destruct H as (T, H); inv_typ; eauto).
    destruct (extract_term _ _ Hp) as [p' Hp'].
    destruct (extract_term _ _ Ha) as [a' Ha'].
    exists (Cstr {n // p',, a'})...

    (* MutCase *)
    destruct t as [ | | | | | | | t1 t1_1]; try trivial_extract.
    destruct t1_1 as [ | | | | | | | pr0 t2]; try trivial_extract.
    assert (Ht1 : exists T1, sg – e ⊢ t1 : T1)
      by (destruct H as (T, H); inv_typ; eauto).
    assert (Ht2 : exists T2, sg – e ⊢ t2 : T2)
      by (destruct H as (T, H); inv_typ; eauto).
    destruct (extract_term _ _ Ht1) as (t1', Ht1').
    destruct (extract_term _ _ Ht2) as (t2', Ht2').
    exists (Match t1' => ## with b — p => t2' end)...

    (* Fix *)
    destruct t; try trivial_extract.
    assert (Ht2 : exists T2, sg – e ⊢ t2 : T2)
      by (destruct H as (T, H); inv_typ; eauto).
    destruct (extract_term _ _ Ht2) as (t2', Ht2').
    exists (Fix {(##,, t2')})...

  (* Lam *)
  assert (Ht2 : exists T2, sg – (e; n:t1) ⊢ t2 : T2)
    by (destruct H as (T, H); inv_typ; eauto).
  destruct (extract_term _ _ Ht2) as (t2', Ht2').
  exists (Fun n : ## => t2')...

  (* Pair *)
  assert (Ht1 : exists T1, sg – e ⊢ t1 : T1)
    by (destruct H as (T, H); inv_typ; eauto).
  assert (Ht2 : exists T2, sg – e ⊢ t2 : T2)
    by (destruct H as (T, H); inv_typ; eauto).
  destruct (extract_term _ _ Ht1) as (t1', Ht1').
  destruct (extract_term _ _ Ht2) as (t2', Ht2').
  exists (t1',, t2')...
Defined.

(*
Fixpoint extract_term (g:genv) t {struct t}: term :=
  if discard g t then
    Mt_mark
  else
    match t with
      | Srt _ => Mt_mark
      | Rel i => t
      | Cst0 _ => Mt_mark
      | t..L => (extract_term g t)..L
      | t..R => (extract_term g t)..R
      | t1 @ t2 => (extract_term g t1) @ (extract_term g t2)
      | Cst1 (Const x) t => Cst1 (Const x) (extract_term g t)
      | Cst1 (Struct s) t => Cst1 (Struct s) (extract_term g t)
      | t..(x) => (extract_term g t)..(x)
      | Fix {_,, t} => Fix {Mt_mark,, extract_term g t}
      | Cst1 (MutCase pats) (c,,(pr,,br)) =>
        Cst1 (MutCase pats) (extract_term g c,,(Mt_mark,,extract_term g br))
      | Cstr{C//p,,a} => Cstr{C//extract_term g p,, extract_term g a}
      | Cst1 _ _ => Mt_mark
      | Prd _ _ _ => Mt_mark
      | Lam x _ t => Lam x Mt_mark (extract_term g t)
      | Sum _ _ => Mt_mark
      | Pair t1 t2 => Pair (extract_term g t1) (extract_term g t2)
    end.

Functional Scheme extract_term_ind :=
  Induction for extract_term Sort Prop.
*)

End ExtractionFunction.
