(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Import NArith.

Set Implicit Arguments.
Unset Strict Implicit.

Local Open Scope N_scope.

Unset Case Analysis Schemes.

Record int := mk_int {
  int_value : N
}.

Definition zero := mk_int 0.
Definition succ x := mk_int (Nsucc (int_value x)).

Lemma succ_inj : forall m n, succ m = succ n -> m = n.
Proof.
  unfold succ; intros [m] [n] Hmn; simpl in *; f_equal.
  apply Nsucc_inj; congruence.
Qed.

Lemma int_discr : forall n, zero <> succ n.
Proof.
  unfold succ, zero; intros [n] Hn; simpl in *.
  apply Nsucc_0 with n; congruence.
Qed.

Lemma int_case : forall n, {m | n = succ m} + {n = zero}.
Proof.
  unfold succ, zero.
  intros [n]; simpl in *.
  replace n with (N_of_nat (nat_of_N n)) by apply N_of_nat_of_N.
  destruct (nat_of_N n) as [ |p]; simpl.
  right; reflexivity.
  left; exists (mk_int (N_of_nat p)); simpl; f_equal.
  destruct p; reflexivity.
Qed.

Fixpoint int_of_nat (n : nat) : int :=
  match n with
    | O => zero
    | S k => succ (int_of_nat k)
  end.

Lemma int_injection :
  forall i j, int_of_nat i = int_of_nat j -> i = j.
Proof.
  simple induction i; simple destruct j; simpl in |- *; intros; auto.
  elim int_discr with (int_of_nat n); auto.

  elim int_discr with (int_of_nat n); auto.

  elim H with n0; auto.
  apply succ_inj; auto.
Qed.
