(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Set Implicit Arguments.
Unset Strict Implicit.
Require Export MlString.

Definition error_string := mk_str "Error: ".


  Definition print_error s := print_endline (app_str error_string s).

  Section PrintErrors.

    Variable A B : Prop.
    Variable X : Set.
    Variable s : ml_string.

    (* Beware: the string is printed *before* a is evaluated. *)
    Definition inleft_str (a : X) :=
      match print_error s with
        | tt => inleft B a
      end.

    Definition inright_str (b : B) :=
      match print_error s with
        | tt => inright X b
      end.

    Definition left_str (a : A) :=
      match print_error s with
        | tt => left B a
      end.

    Definition right_str (b : B) :=
      match print_error s with
        | tt => right A b
      end.

  End PrintErrors.

  Ltac LeftStr str :=
    repeat red in |- *; intros;
     apply inleft_str with (1 := str) || apply left_str with (1 := str).

  Ltac RightStr str :=
    repeat red in |- *; intros;
     apply inright_str with (1 := str) || apply right_str with (1 := str).

  (* Used in Machops *)
  Definition err_not_type := mk_str "the type is not typable with a sort.".
  Definition err_cannot_coerce := mk_str "not convertible.".

  (* Used in CaseDec *)
  Definition err_not_exhaustive := mk_str "non exhaustive pattern-matching.".
  Definition err_bad_predicate := mk_str "ill-typed return predicate in pattern-matching.".
  Definition err_ill_formed_branch := mk_str "ill-typed branch in pattern-matching.".

  (* Used in RecordDec *)
  Definition err_field_name_clash := mk_str "two fields have the same name.".
  Definition err_no_record_arity := mk_str "ill-formed record arity.".
  Definition err_unknown_field := mk_str "unknown field.".
  Definition err_not_record := mk_str "not a record.".

  (* Used in WfSign *)
  Definition err_rec_type_marked := mk_str "the inductive is marked.".
  Definition err_rec_type_bad_param := mk_str "the inductive is applied to a wrong parameter.".
  Definition err_nested_rec_type := mk_str "the inductive is nested.".
  Definition err_negative_occ := mk_str "non strictly positive occurrence.".

  (* Used in TypCase *)
  Definition name_string := mk_str "name ".
  Definition err_not_constructor := mk_str " is not a constructor.".

  (* Used in SignCheck *)
  Definition err_already_defined := mk_str " is already defined.".
  Definition err_ind_name_clash := mk_str "two inductives have the same name.".
  Definition err_ind_ill_formed := mk_str "ill-formed inductive declaration.".
  Definition err_wrong_cstr_kind := mk_str "the constructor is in an incorrect sort".
  Definition err_bad_cstr_args := mk_str "the constructor is applied to an ill-typed argument.".
  Definition err_bad_instance := mk_str "the return type of the constructor is incorrect.".
  Definition err_cstr_name_clash := mk_str "two constructors have the same name.".
  Definition err_constant_ill_typed := mk_str "ill-formed constant or axiom.".

  (* Used in ReductionDec *)
  Definition err_not_sum := mk_str "not a sum.".
  Definition err_not_prod := mk_str "not a product.".
  Definition err_not_ind := mk_str "not an inductive.".
