(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Set Implicit Arguments.
Unset Strict Implicit.

Require Import Omega.

Require Import Arith.
Require Import Wf_nat.
Require Export List.
Require Import MoreLogic.

Section MoreLists.

  Variable A : Set.

  Let List := list A.

  Lemma list_length_ind :
   forall P : List -> Prop,
   (forall m,
    (forall n, n < m -> forall l, length l = n -> P l) ->
    forall l, length l = m -> P l) -> forall l, P l. 
Proof.
intros.
cut (forall n l, length l = n -> P l); eauto.
intro.
elim (lt_wf n).
eauto.
Qed.

  Lemma list_length_rec :
   forall P : List -> Set,
   (forall m,
    (forall n, n < m -> forall l, length l = n -> P l) ->
    forall l, length l = m -> P l) -> forall l, P l. 
Proof.
intros.
cut (forall n l, length l = n -> P l); eauto.
intro.
elim (lt_wf n).
eauto.
Defined.


  Lemma app_inv :
   forall l1 l2 l3 l4 : List,
   l1 ++ l2 = l3 ++ l4 -> length l1 = length l3 -> l2 = l4.
simple induction l1; simple destruct l3; simpl in |- *; intros; auto.
discriminate H0.

discriminate H1.

apply H with l0; auto.
injection H0; auto.
Qed.

  Inductive at_depth (P : List -> Prop) : nat -> List -> Prop :=
    | Ad_hd : forall l, P l -> at_depth P 0 l
    | Ad_tl : forall n l x, at_depth P n l -> at_depth P (S n) (x :: l).



  (* Search a member of a list satisfying P *)
  Variable P : A -> Prop.


Section SearchList.

  Variable P_dec : forall x, decide (P x).

  Lemma search_list :
    forall l, {x : _ | In x l &  P x} + {(forall x, In x l -> ~ P x)}.
  Proof.
    (*Realizer Fix srch{
      srch [l:?]: ? :=
        Cases l of
          nil => (inright A)
        | (cons x m) => if (P_dec x) then (inleft ? x) else (srch m)
        end}.
    *)
    fix srch 1.
    intros [ |x m].
    right; red in |- *; intros x in_nil; inversion in_nil; fail.

    case (P_dec x).
    intro x_ok.
    left; exists x; [left; reflexivity|assumption].

    intros x_nok.
    case (srch m).
    intros (y, y_in_tl, y_ok).
    left; exists y; [right| ]; assumption.

    intros not_in_tl.
    right; intros y y_in_tl.
    inversion_clear y_in_tl; subst; auto.
  Defined.

End SearchList.

Lemma dec_for_all :
  forall l,
    (forall x, In x l -> decide (P x)) -> decide (forall x, In x l -> P x).
Proof.
  (*Realizer Fix for_all { for_all [l:?]: ? :=
    [dec_P:A->bool]Cases l of
      nil => true
    | (cons x l') => if (dec_P x) then (for_all l' dec_P) else false
    end}.
  *)
  fix for_all 1.
  intros [ |x l'] dec_P.
  left; intros x x_in_nil; inversion_clear x_in_nil; fail.

  case (dec_P x).
  left; reflexivity.

  intros x_ok.
  case (for_all l').
  intros; apply dec_P; right; assumption.

  intros all_l'.
  left; intros y y_in_l; inversion_clear y_in_l; subst; auto.

  intros not_all_l'; right; red in |- *.
  contradict not_all_l'; intros; apply not_all_l'; right; assumption.

  intros x_nok; right.
  contradict x_nok; apply x_nok; left; reflexivity.
Defined.

Hint Unfold incl: core.

Lemma incl_app_sym : forall l1 l2 : List, incl (l1 ++ l2) (l2 ++ l1).
Proof.
  intros.
  apply incl_app; [apply incl_appr|apply incl_appl]; apply incl_refl.
Qed.

Inductive item (x : A) : List -> nat -> Prop :=
| item_hd : forall l, item x (x :: l) 0
| item_tl : forall l n y, item x l n -> item x (y :: l) (S n).

Lemma In_item : forall x l, In x l -> exists n : _, item x l n.
Proof.
  induction l; destruct 1.
  subst; repeat econstructor; fail.
  specialize (IHl H); destruct IHl.
  repeat econstructor; eassumption.
Qed.

  Lemma fun_item : forall u v e n, item u e n -> item v e n -> u = v.
    simple induction 1; intros.
    inversion_clear H0; auto.

    inversion_clear H2; auto.
  Qed.

  Lemma item_dom :
   forall (l : List) n, S n <= length l -> exists x : _, item x l n.
Proof.
simple induction l; simpl in |- *; intros.
inversion H.

generalize H0.
clear H0.
case n; intros.
exists a.
constructor.

elim H with n0; intros; auto with arith.
exists x; constructor; trivial.
Qed.

  Lemma list_item :
   forall e n, {t : _ | item t e n} + {(forall t, ~ item t e n)}.
Proof.
fix item_rec 1.
intros [| h l].
right; red in |- *; intros t in_nil; inversion in_nil.

intros [| k].
left; exists h; constructor.

case (item_rec l k).
intros (y, in_tl); left; exists y; constructor; trivial.

intros not_in_tl; right; intros t in_tl_l; inversion_clear in_tl_l;
 red in not_in_tl; eauto.
Defined.
(*Realizer Fix item_rec { item_rec [e:?]: ? :=
  [n]Cases e n of
     (cons _ l) (S k) => (item_rec l k)
   | (cons h _) _     => (inleft ? h)
   | nil        _     => (inright A)
   end}.
*)

  Lemma nth_sound : forall x l n, item x l n -> forall d, nth n l d = x.
Proof.
simple induction 1; simpl in |- *; auto.
Qed.

  Lemma inv_nth_nl : forall x n, ~ item x nil n.
Proof.
unfold not in |- *; intros.
inversion_clear H.
Qed.

  Lemma inv_nth_cs : forall x y l n, item x (y :: l) (S n) -> item x l n.
intros.
inversion_clear H; auto.
Qed.


Section Trunc.

  Inductive split (l : List) : List -> List -> Prop :=
    | Spl_hd : split l l nil
    | Spl_tl : forall l1 l2 x, split l l1 l2 -> split l (x :: l1) (x :: l2).

  Inductive trunc : nat -> List -> List -> Prop :=
    | trunc_O : forall e, trunc 0 e e
    | trunc_S : forall k e f x, trunc k e f -> trunc (S k) (x :: e) f.

  Hint Resolve trunc_O trunc_S: core.

  Lemma trunc_inv_r :
   forall n x l1 l2, trunc n l1 (x :: l2) -> trunc (S n) l1 l2.
intros.
cut (forall l3, trunc n l1 l3 -> l3 = x :: l2 -> trunc (S n) l1 l2).
intros.
apply H0 with (x :: l2); auto.

simple induction 1; intros; auto.
rewrite H1; auto.
Qed.


  Lemma at_depth_trunc :
   forall e f n, at_depth (fun l => e = l) n f -> trunc n f e.
simple induction 1; intros; auto.
elim H0; auto.
Qed.

  Lemma item_trunc :
   forall n e t, item t e n -> exists f : _, trunc (S n) e f.
simple induction n; intros.
inversion_clear H.
exists l; auto.

inversion_clear H0.
elim H with l t; intros; auto.
exists x; auto.
Qed.

  Lemma trunc_it : forall l l' x n, trunc n l (x :: l') -> item x l n.
Proof.
intros.
cut (forall l1, trunc n l l1 -> l1 = x :: l' -> item x l n).
intros.
apply H0 with (1 := H); trivial.

simple induction 1; intros.
rewrite H1; constructor.

constructor; auto.
Qed.


  Lemma fun_trunc :
   forall n l1 l2, trunc n l1 l2 -> forall l3, trunc n l1 l3 -> l2 = l3.
simple induction 1; intros.
inversion H0; auto.

inversion_clear H2; auto.
Qed.

End Trunc.


  (* Insertion *)
  Inductive insert (x : A) : nat -> List -> List -> Prop :=
    | insert_hd : forall l, insert x 0 l (x :: l)
    | insert_tl :
        forall n l il y, insert x n l il -> insert x (S n) (y :: l) (y :: il).

  Lemma ins_le :
   forall k f g d x,
   insert x k f g -> forall n, k <= n -> nth n f d = nth (S n) g d.
simple induction 1; auto.
simple destruct n0; intros.
inversion_clear H2.

simpl in |- *; auto with arith.
Qed.

  Lemma ins_gt :
   forall k f g d x,
   insert x k f g -> forall n, k > n -> nth n f d = nth n g d.
Proof.
simple induction 1; auto.
intros.
inversion_clear H0.

simple destruct n0; intros.
auto.

simpl in |- *; auto with arith.
Qed.

  Lemma ins_eq : forall k f g d x, insert x k f g -> nth k g d = x.
Proof.
simple induction 1; simpl in |- *; auto.
Qed.


Lemma insert_in_inv :
  forall x k l1 l2, insert x k l1 l2 -> forall y, In y l2 -> In y (x :: l1).
Proof.
  simple induction 1; intros; auto.
  inversion H2; subst; clear H2.
  right; left; reflexivity.
  specialize (H1 _ H3); destruct H1.
  subst; left; reflexivity.
  right; right; assumption.
Qed.



  Inductive first_item (x : A) : List -> nat -> Prop :=
    | fit_hd : forall l, first_item x (x :: l) 0
    | fit_tl :
        forall l y n,
        first_item x l n -> x <> y -> first_item x (y :: l) (S n).

  Hint Resolve fit_hd fit_tl: core.

  Lemma first_item_is_item : forall x l n, first_item x l n -> item x l n.
Proof.
simple induction 1; intros.
apply item_hd.

apply item_tl; trivial.
Qed.

  Lemma first_item_unique :
   forall x l n, first_item x l n -> forall m, first_item x l m -> m = n.
Proof.
simple induction 1; intros; auto.
inversion_clear H0; auto.

elim H2; auto.

generalize H2.
inversion_clear H3; intros.
elim H3; auto.

elim H1 with n1; auto.
Qed.



  Hypothesis eq_A : forall x y : A, decide (x = y).

  Lemma eq_list : forall l1 l2 : List, decide (l1 = l2).
Proof.
red in |- *.
decide equality.
apply (eq_A a a0).
Defined.


  Lemma list_index : forall x l, {n : _ | first_item x l n} + {~ In x l}.
Proof.
fix list_index 2.
intros x [| y l1].
right; red in |- *; intros in_nil; inversion in_nil.

case (eq_A x y).
intros eq_x_y; left; exists 0; rewrite eq_x_y; constructor.

intros diff_x_y.
case (list_index x l1).
intros (n, item_n); left; exists (S n); constructor; trivial.

intros not_in_tl; right; red in |- *; intros in_l; generalize diff_x_y;
 inversion_clear in_l; auto.
Defined.
(*Realizer Fix list_index { list_index [x:?; l:?]: ? :=
  Cases l of
    nil => (inright ?)
  | (cons y l1) => if (eq_A x y) then (inleft ? O) else
           Cases (list_index x l1) of
             (inleft k) => (inleft ? (S k))
           | inright => (inright ?)
           end
  end}.
*)

Definition list_disjoint (l1 l2 : List) :=
  forall x, In x l1 -> In x l2 -> False.

Definition list_unique l :=
  forall m n x, item x l m -> item x l n -> m = n.

Lemma item_In : forall x l n, item x l n -> In x l.
Proof.
  induction 1.
  left; reflexivity.
  right; assumption.
Qed.

Lemma unique_ext :
  forall x l, list_unique l -> ~ In x l -> list_unique (x :: l).
Proof.
  intros x l Hl Hx m n y Hy1 Hy2.
  inversion Hy1; subst; inversion Hy2; subst; clear Hy1 Hy2.
  reflexivity.
  apply item_In in H2; contradiction; fail.
  apply item_In in H2; contradiction; fail.
  f_equal.
  eapply Hl; eassumption.
Qed.


  Lemma list_unique_dec : forall l, decide (list_unique l).
Proof.
intros l.
elimtype
 ({list_unique l} +
  {(exists x : _,
      (exists2 m : _, item x l m & (exists2 n : _, item x l n & m <> n)))});
 intros; auto.
right; red in |- *; intros.
inversion_clear b.
inversion_clear H0.
inversion_clear H2.
apply H3.
apply H with x; trivial.

generalize l; clear l.
fix list_unique_dec 1.
intros [| x l'].
left; red in |- *; intros n1 n2 y in_nil1 in_nil2; inversion_clear in_nil1.

case (list_index x l').
intros (n, item_n); right; exists x; exists 0; try constructor.
exists (S n); try discriminate. 
constructor; apply first_item_is_item; trivial.

intros not_in_tl.
case (list_unique_dec l').
intros unique_l'; left; apply unique_ext; trivial.

intros not_unique_l'.
right; generalize not_unique_l'; intros (y, (n1, item1, (n2, item2, diff))).
exists y; exists (S n1); try constructor; trivial.
exists (S n2); auto with arith; constructor (trivial).
Defined.

  Lemma list_unique_first :
   forall x l n, item x l n -> list_unique l -> first_item x l n.
Proof.
simple induction 1; intros; auto.
apply fit_tl; auto.
apply H1.
red in |- *; intros.
cut (S m = S n1); intros.
injection H5; auto.

elim H2 with (S m) (S n1) x0; auto.
constructor; trivial.

constructor; trivial.

red in |- *; intros.
cut (0 = S n0); intros.
discriminate H4.

elim H2 with 0 (S n0) x; auto.
rewrite H3; auto.
constructor.

constructor; trivial.
Qed.


  Lemma unique_app_diff :
   forall x l1 l2, list_unique (l1 ++ l2) -> In x l1 -> In x l2 -> False.
Proof.
intros.
elim In_item with (1 := H0); intros.
elim In_item with (1 := H1); intros.
red in H.
cut (x0 = length l1 + x1).
cut (x0 < length l1).
omega.

elim H2; simpl in |- *; auto with arith.

apply H with x.
elim H2; simpl in |- *; constructor; auto.

elim l1; simpl in |- *; try constructor; auto.
Qed.


  Lemma unique_app_l :
   forall l1 l2 : List, list_unique (l1 ++ l2) -> list_unique l1.
Proof.
unfold list_unique in |- *; intros.
apply H with x.
elim H0; simpl in |- *; constructor; auto.

elim H1; simpl in |- *; constructor; auto.
Qed.

  Lemma unique_app_r :
   forall l1 l2 : List, list_unique (l1 ++ l2) -> list_unique l2.
Proof.
unfold list_unique in |- *; intros.
cut (length l1 + m = length l1 + n).
omega.

apply H with x.
elim l1; simpl in |- *; try constructor; auto.

elim l1; simpl in |- *; try constructor; auto.
Qed.



  Hint Resolve item_hd item_tl insert_hd insert_tl trunc_O trunc_S: core.



  Variable R : List -> A -> A -> Prop.

  Inductive rel_in_list : List -> List -> Prop :=
    | Rel_hd : forall l x y, R l x y -> rel_in_list (x :: l) (y :: l)
    | Rel_tl :
        forall l1 l2 x, rel_in_list l1 l2 -> rel_in_list (x :: l1) (x :: l2).

  Hint Resolve Rel_hd Rel_tl: core.

  Lemma rel_list_item :
   forall n x l,
   item x l n ->
   forall f,
   rel_in_list f l ->
   item x f n \/
   (forall g,
    trunc (S n) l g ->
    (exists2 y : _, R g y x & item y f n) /\ trunc (S n) f g).
Proof.
simple induction 1; intros.
inversion_clear H0; auto.
right; split.
exists x0; auto.
inversion_clear H0.
generalize H1.
inversion_clear H2; auto.

inversion_clear H0; auto.

inversion_clear H2; auto.
elim H1 with l1; intros; auto.
right; intros.
inversion_clear H4.
elim H2 with g; auto.
simple induction 1; split; intros; auto.
exists x0; auto.
Qed.


  Lemma rel_split :
   forall l1 l2,
   rel_in_list l1 l2 ->
   forall l l1',
   split l1 l1' l -> forall l2', split l2 l2' l -> rel_in_list l1' l2'.
Proof.
simple induction 2; intros.
generalize H.
inversion_clear H1; trivial.

inversion_clear H3; auto.
Qed.


End MoreLists.

  Hint Resolve Ad_hd Ad_tl: core.
  Hint Resolve item_hd item_tl insert_hd insert_tl trunc_O trunc_S: core.
  Hint Resolve Spl_hd Spl_tl: core.
  Hint Resolve Rel_hd Rel_tl in_eq in_cons fit_hd fit_tl trunc_O trunc_S: core.
  Hint Unfold incl: core.

  Lemma item_trunc_cons :
   forall (A : Set) n e (t : A),
   item t e n -> exists f : _, trunc n e (t :: f).
Proof.
simple induction 1; intros.
split with l; trivial.

inversion_clear H1.
split with x; auto.
Qed.

  Lemma item_map :
   forall (A B : Set) (f : A -> B) x l n,
   item x l n -> item (f x) (map f l) n.
Proof.
simple induction 1; simpl in |- *; intros; auto.
Qed.

  Lemma item_map' :
   forall (A B : Set) (f : A -> B) x y l n,
   y = f x -> item x l n -> item y (map f l) n.
Proof.
intros.
rewrite H.
apply item_map with (1 := H0).
Qed.


  Lemma unique_map :
   forall (A B : Set) (f : A -> B) l,
   list_unique (map f l) ->
   forall x y, In x l -> In y l -> f x = f y -> x = y.
Proof.
intros.
red in H.
elim In_item with (1 := H0); intros.
elim In_item with (1 := H1); intros.
generalize (item_map f H3) (item_map f H4).
intros.
generalize H6.
elim H2; intros.
generalize H4.
elim H with (1 := H5) (2 := H7); intros.
apply fun_item with (1 := H3) (2 := H8).
Qed.


  Lemma trunc_map :
   forall (A B : Set) (f : A -> B) l l1 n,
   trunc n (map f l) l1 -> exists2 l2 : _, trunc n l l2 & l1 = map f l2.
Proof.
simple induction l; simpl in |- *; intros.
inversion_clear H.
split with (nil (A:=A)); auto.

inversion_clear H0.
split with (a :: l0); auto.

elim H with (1 := H1); intros.
split with x; auto.
Qed.


  Lemma map_inv_item :
   forall (A B : Set) (f : A -> B) x l n,
   item x (map f l) n -> exists2 y : _, item y l n & x = f y.
Proof.
intros.
cut
 (forall ll,
  item x ll n ->
  forall l, ll = map f l -> exists2 y : _, item y l n & x = f y).
eauto.

simple induction 1.
simple destruct l1; simpl in |- *; intros.
discriminate H1.

injection H1.
intros.
split with a; trivial.

simple destruct l1; intros.
discriminate H3.

simpl in H3.
injection H3; intros.
elim H2 with (1 := H4); intros.
split with x0; auto.
Qed.

Lemma map_inv :
  forall (A B : Set) (f : A -> B) x l,
  In x (map f l) -> exists2 y : _, In y l & x = f y.
Proof.
  intros.
  elim In_item with (1 := H); intros.
  elim map_inv_item with (1 := H0); intros.
  split with x1; auto.
  eapply item_In; eassumption.
Qed.

Fixpoint flat_map (A B : Set) (f : A -> list B) (l : list A) {struct l} :
 list B := match l with
           | nil => nil
           | x :: y => f x ++ flat_map f y
           end.


  Lemma list_assoc_raw :
   forall A B : Set,
   (forall x y : A, decide (x = y)) ->
   forall x (l : list (A * B)),
   {y : _ |
   exists2 n : _, item (x, y) l n & first_item x (map (fst (B:=_)) l) n} +
   {(forall y, ~ In (x, y) l)}.
Proof.
intros A B eqA.
fix assoc 2.
intros x [| p l'].
right; red in |- *; intros y y_in_nil; inversion y_in_nil.

case p; intros x' y'.
case (eqA x x').
intros eq_key; left; exists y'; rewrite eq_key; exists 0; simpl in |- *;
 repeat constructor.

intros diff_key.
case (assoc x l').
intros (y, in_tl); left; exists y.
generalize in_tl; intros (n, item_n, item_key); exists (S n); simpl in |- *;
 repeat constructor; trivial.

intros not_in_tl; red in not_in_tl.
right; red in |- *; intros y y_in_l; generalize diff_key;
 inversion_clear y_in_l; eauto.
injection H; intros; subst; destruct diff_key; reflexivity.
Defined.
(*Realizer Fix assoc{assoc [x:?; l:(list (prod A B))]: ? :=
  Cases l of
    nil => (inright ?)
  | (cons (x',y') l') => if (eqA x x') then (inleft ? y') else (assoc x l')
  end}.
*)

Lemma list_assoc :
  forall A B : Set,
  (forall x y : A, decide (x = y)) ->
  forall x (l : list (A * B)), dec_find (fun y => In (x, y) l).
Proof.
  intros A B eq_dec x l.
  case (list_assoc_raw eq_dec x l).
  intros (y, found); left; exists y; case found; intros n item_n _.
  eapply item_In; eassumption.
  intros not_found; right; auto.
Defined.

Section ForAll2.

  Variable A B : Set.
  Variable R : A -> B -> Prop.

  Inductive for_all2 : list A -> list B -> Prop :=
    | Fa2_nil : for_all2 nil nil
    | Fa2_cons :
        forall x y l1 l2,
        R x y -> for_all2 l1 l2 -> for_all2 (x :: l1) (y :: l2).

  Hint Resolve Fa2_nil Fa2_cons: core.

  Lemma for_all2_item_ex :
   forall x l1 n,
   item x l1 n ->
   forall l2, for_all2 l1 l2 -> exists2 y : _, item y l2 n & R x y.
Proof.
simple induction 1; intros; auto.
inversion_clear H0.
split with y; auto.

inversion_clear H2.
elim H1 with l3; intros; auto.
split with x0; auto.
Qed.

  Lemma for_all2_ext :
   forall l1 l2,
   (forall n x y, item x l1 n -> item y l2 n -> R x y) ->
   length l1 = length l2 -> for_all2 l1 l2.
Proof.
simple induction l1; simple destruct l2; simpl in |- *; intros; auto;
 try discriminate.
injection H1; intros.
eauto 10.
Qed.

  Lemma for_all2_app :
   forall l2 l2' l1 l1',
   for_all2 l1 l1' -> for_all2 l2 l2' -> for_all2 (l1 ++ l2) (l1' ++ l2').
simple induction 1; simpl in |- *; intros; auto.
Qed.



  Hypothesis R_dec : forall x y, decide (R x y).

  Lemma for_all2_dec : forall l l', decide (for_all2 l l').
Proof.
simple induction l; simple destruct l'; intros.
left; trivial.

right; red in |- *; intros.
inversion_clear H.

right; red in |- *; intros.
inversion_clear H0.

elim R_dec with a b; intros.
elim H with l1; intros.
left; auto.

right; red in |- *; intros; apply b0.
inversion_clear H0; trivial.

right; red in |- *; intros; apply b0.
inversion_clear H0; trivial.
Defined.

End ForAll2.

  Hint Resolve Fa2_nil Fa2_cons: core.
