(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Set Implicit Arguments.
Unset Strict Implicit.

Require Import Relation_Operators.
Require Import Relation_Definitions.
Require Import Operators_Properties.
Require Import Transitive_Closure.


Hint Resolve rt_step rt_refl rst_step rst_refl t_step: core.
Hint Unfold transp union reflexive transitive: core.


Section MoreRels.

Variables (A : Set) (R : relation A).

Theorem Acc_inv_rt :
 forall x y, clos_refl_trans A R y x -> Acc R x -> Acc R y.
simple induction 1; intros; auto.
apply Acc_inv with y0; auto.
Qed.

Theorem clos_t_clos_rt :
 forall x y, clos_trans _ R x y -> clos_refl_trans _ R x y.
simple induction 1; intros; auto.
apply rt_trans with y0; auto.
Qed.

End MoreRels.


Section WfUnion.

  Variable A : Set.
  Variable R1 R2 : relation A.
  
  Notation Union := (union _ R1 R2) (only parsing).

  Hint Resolve Acc_clos_trans wf_clos_trans: core.

Section UnionMin.

  Hypothesis commutR1R2 : commut _ (clos_refl_trans _ R1) (clos_trans _ R2).


  Lemma Acc_union :
   forall a,
   Acc R2 a ->
   (forall x, clos_refl_trans _ R2 x a -> Acc R1 x) -> Acc (union _ R1 R2) a.
Proof.
intros a H.
elimtype (Acc (clos_trans _ R2) a); intros; auto.
cut (forall y, clos_refl_trans _ R1 y x -> Acc (union _ R1 R2) y); intros;
 auto.
generalize H3.
elimtype (Acc R1 y); intros.
apply Acc_intro; intros.
inversion_clear H7.
apply H5; trivial.
apply rt_trans with x0; auto.

elim commutR1R2 with x x0 y0; intros; auto.
apply Acc_inv_rt with x1.
elim H9; intros; auto.
apply rt_trans with y1; auto.

apply H1; trivial; intros.
apply H2.
apply rt_trans with x1; auto.
apply clos_t_clos_rt; trivial.

apply Acc_inv_rt with x; auto.
Qed.


   Theorem wf_union :
    well_founded R1 -> well_founded R2 -> well_founded (union _ R1 R2).
Proof.
 unfold well_founded in |- *.
 intros.
 apply Acc_union; auto.
Qed.

End UnionMin.


Section WeakerUnion.

  Hypothesis comm : commut _ R1 R2.

  Lemma strip_commut : commut _ (clos_refl_trans _ R1) R2.
Proof.
red in |- *.
simple induction 1; intros; auto.
elim comm with (1 := H0) (2 := H1); intros.
split with x1; auto.

split with z; auto.

elim H1 with (1 := H4); intros.
elim H3 with (1 := H5); intros.
split with x2; auto.
apply rt_trans with x1; trivial.
Qed.

  Lemma strip_commut2 : commut _ (clos_refl_trans _ R1) (clos_trans _ R2).
Proof.
red in |- *; intros.
generalize x H.
clear H x.
elim H0; intros.
elim strip_commut with (1 := H1) (2 := H); intros.
split with x1; auto.

elim H3 with (1 := H4); intros.
elim H1 with (1 := H6); intros.
split with x2; auto.
apply t_trans with x1; trivial.
Qed.

  Lemma Acc_union2 :
   forall a,
   Acc R2 a ->
   (forall x, clos_refl_trans _ R2 x a -> Acc R1 x) -> Acc (union _ R1 R2) a.
Proof Acc_union strip_commut2.

  Theorem wf_union2 :
   well_founded R1 -> well_founded R2 -> well_founded (union _ R1 R2).
Proof wf_union strip_commut2.

End WeakerUnion.

End WfUnion.
