(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Import Rules.

Module Type IConfluence
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes)
  (Import MRules : IRules P MTermes MEnv).

  Definition confluent (R : red_rule) : Prop :=
    forall e : env, commut term (R e) (transp term (R e)).

  Section Confluence.

    Variable f1 f2 g1 g2 : term -> term -> term.

    Definition R1 u v := exists a : _, (exists2 b : _, u = f1 a b & v = f2 a b).
    Definition R2 u v := exists a : _, (exists2 b : _, u = g1 a b & v = g2 a b).

    Hypothesis no_cp :
      forall a b x, R2 (f1 a b) x ->
        exists a' : _,
          (exists2 b' : _,
            clos_refl_trans _ R2 a a' /\ clos_refl_trans _ R2 b b' &
            x = f1 a' b').

    Hypothesis corr1 : forall a a' b, R2 a a' -> R2 (f2 a b) (f2 a' b).
    Hypothesis corr2 : forall a b b', R2 b b' -> R2 (f2 a b) (f2 a b').

    Hypothesis confl_R1 : commut _ R1 (transp _ R1).
    Hypothesis confl_R2 : commut _ R2 (transp _ R2).

    (**)

    Variable R R' : red_rule.

    Hypothesis same_rel : forall e, same_relation term (R e) (R' e).
    Hypothesis R_confl : confluent R.

    Lemma confluent_is_ext : confluent R'.
    Proof.
      unfold confluent, commut, transp in |- *.
      intros.
      elim same_rel with e.
      unfold inclusion in |- *; intros.
      elim R_confl with e x y z; auto; intros.
      exists x0; auto.
    Qed.

    Lemma strip_confl : forall e, commut term (R_rt R e) (transp term (R e)).
    Proof.
      unfold commut, transp in |- *.
      simple induction 1; intros.
      elim R_confl with e y0 x0 z; intros; auto.
      exists x1; auto.

      exists z; auto.

      elim H1 with z0; intros; auto.
      elim H3 with x1; intros; auto.
      exists x2; auto.
      red in |- *.
      apply rt_trans with x1; auto.
    Qed.

    Lemma strongly_confluent : confluent (R_rt R).
      unfold confluent, commut, transp in |- *.
      simple induction 1; intros.
      elim strip_confl with e z x0 y0; intros; auto.
      exists x1; auto.

      exists z; auto.

      elim H1 with z0; intros; auto.
      elim H3 with x1; intros; auto.
      exists x2; auto.
      red in |- *.
      apply rt_trans with x1; auto.
    Qed.

  End Confluence.

  Section Tait_Martin_Lof.

    Variable R R' : red_rule.

    Definition church_rosser :=
      forall e u v, R_rst R e u v -> exists2 t, R_rt R e u t & R_rt R e v t.

    Hypothesis R'_confl : confluent R'.
    Hypothesis incl_R_R' : forall e, inclusion term (R e) (R' e).
    Hypothesis incl_R'_R_rt : forall e, inclusion term (R' e) (R_rt R e).

    Lemma closures_are_equal :
      forall e, same_relation term (R_rt R' e) (R_rt R e).
    Proof.
      unfold same_relation, inclusion, R_rt in |- *.
      split.
      simple induction 1; intros; auto.
      apply (incl_R'_R_rt e x0 y0); auto.

      apply rt_trans with y0; auto.

      simple induction 1; intros; auto.
      apply rt_step.
      apply (incl_R_R' e x0 y0); auto.

      apply rt_trans with y0; auto.
    Qed.

    Lemma confl_closure : confluent (R_rt R).
    Proof.
      apply confluent_is_ext with (R_rt R').
      exact closures_are_equal.

      apply strongly_confluent.
      trivial.
    Qed.

    Lemma TML_Church_rosser : church_rosser.
    Proof.
      unfold church_rosser, R_rt in |- *.
      simple induction 1; intros.
      exists y; auto.

      exists x; auto.

      elim H1; intros.
      exists x0; auto.

      elim H1; intros.
      elim H3; intros.
      elim confl_closure with e x0 y x1; intros; auto.
      exists x2.
      apply rt_trans with x0; auto.

      apply rt_trans with x1; auto.
    Qed.

  End Tait_Martin_Lof.

End IConfluence.
