(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export Rules.

Module Type ICtxt
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes)
  (Import MRules : IRules P MTermes MEnv).

  Section RedConv.

    Variable stab_rel : env -> decl -> decl -> Prop.
    Variable R : red_rule.

    Hypothesis R_lifts : R_lift R.
    Hypothesis R_substs : R_subst_weak R.
    Hypothesis stable : R_stable stab_rel R.

    Definition red := R_rt (ctxt R).
    Definition conv := R_rst (ctxt R).

    Hint Unfold red conv: core.

    (* amenes a disparaitre... ? *)
    Lemma red_lift : R_lift red.
    Proof R_rt_lift _ (ctxt_lift _ R_lifts).

    Lemma conv_lift : R_lift conv.
    Proof R_rst_lift _ (ctxt_lift _ R_lifts).

    Lemma red_subst : R_subst red.
    Proof R_rt_subst _ (ctxt_subst_weak _ R_substs).

    Lemma conv_subst : R_subst conv.
    Proof R_rst_subst _ (ctxt_subst_weak _ R_substs).

    Lemma red_trans : forall e x y z, red e x y -> red e y z -> red e x z.
    Proof fun e : env => rt_trans term (ctxt R e).

    Lemma conv_trans : forall e x y z, conv e x y -> conv e y z -> conv e x z.
    Proof fun e : env => rst_trans term (ctxt R e).

    Lemma conv_sym : forall e x y, conv e x y -> conv e y x.
    Proof fun e : env => rst_sym term (ctxt R e).

    Hint Immediate conv_sym.

    Lemma red_conv : forall e x y, red e x y -> conv e x y.
    Proof fun e : env => clos_rt_clos_rst term (ctxt R e).

    Lemma red_sym_conv : forall e x y, red e x y -> conv e y x.
    Proof.
      intros.
      apply conv_sym; apply red_conv; auto.
    Qed.

    Lemma red_red_cst :
      forall e c a a', red e a a' -> red e (Cst1 c a) (Cst1 c a').
    Proof.
      intros.
      elim H; intros; auto.
      prolog [ red_trans ] 2.
    Qed.

    Lemma red_red_lam : forall e x a a' m m',
      red e a a' -> red (e; x : a') m m' ->
        red e (Fun x : a => m) (Fun x : a' => m').
    Proof.
      intros.
      apply red_trans with (Fun x : a' => m).
      elim H; simpl in |- *; intros; auto.
      prolog [ red_trans ] 2.

      elim H0; intros; auto.
      prolog [ red_trans ] 2.
    Qed.

    Lemma red_red_prd : forall e x a a' m m',
      red e a a' -> red (e; x : a) m m' ->
        red e (Forall x : a, m) (Forall x : a', m').
    Proof.
      intros.
      apply red_trans with (Forall x : a, m').
      elim H0; intros; auto.
      prolog [ red_trans ] 2.

      elim H; simpl in |- *; intros; auto.
      prolog [ red_trans ] 2.
    Qed.

    Lemma red_red_sum :
      forall e m m' t t', red e m m' -> red e t t' -> red e (m ** t) (m' ** t').
    Proof.
      intros.
      apply red_trans with (m ** t').
      elim H0; intros; auto.
      prolog [ red_trans ] 2.

      elim H; intros; auto.
      prolog [ red_trans ] 2.
    Qed.

    Lemma red_red_pair :
      forall e a a' m m', red e a a' -> red e m m' -> red e (a,,m) (a',,m').
    Proof.
      intros.
      apply red_trans with (a',,m).
      elim H; simpl in |- *; intros; auto.
      prolog [ red_trans ] 2.

      elim H0; intros; eauto.
      prolog [ red_trans ] 2.
    Qed.

    Lemma conv_conv_cst :
      forall e c a a', conv e a a' -> conv e (Cst1 c a) (Cst1 c a').
    Proof.
      intros.
      elim H; intros; auto.
      prolog [ conv_trans ] 2.
    Qed.

    Lemma conv_conv_lam :
      forall e x a a' m m',
        conv e a a' ->
        conv (e; x : a) m m' -> conv e (Fun x : a => m) (Fun x : a' => m').
    Proof.
      intros.
      apply conv_trans with (Fun x : a => m').
      elim H0; intros; auto.
      prolog [ conv_trans ] 2.

      elim H; simpl in |- *; intros; auto.
      prolog [ conv_trans ] 2.
    Qed.

    Lemma conv_conv_prd :
      forall e x a a' m m',
        conv e a a' ->
        conv (e; x : a) m m' -> conv e (Forall x : a, m) (Forall x : a', m').
    Proof.
      intros.
      apply conv_trans with (Forall x : a, m').
      elim H0; intros; auto.
      prolog [ conv_trans ] 2.

      elim H; simpl in |- *; intros; auto.
      prolog [ conv_trans ] 2.
    Qed.

    Lemma conv_conv_sum :
      forall e m m' t t',
        conv e m m' -> conv e t t' -> conv e (m ** t) (m' ** t').
    Proof.
      intros.
      apply conv_trans with (m ** t').
      elim H0; intros; auto.
      prolog [ conv_trans ] 2.

      elim H; intros; auto.
      prolog [ conv_trans ] 2.
    Qed.

    Lemma conv_conv_pair :
      forall e a a' m m', conv e a a' -> conv e m m' -> conv e (a,,m) (a',,m').
    Proof.
      intros.
      apply conv_trans with (a',,m).
      elim H; simpl in |- *; intros; auto.
      prolog [ conv_trans ] 2.

      elim H0; intros; eauto.
      prolog [ conv_trans ] 2.
    Qed.

    Hint Resolve red_conv red_sym_conv red_red_cst red_red_lam red_red_prd
      red_red_sum red_red_pair conv_conv_cst conv_conv_lam conv_conv_prd
      conv_conv_sum conv_conv_pair: core.

    (* utile pour ctxt_sound, cas (u v): v-->v' *)
    Lemma red_subst_l :
      forall g u v, ctxt R g u v ->
        forall t n e, trunc n e g -> red e t[n:=u] t[n:=v].
    Proof.
      simple induction t; simpl in |- *; intros; auto.
      elim (lt_eq_lt_dec n0 n); intros; auto.
      case a; intros; auto.
      red in |- *; red in |- *.
      apply rt_step.
      apply iter_R_lift with g; auto.
    Qed.

    Lemma conv_subst_l :
      forall g u v,
        conv g u v -> forall t n e, trunc n e g -> conv e t[n:=u] t[n:=v].
    Proof.
      simple induction t; simpl in |- *; intros; auto.
      elim (lt_eq_lt_dec n0 n); intros; auto.
      case a; intros; auto.
      apply iter_R_lift with g; auto.
      exact conv_lift.
    Qed.

  End RedConv.

  Hint Unfold red conv: core.

  Hint Resolve red_conv red_sym_conv red_red_cst red_red_lam red_red_prd
    red_red_sum red_red_pair conv_conv_cst conv_conv_lam conv_conv_prd
    conv_conv_sum conv_conv_pair: core.

  Hint Immediate conv_sym: core.

  Hint Resolve red_lift conv_lift red_subst conv_subst: core.

End ICtxt.
