(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export Termes.

Module Type IEnv
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P).


(* Termes abstraits:
Variable term:Set.
Variable lift_rec: nat -> term -> nat -> term.
Variable subst_rec: term -> term -> nat -> term.
Definition lift := [n,t](lift_rec n t O).
Hypothesis permute_lift_rec: (n,p,M,k,i:?)(le i k)
       ->(lift_rec p (lift_rec n M k) i)
           =(lift_rec n (lift_rec p M i) (plus p k)).
Hypothesis commut_lift_subst_rec: (n,s,M,p,k:?)(le k p)
            ->(lift_rec n (subst_rec s M p) k)
               =(subst_rec s (lift_rec n M k) (plus n p)).
Hypothesis lift0:.
Hypothesis simpl_lift:.
*)

  (* Components of contexts *)

  Unset Elimination Schemes.

  Record decl : Set := mkdecl {NAME : name; TYPE : term; BODY : option term}.

  Record cdecl n : Set := {
    dNAME : name;
    dTYPE : cterm n;
    dBODY : option (cterm n)
  }.

  Definition cdcl n (x : cdecl n) : decl.
  Proof.
    destruct x as [N [t Ht] [[b Hb]| ]].
    exact (mkdecl N t (Some b)).
    exact (mkdecl N t None).
  Defined.

  Coercion cdcl : cdecl >-> decl.

  Lemma closty n (x : cdecl n) : closed n (TYPE x).
  Proof.
    destruct x as [N [t Ht] [[b Hb]| ]]; exact Ht.
  Qed.

  Lemma closbd n (x : cdecl n) : forall t, BODY x = Some t -> closed n t.
  Proof.
    destruct x as [N [t Ht] [[b Hb]| ]]; intros T HT.
    injection HT; intros; subst; assumption.
    discriminate.
  Qed.

  Implicit Arguments cdcl [ n ].
  Implicit Arguments closty [ n ].
  Implicit Arguments closbd [ n ].
  Hint Resolve closty: core.


  (* Binding structure of terms *)
  Inductive subt_nobind (m : term) : term -> Prop :=
    | Nbsbt_cst : forall c, subt_nobind m (Cst1 c m)
    | Nbsbt_lam : forall x n, subt_nobind m (Fun x : m => n)
    | Nbsbt_prd : forall x n, subt_nobind m (Forall x : m, n)
    | Nbsbt_sum_l : forall n, subt_nobind m (m ** n)
    | Nbsbt_sum_r : forall n, subt_nobind m (n ** m)
    | Nbsbt_pair_l : forall n, subt_nobind m (m,,n)
    | Nbsbt_pair_r : forall n, subt_nobind m (n,,m).

  Inductive subt_bind (m : term) : term -> decl -> Prop :=
    | Bsbt_lam : forall x t, subt_bind m (Fun x : t => m) (mkdecl x t None)
    | Bsbt_prd : forall x t, subt_bind m (Forall x : t, m) (mkdecl x t None).

  Hint Resolve Nbsbt_cst Nbsbt_lam Nbsbt_prd Nbsbt_sum_l Nbsbt_sum_r
    Nbsbt_pair_l Nbsbt_pair_r: core.
  Hint Resolve Bsbt_lam Bsbt_prd: core.


  Definition liftd_rec n d k :=
    mkdecl (NAME d) ^(n//k)(TYPE d) (app_option (fun t => ^(n//k)t) (BODY d)).

  Definition substd_rec s d k :=
    mkdecl (NAME d) (TYPE d)[k:=s] (app_option (fun t => t[k:=s]) (BODY d)).



  Definition env := list decl.

  Definition nil_env : env := nil.
  Definition add_var (e : env) (x : name) (t : term) : env :=
    mkdecl x t None :: e.
  Definition add_def (e : env) (x : name) (m t : term) : env :=
    mkdecl x t (Some m) :: e.
  Definition item_rel e n d := item d (e:env) n.

  Lemma add_rel_hd : forall (e : env) d, item_rel (d :: e) 0 d.
Proof.
unfold item_rel in |- *; constructor.
Qed.

  Lemma add_rel_tl :
   forall e n d d', item_rel e n d -> item_rel (d' :: e) (S n) d. 
Proof.
unfold item_rel in |- *; constructor; trivial.
Qed.

  Hint Resolve add_rel_hd add_rel_tl: core.

  Lemma trunc_add_var :
   forall (e f : env) n x t, trunc n e f -> trunc (S n) (add_var e x t) f.
Proof.
unfold add_var in |- *.
auto.
Qed.

  Lemma trunc_add_def :
   forall (e f : env) n x m t, trunc n e f -> trunc (S n) (add_def e x m t) f.
Proof.
unfold add_def in |- *.
auto.
Qed.

  Hint Resolve trunc_add_var trunc_add_def: core.



  Lemma inj_item_rel :
   forall e n d1 d2, item_rel e n d1 -> item_rel e n d2 -> d1 = d2.
Proof fun e n d1 d2 it1 it2 => fun_item it1 it2.


  Definition val_ok d t :=
    match BODY d with
    | Some t' => t = t'
    | _ => True
    end.

  Hint Unfold val_ok: core.


  Definition type_of_rel e n t :=
    exists2 u : _, t = ^(S n)(TYPE u) & item_rel e n u.



(* insertion in the context with relocation *)

  Set Elimination Schemes.

  Inductive ins_in_env (e : env) (d : decl) : nat -> env -> env -> Prop :=
    | Ins_hd : ins_in_env e d 0 e (d :: e)
    | Ins_tl :
        forall n f g t,
        ins_in_env e d n f g ->
        ins_in_env e d (S n) (t :: f) (liftd_rec 1 t n :: g).

  Hint Resolve Ins_hd Ins_tl: core.

  Lemma ins_tl' :
   forall e d n f g t t',
   ins_in_env e d n f g ->
   t' = liftd_rec 1 t n -> ins_in_env e d (S n) (t :: f) (t' :: g).
Proof.
intros.
rewrite H0; auto.
Qed.


  Lemma ins_item_ge :
   forall d n e f g,
   ins_in_env g d n e f ->
   forall d' v, n <= v -> item_rel e v d' -> item_rel f (S v) d'.
Proof.
simple induction 1; auto.
simple destruct v; intros.
inversion_clear H2.

inversion_clear H3; auto with arith.
Qed.


  Lemma ins_item_lt :
   forall x n e f g,
   ins_in_env g x n e f ->
   forall d v,
   n > v -> item_rel e v d -> item_rel f v (liftd_rec 1 d (n - S v)).
Proof.
simple induction 1; auto.
intros.
inversion_clear H0.

simple destruct v; simpl in |- *; intros.
replace (n0 - 0) with n0; auto with arith.
inversion_clear H3; auto.

inversion_clear H3; auto with arith.
Qed.


  Lemma ins_type_of_rel_lt :
   forall d n e f g,
   ins_in_env g d n e f ->
   forall t v, n > v -> type_of_rel e v t -> type_of_rel f v ^(1//n)t.
Proof.
simple destruct 3; intros.
split with (liftd_rec 1 x (n - S v)).
rewrite H2.
unfold lift in |- *.
pattern n at 1 in |- *.
replace n with (S v + (n - S v)); auto with arith.
rewrite <- permute_lift_rec; auto with arith.

apply ins_item_lt with d e g; auto.
Qed.



(* substitution in context with relocation *)

  Inductive sub_in_env (e : env) (t : term) (d : decl) :
  nat -> env -> env -> Prop :=
    | Sub_hd : val_ok d t -> sub_in_env e t d 0 (d :: e) e
    | Sub_tl :
        forall f g n u,
        sub_in_env e t d n f g ->
        sub_in_env e t d (S n) (u :: f) (substd_rec t u n :: g).

  Hint Resolve Sub_hd Sub_tl: core.


  Lemma sub_env_trunc :
   forall e f g t d n, sub_in_env g t d n e f -> trunc n f g.
simple induction 1; intros; auto.
Qed.


  Lemma nth_sub_sup :
   forall d t n e f g,
   sub_in_env g t d n e f ->
   forall d' v, n <= v -> item_rel e (S v) d' -> item_rel f v d'.
Proof.
simple induction 1.
intros.
inversion_clear H2; auto.

simple destruct v; intros.
inversion_clear H2.

inversion_clear H3; auto with arith.
Qed.


  Lemma sub_decl_eq_def :
   forall e f g t d n, sub_in_env g t d n e f -> item_rel e n d.
Proof.
simple induction 1; auto.
Qed.

  Lemma sub_decl_eq_ok :
   forall e f g t d n, sub_in_env g t d n e f -> val_ok d t.
Proof.
simple induction 1; auto.
Qed.


  Lemma nth_sub_inf :
   forall e f g t d n,
   sub_in_env g t d n e f ->
   forall d' v,
   n > v -> item_rel e v d' -> item_rel f v (substd_rec t d' (n - S v)).
Proof.
simple induction 1; auto.
intros.
inversion_clear H1.

simple destruct v; simpl in |- *; intros.
replace (n0 - 0) with n0; auto with arith.
inversion_clear H3; auto.

inversion_clear H3; auto with arith.
Qed.


  Lemma nth_sub_item_inf :
   forall e f g t d n,
   sub_in_env g t d n e f ->
   forall u v, n > v -> type_of_rel e v u -> type_of_rel f v u[n:=t].
Proof.
simple destruct 3; intros.
split with (substd_rec t x (n - S v)).
rewrite H2.
unfold lift in |- *.
pattern n at 1 in |- *.
replace n with (S v + (n - S v)); auto with arith.
elim commut_lift_subst_rec with (S v) t (TYPE x) (n - S v) 0; auto with arith.

apply nth_sub_inf with e g d; auto.
Qed.




  Lemma ins_in_env_var :
   forall e f g k d x t,
   ins_in_env g d k e f ->
   ins_in_env g d (S k) (add_var e x t) (add_var f x ^(1//k)t).
Proof.
intros.
unfold add_var in |- *.
apply Ins_tl; assumption.
Qed.

  Lemma ins_in_env_def :
   forall e f g k d x m t,
   ins_in_env g d k e f ->
   ins_in_env g d (S k) (add_def e x m t) (add_def f x ^(1//k)m ^(1//k)t).
Proof.
intros.
unfold add_def in |- *.
apply Ins_tl; assumption.
Qed.


  Lemma sub_in_env_var :
   forall e f g k d x s t,
   sub_in_env g s d k e f ->
   sub_in_env g s d (S k) (add_var e x t) (add_var f x t[k:=s]).
Proof.
intros.
unfold add_var in |- *.
apply Sub_tl; assumption.
Qed.

  Lemma sub_in_env_def :
   forall e f g k d x s m t,
   sub_in_env g s d k e f ->
   sub_in_env g s d (S k) (add_def e x m t) (add_def f x m[k:=s] t[k:=s]).
Proof.
intros.
unfold add_def in |- *.
apply Sub_tl; assumption.
Qed.

  Lemma rel_in_env_var :
   forall (R : env -> decl -> decl -> Prop) e f x t,
   rel_in_list R e f -> rel_in_list R (add_var e x t) (add_var f x t).
Proof.
intros.
unfold add_var in |- *.
auto.
Qed.

  Lemma rel_in_env_def :
   forall (R : env -> decl -> decl -> Prop) e f x m t,
   rel_in_list R e f -> rel_in_list R (add_def e x m t) (add_def f x m t).
Proof.
intros.
unfold add_def in |- *; auto.
Qed.

  Hint Resolve ins_in_env_var ins_in_env_def sub_in_env_var sub_in_env_def
    rel_in_env_var rel_in_env_def: core.





  Definition red_rule := env -> term -> term -> Prop.


Section RuleOperators.

  Variable stab_rel : env -> decl -> decl -> Prop.
  Variable R : red_rule.

  (* transitive closures *)
  Definition R_t : red_rule := fun e => clos_trans _ (R e).
  Definition R_rt : red_rule := fun e => clos_refl_trans _ (R e).
  Definition R_rst : red_rule := fun e => clos_refl_sym_trans _ (R e).


(* stability w.r.t. *infered* types of *dynamically* introduced objects.
   Stability requirement sufficient to prove type inference, but not to prove
   subject reduction beta -> PTSO_spec uses rel_typ
 *)
(*  Inductive rel_def_typ [e:env]: (relation obj) :=
    Rty_def: (d,t,u:term)(R e t u)->(rel_def_typ e ((Def d),t) ((Def d),u)).
*)

  (* Does not take names into account *)
  Definition rel_typ e d1 d2 := R e (TYPE d1) (TYPE d2) /\ BODY d1 = BODY d2.


  Definition R_stable :=
    forall e t u, R e t u -> forall f, rel_in_list stab_rel f e -> R f t u.


  Definition R_lift :=
    forall g e d t u,
    R e t u -> forall k f, ins_in_env g d k e f -> R f ^(1//k)t ^(1//k)u.


  Definition R_subst :=
    forall s d g e t u,
    R e t u -> forall f k, sub_in_env g s d k e f -> R f t[k:=s] u[k:=s].


  Definition R_subst_weak :=
    forall m d e g t u,
    R e t u ->
    forall n f,
    sub_in_env g m d n e f -> clos_refl_trans _ (R f) t[n:=m] u[n:=m].


  Hypothesis Rlifts : R_lift.

  Lemma iter_R_lift :
   forall (e f : env) n,
   trunc n f e -> forall t u, R e t u -> R f ^(n)t ^(n)u.
Proof.
simple induction 1; intros.
repeat rewrite lift0; auto.

rewrite simpl_lift.
rewrite (simpl_lift u).
unfold lift at 1 3 in |- *.
apply Rlifts with e0 e0 x; auto.
Qed.

End RuleOperators.

  Hint Unfold R_rt R_rst: core.
  Hint Unfold R_lift R_subst R_subst_weak R_stable: core.



Section LiftProperties.

  Variable R : red_rule.
  Hypothesis R_lifts : R_lift R.

  Lemma Rlift0 :
   forall e a b, R e a b -> forall x t, R (add_var e x t) ^(1)a ^(1)b.
Proof.
intros.
unfold lift in |- *.
apply R_lifts with e e (mkdecl x t None); auto.
unfold add_var in |- *; auto.
Qed.


  Lemma R_rt_lift : R_lift (R_rt R).
Proof.
unfold R_lift, R_rt in |- *.
simple induction 1; intros; eauto.
apply rt_trans with (^(1//k)y); auto.
Qed.

  Lemma R_rst_lift : R_lift (R_rst R).
Proof.
unfold R_lift, R_rst in |- *.
simple induction 1; intros; eauto.
apply rst_sym; eauto.

apply rst_trans with (^(1//k)y); auto.
Qed.

End LiftProperties.


Section SubstProperties.

  Variable R : red_rule.
  Hypothesis R_substs : R_subst R.

  Lemma weak_subst : R_subst_weak R.
Proof.
red in |- *.
eauto.
Qed.

  Lemma Rsubst0 :
   forall e x t a b, R (add_var e x t) a b -> forall m, R e a[0:=m] b[0:=m].
Proof.
intros.
apply R_substs with (mkdecl x t None) e (add_var e x t); auto.
unfold add_var in |- *; constructor.
compute in |- *; trivial.
Qed.


End SubstProperties.


Section SubstWeakProperties.

  Variable R : red_rule.
  Hypothesis R_substs : R_subst_weak R.


  Lemma R_rt_subst : R_subst (R_rt R).
Proof.
red in |- *; red in |- *.
simple induction 1; intros; auto.
eauto.

apply rt_trans with (y[k:=s]); auto.
Qed.


  Lemma R_rst_subst : R_subst (R_rst R).
Proof.
red in |- *; red in |- *.
simple induction 1; intros; auto.
elimtype (clos_refl_trans _ (R f) x[k:=s] y[k:=s]); intros; auto.
apply rst_trans with y0; trivial.

eauto.

apply rst_sym; auto.

apply rst_trans with (y[k:=s]); auto.
Qed.

End SubstWeakProperties.


Section StableProperties.

  Variable R : red_rule.
  Variable stab_rel : env -> decl -> decl -> Prop.
  Hypothesis stable : R_stable stab_rel R.

  Lemma stable_rt : R_stable stab_rel (R_rt R).
Proof.
red in |- *; red in |- *.
simple induction 1; intros; auto.
eauto.

apply rt_trans with y; auto.
Qed.

  Lemma stable_rst : R_stable stab_rel (R_rst R).
Proof.
red in |- *; red in |- *.
simple induction 1; intros; auto.
eauto.

apply rst_sym; auto.

apply rst_trans with y; auto.
Qed.


  Variable stab_rel' : env -> decl -> decl -> Prop.
  Hypothesis incl : forall e, inclusion _ (stab_rel' e) (stab_rel e).

  Lemma stable_incl : R_stable stab_rel' R.
Proof.
red in |- *; intros.
red in incl.
apply stable with e; auto.
elim H0; auto.
Qed.


  Lemma stable_rt_left : R_stable (fun e => clos_refl_trans _ (stab_rel e)) R.
Proof.
red in |- *; intros.
generalize H.
clear H.
elimtype (clos_refl_trans _ (rel_in_list stab_rel) f e); intros; auto.
apply stable with y; auto.

elim H0; intros; auto.
elim H; intros; auto.
apply rt_trans with (y0 :: l); auto.

elim H1; intros; auto.
apply rt_trans with (x :: y); auto.
Qed.

End StableProperties.

  Hint Resolve R_rt_lift R_rst_lift R_rt_subst R_rst_subst: core.
  Hint Resolve weak_subst stable_rt stable_rst stable_rt_left: core.



Section SubtypingRuleDef.

  Unset Elimination Schemes.

  Record Subtyping_rule : Type := 
    {Rule : red_rule;
     Rule_lifts : R_lift Rule;
     Rule_substs : R_subst Rule;
     Rule_stable : R_stable (rel_typ Rule) Rule;
     preord : forall e : env, preorder term (Rule e)}.


End SubtypingRuleDef.



  (* Another stability relation... *)
  Record decl_rel (Req Rsub : red_rule) (e : env) (d1 d2 : decl) : Prop := 
    {sub_ty : Rsub e (TYPE d1) (TYPE d2);
     eq_body :
      match BODY d2, BODY d1 with
      | None, _ => True
      | Some m2, Some m1 => Req e m1 m2
      | _, _ => False
      end}.

  Definition alpha_stab :=
    rel_in_list (decl_rel (fun _ t1 t2 => t1 = t2) (fun _ t1 t2 => t1 = t2)).

(*
  Lemma incl_rel_typ_alpha_stab: (Q:red_rule; e:env; d1,d2:decl)
      (rel_typ Q e d1 d2)
    ->(decl_rel [_;t1,t2]t1=t2 Q e d1 d2).
Intros.
Inversion_clear H.
Generalize H0 H1.
(Case d1; Case d2; Simpl; Intros).
Elim H3.
(Split; Auto).
Simpl.
(Case BODY1; Simpl; Trivial).



Inversion_clear H.
(Split; Auto).
Generalize eq_body0.
(Case (BODY d2); Case (BODY d1); Simpl; Trivial).
(Induction 1; Trivial).

Induction 1.
*)

  Lemma incl_rel_typ_alpha_stab :
   forall e f : env,
   rel_in_list (rel_typ (fun _ t1 t2 => t1 = t2)) e f -> alpha_stab e f.
Proof.
unfold alpha_stab in |- *.
simple induction 1; intros; auto.
constructor.
inversion_clear H0.
split; auto.
elim H2.
case (BODY x); trivial.
Qed.

  Hint Resolve incl_rel_typ_alpha_stab: core.


Notation "(  )" := (nil (A:=decl)) (at level 0).
Notation "e ; d " := ((d:decl) :: e)
  (at level 300, d at level 80, left associativity).
Notation "e ; x : t" := (add_var e x t)
  (at level 300, left associativity, x at level 80).
Notation "e ; x := m : t" := (add_def e x m t)
  (at level 300, left associativity, x at level 80, m at level 80).

End IEnv.
