(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Import MlPrint.
Require Export Metatheory.

Module Type IInfer
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes)
  (Import MMetatheory : IMetatheory P MTermes MEnv).


Open Scope pts_scope.

(* Regles de typage *)

Section IInfer.

  Context {the_PTS : PTSO_spec}.

  Definition typed_sort := in_dom_rel axiom.


  Section TopsortDec.

    Hypothesis
      infer_axiom :
      forall s, {sp : _ | ppal (axiom s) le_sort sp} + {~ typed_sort s}.

    Lemma not_topsort : forall t, decide (forall s, t = %%s -> typed_sort s).
    Proof.
      (*Realizer [t]
         Cases t of
          (Srt s) => Cases (infer_axiom s) of
                       (inleft _) => true
                     | inright => false
                     end
        | _ => true
        end.*)
      intros [s| n| c| c u| x a b| x a m| a b| u v];
        try (left; intros; discriminate).
      case (infer_axiom s).
      intros (s', ty_s); left; intros s0 eq_s0; injection eq_s0; simple destruct 1;
        exists s'; apply (pp_ok ty_s).

      red in |- *; auto.
    Defined.

    Lemma sorted_type :
      forall e,
        ⊦ e -> forall t, wf_type e t -> decide (exists s : _, e ⊦ t : %%s).
    Proof.
      intros e wfe t wfty.
      case (not_topsort t).
      intro not_top; left; generalize not_top; inversion_clear wfty.
      exists s; trivial.

      intros not_top_s.
      case not_top_s with s; trivial; intros s' ty_s.
      exists s'; auto.

      intros tops; right; red in |- *; intros (s, ty_s); apply tops.
      intros s' eq_s'; rewrite eq_s' in ty_s; Inversion_typ ty_s.
      exists s2; trivial.
    Defined.

  End TopsortDec.

  Definition cast_to_sort_dec e t :=
    ppal_dec (fun s => e ⊦ t <= %%s) le_sort.


  Definition weak_to_least_sort (R : sort -> sort -> sort -> Prop) e s1 b
    t :=
    e ⊦ b <= t /\
    (exists2 s2 : _, e ⊦ t : %%s2 & (exists s3 : _, R s1 s2 s3)).


  Class PTSO_algos := {

    (* Assumptions on the PTS and the subtyping relation *)

    (* sorts *)
    ppa_infer_axiom :
      forall s, {sp : _ | ppal (axiom s) le_sort sp} + {~ typed_sort s};

    (* constants *)
    ppa_infer_sign0 :
      forall c,
        ppal_dec (fun u => in_sign0 c u /\ ⊦ (( ); x_ : u)) (eq (A:=_));
    ppa_infer_sign1 :
      forall e c m t,
        e ⊦ m : t -> ppal_dec (in_sign_sat c e m t) (le_type e);


    (* assumptions *)
    ppa_least_sort : forall e t, wf_type e t -> cast_to_sort_dec e t;


    (* products and functions *)
    ppa_infer_rule :
      forall s1 s2, ppal_dec (fun s3 => rl_sat s1 s2 s3) le_sort;
    ppa_weak_rule :
      forall e s1 b,
        (exists s : _, e ⊦ b : %%s) ->
        ppal_dec (weak_to_least_sort rl_sat e s1 b) (le_type e);
    ppa_le_type_prod :
      forall e x a b1 b2,
        (e; x : a) ⊦ b1 <= b2 -> e ⊦ Forall x : a, b1 <= Forall x : a, b2;


    (* sums and pairs (for simplicity, we assume that all sums are allowed) *)
    ppa_infer_prod :
      forall s1 s2, {s3 : _ | ppal (fun s => pr_sat s1 s2 s) le_sort s3};
    ppa_le_type_sum :
      forall e a1 a2 b1 b2,
        e ⊦ a1 <= a2 -> e ⊦ b1 <= b2 -> e ⊦ a1 ** b1 <= a2 ** b2;


    (* subtyping *)
    ppa_topsort_untyped :
      forall e s s' t, e ⊦ %%s <= t -> e ⊦ t : %%s' -> typed_sort s;
    ppa_le_type_dec :
      forall e u v, wf_type e u -> wf_type e v -> decide (e ⊦ u <= v)

  }.

  Section IInferWithAlgos.

    Context {the_algos : PTSO_algos}.


(* Remarks: alternative to ppa_weak_rule

   1- Semi-full is a stronger requirement:

Section SemiFull.

Hypothesis semi_full:
  (s1:sort) { (s2:sort)(EX s3 | (rl_sat s1 s2 s3)) }
            +{ (s2,s3:sort)~(rl_sat s1 s2 s3) }.

Local weak_rule: (e:env)(s1:sort)(b:term)(EX s | e |- b : %%s)
        ->(ppal_dec (weak_to_least_sort rl_sat e s1 b) (le_type e)).
Proof.
Intros.
(Elim semi_full with s1; Intros).
Left .
(Split with b; Intros).
Split.
Split; Auto.
Inversion_clear H.
Split with x; Trivial.

Intros y (leby, (s2, tyy, (s3, rl))); Trivial.

(Right; Red; Intros).
Inversion_clear H0.
Inversion_clear H2.
Inversion_clear H3.
Red in y.
(Apply y with x0 x1; Trivial).
Save.

End SemiFull.


  2- Better: We can show that if the principal type of (Lam x a m) is
     decidable, then we can decide weak_to_least_sort above!
*)

  (* Principal type and sort *)
  Definition typ_ppal e m := ppal (fun t => e ⊦ m : t) (le_type e).

  Definition sort_ppal e m := ppal (fun s => e ⊦ m : %%s) le_sort.


  Unset Elimination Schemes.

  (* judgements *)
  Record ppal_judge (e : env) : Set := mkjpp
    {VAL : term; TYP : term; ppal_ty : typ_ppal e VAL TYP}.

  Implicit Arguments mkjpp [ e VAL TYP ].
  Implicit Arguments VAL [ e ].
  Implicit Arguments TYP [ e ].
  Implicit Arguments ppal_ty [ e ].

  Lemma judge_ok : forall {e} (j : ppal_judge e), e ⊦ VAL j : TYP j.
  Proof fun e (j : ppal_judge e) => pp_ok (ppal_ty j).
  Hint Resolve @judge_ok: core.

  Lemma judge_wft : forall {e} (j : ppal_judge e), wf_type e (TYP j).
  Proof fun e (j : ppal_judge e) => type_correctness _ _ _ (judge_ok j).
  Hint Resolve @judge_wft: core.

  Definition ju e t := {j : ppal_judge e | t = VAL j}.

  Definition ju_intro : forall e m t, typ_ppal e m t -> ju e m.
  Proof.
    intros e m t typp.
    exists (mkjpp typp); trivial.
  Defined.


  Definition infer_typ_ppal e t := ju e t + {(forall T, ~ e ⊦ t : T)}.



  (* Transformation judgement -> declaration *)
  Section MachAssum.

    (* retour a la revision 1.7 *)

    Variable e : env.

    Record assum : Set :=
      {ANAME : name;
       ATYPE : term;
       AKIND : sort;
       ppal_kind : sort_ppal e ATYPE AKIND}.

    Implicit Arguments Build_assum [ ATYPE AKIND ].

    Lemma assum_ok : forall a, e ⊦ ATYPE a : %%(AKIND a).
    Proof fun a => pp_ok (ppal_kind a).
    Hint Resolve assum_ok: core.


    Definition push_assum a := (e; ANAME a : ATYPE a).

    Lemma push_wf : forall a, ⊦ push_assum a.
    Proof.
      intro a.
      unfold push_assum in |- *.
      apply wf_var with (AKIND a); trivial.
    Qed.
    Hint Resolve push_wf: core.


    (* Building an assumption *)
    Definition hyp x t := {a : _ | ATYPE a = t &  ANAME a = x}.

    Lemma mka_var : forall x t s, sort_ppal e t s -> hyp x t.
    Proof.
      intros x t s spp.
      exists (Build_assum x spp); trivial.
    Qed.

    Definition infer_sort_ppal x t := hyp x t + {~ ⊦ (e; x : t)}.

    Lemma infer_assum : forall x (j : ppal_judge e), infer_sort_ppal x (VAL j).
    Proof.
      (*Realizer [x,j]
        Cases (ppa_least_sort the_algos e (TYP e j)) of
          (inleft s1) => (inleft ? (mka_var x (VAL e j) s1))
        | inright => (inright assum)
        end.*)
      intros x j.
      case ppa_least_sort with e (TYP j); trivial.
      intros (s1, pp_sort).
      left.
      assert (s1_pp : sort_ppal e (VAL j) s1).
      split; intros.
      apply Typ_conv_srt with (TYP j); trivial.
      apply (pp_ok pp_sort).

      apply (pp_least pp_sort).
      apply (pp_least (ppal_ty j) H).

      exists (Build_assum x s1_pp); simpl in |- *; trivial.

      right; red in |- *; intros.
      Inversion_typ H.
      red in n.
      apply n with s; trivial.
      apply (pp_least (ppal_ty j)); trivial.
    Defined.

  End MachAssum.

  Hint Resolve @assum_ok @push_wf: core.

  Implicit Arguments ATYPE [ e ].
  Implicit Arguments ANAME [ e ].
  Implicit Arguments AKIND [ e ].

  Section MachDef.

    Variable e : env.

    Lemma wfd_ok :
      forall x m t u s,
        e ⊦ m : t -> e ⊦ t <= u -> e ⊦ u : %%s -> ⊦ (e; x := m : t).
    Proof.
      intros x m t u s j.
      generalize (type_correctness _ _ _ j) j.
      intro wft.
      inversion_clear wft; intros.
      apply wf_def with s0; trivial.

      elim ppa_topsort_untyped with e s0 s u; intros; trivial.
      apply wf_def with x0; auto.
      apply Typ_srt; trivial.
      apply typ_wf with (1 := j).
    Qed.


    Record const : Set :=
      {CNAME : name;
       CBODY : ppal_judge e;
       ty_sorted : exists s : _, e ⊦ TYP CBODY : %%s}.

    Implicit Arguments Build_const [ CBODY ].

    Definition push_const c := (e; CNAME c := VAL (CBODY c) : TYP (CBODY c)).



    Lemma push_const_wf : forall c, ⊦ push_const c.
    Proof.
      intro c.
      unfold push_const in |- *.
      elim (ty_sorted c); intros.
      apply wf_def with x; trivial.
    Qed.


    Lemma infer_def :
      forall x (j : ppal_judge e),
        {c : _ | CBODY c = j &  CNAME c = x} + {~ ⊦ (e; x := VAL j : TYP j)}.
    Proof.
      (*Realizer [x,j]
        if (sorted_type (ppa_infer_axiom the_algos) e (TYP e j))
        then (inleft ? (Build_const x j))
        else (inright const).*)
      intros x j.
      case (sorted_type ppa_infer_axiom) with e (TYP j); trivial.
      apply (typ_wf _ _ _ (pp_ok (ppal_ty j))).
      intros ty_s; left; exists (Build_const x ty_s); simpl in |- *; trivial.

      right; red in |- *; intros.
      apply n.
      Inversion_typ H.
      split with s; auto.
    Defined.

    Lemma define_body :
      forall (a : assum e) (j : ppal_judge e),
        e ⊦ TYP j <= ATYPE a -> ⊦ (e; ANAME a := VAL j : ATYPE a).
    Proof.
      intros.
      apply wf_def with (AKIND a); trivial.
      apply Typ_conv with (TYP j) (AKIND a); trivial.
    Qed.


    Lemma check_def :
      forall (a : assum e) (j : ppal_judge e),
        decide (⊦ (e; ANAME a := VAL j : ATYPE a)).
    Proof.
      (*Realizer [a:assum; j:ppal_judge]
        if (ppa_le_type_dec the_algos e (TYP e j) (ATYPE e a)) then true else false.
      *)
      intros a j.
      case ppa_le_type_dec with e (TYP j) (ATYPE a); trivial.

      left with (AKIND a); trivial.

      left; apply define_body; trivial.

      right; red in |- *; intros.
      apply n.
      Inversion_typ H.
      apply (pp_least (ppal_ty j)); trivial.
    Defined.

  End MachDef.

  End IInferWithAlgos.

End IInfer.

Hint Resolve @judge_ok: core.
Hint Resolve @judge_wft: core.
Hint Resolve @assum_ok: core.
Hint Resolve @push_wf: core.

Implicit Arguments CBODY [ the_PTS e ].
Implicit Arguments VAL [ the_PTS e ].
Implicit Arguments TYP [ the_PTS e ].
Implicit Arguments AKIND [ the_PTS e ].
Implicit Arguments ANAME [ the_PTS e ].
Implicit Arguments ATYPE [ the_PTS e ].

End IInfer.
