(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export Machops.

(* The type-checking module to build *)
Module Type IMach
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes)
  (Import MMetatheory : IMetatheory P MTermes MEnv)
  (Import MInfer : IInfer P MTermes MEnv MMetatheory).

  Include IMachops P MTermes MEnv MMetatheory MInfer.

  Ltac TypeCheck e m j :=
    assert (ty_fun : forall f t, ⊦ f -> infer_typ_ppal f t); trivial;
     elim ty_fun with e m; clear ty_fun;
     [ intros (j, equ_j); rewrite equ_j; clear equ_j
     | intro neg_ty; red in neg_ty; Contrapose_silent (); eauto
     | trivial ].


  Section SPTSO_TC.

    Context {MPTSO_spec : PTSO_spec}.
    Context {MPTSO_algos : PTSO_algos}.

    Class PTSO_TC := {
      inf_ppal_type : forall e t, ⊦ e -> infer_typ_ppal e t;
      inf_ppal_sort : forall e x t, ⊦ e -> infer_sort_ppal e x t;
      chk_typ : forall e t T, ⊦ e -> decide (e ⊦ t : T);
      chk_decl : forall e d, ⊦ e -> decide (⊦ (e; d));
      chk_wk : forall e t T, ⊦ e -> wf_type e T -> decide (e ⊦ t : T);
      chk_wft : forall e t, ⊦ e -> decide (wf_type e t)
    }.

  Lemma inf_ppal_type_1 : forall e t, ⊦ e -> infer_typ_ppal e t.
  Proof.
    (*Refine Fix type_rec
    { type_rec [e:?; t:term]: ? := [wfe]
      <[t](infer_typ_ppal e t)>Cases t of
        (Srt s) => (infer_sort wfe s)
      | (Rel n) => (infer_rel wfe n)
      | (Cst0 c) => (infer_cst0 wfe c)
      | (Cst1 c a) => ?
      | (Lam x a m) => ?
      | (Prd x a b) => ?
      | (Sum a b) => ?
      | (Pair u v) => ?
      end}.*)
    fix type_rec 2; intros e [s| n| c| c a| x a b| x a m| a b| u v] wfe;
     [ apply infer_sort; trivial
     | apply infer_rel; trivial
     | apply infer_cst0; trivial
     | idtac
     | idtac
     | idtac
     | idtac
     | idtac ].
    (* Cst1 *)
    TypeCheck e a ipattern:ja.
    apply infer_cst1.

    (* Lam *)
    TypeCheck e a ipattern:ja.
    MakeAssum x ja ipattern:assa.
    TypeCheck (push_assum _ assa) b ipattern:jb.
    apply infer_prd.

    (* Prd *)
    TypeCheck e a ipattern:ja.
    MakeAssum x ja ipattern:assa.
    TypeCheck (push_assum _ assa) m ipattern:jm.
    apply infer_lam.

    Inversion_typ H0; eauto.

    (* Sum *)
    TypeCheck e a ipattern:ja.
    TypeCheck e b ipattern:jb.
    apply infer_sum.

    (* Pair *)
    TypeCheck e u ipattern:j1.
    TypeCheck e v ipattern:j2.
    apply infer_pair.
  Qed.


  Ltac TypeCheckTerm e m j :=
    elim inf_ppal_type_1 with e m;
     [ intros (j, equ_j); rewrite equ_j; clear equ_j
     | right; red in |- *; intros; eauto
     | trivial ].



  Lemma inf_ppal_sort_1 : forall e x t, ⊦ e -> infer_sort_ppal e x t.
  Proof.
    intros.
    TypeCheckTerm e t ipattern:j.
    apply infer_assum.

    Inversion_typ H0.
    elim b with (1 := H1).
  Qed.



  Lemma chk_wft_1 : forall e t, ⊦ e -> decide (wf_type e t).
  Proof.
    (*Realizer [e,t]
      if (is_a_sort t) then true else
        Cases (typecheck_sort e x_ t) of
          (inleft _) => true
        | inright => false
        end.*)
    intros e t wfe.
    case (is_a_sort t).
    intros is_srt; left; elim is_srt; intros s eq_s; rewrite eq_s; trivial.

    intros not_srt.
    case (inf_ppal_sort_1 _ x_ t wfe).
    intros ((x, t', s, ty_t'), eq_t, eq_x); simpl in eq_t, eq_x; left.
    subst t'.
    left with s; case ty_t'; trivial.

    right; red in |- *; intros.
    generalize n not_srt; clear n not_srt.
    inversion_clear H; intros.
    apply n.
    apply wf_var with s; trivial.

    apply not_srt.
    split with s; trivial.
  Defined.

  Theorem chk_wk_1 :
    forall e t T, ⊦ e -> wf_type e T -> decide (e ⊦ t : T).
  Proof.
    intros.
    TypeCheckTerm e t ipattern:j.
    elim ppa_le_type_dec with e (TYP j) T; trivial; intros.
    left.
    apply typ_conv_wf with (TYP j); trivial.

    right; red in |- *; intros.
    apply b.
    apply (pp_least (ppal_ty _ j)); trivial.

    red in b.
    eauto.
  Qed.


  Lemma chk_typ_1 : forall e t T, ⊦ e -> decide (e ⊦ t : T).
  Proof.
    (*Realizer [e,t,T]
      if (check_wf_type e T) then (typecheck_wft e t T)
      else false.
    *)
    intros e t T wfe.
    elim chk_wft_1 with e T; trivial.
    intros wf_T.
    apply chk_wk_1; trivial.

    intros nwf_T.
    right; red in |- *; intro jt.
    apply nwf_T.
    apply type_correctness with (1 := jt).
  Qed.


  Lemma chk_decl_1 : forall e d, ⊦ e -> decide (⊦ (e; d)).
  Proof.
    intros.
    elim inf_ppal_sort_1 with e x_ (TYPE d); intros; trivial.
    inversion_clear a.
    replace d with (mkdecl (NAME d) (TYPE d) (BODY d)).
    elim H0.
    case (BODY d); intros.
    fold (e; NAME d := t : ATYPE x) in |- *.
    elim chk_typ_1 with e t (ATYPE x); intros; trivial.
    left.
    apply wf_def with (AKIND x); trivial.

    right; red in |- *; intros.
    apply b.
    Inversion_typ H2; trivial.

    fold (e; NAME d : ATYPE x) in |- *.
    left.
    apply wf_var with (AKIND x); trivial.

    case d; trivial.

    right; red in |- *; intros.
    apply b.
    elim (inv_wf_sort _ _ H0); intros.
    apply wf_var with x; trivial.
  Qed.


  Theorem decide_wf : forall e, decide (⊦ e).
  Proof.
    simple induction e; intros; auto.
    elim H; intros.
    apply chk_decl_1; trivial.

    right; red in |- *; intro wfc; apply b; apply inv_wf with (1 := wfc).
  Qed.


  Theorem decide_type : forall e t T, decide (e ⊦ t : T).
  Proof.
    intros.
    elim decide_wf with e; intros.
    apply chk_typ_1; auto.

    right; red in |- *; intros.
    apply b.
    apply typ_wf with t T; auto.
  Qed.

  Instance IPTSO_TC : PTSO_TC := {
    inf_ppal_type := inf_ppal_type_1;
    inf_ppal_sort := inf_ppal_sort_1;
    chk_typ := chk_typ_1;
    chk_decl := chk_decl_1;
    chk_wk := chk_wk_1;
    chk_wft := chk_wft_1
  }.

End SPTSO_TC.

Existing Instance IPTSO_TC.
End IMach.
