(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Import MlPrint.
Require Export Infer.

(* 1- Elementary rules, fully explicit, with and without side conditions *)

Module Type IMachops
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes)
  (Import MMetatheory : IMetatheory P MTermes MEnv)
  (Import MInfer : IInfer P MTermes MEnv MMetatheory).

  Ltac Contrapose_silent _ :=
    right; red in |- *; intros ty jglob; Inversion_typ jglob. Ltac Contrapose H errmsg :=
  red in H; RightStr errmsg; red in |- *; intros ty jglob;
   Inversion_typ jglob.

  Ltac MakeAssum x j a :=
    elim (infer_assum _ x j);
     [ intros (a, equ_1, equ_2); elim equ_1; elim equ_2; clear equ_1 equ_2 j;
        generalize (push_wf _ a); intro
     | intro neg_wf; Contrapose neg_wf err_not_type; apply neg_wf;
        try prolog [ typ_wf wf_var wf_def ] 5 ].

  Ltac MakeConst x j c :=
    elim (infer_def _ x j);
     [ intros (c, equ_1, equ_2); elim equ_1; elim equ_2; clear equ_1 equ_2 j;
        generalize (push_const_wf _ c); intro
     | intro neg_wf; Contrapose neg_wf err_not_type; apply neg_wf;
        try prolog [ typ_wf wf_var wf_def ] 5 ].

  Ltac MakeDef j a :=
    elim (check_def a j);
     [ intro wfdef
     | intro neg_wf; Contrapose neg_wf err_cannot_coerce; apply neg_wf;
        try prolog [ typ_wf ] 5 ].

  Section SMachops.

    Context {MPTSO_spec : PTSO_spec}.
    Context {MPTSO_algos : PTSO_algos}.


  Section ElementaryRules.

    Variable e : env.
    Hypothesis wfe : ⊦ e.


    Section MachSort.

      Theorem infer_sort : forall s, infer_typ_ppal e %%s.
      (*
      Realizer [s:sort]
        Cases (ppa_infer_axiom the_algos s) of
          (inleft sp) => (inleft ju (mkjpp (Srt s) (Srt sp)))
        | _ => (inright ju)
        end.
      *)
      Proof.
        intros.
        elim (ppa_infer_axiom s); intros.
        left.
        inversion_clear a.
        apply ju_intro with (%%x).
        split; intros.
        apply Typ_srt; trivial.
        apply (pp_ok H).

        Inversion_typ H0.
        apply le_type_trans with (2 := H2).
        apply (pp_least H H1).

        right; red in |- *; intros.
        Inversion_typ H.
        apply b.
        split with s2; trivial.
      Qed.


      Lemma mk_sort : forall s, typed_sort s -> ju e %%s.
      Proof.
        (*Realizer [s]
          Cases (infer_sort s) of
            (inleft j) => j
          | inright => (False_rec ju)
          end.*)
        intros s tys.
        case (infer_sort s).
        trivial.

        intros; apply False_rec.
        inversion_clear tys.
        elim n with (%%x); auto.
      Defined.

    End MachSort.


    Section MachRef.

      (* regles de derivation de typ_ppal *)
      Lemma typ_ppal_ref :
        forall e n t, ⊦ e -> type_of_rel e n t -> typ_ppal e #n t.
      Proof.
        split; intros; auto.
        Inversion_typ H1.
        inversion_clear H0.
        rewrite H4.
        elim inj_item_rel with (1 := H2) (2 := H5); auto.
      Qed.


      Theorem infer_rel : forall n, infer_typ_ppal e #n.
      (*
      Realizer [n]
        Cases (list_item ? e n) of
          (inleft d) => (inleft ? (mkjpp (Rel n) (lift (S n) (TYPE d))))
        | _ => (inright ju)
        end.
      *)
      Proof.
        intro n.
        elim (list_item e n); intros.
        inversion_clear a.
        left.
        apply ju_intro with (^(S n)(TYPE x)).
        apply typ_ppal_ref; trivial.
        split with x; trivial.

        right; red in |- *; intros.
        Inversion_typ H.
        red in b.
        apply b with (1 := H0).
      Qed.


      Lemma mk_rel : forall n, n < length e -> ju e #n.
      Proof.
        (*Realizer [n]
          Cases (infer_rel n) of
            (inleft j) => j
          | inright => (False_rec ju)
          end.*)
        intros n bound_n.
        elim (infer_rel n); trivial.

        intros; apply False_rec.
        case item_dom with (1 := bound_n); intros.
        red in b.
        apply b with (^(S n)(TYPE x)).
        constructor; trivial.
        split with x; trivial.
      Defined.
    End MachRef.



    Section MachCst.


      Lemma mk_cst0 :
        forall c u, ⊦ (( ); x_ : u) -> in_sign0 c u -> ju e (Cst0 c).
      Proof.
        intros.
        apply ju_intro with u.
        split; intros.
        apply Typ_cst0; trivial.

        Inversion_typ H1.
        replace u with u0; trivial.
        apply ppal_dec_unique with (1 := ppa_infer_sign0 c); auto.
      Qed.


      Lemma infer_cst0 : forall c, infer_typ_ppal e (Cst0 c).
      Proof.
        (*Realizer
          [c]Cases (ppa_infer_sign0 the_algos c) of
            (inleft u) => (inleft ? (mk_cst0 c u))
          | inright => (inright ppal_judge)
          end.*)
        intros c.
        case (ppa_infer_sign0 c).
        intros (u, ty_u).
        left.
        apply mk_cst0 with u.
        elim (pp_ok ty_u); trivial.

        elim (pp_ok ty_u); trivial.

        right; red in |- *; intros.
        Inversion_typ H.
        elim n with u; auto.
      Defined.


      Lemma mk_cst1 :
        forall c (def : const e) u,
        let cval := VAL (CBODY def) in
        let ctyp := TYP (CBODY def) in
          ppal (fun u => in_sign_sat c e cval ctyp u) (le_type e) u ->
          ju e (Cst1 c cval).
      Proof.
        intros.
        apply ju_intro with u.
        split; intros.
        apply typ_cst1_sat with (TYP (CBODY def)); subst cval; trivial.
        apply (pp_ok H).

        Inversion_typ H0.
        apply le_type_trans with (2 := H5).
        apply (pp_least H).
        split with t; auto.
        split; trivial.
        apply (pp_least (ppal_ty _ (CBODY def))); trivial.
      Qed.


      Theorem infer_cst1 :
        forall c (j : ppal_judge e), infer_typ_ppal e (Cst1 c (VAL j)).
      Proof.
        intros.
        MakeConst x_ j ipattern:jc.
        elim ppa_infer_sign1 with e c (VAL (CBODY jc)) (TYP (CBODY jc));
         intros; trivial.
        inversion_clear a.
        left.
        apply mk_cst1 with x; trivial.

        right; red in |- *; intros.
        red in b.
        Inversion_typ H0.
        apply b with u.
        split with t; auto.
        split; trivial.
        apply (pp_least (ppal_ty _ (CBODY jc))); exact H4.

        Inversion_typ H0.
        apply wfd_ok with (3 := H4); trivial.
        apply (pp_least (ppal_ty _ j)); trivial.
      Qed.

    End MachCst.


    Section MachPrd.

      Lemma mk_prd :
        forall (a : assum e) (b : assum (push_assum _ a)) s3,
          ppal (fun s => rl_sat (AKIND a) (AKIND b) s) le_sort s3 ->
          ju e (Forall (ANAME a) : ATYPE a, ATYPE b).
      Proof.
        intros.
        apply ju_intro with (%%s3).
        split; intros.
        apply typ_prd with (AKIND a) (AKIND b); trivial.
        apply (pp_ok H).

        Inversion_typ H0.
        apply le_type_trans with (2 := H4).
        generalize e.
        fold (le_sort s3 s0) in |- *.
        apply (pp_least H).
        split with s1.
        apply (pp_least (ppal_kind _ a)); trivial.

        split with s2; trivial.
        apply (pp_least (ppal_kind _ b)); trivial.
      Qed.


      Theorem infer_prd :
        forall (a : assum e) (j : ppal_judge (push_assum _ a)),
          infer_typ_ppal e (Forall (ANAME a) : ATYPE a, VAL j).
      Proof.
        (*
        Realizer [a:assum; j:ppal_judge]
          Cases (infer_assum (ANAME a) j) of
            (inleft b) =>
               Cases (ppa_infer_rule the_algos (AKIND e a) (AKIND e b)) of
                 (inleft s3) => (inleft ? (mk_prd a b s3))
               | _ => (inright ppal_judge)
               end
          | _ => (inright ppal_judge)
          end.
        *)
        intros.
        MakeAssum x_ j ipattern:b.
        elim ppa_infer_rule with (AKIND a) (AKIND b); intros.
        inversion_clear a0.
        left.
        apply mk_prd with x; trivial.

        right; red in |- *; intros.
        Inversion_typ H0.
        red in b0.
        apply b0 with s3.
        split with s1.
        apply (pp_least (ppal_kind _ a)); trivial.

        split with s2; trivial.
        apply (pp_least (ppal_kind _ b)); trivial.
      Qed.

    End MachPrd.


    Section MachLam.

      Lemma mk_lam :
        forall (a : assum e) (j : ppal_judge (push_assum _ a)) b,
          ppal (weak_to_least_sort rl_sat (push_assum _ a) (AKIND a) (TYP j))
          (le_type (push_assum _ a)) b -> ju e (Fun (ANAME a) : ATYPE a => VAL j).
      Proof.
        intros.
        apply ju_intro with (Forall (ANAME a) : ATYPE a, b).
        split; intros.
        generalize (pp_ok H).
        intros (lety, (s2, tyb, (s3, rl))).
        apply Typ_lam with s3.
        apply Typ_conv with (TYP j) s2; trivial.
        apply typ_prd with (AKIND a) s2; auto.

        Inversion_typ H0.
        Inversion_typ H2.
        apply le_type_trans with (2 := H3).
        apply ppa_le_type_prod.
        apply (pp_least H).
        split.
        apply (pp_least (ppal_ty _ j)); trivial.

        split with s2; trivial.
        split with s3.
        split with s1.
        apply (pp_least (ppal_kind _ a)); trivial.

        split with s2; auto.
        intro; apply le_type_refl.
      Qed.


      Theorem infer_lam :
        forall (a : assum e) (j : ppal_judge (push_assum _ a)),
          infer_typ_ppal e (Fun (ANAME a) : ATYPE a => VAL j).
      Proof.
        (*
        Realizer [x:name; a:assum; j:judge]
          Cases (ppa_weak_rule the_algos (push_assum e x a) (AKIND e a) (TYP e j)) of
            (inleft b') => (inleft judge (mk_lam x a j b'))
          | inright => (inright judge)
          end.
        *)
        intros.
        MakeConst x_ j ipattern:jc.
        elim ppa_weak_rule with (push_assum _ a) (AKIND a) (TYP (CBODY jc));
         intros; trivial.
        inversion_clear a0.
        left.
        apply mk_lam with (1 := H0).

        right; red in |- *; intros.
        Inversion_typ H0.
        Inversion_typ H2.
        elim b with u.
        red in |- *.
        split.
        apply (pp_least (ppal_ty _ (CBODY jc))); trivial.

        split with s2; trivial.
        split with s3.
        split with s1.
        apply (pp_least (ppal_kind _ a)); trivial.

        split with s2; auto.
        intro; apply le_type_refl.

        apply ty_sorted.

        Inversion_typ H0.
        apply wfd_ok with (3 := H3); trivial.
        apply (pp_least (ppal_ty _ j)); trivial.
      Qed.


    End MachLam.


    Section MachSum.

      Lemma mk_sum : forall a b : assum e, ju e (ATYPE a ** ATYPE b).
      Proof.
        intros.
        elim ppa_infer_prod with (AKIND a) (AKIND b); intros.
        apply ju_intro with (%%x).
        split; intros.
        apply typ_sum with (AKIND a) (AKIND b); trivial.
        apply (pp_ok p).

        Inversion_typ H.
        apply le_type_trans with (2 := H3).
        generalize e.
        fold (le_sort x s3) in |- *.
        apply (pp_least p).
        split with s1.
        apply (pp_least (ppal_kind _ a)); trivial.

        split with s2; trivial.
        apply (pp_least (ppal_kind _ b)); trivial.
      Qed.



      Theorem infer_sum :
        forall j1 j2 : ppal_judge e, infer_typ_ppal e (VAL j1 ** VAL j2).
      Proof.
        intros.
        MakeAssum x_ j1 ipattern:ja1.
        MakeAssum x_ j2 ipattern:ja2.
        left.
        apply mk_sum.
      Qed.

    End MachSum.


    Section MachPair.

      Lemma mk_pair :
        forall c1 c2 : const e, ju e (VAL (CBODY c1),,VAL (CBODY c2)).
      Proof.
        intros.
        apply ju_intro with (TYP (CBODY c1) ** TYP (CBODY c2)).
        split; intros.
        elim (ty_sorted _ c1); intros.
        elim (ty_sorted _ c2); intros.
        elim ppa_infer_prod with x x0; intros.
        apply Typ_pair with x1; trivial.
        apply typ_sum with x x0; trivial.
        apply (pp_ok p).

        Inversion_typ H.
        apply le_type_trans with (2 := H3).
        apply ppa_le_type_sum.
        apply (pp_least (ppal_ty _ (CBODY c1))); trivial.

        apply (pp_least (ppal_ty _ (CBODY c2))); trivial.
      Qed.


      Theorem infer_pair :
        forall j1 j2 : ppal_judge e, infer_typ_ppal e (VAL j1,,VAL j2).
      Proof.
        intros.
        MakeConst x_ j1 ipattern:jc1.
        MakeConst x_ j2 ipattern:jc2.
        left.
        apply mk_pair.

        Inversion_typ H2.
        apply wfd_ok with b s2; trivial.
        apply (pp_least (ppal_ty _ j2)); trivial.

        Inversion_typ H1.
        apply wfd_ok with a s1; trivial.
        apply (pp_least (ppal_ty _ j1)); trivial.
      Qed.


    End MachPair.

  End ElementaryRules.

  End SMachops.
End IMachops.
