(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export Ctxt.
Require Export Confluence.

Module Type INormal
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes)
  (Import MRules : IRules P MTermes MEnv)
  (Import MCtxt : ICtxt P MTermes MEnv MRules).

  Include IConfluence P MTermes MEnv MRules.

  Section Normalization.

    (* Notions generales concernant la normalisation *)

    Variable R : red_rule.
    Variable e : env.

    Definition sn t := Acc (transp _ (R e)) t.

    Definition normal t := forall u : term, ~ R e t u.

    Lemma sn_red_sn : forall a b, R_rt R e a b -> sn a -> sn b.
    Proof.
      simple induction 1; intros; auto.
      red in |- *.
      apply Acc_inv with x; auto.
    Qed.

    Lemma red_normal : forall u v, R_rt R e u v -> normal u -> u = v.
    Proof.
      simple induction 1; intros; auto.
      absurd (R e x y); auto.

      generalize H3.
      elim H1; auto.
    Qed.

    Section NormalChurchRosser.

      Hypothesis CR : church_rosser R.

      Lemma conv_normal_red :
        forall u v, R_rst R e u v -> normal v -> R_rt R e u v.
      Proof.
        intros.
        elim CR with e u v; intros; auto.
        replace v with x; auto.
        symmetry  in |- *.
        apply red_normal; auto.
      Qed.

      Lemma nf_unicity :
        forall u v, R_rst R e u v -> normal u -> normal v -> u = v.
      Proof.
        intros.
        apply red_normal; auto.
        apply conv_normal_red; auto.
      Qed.

    End NormalChurchRosser.

  End Normalization.

  Section HeadNormal.

    Variable R : red_rule.

    Definition head_normal e t := forall u, red R e t u -> normal R e u.

    Lemma hn_red_hn :
      forall e t u, head_normal e t -> red R e t u -> head_normal e u.
    Proof.
      unfold head_normal in |- *.
      intros.
      apply H.
      apply red_trans with u; auto.
    Qed.

  End HeadNormal.

  Section NormalTerms.

    Variable R1 R2 : red_rule.

    Lemma normal_union : forall e x,
      normal R1 e x -> normal R2 e x ->
      normal (fun e => union _ (R1 e) (R2 e)) e x.
    Proof.
      red in |- *; red in |- *; intros.
      inversion_clear H1.
      elim H with u; auto.

      elim H0 with u; auto.
    Qed.

  End NormalTerms.

  Lemma sn_R_Rt : forall R e t, sn R e t -> sn (R_t R) e t.
  Proof.
    induction 1 as [ t _ Ht ].
    constructor; intros u Hu.
    apply trans_t1n in Hu.
    induction Hu as [ | t t' u ].
    apply Ht; auto.
    apply Acc_inv with t'.
    apply Ht; auto.
    apply t1n_trans in Hu; auto.
  Qed.

  Lemma sn_Rt_R : forall R e t, sn (R_t R) e t -> sn R e t.
  Proof.
    induction 1 as [ t _ Ht ].
    constructor; intros u Hu.
    apply Ht; hnf.
    constructor; auto.
  Qed.

End INormal.
