(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export Normal.

Module Type IOrder
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes)
  (Import MRules : IRules P MTermes MEnv)
  (Import MCtxt : ICtxt P MTermes MEnv MRules)
  (Import MNormal : INormal P MTermes MEnv MRules MCtxt).


Section NormalizationOrder.

  Variable R R' : red_rule.

  Hypothesis stable : R_stable (rel_typ R') R.

  Definition value := (env * term)%type.


  Inductive subterm : value -> value -> Prop :=
    | Sbt_nobind : forall e M N, subt_nobind M N -> subterm (e, M) (e, N)
    | Sbt_bind :
        forall e d d' M N,
        subt_bind M N d -> rel_typ R' e d d' -> subterm (d' :: e, M) (e, N).

  Hint Resolve Sbt_nobind Sbt_bind: core.


  Lemma wf_subterm : well_founded subterm.
Proof.
red in |- *.
intros (e, t).
generalize e.
elim t; intros; apply Acc_intro; intros.
inversion_clear H; inversion_clear H0.

inversion_clear H; inversion_clear H0.

inversion_clear H; inversion_clear H0.

inversion_clear H0; inversion_clear H1; trivial.

inversion_clear H1; inversion_clear H2; trivial.

inversion_clear H1; inversion_clear H2; trivial.

inversion_clear H1; inversion_clear H2; trivial.

inversion_clear H1; inversion_clear H2; trivial.
Qed.


  Inductive R_exp : value -> value -> Prop :=
      Rex_intro : forall e x y, ctxt R e x y -> R_exp (e, y) (e, x).

  Hint Resolve Rex_intro: core.

  Lemma acc_R_exp_sn : forall e t, Acc R_exp (e, t) -> sn (ctxt R) e t.
Proof.
intros.
replace e with (fst (e, t)); auto.
pattern t at 2 in |- *.
replace t with (snd (e, t)); auto.
elim H; intros.
constructor; intros.
generalize (H1 (fst x, y)).
simpl in |- *; intros.
apply H3.
generalize H2.
case x; intros; auto.
Qed.



  Lemma commut_subterm_R_exp : commut value subterm R_exp.
Proof.
red in |- *.
simple destruct 1.
simple destruct 1; intros; inversion_clear H1.
split with (e, Cst1 c y0); auto.

split with (e, Fun x0 : y0 => n); auto.

split with (e, Forall x0 : y0, n); auto.

split with (e, y0 ** n); auto.

split with (e, n ** y0); auto.

split with (e, y0,,n); auto.

split with (e, n,,y0); auto.

simple destruct 1; intros; inversion_clear H2.
split with (e, Fun x0 : t => y0); auto.
constructor.
apply Ctx_lam_r.
apply (ctxt_stable _ _ stable) with (1 := H3).
unfold add_var in |- *.
constructor; trivial.

constructor 2 with (mkdecl x0 t None); auto.

split with (e, Forall x0 : t, y0); auto.
constructor.
apply Ctx_prd_r.
apply (ctxt_stable _ _ stable) with (1 := H3).
unfold add_var in |- *.
constructor; trivial.

constructor 2 with (mkdecl x0 t None); auto.
Qed.


  Lemma subterm_sn :
   forall e t,
   sn (ctxt R) e t -> forall f u, subterm (f, u) (e, t) -> sn (ctxt R) f u.
Proof.
unfold sn, transp in |- *.
simple induction 1; intros.
constructor; intros.
elim commut_subterm_R_exp with (e, x) (f, u) (f, y); intros; auto.
inversion_clear H4 in H5.
apply H1 with y0; auto.
Qed.



  Definition ord_norm1 := union _ subterm R_exp.
  Definition ord_norm := clos_trans _ ord_norm1.

  Hint Resolve Rex_intro: core.
  Hint Unfold ord_norm1: core.

  Lemma ord_norm_sub :
   forall (e f : env) (x y z : term),
   red R e x y -> subterm (f, z) (e, y) -> ord_norm (f, z) (e, x).
red in |- *; intros.
apply clos_trans_step with (e, y); auto 10.
elim H; intros; auto 10.
eapply rt_trans; eauto.
Qed.

  Lemma ord_norm_step :
   forall (e : env) (x y z : term),
   red R e x y -> ctxt R e y z -> ord_norm (e, z) (e, x).
red in |- *; intros.
apply clos_trans_step with (e, y); auto 10.
elim H; intros; auto 10.
eapply rt_trans; eauto.
Qed.

(* unused... *)
  Lemma ord_norm_intro :
   forall e f x y z,
   subterm (f, y) (e, x) -> red R f y z -> ord_norm (f, z) (e, x).
Proof.
red in |- *; intros.
apply step_clos_trans with (f, y); auto 10.
elim H0; intros; auto 10.

apply rt_trans with (f, y0); trivial.
Qed.


  Lemma ord_norm_sn : forall e t, Acc ord_norm (e, t) -> sn (ctxt R) e t.
intros.
apply acc_R_exp_sn.
apply Acc_incl with ord_norm; trivial.
red in |- *; red in |- *; auto.
Qed.

  Lemma sn_acc_ord_norm1 :
   forall e t, sn (ctxt R) e t -> Acc ord_norm1 (e, t).
unfold ord_norm1 in |- *.
intros.
apply Acc_union2; intros.
exact commut_subterm_R_exp.

elim H; intros.
constructor; intros.
inversion_clear H2; auto.

apply wf_subterm.
Qed.

  Lemma sn_ord_norm :
    forall e t, sn (ctxt R) e t -> Acc (A:=env*term) ord_norm (e, t). 
intros e t sn_t.
exact (Acc_clos_trans _ _ _ (sn_acc_ord_norm1 _ _ sn_t)).
Qed.


Section RecursionSchemes.

  Variable P : env -> term -> Set.
  Hypothesis
    hnf :
      forall e x,
      sn (ctxt R) e x -> {y : _ | red R e x y &  head_normal R e y}.

  Lemma whnf_induc :
   (forall e x y, red R e x y -> P e y -> P e x) ->
   (forall e x,
    head_normal R e x ->
    (forall f y, subterm (f, y) (e, x) -> P f y) -> P e x) ->
   forall e t, sn (ctxt R) e t -> P e t.
Proof.
intros.
elim (sn_ord_norm _ _ H1) using Acc2_rec.
clear H1 e t.
intros e t sn_t rec_fun.
elim hnf with e t; intros; trivial.
apply H with x; trivial.
apply H0; trivial; intros.
apply (rec_fun f y).
red in |- *; apply clos_trans_step with (e, x); auto.
elim p; intros; auto.
apply rt_step.
right; auto.

apply rt_trans with (e, y0); trivial.

apply ord_norm_sn; trivial.
Qed.

End RecursionSchemes.


  Lemma red_rt_ord :
   forall e f t u rt ru,
   red R e t rt ->
   red R f u ru ->
   clos_refl_trans _ (swapprod _ ord_norm1) (e, rt, (f, ru)) (e, t, (f, u)).
intros.
apply rt_trans with (e, rt, (f, u)).
elim H0; intros; auto 10.
apply rt_trans with (e, rt, (f, y)); auto.

elim H; intros; auto 10.
apply rt_trans with (e, y, (f, u)); auto.
Qed.


End NormalizationOrder.

  Hint Resolve Rex_intro Sbt_nobind Sbt_bind: core.
  Hint Unfold ord_norm1 ord_norm: core.
  Hint Resolve ord_norm_sub ord_norm_step ord_norm_sn: core.

(* ord_conv.*)

(*
Section SubtypingOrder.


  Definition ord_conv: (relation value*value) :=
                   (clos_trans value*value (swapprod value ord_norm1)).

  Lemma Acc_conv_sn_left: (a,b:value)(Acc ? ord_conv (a,b))->(Acc ? R_exp a).
Intros.
(Replace a with (Fst (a,b)); Auto).
(Elim H; Intros).
(Apply Acc_intro; Intros; Auto).
(Replace y with (Fst (y,(Snd x))); Auto).
Apply H1.
Red.
Generalize H2 .
(Case x; Simpl; Intros; Auto 10).
Save.

  Lemma Acc_conv_sn_right: (a,b:value)(Acc ? ord_conv (a,b))->(Acc ? R_exp b).
Intros.
(Replace b with (Snd (a,b)); Auto).
(Elim H; Intros).
(Apply Acc_intro; Intros; Auto).
(Replace y with (Snd ((Fst x),y)); Auto).
Apply H1.
Red.
Generalize H2 .
(Case x; Simpl; Intros; Auto 10).
Save.



  Lemma ord_cv_no_swap: (e,f:env)(u,ru:term)(red R e ru u)
            ->(v,rv:term)(red R f rv v)
              ->(su,sv:value)(ord_norm1 (e,ru) su)->(ord_norm1 (f,rv) sv)
                ->(ord_conv ((e,u),(f,v)) (su,sv)).
(Red; Intros).
Apply t_trans with (su,(f,v)).
(Elim H using clos_refl_trans_ind_left; Intros; Auto).
(Apply t_trans with ((e,P0),(f,v)); Auto 10).

(Elim H0 using clos_refl_trans_ind_left; Intros; Auto).
(Apply t_trans with (su,(f,P0)); Auto 10).
Save.


  Lemma ord_cv_swap: (e:env)(u,ru:term)(red R e ru u)
            ->(v,rv:term)(red R e rv v)
              ->(su,sv:value)(ord_norm1 (e,ru) su)->(ord_norm1 (e,rv) sv)
                ->(ord_conv ((e,u),(e,v)) (sv,su)).
Red;Intros.
Apply t_trans with (((e,u)),sv).
(Elim H0 using clos_refl_trans_ind_left; Intros; Auto).
(Apply t_trans with (((e,u)),((e,P0))); Auto 10).

(Elim H using clos_refl_trans_ind_left; Intros; Auto).
(Apply t_trans with (((e,P0)),sv); Auto 10).
Save.

End SubtypingOrder.
*)

End IOrder.
