(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Import Termes.

Module Type IParSubst
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P).

(* Parallel substitution *)

  Fixpoint psubst_rel (n : nat) (ts : list term) {struct ts} : term :=
    match ts, n with
    | nil, _ => #n
    | t :: _, O => t
    | _ :: ts', S k => psubst_rel k ts'
    end.

  Fixpoint psubst_rec (ts : list term) (t : term) (k : nat) {struct t} :
   term :=
    match t with
    | %%s => %%s
    | #i =>
        match le_gt_dec k i with
        | left _ => ^(k)(psubst_rel (i - k) ts)
        | right _ => #i
        end
    | Cst0 c => Cst0 c
    | Cst1 c a => Cst1 c (psubst_rec ts a k)
    | Fun x : a => m => Fun x : psubst_rec ts a k => psubst_rec ts m (S k)
    | Forall x : a, b => Forall x : psubst_rec ts a k, psubst_rec ts b (S k)
    | a ** b => psubst_rec ts a k ** psubst_rec ts b k
    | (u ,,v) => psubst_rec ts u k,,psubst_rec ts v k
    end.

  Lemma psubst_id : forall t k, psubst_rec nil t k = t.
simple induction t; simpl in |- *; intros; auto.
elim (le_gt_dec k n); intros.
unfold lift in |- *; simpl in |- *.
replace (k + (n - k)) with n; auto with arith.

trivial.

rewrite H; trivial.

rewrite H; rewrite H0; trivial.

rewrite H; rewrite H0; trivial.

rewrite H; rewrite H0; trivial.

rewrite H; rewrite H0; trivial.
Qed.


  Lemma psubst_cons :
   forall m ts t k, psubst_rec (m :: ts) t k = (psubst_rec ts t (S k))[k:=m].
simple induction t; intros.
reflexivity.

unfold psubst_rec in |- *.
case (le_gt_dec k n); case (le_gt_dec (S k) n); intros.
rewrite simpl_subst.
replace (n - k) with (S (n - S k)).
simpl in |- *.
trivial.

destruct n.
inversion l.

rewrite minus_Sn_m; simpl in |- *; auto with arith.

auto.

assert (k = n); try omega.
rewrite H.
replace (n - n) with 0; simpl in |- *; auto.
case (lt_eq_lt_dec n n); intros.
elim s; intros.
apply False_ind; omega.

trivial.

apply False_ind; omega.

auto with arith.

apply False_ind; omega.

simpl in |- *.
elim (lt_eq_lt_dec k n); intros.
apply False_ind.
elim a; intros; omega.

trivial.

simpl in |- *; trivial.

simpl in |- *; rewrite H; trivial.

simpl in |- *; rewrite H; rewrite H0; trivial.

simpl in |- *; rewrite H; rewrite H0; trivial.

simpl in |- *; rewrite H; rewrite H0; trivial.

simpl in |- *; rewrite H; rewrite H0; trivial.
Qed.

  Lemma psubst_commut_rec :
   forall m ts a k i,
   (psubst_rec ts m (S (k + i)))[k:=a] =
   psubst_rec ts m[k:=^(length ts//i)a] (k + i).
simple induction ts; simpl in |- *; intros.
repeat rewrite psubst_id; trivial.
rewrite lift_rec0; trivial.

repeat rewrite psubst_cons.
pattern a0 at 1 in |- *; replace a0 with (^(1//i)a0[i:=a]).
rewrite <- distr_subst_rec.
replace (k + 0) with k; auto with arith.
replace (S (k + i)) with (k + S i); auto with arith.
rewrite H.
rewrite simpl_lift_rec; auto with arith.
replace (length l + 1) with (S (length l)); trivial.
omega.

omega.

rewrite simpl_subst_rec; try omega.
rewrite lift_rec0; trivial.
Qed.

  Lemma psubst_commut :
   forall m ts a k,
   (psubst_rec ts m (S k))[k:=a] = psubst_rec ts m[k:=^(length ts)a] k.
intros.
generalize (psubst_commut_rec m ts a k 0).
replace (k + 0) with k; auto with arith.
Qed.


  Lemma psubst_app :
   forall ts1 ts2 t k,
   psubst_rec (ts1 ++ ts2) t k =
   psubst_rec ts2 (psubst_rec (map (fun t => ^(length ts2)t) ts1) t k) k.
simple induction ts1; simpl in |- *; intros.
rewrite psubst_id; trivial.

repeat rewrite psubst_cons.
rewrite <- psubst_commut.
rewrite H; trivial.
Qed.

End IParSubst.
