(**************************************************************************)
(*               Certified Pure Type Systems with Operators               *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL free software license, version 2.   *)
(**************************************************************************)

Require Export ParCtxt.
Require Export Metatheory.

Module Type ISoundness
  (Import P : PTSPARAMS)
  (Import MTermes : ITermes P)
  (Import MEnv : IEnv P MTermes)
  (Import MMetatheory : IMetatheory P MTermes MEnv)
  (Import MRules: IRules P MTermes MEnv)
  (Import MCtxt : ICtxt P MTermes MEnv MRules)
  (Import MParCtxt : IParCtxt P MTermes MEnv MRules MCtxt).


Open Scope pts_scope.


(* Alternative (un peu plus simple, mais ne convient pas pour HNF):
    la fermeture au contexte parallele (i.e. comme dans TypCtxt)
   (revision 1.6) *)

Section ISoundness.

  Context {the_PTS : PTSO_spec}.

  Section SoundnessDef.

    Variable R : red_rule.

    Definition rule_sound :=
      forall e x y, R e x y -> forall T, e ⊦ x : T -> e ⊦ y : T.

  End SoundnessDef.

  Section OperSound.

    Variable P : red_rule.
    Hypothesis sound : rule_sound P.

    Theorem reunion_sound :
      forall Q, rule_sound Q -> rule_sound (reunion P Q).
    Proof.
      unfold rule_sound, reunion, union in |- *.
      simple destruct 2; intros.
      apply sound with x; auto.

      apply H with x; auto.
    Qed.


    Theorem incl_sound :
      forall Q : red_rule, (forall e x y, Q e x y -> P e x y) -> rule_sound Q.
    Proof.
      red in |- *.
      eauto.
    Qed.



    Theorem clos_trans_sound : rule_sound (fun e => clos_trans term (P e)).
    Proof.
      red in |- *.
      simple induction 1; intros; auto.
      apply sound with x0; auto.
    Qed.


    Theorem clos_refl_trans_sound : rule_sound (R_rt P).
    Proof.
      red in |- *.
      simple induction 1; intros; auto.
      apply sound with x0; auto.
    Qed.


    (*
      Inductive rel_option [A:Set;R:A->A->Prop]: (option A)->(option A)->Prop :=
        Rel_op_some: (x,y:A)(R x y)->(rel_option A R (Some x) (Some y))
      | Rel_op_none: (rel_option A R (None ?) (None ?)).

      Hints Resolve Rel_op_some Rel_op_none : core.
    *)

    Inductive red_stab (R : red_rule) (e : env) : decl -> decl -> Prop :=
        Red_stb_ty :
          forall x t1 t2 b,
          R e t2 t1 -> red_stab R e (mkdecl x t1 b) (mkdecl x t2 b).
    (*  | Red_stb_bd: (x:name)(t,b1,b2:term)(R e b2 b1)
              ->(red_stab R e (mkdecl x t (Some b1)) (mkdecl x t (Some b2)))*)

    Hint Resolve Red_stb_ty (*Red_stb_bd*) : core.

    Lemma rel_list_par :
     forall R e f,
     rel_in_list (red_stab (par_ctxt R)) f e ->
     clos_refl_trans _ (rel_in_list (red_stab (ctxt R))) f e.
    Proof.
      simple induction 1; intros.
      inversion_clear H0.
      elim (incl_par_red _ _ _ _ H1); intros; auto.
      apply rt_trans with (mkdecl x0 y0 b :: l); auto.

      elim H1; intros; auto.
      apply rt_trans with (x :: y); auto.
    Qed.


(*
  Definition stab_deft: env->decl->decl->Prop :=
    [e; d1,d2](EX t | (BODY d1)=(Some t) & (BODY d2)=(Some t)).

    Std_def: (d,t,u:term)(stab_deft e ((Def d),t) ((Def d),u)).

  Hints Resolve Std_def : core.
*)


(* --> lib
    Lemma rel_in_list_sym: (R:env->decl->decl->Prop)(e,f:env)
       (rel_in_list [e;x,y](R e y x) f e)->(rel_in_list R e f).
Proof.
(Induction 1; Intros; Auto).
Save.
*)

    Section ParCtxtSound.

      Variable R : red_rule.


      (* Reduction must be included in the symmetric part of subtyping *)
      Hypothesis incl_R_le_type : forall e x y, par_ctxt R e x y -> e ⊦ y <= x.

      Hypothesis incl_R_ge_type : forall e x y, par_ctxt R e x y -> e ⊦ x <= y.


      Hypothesis
        in_sign1_stable_val :
          forall c e m m' t u,
          par_ctxt R e m m' ->
          in_sign1 c e m t u ->
          exists2 t' : _,
            par_ctxt R e t t' &
            (exists2 u' : _, par_ctxt R e u u' & in_sign1 c e m' t' u').



      Lemma in_sign1_stable :
        forall c f e m t u,
          rel_in_list (red_stab (par_ctxt R)) f e ->
          in_sign1 c e m t u -> in_sign1 c f m t u.
      Proof.
        intros.
        apply in_sign_stable with (2 := H0).
        elim H; intros; auto.
        constructor.
        inversion_clear H1.
        split; simpl in |- *; auto.
        simple apply incl_R_le_type; exact H2.
      Qed.

      (* Subtyping cannot be affected by reduction in the context *)
      Lemma le_type_pstable_red : R_stable (red_stab (par_ctxt R)) le_type.
      Proof.
        apply stable_incl with (rel_typ le_type).
        exact le_type_stable.

        red in |- *; red in |- *; simple destruct 1; simpl in |- *; auto.
      Qed.

      Let red_env :
        forall e f t T,
          e ⊦ t : T ->
            rel_in_list (red_stab (par_ctxt R)) f e -> ⊦ f -> f ⊦ t : T.
      Proof.
        intros.
        apply reduction_env_rec with (red_stab (par_ctxt R)) e; auto.
        simple destruct 1; auto.

        exact in_sign1_stable.

        exact le_type_pstable_red.
      Qed.

      Let red_env_last :
        forall (e : env) d1 d2 t T,
          (e; d1) ⊦ t : T ->
            red_stab (par_ctxt R) e d2 d1 -> ⊦ (e; d2) -> (e; d2) ⊦ t : T.
      Proof.
        intros.
        apply red_env with (d1 :: e); auto.
      Qed.


      Hypothesis R_sound : rule_sound R.



      Ltac InverseCtxt H :=
        inversion_clear H;
          try
            prolog
              [ Typ_srt Typ_rel Typ_cst0 Typ_cst1 Typ_lam Typ_prd Typ_sum Typ_pair ]
              5.


      Theorem par_ctxt_sound : rule_sound (par_ctxt R).
      Proof.
        red in |- *; intros e m m' ctxm t j.
        generalize m' ctxm.
        clear ctxm m'.
        refine (@typ_mind the_PTS (fun e m t => forall m', par_ctxt R e m m' -> e ⊦ m' : t)
          (fun e => forall f, rel_in_list (red_stab (par_ctxt R)) f e -> ⊦ f)
          _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ _ j); intros.
        (* Was originally:
             elim j using
             typ_mind
               with
               (P0 := fun e => forall f, rel_in_list (red_stab (par_ctxt R)) f e -> |- f);
             intros.
           but doesn't work anymore.
           I had to change a lot of names in the rest of the proof... *)
        InverseCtxt H2.

        InverseCtxt H2.

        InverseCtxt H4.

        InverseCtxt H6.
        elim in_sign1_stable_val with (1 := H7) (2 := H5); intros.
        inversion_clear H8.
        apply typ_conv_wf with x0; auto.
        apply typ_cst1 with x; auto.
        cut (⊦ (e0; x_ : x)).
        intro.
        Inversion_typ H8.
        apply wf_def with s; auto.
        apply Typ_conv with t0 s; auto.

        apply H2.
        unfold add_var in |- *; constructor.
        auto.

        apply H0.
        unfold add_var in |- *; constructor; auto.

        apply var_wft.
        trivial.

        auto.

        InverseCtxt H3.
        apply Typ_conv with (Forall x : m'0, u) s;
         auto.
        cut (e0 ⊦ Forall x : m'0, u : %%s); intros; eauto.
        apply Typ_lam with s; trivial.
        unfold add_var in |- *.
        apply red_env_last with (mkdecl x a None); auto.
        fold (e0; x : m'0) in |- *.
        Inversion_typ H3.
        apply wf_var with s1; trivial.

        InverseCtxt H4.
        apply Typ_prd with s1 s2; eauto.
        unfold add_var in |- *.
        apply red_env_last with (mkdecl x a None); auto.
        fold (e0; x : m'0) in |- *.
        apply wf_var with s1; auto.

        InverseCtxt H4.

        InverseCtxt H5.

        apply Typ_conv with U s; auto.

        apply Typ_conv_srt with U; auto.

        inversion_clear H.

        unfold add_var in H3.
        inversion_clear H3.
        inversion_clear H4.
        fold (e0; x : t1) in |- *.
        apply wf_var with s; auto.

        fold (l1; x : T) in |- *.
        apply wf_var with s.
        apply red_env with e0; auto.

        unfold add_def in H5.
        inversion_clear H5.
        inversion_clear H6.
        fold (e0; x := M : t1) in |- *.
        apply wf_def with s; auto.
        apply Typ_conv with T s; auto.

        fold (l1; x := M : T) in |- *.
        apply wf_def with s.
        apply red_env with e0; auto.

        apply red_env with e0; auto.
      Qed.

    End ParCtxtSound.


    Section CtxtSound.

      Variable R : red_rule.

      Hypothesis incl_R_le_type : forall e x y, ctxt R e x y -> e ⊦ y <= x.

      Hypothesis incl_R_ge_type : forall e x y, ctxt R e x y -> e ⊦ x <= y.

      Hypothesis in_sign1_stable_val :
        forall c e m m' t u,
          par_ctxt R e m m' ->
          in_sign1 c e m t u ->
          exists2 t' : _,
          par_ctxt R e t t' &
          (exists2 u' : _, par_ctxt R e u u' & in_sign1 c e m' t' u').

      Hypothesis R_sound : rule_sound R.



      Let incl_pR_le_type : forall e x y, par_ctxt R e x y -> e ⊦ y <= x.
      Proof.
        intros.
        elim incl_par_red with (1 := H); intros; auto.
        apply le_type_refl.
        apply le_type_trans with y0; trivial.
      Qed.

      Let incl_pR_ge_type : forall e x y, par_ctxt R e x y -> e ⊦ x <= y.
      Proof.
        intros.
        elim incl_par_red with (1 := H); intros; auto.
        apply le_type_trans with y0; trivial.
      Qed.


      Let rs_par_ctxt : rule_sound (par_ctxt R).
      Proof
      par_ctxt_sound _ incl_pR_le_type incl_pR_ge_type in_sign1_stable_val R_sound.


      Theorem ctxt_sound : rule_sound (ctxt R).
      Proof.
        red in |- *; intros.
        apply rs_par_ctxt with (2 := H0).
        auto.
      Qed.

      Theorem wctxt_sound : rule_sound (wctxt R).
      Proof.
        red in |- *; intros.
        apply rs_par_ctxt with (2 := H0).
        auto.
      Qed.

    End CtxtSound.

  End OperSound.

End ISoundness.

Ltac InverseCtxt H :=
  inversion_clear H;
    try
      prolog
        [ Typ_srt Typ_rel Typ_cst0 Typ_cst1 Typ_lam Typ_prd Typ_sum Typ_pair ]
        5.

Hint Resolve @Red_stb_ty (*Red_stb_bd*) : core.

End ISoundness.
