(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

(******************** Naturels ********************)

(* plus simple que celle du prelude, et marche aussi (??) *)
(* Difference seulement visible en cas d'elim dependente *)

Inductive nat: Set := O: nat | S: (!nat _ ~)->nat.

(* fonction idiote repondant toujours 0 *)

Definition test: nat->nat :=
  (Fix {X, F | (nat-X)->nat :> nat->nat :=
          [n:nat+X]<[a:Lmark][_:(!nat ~ a)]nat>Cases n of
          | O => [_:Lmark]O
          | S => [pn:(nat-X)*Lmark](F^0 ~ pn^0)
          end}^0 ~).

Print test.
Extraction test.

(* plus usuel *)

Definition plus: nat -> nat -> nat :=
  (Fix {X, F | (nat-X)->nat->nat :> nat->nat->nat :=
       [n:nat+X]<[a:Lmark][_:(!nat ~ a)]nat->nat>Cases n of
          | O => [_:Lmark][m:nat]m
          | S => [pn:(nat-X)*Lmark][m:nat](S (F^0 ~ pn^0 m))
          end}^0 ~).

Print test.
Extraction test.

Definition deuxplusdeux : nat := (plus (S (S O)) (S (S O))).
Definition quatre : nat := (S (S (S (S O)))).

Print quatre.
Extraction quatre.

(* ackermann *)

Definition ack: nat -> nat -> nat :=
  (Fix {X, F | (nat-X)->nat->nat :> nat->nat->nat :=
    [n:nat+X]<[a:Lmark][_:(!nat ~ a)]nat->nat>Cases n of
      | O => [_:Lmark][m:nat](S m)
      | S => [pn:(nat-X)*Lmark]
             (Fix {Y, G | (nat-Y)->nat :> nat->nat :=
               [m:nat+Y]<[a:Lmark][_:(!nat ~ a)]nat>Cases m of
                 | O => [_:Lmark](F^0 ~ pn^0 (S O))
                 | S => [pm:(nat-Y)*Lmark](F^0 ~ pn^0 (G^0 ~ pm^0))
               end}^0 ~)
    end}^0 ~).

Print ack.
Extraction ack.

(******************** Listes ********************)

Inductive list[A:Set]: Set :=
    nil: (list A)
  | cons: A -> (list A) -> (list A).

Definition length: (A:Set)(list A)->nat := [A:Set]
  (Fix {X, F | ((list A)-X)->nat :> (list A)->nat :=
    [l:(list A)+X]<[a:Lmark][_:(!list A a)]nat>Cases l of
      | nil => [_:Lmark]O
      | cons => [l:A*((list A)-X)*Lmark](F^0 ~ l^1)
    end}^0 ~).

Definition map: (A:Set)(B:Set)(A->B)->(list A)->(list B) :=
  [A:Set][B:Set][f:A->B]
  (Fix {X, F | ((list A)-X)->(list B) :> (list A)->(list B) :=
    [l:(list A)+X]<[a:Lmark][_:(!list A a)](list B)>Cases l of
      | nil => [_:Lmark](nil B)
      | cons => [l:A*((list A)-X)*Lmark](cons B (f l^0) (F^0 ~ l^1))
    end}^0 ~).

Definition fold_left: (A:Set)(B:Set)(A->B->A)->A->(list B)->A :=
  [A:Set][B:Set][f:(A->B->A)][a0:A]
  (Fix {X, F | [a0:A] ((list B)-X)->A :> (list B)->A :=
    [l:(list B)+X]<[a:Lmark][_:(!list B a)]A>Cases l of
      | nil => [_:Lmark]a0
      | cons => [l:B*((list B)-X)*Lmark]a0
    end}^0 a0).

(*
Definition F [X:Mark]: ((list nat)-X)->((list nat)-X) := (map nat nat ([n:nat](S n))).

Definition G5: (list nat)->(list nat) :=
  (Fix {X, H | ((list nat)-X)->(list nat) :> (list nat)->(list nat) :=
    [l:(list nat)+X]<[a:Lmark][_:(!list nat a)](list nat)>Cases l of
      | nil => [_:Lmark](nil nat)
      | cons => [l:nat*((list nat)-X)*Lmark](cons nat (l^0) (H^0 ~ (F ~ l^1)))
    end}^0 ~).
*)

Inductive tree4: Set := Node4: (!tree4 _ ~) -> tree4.
Inductive tree6: Set := Node6: (list (!tree6 _ ~)) -> tree6.


Definition size1: (A:Set)(tree A)->nat := [A:Set]
  (Fix {X, F | ((tree A)-X)->nat :> (tree A)->nat :=
    [t:(tree A)+X]<[a:Lmark][_:(!tree A a)]nat>Cases t of
      | Node => [t:((tree A)-X)*Lmark]O
    end}^0 ~).

Fixpoint size t := match t with Node l => S (fold_left plus (map size (t::nil)) O) end.


(******************** Logique ********************)

(*
# convert !_ENV Nil
      (Cst1 (Const "deuxplusdeux", Cst0 Nil_mark))
      (Cst1 (Const "quatre", Cst0 Nil_mark));;
- : Kernel.decide = true
*)

(* False *)

Inductive False: Prop := .

Definition False_ind: (P:Prop)False->P :=
  [P:Prop][f:False]
    <[p:Lmark][_:(!False ~ p)]P>Cases f of end.

(* False_rec: pas typable actuellement ? *)

Definition False_rec: (P:Set)False->P :=
  [P:Set][f:False]
    <[p:Lmark][_:(!False ~ p)]P>Cases f of end.

(* eq *)

Definition ne_set: Type(0) := {< A: Set; x: A >}.
Definition mk_set: (A:Set)A->ne_set :=
  [A:Set][x:A] { A: Set := A; x: A :=x }.

Inductive eq [p:ne_set]: p^A->Prop :=
  refl_eq: (eq p p^x).

Definition eq1: (A:Set)A->A->Prop := [A:Set][x,y:A](eq (mk_set A x) y).

Definition refl_equal: (A:Set)(x:A)(eq1 A x x) :=
  [A:Set][x:A](refl_eq (mk_set A x)).

Definition eq_ind: (A:Set)(x:A)(P:A->Prop)(P x)->(y:A)(eq1 A x y)->(P y) :=
  [A:Set][x:A][P:A->Prop][f:(P x)][y:A][e:(eq1 A x y)]
    <[y':A*Lmark][_:(!eq (mk_set A x) y')](P y'^0)>Cases e of
      refl_eq => [_:Lmark]f end.

(* eq_rec : pas typable actuellement ? *)

Definition eq_rec: (A:Set)(x:A)(P:A->Set)(P x)->(y:A)(eq1 A x y)->(P y) :=
  [A:Set][x:A][P:A->Set][f:(P x)][y:A][e:(eq1 A x y)]
    <[y':A*Lmark][_:(!eq (mk_set A x) y')](P y'^0)>Cases e of
      refl_eq => [_:Lmark]f end.

(* Acc *)

Definition related_set : Type(0) := {< A:Set; R:A->A->Prop >}.

Inductive Acc [AR:related_set]: AR^A -> Prop :=
  Acc_intro : (x:AR^A)((y:AR^A)(AR^R y x)->(Acc AR y)) -> (Acc AR x).

(* Erreur: bad constructor type *)

(******************** acc_iter ********************)

Definition related_set : Type(0) := {< A:Set; R:A->A->Prop >}.

(* Version petit bout par petit bout *)

Inductive Acc [AR:related_set]: AR^A -> Prop / Set :=
  Acc_intro : {<x:AR^A;Hx:(y:AR^A)(AR^R y x) -> (Acc AR y)>} -> (Acc
  AR #0^x).

Axiom AR:related_set.

Definition Acc_inv [M:Mark] :
  (x:AR^A)((Acc AR x)+M)->(y:AR^A)(AR^R y x)->((Acc AR y)-M) :=
  [x:AR^A][a:(Acc AR x)+M]
  <[x':AR^A*Lmark][_:(!Acc AR x')](y:AR^A)(AR^R y x'^^1)->((Acc AR y)-M)>
  Cases a of Acc_intro =>
   [arg:{<x:AR^A;Hx:(y:AR^A)(AR^R y x)->((Acc AR y)-M)>}](arg^Hx)
  end.

Axiom P:AR^A->Set.
Axiom F:(x:AR^A)((y:AR^A)(AR^R y x) -> (P y))->(P x).

Definition Acc_iter [x:AR^A] : (Acc AR x) -> (P x) :=
 (Fix {M, Acc_iter | [x:AR^A] ((Acc AR x)-M)->(P x) :> (Acc AR x)->(P x) :=
        [a:(Acc AR x)+M]
          (F x ([y:AR^A][h:AR^R y x](Acc_iter^0 y (Acc_inv M x a y h))))
      }^0 x).

(* Version tout-en-un : *)

Definition Acc_iter_standalone :
 (AR:related_set)
 (P:AR^A->Set)
 (F:(x:AR^A)((y:AR^A)(AR^R y x) -> (P y))->(P x))
 (x:AR^A)
 (Acc AR x) -> (P x)
:=
 [AR:related_set]
 [P:AR^A->Set]
 [F:(x:AR^A)((y:AR^A)(AR^R y x) -> (P y))->(P x)]
 (Fix {M, Acc_iter | [x:AR^A] ((Acc AR x)-M)->(P x) :> (Acc AR x)->(P x) :=
   [a:(Acc AR x)+M]
     (F x ([y:AR^A][h:AR^R y x](Acc_iter^0 y
        (<[x':AR^A*Lmark][_:(!Acc AR x')](y:AR^A)(AR^R y x'^^1)->((Acc AR y)-M)>
         Cases a of Acc_intro =>
           [arg:{<x:AR^A;Hx:(y:AR^A)(AR^R y x)->((Acc AR y)-M)>}](arg^Hx)
         end y h))))
   }^0).
