(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Require Import Extraction.

Set Extraction AccessOpaque.

(* Use OCaml types *)
Extract Inductive Coq.Init.Datatypes.unit => "unit" [ "()" ].
Extract Inductive bool => "bool" [ "true" "false" ].
Extract Inductive sumbool => "bool" [ "true" "false" ].
Extract Inductive sumor => "option" [ "Some" "None" ].
Extract Inductive option => "option" [ "Some" "None" ].
Extract Inductive prod => "( * )"  [ "( , )" ].
Extract Inductive sigT => "( * )" [ "( , )" ].
Extract Inductive list => "list" [ "[]" "( :: )" ].

(* Coq's strings to OCaml's strings. Should probably be moved to a
   plugin, so that the conversion is done at extraction time (at least
   for litterals)... *)
Extract Inductive ml_string => "String.t" [ "Magic.mk_str" ].
Extract Constant string_value => "fun x -> failwith ""no reification defined for string_value""".

(* Coq's ints to OCaml's Big_int *)
Extract Inductive MlInteger.int => "Big_int.big_int" [ "Magic.mk_int" ].
Extract Constant MlInteger.int_value => "fun x -> failwith ""no reification defined for int_value""".

Extract Inlined Constant app_str => "( ^ )".
Extract Inlined Constant eq_string => "( = )".
Extract Inlined Constant string_of_int => "Big_int.string_of_big_int".
Extract Inlined Constant print_endline => "print_endline".

(* Do not wrap extracted code *)
Set Printing Width 1000.

Unset Elimination Schemes.

Module Public.

  Definition env := CicCore.env.
  Definition genv := genv.

  Definition flat_map :
    forall A B : Set, (A -> list B) -> list A -> list B
      := @flat_map.

  Definition find_free_var :
    ml_string -> forall l : list ml_string, {x : ml_string | ~ In x l}
      := @find_free_var.

  Definition cstr_one_ind : one_ind -> list name := @cstr_one_ind.
  Definition nil_genv : genv := @nil_genv.

  Definition enter_decl (g : genv) (s : sig_item) : option genv :=
    match enter_decl g s with
      | inleft (exist x _) => Some x
      | inright _ => None
    end.

  Definition infer_type (g : genv) (t : term) :=
    match infer_type g t with
      | inleft (exist x _) => Some x
      | inright _ => None
    end.

  Definition check_type (g : genv) (t : term) (T : term) : bool :=
    if check_type g t T then true else false.

  Definition extract_term (g : genv) (e : env) (t : term) H : term :=
    match @extract_term g e t H with
      | exist x _ => x
    end.

End Public.

Extraction "bcoq/kernel" Public.
