(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

open Kernel
open CicCore

(* restore pervasive string *)
type string = String.t

(* Definitions auxiliaires utilisant le noyau *)

let rec conv_list = function
    [] -> []
  | x::l -> x::(conv_list l)
;;

let rec nat_of_int = function
  0 -> O
| k -> S (nat_of_int (k-1))
;;

let rec string_conc sep = function
    [] -> ""
  | [x] -> x
  | x::l -> x^sep^string_conc sep l
;;

let names_of = function
    Constant {c_decl={dNAME=x}} -> [x]
  | MutInductive ip ->
      (flat_map (fun is -> (i_name is)::(cstr_one_ind is)) ip)
;;


let split_arrow t =
  let rec split stk = function
      Prd(_,ty,rng) -> split (ty::stk) rng
    | t -> (List.fold_left (fun t ty -> Sum(ty,t)) (Cst0 Lmark) stk, t)
  in split [] t
;;

let split_arity t =
  match split_arrow t with
    (arg, Srt s) -> (arg,s)
  | _ -> failwith "bad arity"
;;

let define_constr is cl =
  {is with i_constrs=Some cl}
;;


let mk_nilmark = Cst0 Nil_mark;;
let mk_app a b = Cst1(App,Pair0(a,b));;
let mk_appl = List.fold_left mk_app;;

let mk_v args = List.fold_right (fun ai atl -> Pair0(ai,atl)) args mk_nilmark;;
let mk_const x (args,oargs) = mk_appl (Cst1(Const x, args)) oargs;;
let mk_ind i (par,arg,(k,l)) =
  Cst1(MutInd(i,k), Pair0(par, Pair0(arg, l)));;
let mk_cstr c (par,arg) = Cst1(MutConstr c, Pair0(par, arg));;


let s_args exp ty args =
  let rec sl stk = function
      (Sum(_,v), x::l) -> sl (x::stk) (v, l)
    | (Cst0 Lmark, l) when not exp -> (mk_v (List.rev stk), l)
    | (t,x::l) -> (List.fold_left (fun atl ai -> Pair0(ai,atl)) x stk, l) 
    | _ -> failwith "constante partiellement appliquee" in
  match (ty,args) with
    (_, x::l) when exp -> (x,l)
  | p -> sl [] p
;;


let rec s_mark_neg = function
    [] -> mk_nilmark
  | (Cst1 (Lmark, m))::l -> Cst1(Cons_mark, Pair0(m, s_mark_neg l))
  | (Cst1 (Mark, _))::_ -> failwith "marque pos. apres neg."
  | [l] -> l
  | _ -> failwith "type inductif applique a trop d'arguments"
;;
let rec s_mark = function
    [] -> (O,mk_nilmark)
  | (Cst1 (Mark, m))::l ->
      let (k,ml) = s_mark l in (S k, Cst1(Cons_mark, Pair0(m,ml)))
  | [Cst1 (Nil_mark, Prd(i,ml,_))] -> (nat_of_int (int_of_string i), ml)
  | l -> (O, s_mark_neg l)
;;
let s_ind exp is t =
  let (p,r) = s_args exp (i_param is) t in
  let (a,l) = s_args exp (i_arg is) r in
  (p, a, s_mark l)
;;
let s_cstr exp is c t =
  let (p,r) = s_args exp (i_param is) t in
  match s_args exp (c_arg c) r with
    (a, []) -> (p,a)
  | _ -> failwith "constructeur applique a trop d'arguments"
;;
let s_cst exp co t = s_args exp (c_par co) t
;;

let glob_cst exp sg x args =
  match search_global sg x with
    Some (GlobConst co) -> mk_const x (s_cst exp co args)
  | Some (GlobInd (_,is)) -> mk_ind x (s_ind exp is args)
  | Some (GlobCstr (is,c)) -> mk_cstr x (s_cstr exp is c args)
  | _ -> failwith ("identificateur "^x^" inconnu")
;;


(* non utilise *)
let globalize0 (sg,lenv) x =
  match (list_index (=) x lenv) with
    Some n -> Rel n
  | None -> glob_cst false sg x []
;;

let globalize1 exp (sg,lenv) x args =
  match (list_index (=) x lenv) with
    Some n -> mk_appl (Rel n) args
  | None -> glob_cst exp sg x args
;;

type gen_sort =
    Gprop
  | Gset
  | Gtype of nat
  | Gtypeset of nat

let gen_of_sort = function
  | Sprop Neg -> Gprop
  | Sprop Pos -> Gset
  | Stype n -> Gtype n

let sort_of_gen = function
  | Gprop -> Some (Sprop Neg)
  | Gset -> Some (Sprop Pos)
  | Gtype n -> Some (Stype n)
  | _ -> None

type expr =
    SRT of gen_sort
  | REL of string
  | OP0 of cci_op
  | OP1 of cci_op * expr
  | LAM of string * expr * expr
  | PRD of string * expr * expr
  | PAIR of expr * expr
  | SUM of expr * expr
;;

type command =
  | Check of term * term option
  | Definition of (string * (string * term) * term) * term option
  | Extraction of string
  | Inductive of ind_pack
  | Print of string
  | Quit
  | Delete
  | List
;;
