(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

(* should be kept in sync with the extraction of string *)

type ascii =
  | Ascii of bool * bool * bool * bool * bool * bool * bool * bool

type coq_string =
  | EmptyString
  | String of ascii * coq_string

type positive =
  | XI of positive
  | XO of positive
  | XH

type n =
  | N0
  | Npos of positive

(* /should be kept in sync with the extraction of string *)

let mk_str x =
  let buf = Buffer.create 256 in
  let rec build_string = function
    | EmptyString -> Buffer.contents buf
    | String (Ascii (_,_,_,_,_,_,_,_) as x, tl) ->
      let x = Obj.repr x in
      let rec build_char i accu =
        if i >= 0 then
          build_char (i-1) (accu lor (Obj.obj (Obj.field x i) lsl i))
        else
          char_of_int accu
      in
      Buffer.add_char buf (build_char 7 0);
      build_string tl
  in build_string (Obj.magic x)

open Big_int

let ( * ) = Big_int.mult_int_big_int
let ( + ) = Big_int.add_int_big_int

let mk_int x = match Obj.magic x with
  | N0 -> Big_int.zero_big_int
  | Npos x ->
    let rec build_int accu = function
      | XH -> accu
      | XI x -> build_int (1 + 2 * accu) x
      | XO x -> build_int (2 * accu) x
    in
    build_int Big_int.unit_big_int x
