(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

Inductive Empty_set: Set := .

Definition absurd_set: (P:Prop)Empty_set->P :=
  [P:Prop][f:Empty_set]
    <[p:Lmark][_:(!Empty_set ~ p)]P>Cases f of end.

Inductive unit: Set := tt: unit.

Inductive bool: Set :=
    true: bool
  | false: bool.

Inductive option [A:Set]: Set :=
    Some: A->(option A)
  | None: (option A).


Inductive nat: Set := O: (!nat _ #0) | S: (!nat _ ~)->nat.

Definition pred: nat->nat :=
  [k:nat]
    <[a:Lmark][_:(!nat ~ a)]nat>Cases k of
      O => [_:Lmark]O
    | S => [n:nat*Lmark]n^0
    end.
(*
Definition nat_eq: nat->nat->Prop :=
  (Fix {X, F | [m:nat] (nat -X)->Prop :> nat->Prop :=
          [n:nat+X]<[a:Lmark][_:(!nat ~ a)]Prop>Cases m of 
          | O => [_:Lmark]<[a:Lmark][_:(!nat ~ a)]Prop>Cases n of
                 | O => [_:Lmark]True
                 | S => [_:nat*Lmark]False
                 end
          | S => [pm:nat*Lmark]<[a:Lmark][_:(!nat ~ a)]Prop>Cases n of
                 | O => [_:Lmark]False
                 | S => [pn:(nat-X)*Lmark](F^0 pm^0 pn^0)
		 end
	  end}^0 ~).

Definition nat_eq: nat->nat->Prop :=
  (Fix {X, F | [m:nat] True :> True:= True
	  }^0 ~).

*)

Inductive list [A:Set]: Set :=
    nil: (list A)
  | cons: A->(list A)->(list A).


Mutual Inductive tree [A:Set]: Set :=
    Node : A -> (forest A) -> (tree A)

with forest [A:Set]: Set :=
    Empty : unit -> (forest A)
  | Fcons : (tree A) -> (forest A) -> (forest A).


Inductive ord [p:Lmark]: Set :=
    Oo: ord
  | Olim: (nat->(!ord p ~))->ord.


Definition sig [A:Set]: (P:A->Prop)Set := [P:A->Prop] {< x:A ; prf: (P x) >}.

Definition ne_set: Type(0) := {< A: Set; x: A >}.
Definition mk_set: (A:Set)A->ne_set :=
  [A:Set][x:A] { A: Set := A; x: A :=x }.


(* Logique *)

Inductive True: Prop := I: True.
Inductive False: Prop := .

Definition absurd: (P:Prop)False->P :=
  [P:Prop][f:False]
    <[p:Lmark][_:(!False ~ p)]P>Cases f of end.
Axiom absurds: (P:Set)False->P.




Inductive or [p:Prop*Prop]: Prop :=
    or_introl: p^^1->(or p^^1 p^^2)
  | or_intror: p^^2->(or p^^1 p^^2).

Definition or_intro1: (A,B:Prop)A->(or A B) :=
  [A,B:Prop][a:A](or_introl A B a).
Definition or_intro2: (A,B:Prop)B->(or A B) :=
  [A,B:Prop][b:B](or_intror A B b).

Definition or_ind: (A,B,P:Prop)(A->P)->(B->P)->(or A B)->P :=
  [A,B,P:Prop][f0:A->P][f1:B->P][p:(or A B)]
    <[x:Lmark][_:(!or (A,B) x)]P>Cases p of
    | or_introl => [a:A*Lmark](f0 a^0)
    | or_intror => [b:B*Lmark](f1 b^0)
    end.


Inductive eq [p:ne_set]: p^A->Prop :=
  refl_eq: (eq p p^x).

Definition eq1: (A:Set)A->A->Prop := [A:Set][x,y:A](eq (mk_set A x) y).
Definition refl_equal: (A:Set)(x:A)(eq1 A x x) := 
  [A:Set][x:A](refl_eq (mk_set A x)).
Definition eq_ind: (A:Set)(x:A)(P:A->Prop)(P x)->(y:A)(eq1 A x y)->(P y) :=
  [A:Set][x:A][P:A->Prop][f:(P x)][y:A][e:(eq1 A x y)]
    <[y':A*Lmark][_:(!eq (mk_set A x) y')](P y'^0)>Cases e of
      refl_eq => [_:Lmark]f end.

