(**************************************************************************)
(*             Certified Calculus of Inductive Constructions              *)
(*                                                                        *)
(*                     Bruno Barras, Stéphane Glondu                      *)
(*                                                                        *)
(*                           © 1999-2011 INRIA                            *)
(*                                                                        *)
(*  This program is free software; you can redistribute it and/or modify  *)
(*   it under the terms of the CeCILL FREE SOFTWARE LICENSE, version 2.   *)
(**************************************************************************)

(* This is a manually restricted version of the kernel.mli generated
   by extraction. Its content comes from extraction, but is manually
   managed (and tuned) on purpose. *)

type nat =
  | O
  | S of nat

type calc =
  | Pos
  | Neg

type srt_v6 =
  | Sprop of calc
  | Stype of nat

module CicCore : sig
  type name = string
  type le_marks = nat
  type patterns = name list

  type cci_op =
    | Mark
    | Mt_mark
    | Lmark
    | Nil_mark
    | Cons_mark
    | App
    | Proj1
    | Proj2
    | Struct of name list
    | Record of name list
    | Field of name
    | Const of name
    | MutInd of name * le_marks
    | MutConstr of name
    | MutCase of bool * patterns
    | Fixp

  type oper = cci_op
  type sort = srt_v6

  type term =
    | Srt of sort
    | Rel of nat
    | Cst0 of oper
    | Cst1 of oper * term
    | Prd of name * term * term
    | Lam of name * term * term
    | Sum of term * term
    | Pair0 of term * term

  type cdecl = {
    dNAME : name;
    dTYPE : term;
    dBODY : term option;
  }

  type ppal_judge = {
    coq_VAL : term;
    coq_TYP : term;
  }
end

open CicCore

type constructor = {
  c_name : name;
  c_arg : term;
  c_inst : term;
}

type one_ind = {
  i_name : name;
  i_param : term;
  i_arg : term;
  i_ckind : sort;
  i_kind : sort;
  i_constrs : constructor list option;
}

type constant_obj = {
  c_par : term;
  c_decl : cdecl;
}

type sig_item =
  | MutInductive of one_ind list
  | Constant of constant_obj

module Public : sig
  type env
  type genv

  val flat_map : ('a -> 'b list) -> 'a list -> 'b list
  val find_free_var : String.t -> String.t list -> String.t
  val cstr_one_ind : one_ind -> name list
  val nil_genv : genv
  val enter_decl : genv -> sig_item -> genv option
  val infer_type : genv -> term -> ppal_judge option
  val check_type : genv -> term -> term -> bool
  val extract_term : genv -> env -> term -> term
end
