open Datatypes

let list_nat = list nat

let xenv = ref Kernel.Public.nil_genv

let () = Summary.(
  declare_summary "Global extracted environment" {
    freeze_function = (fun () -> !xenv);
    unfreeze_function = (fun fr -> xenv := fr);
    init_function = (fun () -> xenv := Kernel.Public.nil_genv);
  }
)

let vernac_enter_decl si =
  match sig_item.interp si with
    | Some si ->
      begin match Kernel.Public.enter_decl !xenv si with
        | Some genv ->
          xenv := genv;
          Pp.msg (Pp.str "OK")
        | None ->
          Util.error "Ill-typed declaration"
      end
    | None ->
      Util.error "Not a declaration"

let vernac_infer_type t =
  match term.interp t with
    | Some t ->
      begin match Kernel.Public.infer_type !xenv t with
        | Some j -> j
        | None -> Util.error "Ill-typed term"
      end
    | None ->
      Util.error "Not a term"

let bootstrap x = match nat.interp x with
  | Some y -> nat.reify y
  | None -> match list_nat.interp x with
  | Some y -> list_nat.reify y
  | None -> match string.interp x with
  | Some y -> string.reify y
  | None -> match term.interp x with
  | Some y -> term.reify y
  | None -> match sig_item.interp x with
  | Some y -> sig_item.reify y
  | _ -> Util.error "Type not supported"

let normalize rc =
  let env = Global.env () in
  let c = Constrintern.interp_constr Evd.empty env rc in
   Redexpr.cbv_vm env Evd.empty c

open Pp

let print_judgment env j =
  let trm = term.reify j.Kernel.CicCore.coq_VAL in
  let typ = term.reify j.Kernel.CicCore.coq_TYP in
  Printer.pr_lconstr_env env trm ++ fnl () ++
    str "     : " ++ Printer.pr_ltype_env env typ ++ fnl ()

VERNAC COMMAND EXTEND XDeclare
| [ "XDeclare" constr(rc) ] -> [
  vernac_enter_decl (normalize rc)
]
END;;

VERNAC COMMAND EXTEND XCheck
| [ "XCheck" constr(rc) ] -> [
  let j = vernac_infer_type (normalize rc) in
  let env = Global.env () in
  Pp.msg (print_judgment env j)
]
END;;

VERNAC COMMAND EXTEND XBoot
| [ "XBoot" constr(rc) ] -> [
  let c = bootstrap (normalize rc) in
  let env = Global.env () in
  let j = Typeops.typing env c in
  Pp.msg (Prettyp.print_judgment env j)
]
END;;
